/* Generated by Cython 0.25.2 */

/* BEGIN: Cython Metadata
{
    "distutils": {},
    "module_name": "pymola.generated.ModelicaListener"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_25_2"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x03030000 || (PY_MAJOR_VERSION == 2 && PY_VERSION_HEX >= 0x02070000)
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_FASTCALL
  #define METH_FASTCALL 0x80
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject **args,
                                              Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__pymola__generated__ModelicaListener
#define __PYX_HAVE_API__pymola__generated__ModelicaListener
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;


static const char *__pyx_f[] = {
  "pymola/generated/ModelicaListener.pyx",
};

/*--- Type declarations ---*/

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CalculateMetaclass.proto */
static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* CythonFunction.proto */
#define __Pyx_CyFunction_USED 1
#include <structmember.h>
#define __Pyx_CYFUNCTION_STATICMETHOD  0x01
#define __Pyx_CYFUNCTION_CLASSMETHOD   0x02
#define __Pyx_CYFUNCTION_CCLASS        0x04
#define __Pyx_CyFunction_GetClosure(f)\
    (((__pyx_CyFunctionObject *) (f))->func_closure)
#define __Pyx_CyFunction_GetClassObj(f)\
    (((__pyx_CyFunctionObject *) (f))->func_classobj)
#define __Pyx_CyFunction_Defaults(type, f)\
    ((type *)(((__pyx_CyFunctionObject *) (f))->defaults))
#define __Pyx_CyFunction_SetDefaultsGetter(f, g)\
    ((__pyx_CyFunctionObject *) (f))->defaults_getter = (g)
typedef struct {
    PyCFunctionObject func;
#if PY_VERSION_HEX < 0x030500A0
    PyObject *func_weakreflist;
#endif
    PyObject *func_dict;
    PyObject *func_name;
    PyObject *func_qualname;
    PyObject *func_doc;
    PyObject *func_globals;
    PyObject *func_code;
    PyObject *func_closure;
    PyObject *func_classobj;
    void *defaults;
    int defaults_pyobjects;
    int flags;
    PyObject *defaults_tuple;
    PyObject *defaults_kwdict;
    PyObject *(*defaults_getter)(PyObject *);
    PyObject *func_annotations;
} __pyx_CyFunctionObject;
static PyTypeObject *__pyx_CyFunctionType = 0;
#define __Pyx_CyFunction_NewEx(ml, flags, qualname, self, module, globals, code)\
    __Pyx_CyFunction_New(__pyx_CyFunctionType, ml, flags, qualname, self, module, globals, code)
static PyObject *__Pyx_CyFunction_New(PyTypeObject *, PyMethodDef *ml,
                                      int flags, PyObject* qualname,
                                      PyObject *self,
                                      PyObject *module, PyObject *globals,
                                      PyObject* code);
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *m,
                                                         size_t size,
                                                         int pyobjects);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *m,
                                                            PyObject *tuple);
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *m,
                                                             PyObject *dict);
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *m,
                                                              PyObject *dict);
static int __pyx_CyFunction_init(void);

/* Py3ClassCreate.proto */
static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name, PyObject *qualname,
                                           PyObject *mkw, PyObject *modname, PyObject *doc);
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases, PyObject *dict,
                                      PyObject *mkw, int calculate_metaclass, int allow_py2_metaclass);

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CStringEquals.proto */
static CYTHON_INLINE int __Pyx_StrEq(const char *, const char *);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);


/* Module declarations from 'pymola.generated.ModelicaListener' */
#define __Pyx_MODULE_NAME "pymola.generated.ModelicaListener"
int __pyx_module_is_main_pymola__generated__ModelicaListener = 0;

/* Implementation of 'pymola.generated.ModelicaListener' */
static const char __pyx_k_[] = "*";
static const char __pyx_k__2[] = ".";
static const char __pyx_k_ctx[] = "ctx";
static const char __pyx_k_doc[] = "__doc__";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_name[] = "__name__";
static const char __pyx_k_self[] = "self";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_antlr4[] = "antlr4";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_module[] = "__module__";
static const char __pyx_k_prepare[] = "__prepare__";
static const char __pyx_k_exitName[] = "exitName";
static const char __pyx_k_qualname[] = "__qualname__";
static const char __pyx_k_enterName[] = "enterName";
static const char __pyx_k_metaclass[] = "__metaclass__";
static const char __pyx_k_NameContext[] = "NameContext";
static const char __pyx_k_exitComment[] = "exitComment";
static const char __pyx_k_exitElement[] = "exitElement";
static const char __pyx_k_exitExpr_or[] = "exitExpr_or";
static const char __pyx_k_enterComment[] = "enterComment";
static const char __pyx_k_enterElement[] = "enterElement";
static const char __pyx_k_enterExpr_or[] = "enterExpr_or";
static const char __pyx_k_exitArgument[] = "exitArgument";
static const char __pyx_k_exitEquation[] = "exitEquation";
static const char __pyx_k_exitExpr_add[] = "exitExpr_add";
static const char __pyx_k_exitExpr_and[] = "exitExpr_and";
static const char __pyx_k_exitExpr_exp[] = "exitExpr_exp";
static const char __pyx_k_exitExpr_mul[] = "exitExpr_mul";
static const char __pyx_k_exitExpr_neg[] = "exitExpr_neg";
static const char __pyx_k_exitExpr_not[] = "exitExpr_not";
static const char __pyx_k_exitExpr_rel[] = "exitExpr_rel";
static const char __pyx_k_enterArgument[] = "enterArgument";
static const char __pyx_k_enterEquation[] = "enterEquation";
static const char __pyx_k_enterExpr_add[] = "enterExpr_add";
static const char __pyx_k_enterExpr_and[] = "enterExpr_and";
static const char __pyx_k_enterExpr_exp[] = "enterExpr_exp";
static const char __pyx_k_enterExpr_mul[] = "enterExpr_mul";
static const char __pyx_k_enterExpr_neg[] = "enterExpr_neg";
static const char __pyx_k_enterExpr_not[] = "enterExpr_not";
static const char __pyx_k_enterExpr_rel[] = "enterExpr_rel";
static const char __pyx_k_exitEnum_list[] = "exitEnum_list";
static const char __pyx_k_exitFor_index[] = "exitFor_index";
static const char __pyx_k_exitStatement[] = "exitStatement";
static const char __pyx_k_exitSubscript[] = "exitSubscript";
static const char __pyx_k_CommentContext[] = "CommentContext";
static const char __pyx_k_ElementContext[] = "ElementContext";
static const char __pyx_k_Expr_orContext[] = "Expr_orContext";
static const char __pyx_k_ModelicaParser[] = "ModelicaParser";
static const char __pyx_k_enterEnum_list[] = "enterEnum_list";
static const char __pyx_k_enterFor_index[] = "enterFor_index";
static const char __pyx_k_enterStatement[] = "enterStatement";
static const char __pyx_k_enterSubscript[] = "enterSubscript";
static const char __pyx_k_exitAnnotation[] = "exitAnnotation";
static const char __pyx_k_exitClass_type[] = "exitClass_type";
static const char __pyx_k_ArgumentContext[] = "ArgumentContext";
static const char __pyx_k_EquationContext[] = "EquationContext";
static const char __pyx_k_Expr_addContext[] = "Expr_addContext";
static const char __pyx_k_Expr_andContext[] = "Expr_andContext";
static const char __pyx_k_Expr_expContext[] = "Expr_expContext";
static const char __pyx_k_Expr_mulContext[] = "Expr_mulContext";
static const char __pyx_k_Expr_negContext[] = "Expr_negContext";
static const char __pyx_k_Expr_notContext[] = "Expr_notContext";
static const char __pyx_k_Expr_relContext[] = "Expr_relContext";
static const char __pyx_k_enterAnnotation[] = "enterAnnotation";
static const char __pyx_k_enterClass_type[] = "enterClass_type";
static const char __pyx_k_exitBase_prefix[] = "exitBase_prefix";
static const char __pyx_k_exitComposition[] = "exitComposition";
static const char __pyx_k_exitDeclaration[] = "exitDeclaration";
static const char __pyx_k_exitEquation_if[] = "exitEquation_if";
static const char __pyx_k_exitFor_indices[] = "exitFor_indices";
static const char __pyx_k_exitIf_equation[] = "exitIf_equation";
static const char __pyx_k_exitImport_list[] = "exitImport_list";
static const char __pyx_k_exitPrimary_end[] = "exitPrimary_end";
static const char __pyx_k_exitType_prefix[] = "exitType_prefix";
static const char __pyx_k_Enum_listContext[] = "Enum_listContext";
static const char __pyx_k_For_indexContext[] = "For_indexContext";
static const char __pyx_k_ModelicaListener[] = "ModelicaListener";
static const char __pyx_k_StatementContext[] = "StatementContext";
static const char __pyx_k_SubscriptContext[] = "SubscriptContext";
static const char __pyx_k_enterBase_prefix[] = "enterBase_prefix";
static const char __pyx_k_enterComposition[] = "enterComposition";
static const char __pyx_k_enterDeclaration[] = "enterDeclaration";
static const char __pyx_k_enterEquation_if[] = "enterEquation_if";
static const char __pyx_k_enterFor_indices[] = "enterFor_indices";
static const char __pyx_k_enterIf_equation[] = "enterIf_equation";
static const char __pyx_k_enterImport_list[] = "enterImport_list";
static const char __pyx_k_enterPrimary_end[] = "enterPrimary_end";
static const char __pyx_k_enterType_prefix[] = "enterType_prefix";
static const char __pyx_k_exitElement_list[] = "exitElement_list";
static const char __pyx_k_exitEquation_for[] = "exitEquation_for";
static const char __pyx_k_exitExpr_primary[] = "exitExpr_primary";
static const char __pyx_k_exitFor_equation[] = "exitFor_equation";
static const char __pyx_k_exitIf_statement[] = "exitIf_statement";
static const char __pyx_k_exitPrimary_true[] = "exitPrimary_true";
static const char __pyx_k_exitStatement_if[] = "exitStatement_if";
static const char __pyx_k_AnnotationContext[] = "AnnotationContext";
static const char __pyx_k_Class_typeContext[] = "Class_typeContext";
static const char __pyx_k_ParseTreeListener[] = "ParseTreeListener";
static const char __pyx_k_enterElement_list[] = "enterElement_list";
static const char __pyx_k_enterEquation_for[] = "enterEquation_for";
static const char __pyx_k_enterExpr_primary[] = "enterExpr_primary";
static const char __pyx_k_enterFor_equation[] = "enterFor_equation";
static const char __pyx_k_enterIf_statement[] = "enterIf_statement";
static const char __pyx_k_enterPrimary_true[] = "enterPrimary_true";
static const char __pyx_k_enterStatement_if[] = "enterStatement_if";
static const char __pyx_k_exitArgument_list[] = "exitArgument_list";
static const char __pyx_k_exitEquation_when[] = "exitEquation_when";
static const char __pyx_k_exitExpression_if[] = "exitExpression_if";
static const char __pyx_k_exitFor_statement[] = "exitFor_statement";
static const char __pyx_k_exitImport_clause[] = "exitImport_clause";
static const char __pyx_k_exitPrimary_false[] = "exitPrimary_false";
static const char __pyx_k_exitStatement_for[] = "exitStatement_for";
static const char __pyx_k_exitWhen_equation[] = "exitWhen_equation";
static const char __pyx_k_Base_prefixContext[] = "Base_prefixContext";
static const char __pyx_k_CompositionContext[] = "CompositionContext";
static const char __pyx_k_DeclarationContext[] = "DeclarationContext";
static const char __pyx_k_Equation_ifContext[] = "Equation_ifContext";
static const char __pyx_k_For_indicesContext[] = "For_indicesContext";
static const char __pyx_k_If_equationContext[] = "If_equationContext";
static const char __pyx_k_Import_listContext[] = "Import_listContext";
static const char __pyx_k_Primary_endContext[] = "Primary_endContext";
static const char __pyx_k_Type_prefixContext[] = "Type_prefixContext";
static const char __pyx_k_enterArgument_list[] = "enterArgument_list";
static const char __pyx_k_enterEquation_when[] = "enterEquation_when";
static const char __pyx_k_enterExpression_if[] = "enterExpression_if";
static const char __pyx_k_enterFor_statement[] = "enterFor_statement";
static const char __pyx_k_enterImport_clause[] = "enterImport_clause";
static const char __pyx_k_enterPrimary_false[] = "enterPrimary_false";
static const char __pyx_k_enterStatement_for[] = "enterStatement_for";
static const char __pyx_k_enterWhen_equation[] = "enterWhen_equation";
static const char __pyx_k_exitClass_prefixes[] = "exitClass_prefixes";
static const char __pyx_k_exitClass_spec_der[] = "exitClass_spec_der";
static const char __pyx_k_exitComponent_list[] = "exitComponent_list";
static const char __pyx_k_exitConnect_clause[] = "exitConnect_clause";
static const char __pyx_k_exitExtends_clause[] = "exitExtends_clause";
static const char __pyx_k_exitNamed_argument[] = "exitNamed_argument";
static const char __pyx_k_exitPrimary_string[] = "exitPrimary_string";
static const char __pyx_k_exitStatement_when[] = "exitStatement_when";
static const char __pyx_k_exitString_comment[] = "exitString_comment";
static const char __pyx_k_exitType_specifier[] = "exitType_specifier";
static const char __pyx_k_exitWhen_statement[] = "exitWhen_statement";
static const char __pyx_k_Element_listContext[] = "Element_listContext";
static const char __pyx_k_Equation_forContext[] = "Equation_forContext";
static const char __pyx_k_Expr_primaryContext[] = "Expr_primaryContext";
static const char __pyx_k_For_equationContext[] = "For_equationContext";
static const char __pyx_k_If_statementContext[] = "If_statementContext";
static const char __pyx_k_Primary_trueContext[] = "Primary_trueContext";
static const char __pyx_k_Statement_ifContext[] = "Statement_ifContext";
static const char __pyx_k_enterClass_prefixes[] = "enterClass_prefixes";
static const char __pyx_k_enterClass_spec_der[] = "enterClass_spec_der";
static const char __pyx_k_enterComponent_list[] = "enterComponent_list";
static const char __pyx_k_enterConnect_clause[] = "enterConnect_clause";
static const char __pyx_k_enterExtends_clause[] = "enterExtends_clause";
static const char __pyx_k_enterNamed_argument[] = "enterNamed_argument";
static const char __pyx_k_enterPrimary_string[] = "enterPrimary_string";
static const char __pyx_k_enterStatement_when[] = "enterStatement_when";
static const char __pyx_k_enterString_comment[] = "enterString_comment";
static const char __pyx_k_enterType_specifier[] = "enterType_specifier";
static const char __pyx_k_enterWhen_statement[] = "enterWhen_statement";
static const char __pyx_k_exitClass_spec_base[] = "exitClass_spec_base";
static const char __pyx_k_exitClass_spec_comp[] = "exitClass_spec_comp";
static const char __pyx_k_exitClass_spec_enum[] = "exitClass_spec_enum";
static const char __pyx_k_exitEquation_simple[] = "exitEquation_simple";
static const char __pyx_k_exitExpression_list[] = "exitExpression_list";
static const char __pyx_k_exitNamed_arguments[] = "exitNamed_arguments";
static const char __pyx_k_exitPrimary_initial[] = "exitPrimary_initial";
static const char __pyx_k_exitRegular_element[] = "exitRegular_element";
static const char __pyx_k_exitStatement_break[] = "exitStatement_break";
static const char __pyx_k_exitStatement_while[] = "exitStatement_while";
static const char __pyx_k_exitWhile_statement[] = "exitWhile_statement";
static const char __pyx_k_Argument_listContext[] = "Argument_listContext";
static const char __pyx_k_Equation_whenContext[] = "Equation_whenContext";
static const char __pyx_k_Expression_ifContext[] = "Expression_ifContext";
static const char __pyx_k_For_statementContext[] = "For_statementContext";
static const char __pyx_k_Import_clauseContext[] = "Import_clauseContext";
static const char __pyx_k_Primary_falseContext[] = "Primary_falseContext";
static const char __pyx_k_Statement_forContext[] = "Statement_forContext";
static const char __pyx_k_When_equationContext[] = "When_equationContext";
static const char __pyx_k_enterClass_spec_base[] = "enterClass_spec_base";
static const char __pyx_k_enterClass_spec_comp[] = "enterClass_spec_comp";
static const char __pyx_k_enterClass_spec_enum[] = "enterClass_spec_enum";
static const char __pyx_k_enterEquation_simple[] = "enterEquation_simple";
static const char __pyx_k_enterExpression_list[] = "enterExpression_list";
static const char __pyx_k_enterNamed_arguments[] = "enterNamed_arguments";
static const char __pyx_k_enterPrimary_initial[] = "enterPrimary_initial";
static const char __pyx_k_enterRegular_element[] = "enterRegular_element";
static const char __pyx_k_enterStatement_break[] = "enterStatement_break";
static const char __pyx_k_enterStatement_while[] = "enterStatement_while";
static const char __pyx_k_enterWhile_statement[] = "enterWhile_statement";
static const char __pyx_k_exitArray_subscripts[] = "exitArray_subscripts";
static const char __pyx_k_exitClass_definition[] = "exitClass_definition";
static const char __pyx_k_exitComponent_clause[] = "exitComponent_clause";
static const char __pyx_k_exitEquation_section[] = "exitEquation_section";
static const char __pyx_k_exitPrimary_function[] = "exitPrimary_function";
static const char __pyx_k_exitStatement_return[] = "exitStatement_return";
static const char __pyx_k_Class_prefixesContext[] = "Class_prefixesContext";
static const char __pyx_k_Class_spec_derContext[] = "Class_spec_derContext";
static const char __pyx_k_Component_listContext[] = "Component_listContext";
static const char __pyx_k_Connect_clauseContext[] = "Connect_clauseContext";
static const char __pyx_k_Extends_clauseContext[] = "Extends_clauseContext";
static const char __pyx_k_Named_argumentContext[] = "Named_argumentContext";
static const char __pyx_k_Primary_stringContext[] = "Primary_stringContext";
static const char __pyx_k_Statement_whenContext[] = "Statement_whenContext";
static const char __pyx_k_String_commentContext[] = "String_commentContext";
static const char __pyx_k_Type_specifierContext[] = "Type_specifierContext";
static const char __pyx_k_When_statementContext[] = "When_statementContext";
static const char __pyx_k_enterArray_subscripts[] = "enterArray_subscripts";
static const char __pyx_k_enterClass_definition[] = "enterClass_definition";
static const char __pyx_k_enterComponent_clause[] = "enterComponent_clause";
static const char __pyx_k_enterEquation_section[] = "enterEquation_section";
static const char __pyx_k_enterPrimary_function[] = "enterPrimary_function";
static const char __pyx_k_enterStatement_return[] = "enterStatement_return";
static const char __pyx_k_exitAlgorithm_section[] = "exitAlgorithm_section";
static const char __pyx_k_exitArgument_function[] = "exitArgument_function";
static const char __pyx_k_exitComponent_clause1[] = "exitComponent_clause1";
static const char __pyx_k_exitEquation_function[] = "exitEquation_function";
static const char __pyx_k_exitExpression_simple[] = "exitExpression_simple";
static const char __pyx_k_exitSimple_expression[] = "exitSimple_expression";
static const char __pyx_k_exitStored_definition[] = "exitStored_definition";
static const char __pyx_k_Class_spec_baseContext[] = "Class_spec_baseContext";
static const char __pyx_k_Class_spec_compContext[] = "Class_spec_compContext";
static const char __pyx_k_Class_spec_enumContext[] = "Class_spec_enumContext";
static const char __pyx_k_Equation_simpleContext[] = "Equation_simpleContext";
static const char __pyx_k_Expression_listContext[] = "Expression_listContext";
static const char __pyx_k_Named_argumentsContext[] = "Named_argumentsContext";
static const char __pyx_k_Primary_initialContext[] = "Primary_initialContext";
static const char __pyx_k_Regular_elementContext[] = "Regular_elementContext";
static const char __pyx_k_Statement_breakContext[] = "Statement_breakContext";
static const char __pyx_k_Statement_whileContext[] = "Statement_whileContext";
static const char __pyx_k_While_statementContext[] = "While_statementContext";
static const char __pyx_k_enterAlgorithm_section[] = "enterAlgorithm_section";
static const char __pyx_k_enterArgument_function[] = "enterArgument_function";
static const char __pyx_k_enterComponent_clause1[] = "enterComponent_clause1";
static const char __pyx_k_enterEquation_function[] = "enterEquation_function";
static const char __pyx_k_enterExpression_simple[] = "enterExpression_simple";
static const char __pyx_k_enterSimple_expression[] = "enterSimple_expression";
static const char __pyx_k_enterStored_definition[] = "enterStored_definition";
static const char __pyx_k_exitClass_modification[] = "exitClass_modification";
static const char __pyx_k_exitClass_spec_extends[] = "exitClass_spec_extends";
static const char __pyx_k_exitFunction_arguments[] = "exitFunction_arguments";
static const char __pyx_k_exitFunction_call_args[] = "exitFunction_call_args";
static const char __pyx_k_exitModification_class[] = "exitModification_class";
static const char __pyx_k_exitPrimary_derivative[] = "exitPrimary_derivative";
static const char __pyx_k_Array_subscriptsContext[] = "Array_subscriptsContext";
static const char __pyx_k_Class_definitionContext[] = "Class_definitionContext";
static const char __pyx_k_Component_clauseContext[] = "Component_clauseContext";
static const char __pyx_k_Equation_sectionContext[] = "Equation_sectionContext";
static const char __pyx_k_Primary_functionContext[] = "Primary_functionContext";
static const char __pyx_k_Statement_returnContext[] = "Statement_returnContext";
static const char __pyx_k_enterClass_modification[] = "enterClass_modification";
static const char __pyx_k_enterClass_spec_extends[] = "enterClass_spec_extends";
static const char __pyx_k_enterFunction_arguments[] = "enterFunction_arguments";
static const char __pyx_k_enterFunction_call_args[] = "enterFunction_call_args";
static const char __pyx_k_enterModification_class[] = "enterModification_class";
static const char __pyx_k_enterPrimary_derivative[] = "enterPrimary_derivative";
static const char __pyx_k_exitArgument_expression[] = "exitArgument_expression";
static const char __pyx_k_exitComponent_reference[] = "exitComponent_reference";
static const char __pyx_k_exitCondition_attribute[] = "exitCondition_attribute";
static const char __pyx_k_exitConstraining_clause[] = "exitConstraining_clause";
static const char __pyx_k_exitElement_replaceable[] = "exitElement_replaceable";
static const char __pyx_k_exitEnumeration_literal[] = "exitEnumeration_literal";
static const char __pyx_k_exitReplaceable_element[] = "exitReplaceable_element";
static const char __pyx_k_Algorithm_sectionContext[] = "Algorithm_sectionContext";
static const char __pyx_k_Argument_functionContext[] = "Argument_functionContext";
static const char __pyx_k_Component_clause1Context[] = "Component_clause1Context";
static const char __pyx_k_Equation_functionContext[] = "Equation_functionContext";
static const char __pyx_k_Expression_simpleContext[] = "Expression_simpleContext";
static const char __pyx_k_Simple_expressionContext[] = "Simple_expressionContext";
static const char __pyx_k_Stored_definitionContext[] = "Stored_definitionContext";
static const char __pyx_k_enterArgument_expression[] = "enterArgument_expression";
static const char __pyx_k_enterComponent_reference[] = "enterComponent_reference";
static const char __pyx_k_enterCondition_attribute[] = "enterCondition_attribute";
static const char __pyx_k_enterConstraining_clause[] = "enterConstraining_clause";
static const char __pyx_k_enterElement_replaceable[] = "enterElement_replaceable";
static const char __pyx_k_enterEnumeration_literal[] = "enterEnumeration_literal";
static const char __pyx_k_enterReplaceable_element[] = "enterReplaceable_element";
static const char __pyx_k_exitElement_modification[] = "exitElement_modification";
static const char __pyx_k_Class_modificationContext[] = "Class_modificationContext";
static const char __pyx_k_Class_spec_extendsContext[] = "Class_spec_extendsContext";
static const char __pyx_k_Function_argumentsContext[] = "Function_argumentsContext";
static const char __pyx_k_Function_call_argsContext[] = "Function_call_argsContext";
static const char __pyx_k_ModelicaListener_exitName[] = "ModelicaListener.exitName";
static const char __pyx_k_Modification_classContext[] = "Modification_classContext";
static const char __pyx_k_Primary_derivativeContext[] = "Primary_derivativeContext";
static const char __pyx_k_enterElement_modification[] = "enterElement_modification";
static const char __pyx_k_exitComponent_declaration[] = "exitComponent_declaration";
static const char __pyx_k_exitElement_redeclaration[] = "exitElement_redeclaration";
static const char __pyx_k_Argument_expressionContext[] = "Argument_expressionContext";
static const char __pyx_k_Component_referenceContext[] = "Component_referenceContext";
static const char __pyx_k_Condition_attributeContext[] = "Condition_attributeContext";
static const char __pyx_k_Constraining_clauseContext[] = "Constraining_clauseContext";
static const char __pyx_k_Element_replaceableContext[] = "Element_replaceableContext";
static const char __pyx_k_Enumeration_literalContext[] = "Enumeration_literalContext";
static const char __pyx_k_ModelicaListener_enterName[] = "ModelicaListener.enterName";
static const char __pyx_k_Replaceable_elementContext[] = "Replaceable_elementContext";
static const char __pyx_k_enterComponent_declaration[] = "enterComponent_declaration";
static const char __pyx_k_enterElement_redeclaration[] = "enterElement_redeclaration";
static const char __pyx_k_exitComponent_declaration1[] = "exitComponent_declaration1";
static const char __pyx_k_exitExternal_function_call[] = "exitExternal_function_call";
static const char __pyx_k_exitLanguage_specification[] = "exitLanguage_specification";
static const char __pyx_k_exitOutput_expression_list[] = "exitOutput_expression_list";
static const char __pyx_k_exitShort_class_definition[] = "exitShort_class_definition";
static const char __pyx_k_Element_modificationContext[] = "Element_modificationContext";
static const char __pyx_k_enterComponent_declaration1[] = "enterComponent_declaration1";
static const char __pyx_k_enterExternal_function_call[] = "enterExternal_function_call";
static const char __pyx_k_enterLanguage_specification[] = "enterLanguage_specification";
static const char __pyx_k_enterOutput_expression_list[] = "enterOutput_expression_list";
static const char __pyx_k_enterShort_class_definition[] = "enterShort_class_definition";
static const char __pyx_k_exitEquation_connect_clause[] = "exitEquation_connect_clause";
static const char __pyx_k_exitModification_assignment[] = "exitModification_assignment";
static const char __pyx_k_exitPrimary_expression_list[] = "exitPrimary_expression_list";
static const char __pyx_k_exitPrimary_unsigned_number[] = "exitPrimary_unsigned_number";
static const char __pyx_k_exitStored_definition_class[] = "exitStored_definition_class";
static const char __pyx_k_Component_declarationContext[] = "Component_declarationContext";
static const char __pyx_k_Element_redeclarationContext[] = "Element_redeclarationContext";
static const char __pyx_k_ModelicaListener_exitComment[] = "ModelicaListener.exitComment";
static const char __pyx_k_ModelicaListener_exitElement[] = "ModelicaListener.exitElement";
static const char __pyx_k_ModelicaListener_exitExpr_or[] = "ModelicaListener.exitExpr_or";
static const char __pyx_k_enterEquation_connect_clause[] = "enterEquation_connect_clause";
static const char __pyx_k_enterModification_assignment[] = "enterModification_assignment";
static const char __pyx_k_enterPrimary_expression_list[] = "enterPrimary_expression_list";
static const char __pyx_k_enterPrimary_unsigned_number[] = "enterPrimary_unsigned_number";
static const char __pyx_k_enterStored_definition_class[] = "enterStored_definition_class";
static const char __pyx_k_exitModification_assignment2[] = "exitModification_assignment2";
static const char __pyx_k_Component_declaration1Context[] = "Component_declaration1Context";
static const char __pyx_k_External_function_callContext[] = "External_function_callContext";
static const char __pyx_k_Language_specificationContext[] = "Language_specificationContext";
static const char __pyx_k_ModelicaListener_enterComment[] = "ModelicaListener.enterComment";
static const char __pyx_k_ModelicaListener_enterElement[] = "ModelicaListener.enterElement";
static const char __pyx_k_ModelicaListener_enterExpr_or[] = "ModelicaListener.enterExpr_or";
static const char __pyx_k_ModelicaListener_exitArgument[] = "ModelicaListener.exitArgument";
static const char __pyx_k_ModelicaListener_exitEquation[] = "ModelicaListener.exitEquation";
static const char __pyx_k_ModelicaListener_exitExpr_add[] = "ModelicaListener.exitExpr_add";
static const char __pyx_k_ModelicaListener_exitExpr_and[] = "ModelicaListener.exitExpr_and";
static const char __pyx_k_ModelicaListener_exitExpr_exp[] = "ModelicaListener.exitExpr_exp";
static const char __pyx_k_ModelicaListener_exitExpr_mul[] = "ModelicaListener.exitExpr_mul";
static const char __pyx_k_ModelicaListener_exitExpr_neg[] = "ModelicaListener.exitExpr_neg";
static const char __pyx_k_ModelicaListener_exitExpr_not[] = "ModelicaListener.exitExpr_not";
static const char __pyx_k_ModelicaListener_exitExpr_rel[] = "ModelicaListener.exitExpr_rel";
static const char __pyx_k_Output_expression_listContext[] = "Output_expression_listContext";
static const char __pyx_k_Short_class_definitionContext[] = "Short_class_definitionContext";
static const char __pyx_k_enterModification_assignment2[] = "enterModification_assignment2";
static const char __pyx_k_Equation_connect_clauseContext[] = "Equation_connect_clauseContext";
static const char __pyx_k_ModelicaListener_enterArgument[] = "ModelicaListener.enterArgument";
static const char __pyx_k_ModelicaListener_enterEquation[] = "ModelicaListener.enterEquation";
static const char __pyx_k_ModelicaListener_enterExpr_add[] = "ModelicaListener.enterExpr_add";
static const char __pyx_k_ModelicaListener_enterExpr_and[] = "ModelicaListener.enterExpr_and";
static const char __pyx_k_ModelicaListener_enterExpr_exp[] = "ModelicaListener.enterExpr_exp";
static const char __pyx_k_ModelicaListener_enterExpr_mul[] = "ModelicaListener.enterExpr_mul";
static const char __pyx_k_ModelicaListener_enterExpr_neg[] = "ModelicaListener.enterExpr_neg";
static const char __pyx_k_ModelicaListener_enterExpr_not[] = "ModelicaListener.enterExpr_not";
static const char __pyx_k_ModelicaListener_enterExpr_rel[] = "ModelicaListener.enterExpr_rel";
static const char __pyx_k_ModelicaListener_exitEnum_list[] = "ModelicaListener.exitEnum_list";
static const char __pyx_k_ModelicaListener_exitFor_index[] = "ModelicaListener.exitFor_index";
static const char __pyx_k_ModelicaListener_exitStatement[] = "ModelicaListener.exitStatement";
static const char __pyx_k_ModelicaListener_exitSubscript[] = "ModelicaListener.exitSubscript";
static const char __pyx_k_Modification_assignmentContext[] = "Modification_assignmentContext";
static const char __pyx_k_Primary_expression_listContext[] = "Primary_expression_listContext";
static const char __pyx_k_Primary_unsigned_numberContext[] = "Primary_unsigned_numberContext";
static const char __pyx_k_Stored_definition_classContext[] = "Stored_definition_classContext";
static const char __pyx_k_exitPrimary_function_arguments[] = "exitPrimary_function_arguments";
static const char __pyx_k_ModelicaListener_enterAlgorithm[] = "ModelicaListener.enterAlgorithm_section";
static const char __pyx_k_ModelicaListener_enterComponent[] = "ModelicaListener.enterComponent_clause";
static const char __pyx_k_ModelicaListener_enterCondition[] = "ModelicaListener.enterCondition_attribute";
static const char __pyx_k_ModelicaListener_enterEnum_list[] = "ModelicaListener.enterEnum_list";
static const char __pyx_k_ModelicaListener_enterFor_index[] = "ModelicaListener.enterFor_index";
static const char __pyx_k_ModelicaListener_enterStatement[] = "ModelicaListener.enterStatement_component_reference";
static const char __pyx_k_ModelicaListener_enterSubscript[] = "ModelicaListener.enterSubscript";
static const char __pyx_k_ModelicaListener_exitAnnotation[] = "ModelicaListener.exitAnnotation";
static const char __pyx_k_ModelicaListener_exitClass_spec[] = "ModelicaListener.exitClass_spec_comp";
static const char __pyx_k_ModelicaListener_exitClass_type[] = "ModelicaListener.exitClass_type";
static const char __pyx_k_ModelicaListener_exitExpression[] = "ModelicaListener.exitExpression_simple";
static const char __pyx_k_Modification_assignment2Context[] = "Modification_assignment2Context";
static const char __pyx_k_enterPrimary_function_arguments[] = "enterPrimary_function_arguments";
static const char __pyx_k_exitComponent_reference_element[] = "exitComponent_reference_element";
static const char __pyx_k_exitPrimary_component_reference[] = "exitPrimary_component_reference";
static const char __pyx_k_home_jgoppert_git_pymola_pymola[] = "/home/jgoppert/git/pymola/pymola/generated/ModelicaListener.pyx";
static const char __pyx_k_Component_reference_elementConte[] = "Component_reference_elementContext";
static const char __pyx_k_Element_modification_or_replacea[] = "Element_modification_or_replaceableContext";
static const char __pyx_k_ModelicaListener_enterAnnotation[] = "ModelicaListener.enterAnnotation";
static const char __pyx_k_ModelicaListener_enterArgument_e[] = "ModelicaListener.enterArgument_expression";
static const char __pyx_k_ModelicaListener_enterArgument_f[] = "ModelicaListener.enterArgument_function";
static const char __pyx_k_ModelicaListener_enterArgument_l[] = "ModelicaListener.enterArgument_list";
static const char __pyx_k_ModelicaListener_enterArray_subs[] = "ModelicaListener.enterArray_subscripts";
static const char __pyx_k_ModelicaListener_enterBase_prefi[] = "ModelicaListener.enterBase_prefix";
static const char __pyx_k_ModelicaListener_enterClass_defi[] = "ModelicaListener.enterClass_definition";
static const char __pyx_k_ModelicaListener_enterClass_modi[] = "ModelicaListener.enterClass_modification";
static const char __pyx_k_ModelicaListener_enterClass_pref[] = "ModelicaListener.enterClass_prefixes";
static const char __pyx_k_ModelicaListener_enterClass_spec[] = "ModelicaListener.enterClass_spec_comp";
static const char __pyx_k_ModelicaListener_enterClass_type[] = "ModelicaListener.enterClass_type";
static const char __pyx_k_ModelicaListener_enterCompositio[] = "ModelicaListener.enterComposition";
static const char __pyx_k_ModelicaListener_enterConnect_cl[] = "ModelicaListener.enterConnect_clause";
static const char __pyx_k_ModelicaListener_enterConstraini[] = "ModelicaListener.enterConstraining_clause";
static const char __pyx_k_ModelicaListener_enterDeclaratio[] = "ModelicaListener.enterDeclaration";
static const char __pyx_k_ModelicaListener_enterElement_li[] = "ModelicaListener.enterElement_list";
static const char __pyx_k_ModelicaListener_enterElement_mo[] = "ModelicaListener.enterElement_modification_or_replaceable";
static const char __pyx_k_ModelicaListener_enterElement_re[] = "ModelicaListener.enterElement_redeclaration";
static const char __pyx_k_ModelicaListener_enterEnumeratio[] = "ModelicaListener.enterEnumeration_literal";
static const char __pyx_k_ModelicaListener_enterEquation_c[] = "ModelicaListener.enterEquation_connect_clause";
static const char __pyx_k_ModelicaListener_enterEquation_f[] = "ModelicaListener.enterEquation_for";
static const char __pyx_k_ModelicaListener_enterEquation_i[] = "ModelicaListener.enterEquation_if";
static const char __pyx_k_ModelicaListener_enterEquation_s[] = "ModelicaListener.enterEquation_section";
static const char __pyx_k_ModelicaListener_enterEquation_w[] = "ModelicaListener.enterEquation_when";
static const char __pyx_k_ModelicaListener_enterExpr_prima[] = "ModelicaListener.enterExpr_primary";
static const char __pyx_k_ModelicaListener_enterExpression[] = "ModelicaListener.enterExpression_simple";
static const char __pyx_k_ModelicaListener_enterExtends_cl[] = "ModelicaListener.enterExtends_clause";
static const char __pyx_k_ModelicaListener_enterExternal_f[] = "ModelicaListener.enterExternal_function_call";
static const char __pyx_k_ModelicaListener_enterFor_equati[] = "ModelicaListener.enterFor_equation";
static const char __pyx_k_ModelicaListener_enterFor_indice[] = "ModelicaListener.enterFor_indices";
static const char __pyx_k_ModelicaListener_enterFor_statem[] = "ModelicaListener.enterFor_statement";
static const char __pyx_k_ModelicaListener_enterFunction_a[] = "ModelicaListener.enterFunction_arguments";
static const char __pyx_k_ModelicaListener_enterFunction_c[] = "ModelicaListener.enterFunction_call_args";
static const char __pyx_k_ModelicaListener_enterIf_equatio[] = "ModelicaListener.enterIf_equation";
static const char __pyx_k_ModelicaListener_enterIf_stateme[] = "ModelicaListener.enterIf_statement";
static const char __pyx_k_ModelicaListener_enterImport_cla[] = "ModelicaListener.enterImport_clause";
static const char __pyx_k_ModelicaListener_enterImport_lis[] = "ModelicaListener.enterImport_list";
static const char __pyx_k_ModelicaListener_enterLanguage_s[] = "ModelicaListener.enterLanguage_specification";
static const char __pyx_k_ModelicaListener_enterModificati[] = "ModelicaListener.enterModification_class";
static const char __pyx_k_ModelicaListener_enterNamed_argu[] = "ModelicaListener.enterNamed_arguments";
static const char __pyx_k_ModelicaListener_enterOutput_exp[] = "ModelicaListener.enterOutput_expression_list";
static const char __pyx_k_ModelicaListener_enterPrimary_co[] = "ModelicaListener.enterPrimary_component_reference";
static const char __pyx_k_ModelicaListener_enterPrimary_de[] = "ModelicaListener.enterPrimary_derivative";
static const char __pyx_k_ModelicaListener_enterPrimary_en[] = "ModelicaListener.enterPrimary_end";
static const char __pyx_k_ModelicaListener_enterPrimary_ex[] = "ModelicaListener.enterPrimary_expression_list";
static const char __pyx_k_ModelicaListener_enterPrimary_fa[] = "ModelicaListener.enterPrimary_false";
static const char __pyx_k_ModelicaListener_enterPrimary_fu[] = "ModelicaListener.enterPrimary_function";
static const char __pyx_k_ModelicaListener_enterPrimary_in[] = "ModelicaListener.enterPrimary_initial";
static const char __pyx_k_ModelicaListener_enterPrimary_ou[] = "ModelicaListener.enterPrimary_output_expression_list";
static const char __pyx_k_ModelicaListener_enterPrimary_st[] = "ModelicaListener.enterPrimary_string";
static const char __pyx_k_ModelicaListener_enterPrimary_tr[] = "ModelicaListener.enterPrimary_true";
static const char __pyx_k_ModelicaListener_enterPrimary_un[] = "ModelicaListener.enterPrimary_unsigned_number";
static const char __pyx_k_ModelicaListener_enterRegular_el[] = "ModelicaListener.enterRegular_element";
static const char __pyx_k_ModelicaListener_enterReplaceabl[] = "ModelicaListener.enterReplaceable_element";
static const char __pyx_k_ModelicaListener_enterShort_clas[] = "ModelicaListener.enterShort_class_definition";
static const char __pyx_k_ModelicaListener_enterSimple_exp[] = "ModelicaListener.enterSimple_expression";
static const char __pyx_k_ModelicaListener_enterStored_def[] = "ModelicaListener.enterStored_definition";
static const char __pyx_k_ModelicaListener_enterString_com[] = "ModelicaListener.enterString_comment";
static const char __pyx_k_ModelicaListener_enterType_prefi[] = "ModelicaListener.enterType_prefix";
static const char __pyx_k_ModelicaListener_enterType_speci[] = "ModelicaListener.enterType_specifier";
static const char __pyx_k_ModelicaListener_enterWhen_equat[] = "ModelicaListener.enterWhen_equation";
static const char __pyx_k_ModelicaListener_enterWhen_state[] = "ModelicaListener.enterWhen_statement";
static const char __pyx_k_ModelicaListener_enterWhile_stat[] = "ModelicaListener.enterWhile_statement";
static const char __pyx_k_ModelicaListener_exitAlgorithm_s[] = "ModelicaListener.exitAlgorithm_section";
static const char __pyx_k_ModelicaListener_exitArgument_ex[] = "ModelicaListener.exitArgument_expression";
static const char __pyx_k_ModelicaListener_exitArgument_fu[] = "ModelicaListener.exitArgument_function";
static const char __pyx_k_ModelicaListener_exitArgument_li[] = "ModelicaListener.exitArgument_list";
static const char __pyx_k_ModelicaListener_exitArray_subsc[] = "ModelicaListener.exitArray_subscripts";
static const char __pyx_k_ModelicaListener_exitBase_prefix[] = "ModelicaListener.exitBase_prefix";
static const char __pyx_k_ModelicaListener_exitClass_defin[] = "ModelicaListener.exitClass_definition";
static const char __pyx_k_ModelicaListener_exitClass_modif[] = "ModelicaListener.exitClass_modification";
static const char __pyx_k_ModelicaListener_exitClass_prefi[] = "ModelicaListener.exitClass_prefixes";
static const char __pyx_k_ModelicaListener_exitComponent_c[] = "ModelicaListener.exitComponent_clause";
static const char __pyx_k_ModelicaListener_exitComponent_d[] = "ModelicaListener.exitComponent_declaration";
static const char __pyx_k_ModelicaListener_exitComponent_l[] = "ModelicaListener.exitComponent_list";
static const char __pyx_k_ModelicaListener_exitComponent_r[] = "ModelicaListener.exitComponent_reference_element";
static const char __pyx_k_ModelicaListener_exitComposition[] = "ModelicaListener.exitComposition";
static const char __pyx_k_ModelicaListener_exitCondition_a[] = "ModelicaListener.exitCondition_attribute";
static const char __pyx_k_ModelicaListener_exitConnect_cla[] = "ModelicaListener.exitConnect_clause";
static const char __pyx_k_ModelicaListener_exitConstrainin[] = "ModelicaListener.exitConstraining_clause";
static const char __pyx_k_ModelicaListener_exitDeclaration[] = "ModelicaListener.exitDeclaration";
static const char __pyx_k_ModelicaListener_exitElement_lis[] = "ModelicaListener.exitElement_list";
static const char __pyx_k_ModelicaListener_exitElement_mod[] = "ModelicaListener.exitElement_modification_or_replaceable";
static const char __pyx_k_ModelicaListener_exitElement_red[] = "ModelicaListener.exitElement_redeclaration";
static const char __pyx_k_ModelicaListener_exitElement_rep[] = "ModelicaListener.exitElement_replaceable";
static const char __pyx_k_ModelicaListener_exitEnumeration[] = "ModelicaListener.exitEnumeration_literal";
static const char __pyx_k_ModelicaListener_exitEquation_co[] = "ModelicaListener.exitEquation_connect_clause";
static const char __pyx_k_ModelicaListener_exitEquation_fo[] = "ModelicaListener.exitEquation_for";
static const char __pyx_k_ModelicaListener_exitEquation_fu[] = "ModelicaListener.exitEquation_function";
static const char __pyx_k_ModelicaListener_exitEquation_if[] = "ModelicaListener.exitEquation_if";
static const char __pyx_k_ModelicaListener_exitEquation_se[] = "ModelicaListener.exitEquation_section";
static const char __pyx_k_ModelicaListener_exitEquation_si[] = "ModelicaListener.exitEquation_simple";
static const char __pyx_k_ModelicaListener_exitEquation_wh[] = "ModelicaListener.exitEquation_when";
static const char __pyx_k_ModelicaListener_exitExpr_primar[] = "ModelicaListener.exitExpr_primary";
static const char __pyx_k_ModelicaListener_exitExtends_cla[] = "ModelicaListener.exitExtends_clause";
static const char __pyx_k_ModelicaListener_exitExternal_fu[] = "ModelicaListener.exitExternal_function_call";
static const char __pyx_k_ModelicaListener_exitFor_equatio[] = "ModelicaListener.exitFor_equation";
static const char __pyx_k_ModelicaListener_exitFor_indices[] = "ModelicaListener.exitFor_indices";
static const char __pyx_k_ModelicaListener_exitFor_stateme[] = "ModelicaListener.exitFor_statement";
static const char __pyx_k_ModelicaListener_exitFunction_ar[] = "ModelicaListener.exitFunction_arguments";
static const char __pyx_k_ModelicaListener_exitFunction_ca[] = "ModelicaListener.exitFunction_call_args";
static const char __pyx_k_ModelicaListener_exitIf_equation[] = "ModelicaListener.exitIf_equation";
static const char __pyx_k_ModelicaListener_exitIf_statemen[] = "ModelicaListener.exitIf_statement";
static const char __pyx_k_ModelicaListener_exitImport_clau[] = "ModelicaListener.exitImport_clause";
static const char __pyx_k_ModelicaListener_exitImport_list[] = "ModelicaListener.exitImport_list";
static const char __pyx_k_ModelicaListener_exitLanguage_sp[] = "ModelicaListener.exitLanguage_specification";
static const char __pyx_k_ModelicaListener_exitModificatio[] = "ModelicaListener.exitModification_class";
static const char __pyx_k_ModelicaListener_exitNamed_argum[] = "ModelicaListener.exitNamed_arguments";
static const char __pyx_k_ModelicaListener_exitOutput_expr[] = "ModelicaListener.exitOutput_expression_list";
static const char __pyx_k_ModelicaListener_exitPrimary_com[] = "ModelicaListener.exitPrimary_component_reference";
static const char __pyx_k_ModelicaListener_exitPrimary_der[] = "ModelicaListener.exitPrimary_derivative";
static const char __pyx_k_ModelicaListener_exitPrimary_end[] = "ModelicaListener.exitPrimary_end";
static const char __pyx_k_ModelicaListener_exitPrimary_exp[] = "ModelicaListener.exitPrimary_expression_list";
static const char __pyx_k_ModelicaListener_exitPrimary_fal[] = "ModelicaListener.exitPrimary_false";
static const char __pyx_k_ModelicaListener_exitPrimary_fun[] = "ModelicaListener.exitPrimary_function";
static const char __pyx_k_ModelicaListener_exitPrimary_ini[] = "ModelicaListener.exitPrimary_initial";
static const char __pyx_k_ModelicaListener_exitPrimary_out[] = "ModelicaListener.exitPrimary_output_expression_list";
static const char __pyx_k_ModelicaListener_exitPrimary_str[] = "ModelicaListener.exitPrimary_string";
static const char __pyx_k_ModelicaListener_exitPrimary_tru[] = "ModelicaListener.exitPrimary_true";
static const char __pyx_k_ModelicaListener_exitPrimary_uns[] = "ModelicaListener.exitPrimary_unsigned_number";
static const char __pyx_k_ModelicaListener_exitRegular_ele[] = "ModelicaListener.exitRegular_element";
static const char __pyx_k_ModelicaListener_exitReplaceable[] = "ModelicaListener.exitReplaceable_element";
static const char __pyx_k_ModelicaListener_exitShort_class[] = "ModelicaListener.exitShort_class_definition";
static const char __pyx_k_ModelicaListener_exitSimple_expr[] = "ModelicaListener.exitSimple_expression";
static const char __pyx_k_ModelicaListener_exitStatement_b[] = "ModelicaListener.exitStatement_break";
static const char __pyx_k_ModelicaListener_exitStatement_c[] = "ModelicaListener.exitStatement_component_reference";
static const char __pyx_k_ModelicaListener_exitStatement_f[] = "ModelicaListener.exitStatement_for";
static const char __pyx_k_ModelicaListener_exitStatement_i[] = "ModelicaListener.exitStatement_if";
static const char __pyx_k_ModelicaListener_exitStatement_r[] = "ModelicaListener.exitStatement_return";
static const char __pyx_k_ModelicaListener_exitStatement_w[] = "ModelicaListener.exitStatement_while";
static const char __pyx_k_ModelicaListener_exitStored_defi[] = "ModelicaListener.exitStored_definition";
static const char __pyx_k_ModelicaListener_exitString_comm[] = "ModelicaListener.exitString_comment";
static const char __pyx_k_ModelicaListener_exitType_prefix[] = "ModelicaListener.exitType_prefix";
static const char __pyx_k_ModelicaListener_exitType_specif[] = "ModelicaListener.exitType_specifier";
static const char __pyx_k_ModelicaListener_exitWhen_equati[] = "ModelicaListener.exitWhen_equation";
static const char __pyx_k_ModelicaListener_exitWhen_statem[] = "ModelicaListener.exitWhen_statement";
static const char __pyx_k_ModelicaListener_exitWhile_state[] = "ModelicaListener.exitWhile_statement";
static const char __pyx_k_Primary_component_referenceConte[] = "Primary_component_referenceContext";
static const char __pyx_k_Primary_function_argumentsContex[] = "Primary_function_argumentsContext";
static const char __pyx_k_Primary_output_expression_listCo[] = "Primary_output_expression_listContext";
static const char __pyx_k_Statement_component_functionCont[] = "Statement_component_functionContext";
static const char __pyx_k_Statement_component_referenceCon[] = "Statement_component_referenceContext";
static const char __pyx_k_enterComponent_reference_element[] = "enterComponent_reference_element";
static const char __pyx_k_enterElement_modification_or_rep[] = "enterElement_modification_or_replaceable";
static const char __pyx_k_enterPrimary_component_reference[] = "enterPrimary_component_reference";
static const char __pyx_k_enterPrimary_output_expression_l[] = "enterPrimary_output_expression_list";
static const char __pyx_k_enterStatement_component_functio[] = "enterStatement_component_function";
static const char __pyx_k_enterStatement_component_referen[] = "enterStatement_component_reference";
static const char __pyx_k_exitElement_modification_or_repl[] = "exitElement_modification_or_replaceable";
static const char __pyx_k_exitPrimary_output_expression_li[] = "exitPrimary_output_expression_list";
static const char __pyx_k_exitStatement_component_function[] = "exitStatement_component_function";
static const char __pyx_k_exitStatement_component_referenc[] = "exitStatement_component_reference";
static const char __pyx_k_pymola_generated_ModelicaListene[] = "pymola.generated.ModelicaListener";
static const char __pyx_k_ModelicaListener_enterComponent_2[] = "ModelicaListener.enterComponent_list";
static const char __pyx_k_ModelicaListener_enterComponent_3[] = "ModelicaListener.enterComponent_declaration";
static const char __pyx_k_ModelicaListener_enterComponent_4[] = "ModelicaListener.enterComponent_clause1";
static const char __pyx_k_ModelicaListener_enterComponent_5[] = "ModelicaListener.enterComponent_declaration1";
static const char __pyx_k_ModelicaListener_enterComponent_6[] = "ModelicaListener.enterComponent_reference_element";
static const char __pyx_k_ModelicaListener_enterComponent_7[] = "ModelicaListener.enterComponent_reference";
static const char __pyx_k_ModelicaListener_enterStatement_2[] = "ModelicaListener.enterStatement_component_function";
static const char __pyx_k_ModelicaListener_enterStatement_3[] = "ModelicaListener.enterStatement_break";
static const char __pyx_k_ModelicaListener_enterStatement_4[] = "ModelicaListener.enterStatement_return";
static const char __pyx_k_ModelicaListener_enterStatement_5[] = "ModelicaListener.enterStatement_if";
static const char __pyx_k_ModelicaListener_enterStatement_6[] = "ModelicaListener.enterStatement_for";
static const char __pyx_k_ModelicaListener_enterStatement_7[] = "ModelicaListener.enterStatement_while";
static const char __pyx_k_ModelicaListener_enterStatement_8[] = "ModelicaListener.enterStatement_when";
static const char __pyx_k_ModelicaListener_enterStatement_9[] = "ModelicaListener.enterStatement";
static const char __pyx_k_ModelicaListener_exitClass_spec_2[] = "ModelicaListener.exitClass_spec_base";
static const char __pyx_k_ModelicaListener_exitClass_spec_3[] = "ModelicaListener.exitClass_spec_enum";
static const char __pyx_k_ModelicaListener_exitClass_spec_4[] = "ModelicaListener.exitClass_spec_der";
static const char __pyx_k_ModelicaListener_exitClass_spec_5[] = "ModelicaListener.exitClass_spec_extends";
static const char __pyx_k_ModelicaListener_exitExpression_2[] = "ModelicaListener.exitExpression_if";
static const char __pyx_k_ModelicaListener_exitExpression_3[] = "ModelicaListener.exitExpression_list";
static const char __pyx_k_ModelicaListener_enterClass_spec_2[] = "ModelicaListener.enterClass_spec_base";
static const char __pyx_k_ModelicaListener_enterClass_spec_3[] = "ModelicaListener.enterClass_spec_enum";
static const char __pyx_k_ModelicaListener_enterClass_spec_4[] = "ModelicaListener.enterClass_spec_der";
static const char __pyx_k_ModelicaListener_enterClass_spec_5[] = "ModelicaListener.enterClass_spec_extends";
static const char __pyx_k_ModelicaListener_enterElement_mo_2[] = "ModelicaListener.enterElement_modification";
static const char __pyx_k_ModelicaListener_enterElement_re_2[] = "ModelicaListener.enterElement_replaceable";
static const char __pyx_k_ModelicaListener_enterEquation_f_2[] = "ModelicaListener.enterEquation_function";
static const char __pyx_k_ModelicaListener_enterEquation_s_2[] = "ModelicaListener.enterEquation_simple";
static const char __pyx_k_ModelicaListener_enterExpression_2[] = "ModelicaListener.enterExpression_if";
static const char __pyx_k_ModelicaListener_enterExpression_3[] = "ModelicaListener.enterExpression_list";
static const char __pyx_k_ModelicaListener_enterModificati_2[] = "ModelicaListener.enterModification_assignment";
static const char __pyx_k_ModelicaListener_enterModificati_3[] = "ModelicaListener.enterModification_assignment2";
static const char __pyx_k_ModelicaListener_enterNamed_argu_2[] = "ModelicaListener.enterNamed_argument";
static const char __pyx_k_ModelicaListener_enterPrimary_fu_2[] = "ModelicaListener.enterPrimary_function_arguments";
static const char __pyx_k_ModelicaListener_enterStored_def_2[] = "ModelicaListener.enterStored_definition_class";
static const char __pyx_k_ModelicaListener_exitComponent_c_2[] = "ModelicaListener.exitComponent_clause1";
static const char __pyx_k_ModelicaListener_exitComponent_d_2[] = "ModelicaListener.exitComponent_declaration1";
static const char __pyx_k_ModelicaListener_exitComponent_r_2[] = "ModelicaListener.exitComponent_reference";
static const char __pyx_k_ModelicaListener_exitElement_mod_2[] = "ModelicaListener.exitElement_modification";
static const char __pyx_k_ModelicaListener_exitModificatio_2[] = "ModelicaListener.exitModification_assignment";
static const char __pyx_k_ModelicaListener_exitModificatio_3[] = "ModelicaListener.exitModification_assignment2";
static const char __pyx_k_ModelicaListener_exitNamed_argum_2[] = "ModelicaListener.exitNamed_argument";
static const char __pyx_k_ModelicaListener_exitPrimary_fun_2[] = "ModelicaListener.exitPrimary_function_arguments";
static const char __pyx_k_ModelicaListener_exitStatement_c_2[] = "ModelicaListener.exitStatement_component_function";
static const char __pyx_k_ModelicaListener_exitStatement_w_2[] = "ModelicaListener.exitStatement_when";
static const char __pyx_k_ModelicaListener_exitStored_defi_2[] = "ModelicaListener.exitStored_definition_class";
static PyObject *__pyx_n_s_;
static PyObject *__pyx_n_s_Algorithm_sectionContext;
static PyObject *__pyx_n_s_AnnotationContext;
static PyObject *__pyx_n_s_ArgumentContext;
static PyObject *__pyx_n_s_Argument_expressionContext;
static PyObject *__pyx_n_s_Argument_functionContext;
static PyObject *__pyx_n_s_Argument_listContext;
static PyObject *__pyx_n_s_Array_subscriptsContext;
static PyObject *__pyx_n_s_Base_prefixContext;
static PyObject *__pyx_n_s_Class_definitionContext;
static PyObject *__pyx_n_s_Class_modificationContext;
static PyObject *__pyx_n_s_Class_prefixesContext;
static PyObject *__pyx_n_s_Class_spec_baseContext;
static PyObject *__pyx_n_s_Class_spec_compContext;
static PyObject *__pyx_n_s_Class_spec_derContext;
static PyObject *__pyx_n_s_Class_spec_enumContext;
static PyObject *__pyx_n_s_Class_spec_extendsContext;
static PyObject *__pyx_n_s_Class_typeContext;
static PyObject *__pyx_n_s_CommentContext;
static PyObject *__pyx_n_s_Component_clause1Context;
static PyObject *__pyx_n_s_Component_clauseContext;
static PyObject *__pyx_n_s_Component_declaration1Context;
static PyObject *__pyx_n_s_Component_declarationContext;
static PyObject *__pyx_n_s_Component_listContext;
static PyObject *__pyx_n_s_Component_referenceContext;
static PyObject *__pyx_n_s_Component_reference_elementConte;
static PyObject *__pyx_n_s_CompositionContext;
static PyObject *__pyx_n_s_Condition_attributeContext;
static PyObject *__pyx_n_s_Connect_clauseContext;
static PyObject *__pyx_n_s_Constraining_clauseContext;
static PyObject *__pyx_n_s_DeclarationContext;
static PyObject *__pyx_n_s_ElementContext;
static PyObject *__pyx_n_s_Element_listContext;
static PyObject *__pyx_n_s_Element_modificationContext;
static PyObject *__pyx_n_s_Element_modification_or_replacea;
static PyObject *__pyx_n_s_Element_redeclarationContext;
static PyObject *__pyx_n_s_Element_replaceableContext;
static PyObject *__pyx_n_s_Enum_listContext;
static PyObject *__pyx_n_s_Enumeration_literalContext;
static PyObject *__pyx_n_s_EquationContext;
static PyObject *__pyx_n_s_Equation_connect_clauseContext;
static PyObject *__pyx_n_s_Equation_forContext;
static PyObject *__pyx_n_s_Equation_functionContext;
static PyObject *__pyx_n_s_Equation_ifContext;
static PyObject *__pyx_n_s_Equation_sectionContext;
static PyObject *__pyx_n_s_Equation_simpleContext;
static PyObject *__pyx_n_s_Equation_whenContext;
static PyObject *__pyx_n_s_Expr_addContext;
static PyObject *__pyx_n_s_Expr_andContext;
static PyObject *__pyx_n_s_Expr_expContext;
static PyObject *__pyx_n_s_Expr_mulContext;
static PyObject *__pyx_n_s_Expr_negContext;
static PyObject *__pyx_n_s_Expr_notContext;
static PyObject *__pyx_n_s_Expr_orContext;
static PyObject *__pyx_n_s_Expr_primaryContext;
static PyObject *__pyx_n_s_Expr_relContext;
static PyObject *__pyx_n_s_Expression_ifContext;
static PyObject *__pyx_n_s_Expression_listContext;
static PyObject *__pyx_n_s_Expression_simpleContext;
static PyObject *__pyx_n_s_Extends_clauseContext;
static PyObject *__pyx_n_s_External_function_callContext;
static PyObject *__pyx_n_s_For_equationContext;
static PyObject *__pyx_n_s_For_indexContext;
static PyObject *__pyx_n_s_For_indicesContext;
static PyObject *__pyx_n_s_For_statementContext;
static PyObject *__pyx_n_s_Function_argumentsContext;
static PyObject *__pyx_n_s_Function_call_argsContext;
static PyObject *__pyx_n_s_If_equationContext;
static PyObject *__pyx_n_s_If_statementContext;
static PyObject *__pyx_n_s_Import_clauseContext;
static PyObject *__pyx_n_s_Import_listContext;
static PyObject *__pyx_n_s_Language_specificationContext;
static PyObject *__pyx_n_s_ModelicaListener;
static PyObject *__pyx_n_s_ModelicaListener_enterAlgorithm;
static PyObject *__pyx_n_s_ModelicaListener_enterAnnotation;
static PyObject *__pyx_n_s_ModelicaListener_enterArgument;
static PyObject *__pyx_n_s_ModelicaListener_enterArgument_e;
static PyObject *__pyx_n_s_ModelicaListener_enterArgument_f;
static PyObject *__pyx_n_s_ModelicaListener_enterArgument_l;
static PyObject *__pyx_n_s_ModelicaListener_enterArray_subs;
static PyObject *__pyx_n_s_ModelicaListener_enterBase_prefi;
static PyObject *__pyx_n_s_ModelicaListener_enterClass_defi;
static PyObject *__pyx_n_s_ModelicaListener_enterClass_modi;
static PyObject *__pyx_n_s_ModelicaListener_enterClass_pref;
static PyObject *__pyx_n_s_ModelicaListener_enterClass_spec;
static PyObject *__pyx_n_s_ModelicaListener_enterClass_spec_2;
static PyObject *__pyx_n_s_ModelicaListener_enterClass_spec_3;
static PyObject *__pyx_n_s_ModelicaListener_enterClass_spec_4;
static PyObject *__pyx_n_s_ModelicaListener_enterClass_spec_5;
static PyObject *__pyx_n_s_ModelicaListener_enterClass_type;
static PyObject *__pyx_n_s_ModelicaListener_enterComment;
static PyObject *__pyx_n_s_ModelicaListener_enterComponent;
static PyObject *__pyx_n_s_ModelicaListener_enterComponent_2;
static PyObject *__pyx_n_s_ModelicaListener_enterComponent_3;
static PyObject *__pyx_n_s_ModelicaListener_enterComponent_4;
static PyObject *__pyx_n_s_ModelicaListener_enterComponent_5;
static PyObject *__pyx_n_s_ModelicaListener_enterComponent_6;
static PyObject *__pyx_n_s_ModelicaListener_enterComponent_7;
static PyObject *__pyx_n_s_ModelicaListener_enterCompositio;
static PyObject *__pyx_n_s_ModelicaListener_enterCondition;
static PyObject *__pyx_n_s_ModelicaListener_enterConnect_cl;
static PyObject *__pyx_n_s_ModelicaListener_enterConstraini;
static PyObject *__pyx_n_s_ModelicaListener_enterDeclaratio;
static PyObject *__pyx_n_s_ModelicaListener_enterElement;
static PyObject *__pyx_n_s_ModelicaListener_enterElement_li;
static PyObject *__pyx_n_s_ModelicaListener_enterElement_mo;
static PyObject *__pyx_n_s_ModelicaListener_enterElement_mo_2;
static PyObject *__pyx_n_s_ModelicaListener_enterElement_re;
static PyObject *__pyx_n_s_ModelicaListener_enterElement_re_2;
static PyObject *__pyx_n_s_ModelicaListener_enterEnum_list;
static PyObject *__pyx_n_s_ModelicaListener_enterEnumeratio;
static PyObject *__pyx_n_s_ModelicaListener_enterEquation;
static PyObject *__pyx_n_s_ModelicaListener_enterEquation_c;
static PyObject *__pyx_n_s_ModelicaListener_enterEquation_f;
static PyObject *__pyx_n_s_ModelicaListener_enterEquation_f_2;
static PyObject *__pyx_n_s_ModelicaListener_enterEquation_i;
static PyObject *__pyx_n_s_ModelicaListener_enterEquation_s;
static PyObject *__pyx_n_s_ModelicaListener_enterEquation_s_2;
static PyObject *__pyx_n_s_ModelicaListener_enterEquation_w;
static PyObject *__pyx_n_s_ModelicaListener_enterExpr_add;
static PyObject *__pyx_n_s_ModelicaListener_enterExpr_and;
static PyObject *__pyx_n_s_ModelicaListener_enterExpr_exp;
static PyObject *__pyx_n_s_ModelicaListener_enterExpr_mul;
static PyObject *__pyx_n_s_ModelicaListener_enterExpr_neg;
static PyObject *__pyx_n_s_ModelicaListener_enterExpr_not;
static PyObject *__pyx_n_s_ModelicaListener_enterExpr_or;
static PyObject *__pyx_n_s_ModelicaListener_enterExpr_prima;
static PyObject *__pyx_n_s_ModelicaListener_enterExpr_rel;
static PyObject *__pyx_n_s_ModelicaListener_enterExpression;
static PyObject *__pyx_n_s_ModelicaListener_enterExpression_2;
static PyObject *__pyx_n_s_ModelicaListener_enterExpression_3;
static PyObject *__pyx_n_s_ModelicaListener_enterExtends_cl;
static PyObject *__pyx_n_s_ModelicaListener_enterExternal_f;
static PyObject *__pyx_n_s_ModelicaListener_enterFor_equati;
static PyObject *__pyx_n_s_ModelicaListener_enterFor_index;
static PyObject *__pyx_n_s_ModelicaListener_enterFor_indice;
static PyObject *__pyx_n_s_ModelicaListener_enterFor_statem;
static PyObject *__pyx_n_s_ModelicaListener_enterFunction_a;
static PyObject *__pyx_n_s_ModelicaListener_enterFunction_c;
static PyObject *__pyx_n_s_ModelicaListener_enterIf_equatio;
static PyObject *__pyx_n_s_ModelicaListener_enterIf_stateme;
static PyObject *__pyx_n_s_ModelicaListener_enterImport_cla;
static PyObject *__pyx_n_s_ModelicaListener_enterImport_lis;
static PyObject *__pyx_n_s_ModelicaListener_enterLanguage_s;
static PyObject *__pyx_n_s_ModelicaListener_enterModificati;
static PyObject *__pyx_n_s_ModelicaListener_enterModificati_2;
static PyObject *__pyx_n_s_ModelicaListener_enterModificati_3;
static PyObject *__pyx_n_s_ModelicaListener_enterName;
static PyObject *__pyx_n_s_ModelicaListener_enterNamed_argu;
static PyObject *__pyx_n_s_ModelicaListener_enterNamed_argu_2;
static PyObject *__pyx_n_s_ModelicaListener_enterOutput_exp;
static PyObject *__pyx_n_s_ModelicaListener_enterPrimary_co;
static PyObject *__pyx_n_s_ModelicaListener_enterPrimary_de;
static PyObject *__pyx_n_s_ModelicaListener_enterPrimary_en;
static PyObject *__pyx_n_s_ModelicaListener_enterPrimary_ex;
static PyObject *__pyx_n_s_ModelicaListener_enterPrimary_fa;
static PyObject *__pyx_n_s_ModelicaListener_enterPrimary_fu;
static PyObject *__pyx_n_s_ModelicaListener_enterPrimary_fu_2;
static PyObject *__pyx_n_s_ModelicaListener_enterPrimary_in;
static PyObject *__pyx_n_s_ModelicaListener_enterPrimary_ou;
static PyObject *__pyx_n_s_ModelicaListener_enterPrimary_st;
static PyObject *__pyx_n_s_ModelicaListener_enterPrimary_tr;
static PyObject *__pyx_n_s_ModelicaListener_enterPrimary_un;
static PyObject *__pyx_n_s_ModelicaListener_enterRegular_el;
static PyObject *__pyx_n_s_ModelicaListener_enterReplaceabl;
static PyObject *__pyx_n_s_ModelicaListener_enterShort_clas;
static PyObject *__pyx_n_s_ModelicaListener_enterSimple_exp;
static PyObject *__pyx_n_s_ModelicaListener_enterStatement;
static PyObject *__pyx_n_s_ModelicaListener_enterStatement_2;
static PyObject *__pyx_n_s_ModelicaListener_enterStatement_3;
static PyObject *__pyx_n_s_ModelicaListener_enterStatement_4;
static PyObject *__pyx_n_s_ModelicaListener_enterStatement_5;
static PyObject *__pyx_n_s_ModelicaListener_enterStatement_6;
static PyObject *__pyx_n_s_ModelicaListener_enterStatement_7;
static PyObject *__pyx_n_s_ModelicaListener_enterStatement_8;
static PyObject *__pyx_n_s_ModelicaListener_enterStatement_9;
static PyObject *__pyx_n_s_ModelicaListener_enterStored_def;
static PyObject *__pyx_n_s_ModelicaListener_enterStored_def_2;
static PyObject *__pyx_n_s_ModelicaListener_enterString_com;
static PyObject *__pyx_n_s_ModelicaListener_enterSubscript;
static PyObject *__pyx_n_s_ModelicaListener_enterType_prefi;
static PyObject *__pyx_n_s_ModelicaListener_enterType_speci;
static PyObject *__pyx_n_s_ModelicaListener_enterWhen_equat;
static PyObject *__pyx_n_s_ModelicaListener_enterWhen_state;
static PyObject *__pyx_n_s_ModelicaListener_enterWhile_stat;
static PyObject *__pyx_n_s_ModelicaListener_exitAlgorithm_s;
static PyObject *__pyx_n_s_ModelicaListener_exitAnnotation;
static PyObject *__pyx_n_s_ModelicaListener_exitArgument;
static PyObject *__pyx_n_s_ModelicaListener_exitArgument_ex;
static PyObject *__pyx_n_s_ModelicaListener_exitArgument_fu;
static PyObject *__pyx_n_s_ModelicaListener_exitArgument_li;
static PyObject *__pyx_n_s_ModelicaListener_exitArray_subsc;
static PyObject *__pyx_n_s_ModelicaListener_exitBase_prefix;
static PyObject *__pyx_n_s_ModelicaListener_exitClass_defin;
static PyObject *__pyx_n_s_ModelicaListener_exitClass_modif;
static PyObject *__pyx_n_s_ModelicaListener_exitClass_prefi;
static PyObject *__pyx_n_s_ModelicaListener_exitClass_spec;
static PyObject *__pyx_n_s_ModelicaListener_exitClass_spec_2;
static PyObject *__pyx_n_s_ModelicaListener_exitClass_spec_3;
static PyObject *__pyx_n_s_ModelicaListener_exitClass_spec_4;
static PyObject *__pyx_n_s_ModelicaListener_exitClass_spec_5;
static PyObject *__pyx_n_s_ModelicaListener_exitClass_type;
static PyObject *__pyx_n_s_ModelicaListener_exitComment;
static PyObject *__pyx_n_s_ModelicaListener_exitComponent_c;
static PyObject *__pyx_n_s_ModelicaListener_exitComponent_c_2;
static PyObject *__pyx_n_s_ModelicaListener_exitComponent_d;
static PyObject *__pyx_n_s_ModelicaListener_exitComponent_d_2;
static PyObject *__pyx_n_s_ModelicaListener_exitComponent_l;
static PyObject *__pyx_n_s_ModelicaListener_exitComponent_r;
static PyObject *__pyx_n_s_ModelicaListener_exitComponent_r_2;
static PyObject *__pyx_n_s_ModelicaListener_exitComposition;
static PyObject *__pyx_n_s_ModelicaListener_exitCondition_a;
static PyObject *__pyx_n_s_ModelicaListener_exitConnect_cla;
static PyObject *__pyx_n_s_ModelicaListener_exitConstrainin;
static PyObject *__pyx_n_s_ModelicaListener_exitDeclaration;
static PyObject *__pyx_n_s_ModelicaListener_exitElement;
static PyObject *__pyx_n_s_ModelicaListener_exitElement_lis;
static PyObject *__pyx_n_s_ModelicaListener_exitElement_mod;
static PyObject *__pyx_n_s_ModelicaListener_exitElement_mod_2;
static PyObject *__pyx_n_s_ModelicaListener_exitElement_red;
static PyObject *__pyx_n_s_ModelicaListener_exitElement_rep;
static PyObject *__pyx_n_s_ModelicaListener_exitEnum_list;
static PyObject *__pyx_n_s_ModelicaListener_exitEnumeration;
static PyObject *__pyx_n_s_ModelicaListener_exitEquation;
static PyObject *__pyx_n_s_ModelicaListener_exitEquation_co;
static PyObject *__pyx_n_s_ModelicaListener_exitEquation_fo;
static PyObject *__pyx_n_s_ModelicaListener_exitEquation_fu;
static PyObject *__pyx_n_s_ModelicaListener_exitEquation_if;
static PyObject *__pyx_n_s_ModelicaListener_exitEquation_se;
static PyObject *__pyx_n_s_ModelicaListener_exitEquation_si;
static PyObject *__pyx_n_s_ModelicaListener_exitEquation_wh;
static PyObject *__pyx_n_s_ModelicaListener_exitExpr_add;
static PyObject *__pyx_n_s_ModelicaListener_exitExpr_and;
static PyObject *__pyx_n_s_ModelicaListener_exitExpr_exp;
static PyObject *__pyx_n_s_ModelicaListener_exitExpr_mul;
static PyObject *__pyx_n_s_ModelicaListener_exitExpr_neg;
static PyObject *__pyx_n_s_ModelicaListener_exitExpr_not;
static PyObject *__pyx_n_s_ModelicaListener_exitExpr_or;
static PyObject *__pyx_n_s_ModelicaListener_exitExpr_primar;
static PyObject *__pyx_n_s_ModelicaListener_exitExpr_rel;
static PyObject *__pyx_n_s_ModelicaListener_exitExpression;
static PyObject *__pyx_n_s_ModelicaListener_exitExpression_2;
static PyObject *__pyx_n_s_ModelicaListener_exitExpression_3;
static PyObject *__pyx_n_s_ModelicaListener_exitExtends_cla;
static PyObject *__pyx_n_s_ModelicaListener_exitExternal_fu;
static PyObject *__pyx_n_s_ModelicaListener_exitFor_equatio;
static PyObject *__pyx_n_s_ModelicaListener_exitFor_index;
static PyObject *__pyx_n_s_ModelicaListener_exitFor_indices;
static PyObject *__pyx_n_s_ModelicaListener_exitFor_stateme;
static PyObject *__pyx_n_s_ModelicaListener_exitFunction_ar;
static PyObject *__pyx_n_s_ModelicaListener_exitFunction_ca;
static PyObject *__pyx_n_s_ModelicaListener_exitIf_equation;
static PyObject *__pyx_n_s_ModelicaListener_exitIf_statemen;
static PyObject *__pyx_n_s_ModelicaListener_exitImport_clau;
static PyObject *__pyx_n_s_ModelicaListener_exitImport_list;
static PyObject *__pyx_n_s_ModelicaListener_exitLanguage_sp;
static PyObject *__pyx_n_s_ModelicaListener_exitModificatio;
static PyObject *__pyx_n_s_ModelicaListener_exitModificatio_2;
static PyObject *__pyx_n_s_ModelicaListener_exitModificatio_3;
static PyObject *__pyx_n_s_ModelicaListener_exitName;
static PyObject *__pyx_n_s_ModelicaListener_exitNamed_argum;
static PyObject *__pyx_n_s_ModelicaListener_exitNamed_argum_2;
static PyObject *__pyx_n_s_ModelicaListener_exitOutput_expr;
static PyObject *__pyx_n_s_ModelicaListener_exitPrimary_com;
static PyObject *__pyx_n_s_ModelicaListener_exitPrimary_der;
static PyObject *__pyx_n_s_ModelicaListener_exitPrimary_end;
static PyObject *__pyx_n_s_ModelicaListener_exitPrimary_exp;
static PyObject *__pyx_n_s_ModelicaListener_exitPrimary_fal;
static PyObject *__pyx_n_s_ModelicaListener_exitPrimary_fun;
static PyObject *__pyx_n_s_ModelicaListener_exitPrimary_fun_2;
static PyObject *__pyx_n_s_ModelicaListener_exitPrimary_ini;
static PyObject *__pyx_n_s_ModelicaListener_exitPrimary_out;
static PyObject *__pyx_n_s_ModelicaListener_exitPrimary_str;
static PyObject *__pyx_n_s_ModelicaListener_exitPrimary_tru;
static PyObject *__pyx_n_s_ModelicaListener_exitPrimary_uns;
static PyObject *__pyx_n_s_ModelicaListener_exitRegular_ele;
static PyObject *__pyx_n_s_ModelicaListener_exitReplaceable;
static PyObject *__pyx_n_s_ModelicaListener_exitShort_class;
static PyObject *__pyx_n_s_ModelicaListener_exitSimple_expr;
static PyObject *__pyx_n_s_ModelicaListener_exitStatement;
static PyObject *__pyx_n_s_ModelicaListener_exitStatement_b;
static PyObject *__pyx_n_s_ModelicaListener_exitStatement_c;
static PyObject *__pyx_n_s_ModelicaListener_exitStatement_c_2;
static PyObject *__pyx_n_s_ModelicaListener_exitStatement_f;
static PyObject *__pyx_n_s_ModelicaListener_exitStatement_i;
static PyObject *__pyx_n_s_ModelicaListener_exitStatement_r;
static PyObject *__pyx_n_s_ModelicaListener_exitStatement_w;
static PyObject *__pyx_n_s_ModelicaListener_exitStatement_w_2;
static PyObject *__pyx_n_s_ModelicaListener_exitStored_defi;
static PyObject *__pyx_n_s_ModelicaListener_exitStored_defi_2;
static PyObject *__pyx_n_s_ModelicaListener_exitString_comm;
static PyObject *__pyx_n_s_ModelicaListener_exitSubscript;
static PyObject *__pyx_n_s_ModelicaListener_exitType_prefix;
static PyObject *__pyx_n_s_ModelicaListener_exitType_specif;
static PyObject *__pyx_n_s_ModelicaListener_exitWhen_equati;
static PyObject *__pyx_n_s_ModelicaListener_exitWhen_statem;
static PyObject *__pyx_n_s_ModelicaListener_exitWhile_state;
static PyObject *__pyx_n_s_ModelicaParser;
static PyObject *__pyx_n_s_Modification_assignment2Context;
static PyObject *__pyx_n_s_Modification_assignmentContext;
static PyObject *__pyx_n_s_Modification_classContext;
static PyObject *__pyx_n_s_NameContext;
static PyObject *__pyx_n_s_Named_argumentContext;
static PyObject *__pyx_n_s_Named_argumentsContext;
static PyObject *__pyx_n_s_Output_expression_listContext;
static PyObject *__pyx_n_s_ParseTreeListener;
static PyObject *__pyx_n_s_Primary_component_referenceConte;
static PyObject *__pyx_n_s_Primary_derivativeContext;
static PyObject *__pyx_n_s_Primary_endContext;
static PyObject *__pyx_n_s_Primary_expression_listContext;
static PyObject *__pyx_n_s_Primary_falseContext;
static PyObject *__pyx_n_s_Primary_functionContext;
static PyObject *__pyx_n_s_Primary_function_argumentsContex;
static PyObject *__pyx_n_s_Primary_initialContext;
static PyObject *__pyx_n_s_Primary_output_expression_listCo;
static PyObject *__pyx_n_s_Primary_stringContext;
static PyObject *__pyx_n_s_Primary_trueContext;
static PyObject *__pyx_n_s_Primary_unsigned_numberContext;
static PyObject *__pyx_n_s_Regular_elementContext;
static PyObject *__pyx_n_s_Replaceable_elementContext;
static PyObject *__pyx_n_s_Short_class_definitionContext;
static PyObject *__pyx_n_s_Simple_expressionContext;
static PyObject *__pyx_n_s_StatementContext;
static PyObject *__pyx_n_s_Statement_breakContext;
static PyObject *__pyx_n_s_Statement_component_functionCont;
static PyObject *__pyx_n_s_Statement_component_referenceCon;
static PyObject *__pyx_n_s_Statement_forContext;
static PyObject *__pyx_n_s_Statement_ifContext;
static PyObject *__pyx_n_s_Statement_returnContext;
static PyObject *__pyx_n_s_Statement_whenContext;
static PyObject *__pyx_n_s_Statement_whileContext;
static PyObject *__pyx_n_s_Stored_definitionContext;
static PyObject *__pyx_n_s_Stored_definition_classContext;
static PyObject *__pyx_n_s_String_commentContext;
static PyObject *__pyx_n_s_SubscriptContext;
static PyObject *__pyx_n_s_Type_prefixContext;
static PyObject *__pyx_n_s_Type_specifierContext;
static PyObject *__pyx_n_s_When_equationContext;
static PyObject *__pyx_n_s_When_statementContext;
static PyObject *__pyx_n_s_While_statementContext;
static PyObject *__pyx_kp_u__2;
static PyObject *__pyx_n_s_antlr4;
static PyObject *__pyx_n_s_ctx;
static PyObject *__pyx_n_s_doc;
static PyObject *__pyx_n_s_enterAlgorithm_section;
static PyObject *__pyx_n_s_enterAnnotation;
static PyObject *__pyx_n_s_enterArgument;
static PyObject *__pyx_n_s_enterArgument_expression;
static PyObject *__pyx_n_s_enterArgument_function;
static PyObject *__pyx_n_s_enterArgument_list;
static PyObject *__pyx_n_s_enterArray_subscripts;
static PyObject *__pyx_n_s_enterBase_prefix;
static PyObject *__pyx_n_s_enterClass_definition;
static PyObject *__pyx_n_s_enterClass_modification;
static PyObject *__pyx_n_s_enterClass_prefixes;
static PyObject *__pyx_n_s_enterClass_spec_base;
static PyObject *__pyx_n_s_enterClass_spec_comp;
static PyObject *__pyx_n_s_enterClass_spec_der;
static PyObject *__pyx_n_s_enterClass_spec_enum;
static PyObject *__pyx_n_s_enterClass_spec_extends;
static PyObject *__pyx_n_s_enterClass_type;
static PyObject *__pyx_n_s_enterComment;
static PyObject *__pyx_n_s_enterComponent_clause;
static PyObject *__pyx_n_s_enterComponent_clause1;
static PyObject *__pyx_n_s_enterComponent_declaration;
static PyObject *__pyx_n_s_enterComponent_declaration1;
static PyObject *__pyx_n_s_enterComponent_list;
static PyObject *__pyx_n_s_enterComponent_reference;
static PyObject *__pyx_n_s_enterComponent_reference_element;
static PyObject *__pyx_n_s_enterComposition;
static PyObject *__pyx_n_s_enterCondition_attribute;
static PyObject *__pyx_n_s_enterConnect_clause;
static PyObject *__pyx_n_s_enterConstraining_clause;
static PyObject *__pyx_n_s_enterDeclaration;
static PyObject *__pyx_n_s_enterElement;
static PyObject *__pyx_n_s_enterElement_list;
static PyObject *__pyx_n_s_enterElement_modification;
static PyObject *__pyx_n_s_enterElement_modification_or_rep;
static PyObject *__pyx_n_s_enterElement_redeclaration;
static PyObject *__pyx_n_s_enterElement_replaceable;
static PyObject *__pyx_n_s_enterEnum_list;
static PyObject *__pyx_n_s_enterEnumeration_literal;
static PyObject *__pyx_n_s_enterEquation;
static PyObject *__pyx_n_s_enterEquation_connect_clause;
static PyObject *__pyx_n_s_enterEquation_for;
static PyObject *__pyx_n_s_enterEquation_function;
static PyObject *__pyx_n_s_enterEquation_if;
static PyObject *__pyx_n_s_enterEquation_section;
static PyObject *__pyx_n_s_enterEquation_simple;
static PyObject *__pyx_n_s_enterEquation_when;
static PyObject *__pyx_n_s_enterExpr_add;
static PyObject *__pyx_n_s_enterExpr_and;
static PyObject *__pyx_n_s_enterExpr_exp;
static PyObject *__pyx_n_s_enterExpr_mul;
static PyObject *__pyx_n_s_enterExpr_neg;
static PyObject *__pyx_n_s_enterExpr_not;
static PyObject *__pyx_n_s_enterExpr_or;
static PyObject *__pyx_n_s_enterExpr_primary;
static PyObject *__pyx_n_s_enterExpr_rel;
static PyObject *__pyx_n_s_enterExpression_if;
static PyObject *__pyx_n_s_enterExpression_list;
static PyObject *__pyx_n_s_enterExpression_simple;
static PyObject *__pyx_n_s_enterExtends_clause;
static PyObject *__pyx_n_s_enterExternal_function_call;
static PyObject *__pyx_n_s_enterFor_equation;
static PyObject *__pyx_n_s_enterFor_index;
static PyObject *__pyx_n_s_enterFor_indices;
static PyObject *__pyx_n_s_enterFor_statement;
static PyObject *__pyx_n_s_enterFunction_arguments;
static PyObject *__pyx_n_s_enterFunction_call_args;
static PyObject *__pyx_n_s_enterIf_equation;
static PyObject *__pyx_n_s_enterIf_statement;
static PyObject *__pyx_n_s_enterImport_clause;
static PyObject *__pyx_n_s_enterImport_list;
static PyObject *__pyx_n_s_enterLanguage_specification;
static PyObject *__pyx_n_s_enterModification_assignment;
static PyObject *__pyx_n_s_enterModification_assignment2;
static PyObject *__pyx_n_s_enterModification_class;
static PyObject *__pyx_n_s_enterName;
static PyObject *__pyx_n_s_enterNamed_argument;
static PyObject *__pyx_n_s_enterNamed_arguments;
static PyObject *__pyx_n_s_enterOutput_expression_list;
static PyObject *__pyx_n_s_enterPrimary_component_reference;
static PyObject *__pyx_n_s_enterPrimary_derivative;
static PyObject *__pyx_n_s_enterPrimary_end;
static PyObject *__pyx_n_s_enterPrimary_expression_list;
static PyObject *__pyx_n_s_enterPrimary_false;
static PyObject *__pyx_n_s_enterPrimary_function;
static PyObject *__pyx_n_s_enterPrimary_function_arguments;
static PyObject *__pyx_n_s_enterPrimary_initial;
static PyObject *__pyx_n_s_enterPrimary_output_expression_l;
static PyObject *__pyx_n_s_enterPrimary_string;
static PyObject *__pyx_n_s_enterPrimary_true;
static PyObject *__pyx_n_s_enterPrimary_unsigned_number;
static PyObject *__pyx_n_s_enterRegular_element;
static PyObject *__pyx_n_s_enterReplaceable_element;
static PyObject *__pyx_n_s_enterShort_class_definition;
static PyObject *__pyx_n_s_enterSimple_expression;
static PyObject *__pyx_n_s_enterStatement;
static PyObject *__pyx_n_s_enterStatement_break;
static PyObject *__pyx_n_s_enterStatement_component_functio;
static PyObject *__pyx_n_s_enterStatement_component_referen;
static PyObject *__pyx_n_s_enterStatement_for;
static PyObject *__pyx_n_s_enterStatement_if;
static PyObject *__pyx_n_s_enterStatement_return;
static PyObject *__pyx_n_s_enterStatement_when;
static PyObject *__pyx_n_s_enterStatement_while;
static PyObject *__pyx_n_s_enterStored_definition;
static PyObject *__pyx_n_s_enterStored_definition_class;
static PyObject *__pyx_n_s_enterString_comment;
static PyObject *__pyx_n_s_enterSubscript;
static PyObject *__pyx_n_s_enterType_prefix;
static PyObject *__pyx_n_s_enterType_specifier;
static PyObject *__pyx_n_s_enterWhen_equation;
static PyObject *__pyx_n_s_enterWhen_statement;
static PyObject *__pyx_n_s_enterWhile_statement;
static PyObject *__pyx_n_s_exitAlgorithm_section;
static PyObject *__pyx_n_s_exitAnnotation;
static PyObject *__pyx_n_s_exitArgument;
static PyObject *__pyx_n_s_exitArgument_expression;
static PyObject *__pyx_n_s_exitArgument_function;
static PyObject *__pyx_n_s_exitArgument_list;
static PyObject *__pyx_n_s_exitArray_subscripts;
static PyObject *__pyx_n_s_exitBase_prefix;
static PyObject *__pyx_n_s_exitClass_definition;
static PyObject *__pyx_n_s_exitClass_modification;
static PyObject *__pyx_n_s_exitClass_prefixes;
static PyObject *__pyx_n_s_exitClass_spec_base;
static PyObject *__pyx_n_s_exitClass_spec_comp;
static PyObject *__pyx_n_s_exitClass_spec_der;
static PyObject *__pyx_n_s_exitClass_spec_enum;
static PyObject *__pyx_n_s_exitClass_spec_extends;
static PyObject *__pyx_n_s_exitClass_type;
static PyObject *__pyx_n_s_exitComment;
static PyObject *__pyx_n_s_exitComponent_clause;
static PyObject *__pyx_n_s_exitComponent_clause1;
static PyObject *__pyx_n_s_exitComponent_declaration;
static PyObject *__pyx_n_s_exitComponent_declaration1;
static PyObject *__pyx_n_s_exitComponent_list;
static PyObject *__pyx_n_s_exitComponent_reference;
static PyObject *__pyx_n_s_exitComponent_reference_element;
static PyObject *__pyx_n_s_exitComposition;
static PyObject *__pyx_n_s_exitCondition_attribute;
static PyObject *__pyx_n_s_exitConnect_clause;
static PyObject *__pyx_n_s_exitConstraining_clause;
static PyObject *__pyx_n_s_exitDeclaration;
static PyObject *__pyx_n_s_exitElement;
static PyObject *__pyx_n_s_exitElement_list;
static PyObject *__pyx_n_s_exitElement_modification;
static PyObject *__pyx_n_s_exitElement_modification_or_repl;
static PyObject *__pyx_n_s_exitElement_redeclaration;
static PyObject *__pyx_n_s_exitElement_replaceable;
static PyObject *__pyx_n_s_exitEnum_list;
static PyObject *__pyx_n_s_exitEnumeration_literal;
static PyObject *__pyx_n_s_exitEquation;
static PyObject *__pyx_n_s_exitEquation_connect_clause;
static PyObject *__pyx_n_s_exitEquation_for;
static PyObject *__pyx_n_s_exitEquation_function;
static PyObject *__pyx_n_s_exitEquation_if;
static PyObject *__pyx_n_s_exitEquation_section;
static PyObject *__pyx_n_s_exitEquation_simple;
static PyObject *__pyx_n_s_exitEquation_when;
static PyObject *__pyx_n_s_exitExpr_add;
static PyObject *__pyx_n_s_exitExpr_and;
static PyObject *__pyx_n_s_exitExpr_exp;
static PyObject *__pyx_n_s_exitExpr_mul;
static PyObject *__pyx_n_s_exitExpr_neg;
static PyObject *__pyx_n_s_exitExpr_not;
static PyObject *__pyx_n_s_exitExpr_or;
static PyObject *__pyx_n_s_exitExpr_primary;
static PyObject *__pyx_n_s_exitExpr_rel;
static PyObject *__pyx_n_s_exitExpression_if;
static PyObject *__pyx_n_s_exitExpression_list;
static PyObject *__pyx_n_s_exitExpression_simple;
static PyObject *__pyx_n_s_exitExtends_clause;
static PyObject *__pyx_n_s_exitExternal_function_call;
static PyObject *__pyx_n_s_exitFor_equation;
static PyObject *__pyx_n_s_exitFor_index;
static PyObject *__pyx_n_s_exitFor_indices;
static PyObject *__pyx_n_s_exitFor_statement;
static PyObject *__pyx_n_s_exitFunction_arguments;
static PyObject *__pyx_n_s_exitFunction_call_args;
static PyObject *__pyx_n_s_exitIf_equation;
static PyObject *__pyx_n_s_exitIf_statement;
static PyObject *__pyx_n_s_exitImport_clause;
static PyObject *__pyx_n_s_exitImport_list;
static PyObject *__pyx_n_s_exitLanguage_specification;
static PyObject *__pyx_n_s_exitModification_assignment;
static PyObject *__pyx_n_s_exitModification_assignment2;
static PyObject *__pyx_n_s_exitModification_class;
static PyObject *__pyx_n_s_exitName;
static PyObject *__pyx_n_s_exitNamed_argument;
static PyObject *__pyx_n_s_exitNamed_arguments;
static PyObject *__pyx_n_s_exitOutput_expression_list;
static PyObject *__pyx_n_s_exitPrimary_component_reference;
static PyObject *__pyx_n_s_exitPrimary_derivative;
static PyObject *__pyx_n_s_exitPrimary_end;
static PyObject *__pyx_n_s_exitPrimary_expression_list;
static PyObject *__pyx_n_s_exitPrimary_false;
static PyObject *__pyx_n_s_exitPrimary_function;
static PyObject *__pyx_n_s_exitPrimary_function_arguments;
static PyObject *__pyx_n_s_exitPrimary_initial;
static PyObject *__pyx_n_s_exitPrimary_output_expression_li;
static PyObject *__pyx_n_s_exitPrimary_string;
static PyObject *__pyx_n_s_exitPrimary_true;
static PyObject *__pyx_n_s_exitPrimary_unsigned_number;
static PyObject *__pyx_n_s_exitRegular_element;
static PyObject *__pyx_n_s_exitReplaceable_element;
static PyObject *__pyx_n_s_exitShort_class_definition;
static PyObject *__pyx_n_s_exitSimple_expression;
static PyObject *__pyx_n_s_exitStatement;
static PyObject *__pyx_n_s_exitStatement_break;
static PyObject *__pyx_n_s_exitStatement_component_function;
static PyObject *__pyx_n_s_exitStatement_component_referenc;
static PyObject *__pyx_n_s_exitStatement_for;
static PyObject *__pyx_n_s_exitStatement_if;
static PyObject *__pyx_n_s_exitStatement_return;
static PyObject *__pyx_n_s_exitStatement_when;
static PyObject *__pyx_n_s_exitStatement_while;
static PyObject *__pyx_n_s_exitStored_definition;
static PyObject *__pyx_n_s_exitStored_definition_class;
static PyObject *__pyx_n_s_exitString_comment;
static PyObject *__pyx_n_s_exitSubscript;
static PyObject *__pyx_n_s_exitType_prefix;
static PyObject *__pyx_n_s_exitType_specifier;
static PyObject *__pyx_n_s_exitWhen_equation;
static PyObject *__pyx_n_s_exitWhen_statement;
static PyObject *__pyx_n_s_exitWhile_statement;
static PyObject *__pyx_kp_s_home_jgoppert_git_pymola_pymola;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_metaclass;
static PyObject *__pyx_n_s_module;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_prepare;
static PyObject *__pyx_n_s_pymola_generated_ModelicaListene;
static PyObject *__pyx_n_s_qualname;
static PyObject *__pyx_n_s_self;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_enterStored_definition(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_2exitStored_definition(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_4enterStored_definition_class(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_6exitStored_definition_class(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_8enterClass_definition(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_10exitClass_definition(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_12enterClass_prefixes(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_14exitClass_prefixes(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_16enterClass_type(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_18exitClass_type(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_20enterClass_spec_comp(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_22exitClass_spec_comp(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_24enterClass_spec_base(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_26exitClass_spec_base(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_28enterClass_spec_enum(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_30exitClass_spec_enum(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_32enterClass_spec_der(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_34exitClass_spec_der(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_36enterClass_spec_extends(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_38exitClass_spec_extends(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_40enterBase_prefix(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_42exitBase_prefix(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_44enterEnum_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_46exitEnum_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_48enterEnumeration_literal(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_50exitEnumeration_literal(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_52enterComposition(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_54exitComposition(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_56enterLanguage_specification(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_58exitLanguage_specification(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_60enterExternal_function_call(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_62exitExternal_function_call(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_64enterElement_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_66exitElement_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_68enterElement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_70exitElement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_72enterRegular_element(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_74exitRegular_element(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_76enterReplaceable_element(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_78exitReplaceable_element(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_80enterImport_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_82exitImport_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_84enterImport_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_86exitImport_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_88enterExtends_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_90exitExtends_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_92enterConstraining_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_94exitConstraining_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_96enterComponent_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_98exitComponent_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_100enterType_prefix(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_102exitType_prefix(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_104enterType_specifier(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_106exitType_specifier(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_108enterComponent_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_110exitComponent_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_112enterComponent_declaration(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_114exitComponent_declaration(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_116enterCondition_attribute(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_118exitCondition_attribute(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_120enterDeclaration(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_122exitDeclaration(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_124enterModification_class(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_126exitModification_class(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_128enterModification_assignment(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_130exitModification_assignment(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_132enterModification_assignment2(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_134exitModification_assignment2(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_136enterClass_modification(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_138exitClass_modification(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_140enterArgument_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_142exitArgument_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_144enterArgument(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_146exitArgument(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_148enterElement_modification_or_replaceable(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_150exitElement_modification_or_replaceable(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_152enterElement_modification(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_154exitElement_modification(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_156enterElement_redeclaration(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_158exitElement_redeclaration(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_160enterElement_replaceable(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_162exitElement_replaceable(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_164enterComponent_clause1(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_166exitComponent_clause1(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_168enterComponent_declaration1(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_170exitComponent_declaration1(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_172enterShort_class_definition(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_174exitShort_class_definition(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_176enterEquation_section(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_178exitEquation_section(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_180enterAlgorithm_section(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_182exitAlgorithm_section(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_184enterEquation_simple(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_186exitEquation_simple(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_188enterEquation_if(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_190exitEquation_if(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_192enterEquation_for(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_194exitEquation_for(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_196enterEquation_connect_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_198exitEquation_connect_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_200enterEquation_when(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_202exitEquation_when(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_204enterEquation_function(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_206exitEquation_function(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_208enterEquation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_210exitEquation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_212enterStatement_component_reference(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_214exitStatement_component_reference(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_216enterStatement_component_function(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_218exitStatement_component_function(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_220enterStatement_break(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_222exitStatement_break(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_224enterStatement_return(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_226exitStatement_return(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_228enterStatement_if(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_230exitStatement_if(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_232enterStatement_for(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_234exitStatement_for(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_236enterStatement_while(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_238exitStatement_while(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_240enterStatement_when(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_242exitStatement_when(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_244enterStatement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_246exitStatement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_248enterIf_equation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_250exitIf_equation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_252enterIf_statement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_254exitIf_statement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_256enterFor_equation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_258exitFor_equation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_260enterFor_statement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_262exitFor_statement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_264enterFor_indices(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_266exitFor_indices(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_268enterFor_index(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_270exitFor_index(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_272enterWhile_statement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_274exitWhile_statement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_276enterWhen_equation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_278exitWhen_equation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_280enterWhen_statement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_282exitWhen_statement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_284enterConnect_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_286exitConnect_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_288enterExpression_simple(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_290exitExpression_simple(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_292enterExpression_if(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_294exitExpression_if(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_296enterSimple_expression(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_298exitSimple_expression(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_300enterExpr_neg(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_302exitExpr_neg(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_304enterExpr_add(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_306exitExpr_add(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_308enterExpr_exp(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_310exitExpr_exp(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_312enterExpr_or(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_314exitExpr_or(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_316enterExpr_primary(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_318exitExpr_primary(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_320enterExpr_and(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_322exitExpr_and(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_324enterExpr_rel(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_326exitExpr_rel(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_328enterExpr_not(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_330exitExpr_not(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_332enterExpr_mul(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_334exitExpr_mul(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_336enterPrimary_unsigned_number(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_338exitPrimary_unsigned_number(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_340enterPrimary_string(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_342exitPrimary_string(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_344enterPrimary_false(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_346exitPrimary_false(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_348enterPrimary_true(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_350exitPrimary_true(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_352enterPrimary_function(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_354exitPrimary_function(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_356enterPrimary_derivative(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_358exitPrimary_derivative(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_360enterPrimary_initial(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_362exitPrimary_initial(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_364enterPrimary_component_reference(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_366exitPrimary_component_reference(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_368enterPrimary_output_expression_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_370exitPrimary_output_expression_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_372enterPrimary_expression_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_374exitPrimary_expression_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_376enterPrimary_function_arguments(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_378exitPrimary_function_arguments(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_380enterPrimary_end(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_382exitPrimary_end(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_384enterName(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_386exitName(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_388enterComponent_reference_element(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_390exitComponent_reference_element(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_392enterComponent_reference(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_394exitComponent_reference(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_396enterFunction_call_args(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_398exitFunction_call_args(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_400enterFunction_arguments(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_402exitFunction_arguments(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_404enterNamed_arguments(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_406exitNamed_arguments(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_408enterNamed_argument(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_410exitNamed_argument(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_412enterArgument_function(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_414exitArgument_function(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_416enterArgument_expression(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_418exitArgument_expression(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_420enterOutput_expression_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_422exitOutput_expression_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_424enterExpression_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_426exitExpression_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_428enterArray_subscripts(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_430exitArray_subscripts(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_432enterSubscript(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_434exitSubscript(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_436enterComment(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_438exitComment(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_440enterString_comment(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_442exitString_comment(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_444enterAnnotation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_446exitAnnotation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx); /* proto */
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_tuple__51;
static PyObject *__pyx_tuple__53;
static PyObject *__pyx_tuple__55;
static PyObject *__pyx_tuple__57;
static PyObject *__pyx_tuple__59;
static PyObject *__pyx_tuple__61;
static PyObject *__pyx_tuple__63;
static PyObject *__pyx_tuple__65;
static PyObject *__pyx_tuple__67;
static PyObject *__pyx_tuple__69;
static PyObject *__pyx_tuple__71;
static PyObject *__pyx_tuple__73;
static PyObject *__pyx_tuple__75;
static PyObject *__pyx_tuple__77;
static PyObject *__pyx_tuple__79;
static PyObject *__pyx_tuple__81;
static PyObject *__pyx_tuple__83;
static PyObject *__pyx_tuple__85;
static PyObject *__pyx_tuple__87;
static PyObject *__pyx_tuple__89;
static PyObject *__pyx_tuple__91;
static PyObject *__pyx_tuple__93;
static PyObject *__pyx_tuple__95;
static PyObject *__pyx_tuple__97;
static PyObject *__pyx_tuple__99;
static PyObject *__pyx_codeobj__4;
static PyObject *__pyx_codeobj__6;
static PyObject *__pyx_codeobj__8;
static PyObject *__pyx_tuple__101;
static PyObject *__pyx_tuple__103;
static PyObject *__pyx_tuple__105;
static PyObject *__pyx_tuple__107;
static PyObject *__pyx_tuple__109;
static PyObject *__pyx_tuple__111;
static PyObject *__pyx_tuple__113;
static PyObject *__pyx_tuple__115;
static PyObject *__pyx_tuple__117;
static PyObject *__pyx_tuple__119;
static PyObject *__pyx_tuple__121;
static PyObject *__pyx_tuple__123;
static PyObject *__pyx_tuple__125;
static PyObject *__pyx_tuple__127;
static PyObject *__pyx_tuple__129;
static PyObject *__pyx_tuple__131;
static PyObject *__pyx_tuple__133;
static PyObject *__pyx_tuple__135;
static PyObject *__pyx_tuple__137;
static PyObject *__pyx_tuple__139;
static PyObject *__pyx_tuple__141;
static PyObject *__pyx_tuple__143;
static PyObject *__pyx_tuple__145;
static PyObject *__pyx_tuple__147;
static PyObject *__pyx_tuple__149;
static PyObject *__pyx_tuple__151;
static PyObject *__pyx_tuple__153;
static PyObject *__pyx_tuple__155;
static PyObject *__pyx_tuple__157;
static PyObject *__pyx_tuple__159;
static PyObject *__pyx_tuple__161;
static PyObject *__pyx_tuple__163;
static PyObject *__pyx_tuple__165;
static PyObject *__pyx_tuple__167;
static PyObject *__pyx_tuple__169;
static PyObject *__pyx_tuple__171;
static PyObject *__pyx_tuple__173;
static PyObject *__pyx_tuple__175;
static PyObject *__pyx_tuple__177;
static PyObject *__pyx_tuple__179;
static PyObject *__pyx_tuple__181;
static PyObject *__pyx_tuple__183;
static PyObject *__pyx_tuple__185;
static PyObject *__pyx_tuple__187;
static PyObject *__pyx_tuple__189;
static PyObject *__pyx_tuple__191;
static PyObject *__pyx_tuple__193;
static PyObject *__pyx_tuple__195;
static PyObject *__pyx_tuple__197;
static PyObject *__pyx_tuple__199;
static PyObject *__pyx_tuple__201;
static PyObject *__pyx_tuple__203;
static PyObject *__pyx_tuple__205;
static PyObject *__pyx_tuple__207;
static PyObject *__pyx_tuple__209;
static PyObject *__pyx_tuple__211;
static PyObject *__pyx_tuple__213;
static PyObject *__pyx_tuple__215;
static PyObject *__pyx_tuple__217;
static PyObject *__pyx_tuple__219;
static PyObject *__pyx_tuple__221;
static PyObject *__pyx_tuple__223;
static PyObject *__pyx_tuple__225;
static PyObject *__pyx_tuple__227;
static PyObject *__pyx_tuple__229;
static PyObject *__pyx_tuple__231;
static PyObject *__pyx_tuple__233;
static PyObject *__pyx_tuple__235;
static PyObject *__pyx_tuple__237;
static PyObject *__pyx_tuple__239;
static PyObject *__pyx_tuple__241;
static PyObject *__pyx_tuple__243;
static PyObject *__pyx_tuple__245;
static PyObject *__pyx_tuple__247;
static PyObject *__pyx_tuple__249;
static PyObject *__pyx_tuple__251;
static PyObject *__pyx_tuple__253;
static PyObject *__pyx_tuple__255;
static PyObject *__pyx_tuple__257;
static PyObject *__pyx_tuple__259;
static PyObject *__pyx_tuple__261;
static PyObject *__pyx_tuple__263;
static PyObject *__pyx_tuple__265;
static PyObject *__pyx_tuple__267;
static PyObject *__pyx_tuple__269;
static PyObject *__pyx_tuple__271;
static PyObject *__pyx_tuple__273;
static PyObject *__pyx_tuple__275;
static PyObject *__pyx_tuple__277;
static PyObject *__pyx_tuple__279;
static PyObject *__pyx_tuple__281;
static PyObject *__pyx_tuple__283;
static PyObject *__pyx_tuple__285;
static PyObject *__pyx_tuple__287;
static PyObject *__pyx_tuple__289;
static PyObject *__pyx_tuple__291;
static PyObject *__pyx_tuple__293;
static PyObject *__pyx_tuple__295;
static PyObject *__pyx_tuple__297;
static PyObject *__pyx_tuple__299;
static PyObject *__pyx_tuple__301;
static PyObject *__pyx_tuple__303;
static PyObject *__pyx_tuple__305;
static PyObject *__pyx_tuple__307;
static PyObject *__pyx_tuple__309;
static PyObject *__pyx_tuple__311;
static PyObject *__pyx_tuple__313;
static PyObject *__pyx_tuple__315;
static PyObject *__pyx_tuple__317;
static PyObject *__pyx_tuple__319;
static PyObject *__pyx_tuple__321;
static PyObject *__pyx_tuple__323;
static PyObject *__pyx_tuple__325;
static PyObject *__pyx_tuple__327;
static PyObject *__pyx_tuple__329;
static PyObject *__pyx_tuple__331;
static PyObject *__pyx_tuple__333;
static PyObject *__pyx_tuple__335;
static PyObject *__pyx_tuple__337;
static PyObject *__pyx_tuple__339;
static PyObject *__pyx_tuple__341;
static PyObject *__pyx_tuple__343;
static PyObject *__pyx_tuple__345;
static PyObject *__pyx_tuple__347;
static PyObject *__pyx_tuple__349;
static PyObject *__pyx_tuple__351;
static PyObject *__pyx_tuple__353;
static PyObject *__pyx_tuple__355;
static PyObject *__pyx_tuple__357;
static PyObject *__pyx_tuple__359;
static PyObject *__pyx_tuple__361;
static PyObject *__pyx_tuple__363;
static PyObject *__pyx_tuple__365;
static PyObject *__pyx_tuple__367;
static PyObject *__pyx_tuple__369;
static PyObject *__pyx_tuple__371;
static PyObject *__pyx_tuple__373;
static PyObject *__pyx_tuple__375;
static PyObject *__pyx_tuple__377;
static PyObject *__pyx_tuple__379;
static PyObject *__pyx_tuple__381;
static PyObject *__pyx_tuple__383;
static PyObject *__pyx_tuple__385;
static PyObject *__pyx_tuple__387;
static PyObject *__pyx_tuple__389;
static PyObject *__pyx_tuple__391;
static PyObject *__pyx_tuple__393;
static PyObject *__pyx_tuple__395;
static PyObject *__pyx_tuple__397;
static PyObject *__pyx_tuple__399;
static PyObject *__pyx_tuple__401;
static PyObject *__pyx_tuple__403;
static PyObject *__pyx_tuple__405;
static PyObject *__pyx_tuple__407;
static PyObject *__pyx_tuple__409;
static PyObject *__pyx_tuple__411;
static PyObject *__pyx_tuple__413;
static PyObject *__pyx_tuple__415;
static PyObject *__pyx_tuple__417;
static PyObject *__pyx_tuple__419;
static PyObject *__pyx_tuple__421;
static PyObject *__pyx_tuple__423;
static PyObject *__pyx_tuple__425;
static PyObject *__pyx_tuple__427;
static PyObject *__pyx_tuple__429;
static PyObject *__pyx_tuple__431;
static PyObject *__pyx_tuple__433;
static PyObject *__pyx_tuple__435;
static PyObject *__pyx_tuple__437;
static PyObject *__pyx_tuple__439;
static PyObject *__pyx_tuple__441;
static PyObject *__pyx_tuple__443;
static PyObject *__pyx_tuple__445;
static PyObject *__pyx_tuple__447;
static PyObject *__pyx_tuple__449;
static PyObject *__pyx_codeobj__10;
static PyObject *__pyx_codeobj__12;
static PyObject *__pyx_codeobj__14;
static PyObject *__pyx_codeobj__16;
static PyObject *__pyx_codeobj__18;
static PyObject *__pyx_codeobj__20;
static PyObject *__pyx_codeobj__22;
static PyObject *__pyx_codeobj__24;
static PyObject *__pyx_codeobj__26;
static PyObject *__pyx_codeobj__28;
static PyObject *__pyx_codeobj__30;
static PyObject *__pyx_codeobj__32;
static PyObject *__pyx_codeobj__34;
static PyObject *__pyx_codeobj__36;
static PyObject *__pyx_codeobj__38;
static PyObject *__pyx_codeobj__40;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__44;
static PyObject *__pyx_codeobj__46;
static PyObject *__pyx_codeobj__48;
static PyObject *__pyx_codeobj__50;
static PyObject *__pyx_codeobj__52;
static PyObject *__pyx_codeobj__54;
static PyObject *__pyx_codeobj__56;
static PyObject *__pyx_codeobj__58;
static PyObject *__pyx_codeobj__60;
static PyObject *__pyx_codeobj__62;
static PyObject *__pyx_codeobj__64;
static PyObject *__pyx_codeobj__66;
static PyObject *__pyx_codeobj__68;
static PyObject *__pyx_codeobj__70;
static PyObject *__pyx_codeobj__72;
static PyObject *__pyx_codeobj__74;
static PyObject *__pyx_codeobj__76;
static PyObject *__pyx_codeobj__78;
static PyObject *__pyx_codeobj__80;
static PyObject *__pyx_codeobj__82;
static PyObject *__pyx_codeobj__84;
static PyObject *__pyx_codeobj__86;
static PyObject *__pyx_codeobj__88;
static PyObject *__pyx_codeobj__90;
static PyObject *__pyx_codeobj__92;
static PyObject *__pyx_codeobj__94;
static PyObject *__pyx_codeobj__96;
static PyObject *__pyx_codeobj__98;
static PyObject *__pyx_codeobj__100;
static PyObject *__pyx_codeobj__102;
static PyObject *__pyx_codeobj__104;
static PyObject *__pyx_codeobj__106;
static PyObject *__pyx_codeobj__108;
static PyObject *__pyx_codeobj__110;
static PyObject *__pyx_codeobj__112;
static PyObject *__pyx_codeobj__114;
static PyObject *__pyx_codeobj__116;
static PyObject *__pyx_codeobj__118;
static PyObject *__pyx_codeobj__120;
static PyObject *__pyx_codeobj__122;
static PyObject *__pyx_codeobj__124;
static PyObject *__pyx_codeobj__126;
static PyObject *__pyx_codeobj__128;
static PyObject *__pyx_codeobj__130;
static PyObject *__pyx_codeobj__132;
static PyObject *__pyx_codeobj__134;
static PyObject *__pyx_codeobj__136;
static PyObject *__pyx_codeobj__138;
static PyObject *__pyx_codeobj__140;
static PyObject *__pyx_codeobj__142;
static PyObject *__pyx_codeobj__144;
static PyObject *__pyx_codeobj__146;
static PyObject *__pyx_codeobj__148;
static PyObject *__pyx_codeobj__150;
static PyObject *__pyx_codeobj__152;
static PyObject *__pyx_codeobj__154;
static PyObject *__pyx_codeobj__156;
static PyObject *__pyx_codeobj__158;
static PyObject *__pyx_codeobj__160;
static PyObject *__pyx_codeobj__162;
static PyObject *__pyx_codeobj__164;
static PyObject *__pyx_codeobj__166;
static PyObject *__pyx_codeobj__168;
static PyObject *__pyx_codeobj__170;
static PyObject *__pyx_codeobj__172;
static PyObject *__pyx_codeobj__174;
static PyObject *__pyx_codeobj__176;
static PyObject *__pyx_codeobj__178;
static PyObject *__pyx_codeobj__180;
static PyObject *__pyx_codeobj__182;
static PyObject *__pyx_codeobj__184;
static PyObject *__pyx_codeobj__186;
static PyObject *__pyx_codeobj__188;
static PyObject *__pyx_codeobj__190;
static PyObject *__pyx_codeobj__192;
static PyObject *__pyx_codeobj__194;
static PyObject *__pyx_codeobj__196;
static PyObject *__pyx_codeobj__198;
static PyObject *__pyx_codeobj__200;
static PyObject *__pyx_codeobj__202;
static PyObject *__pyx_codeobj__204;
static PyObject *__pyx_codeobj__206;
static PyObject *__pyx_codeobj__208;
static PyObject *__pyx_codeobj__210;
static PyObject *__pyx_codeobj__212;
static PyObject *__pyx_codeobj__214;
static PyObject *__pyx_codeobj__216;
static PyObject *__pyx_codeobj__218;
static PyObject *__pyx_codeobj__220;
static PyObject *__pyx_codeobj__222;
static PyObject *__pyx_codeobj__224;
static PyObject *__pyx_codeobj__226;
static PyObject *__pyx_codeobj__228;
static PyObject *__pyx_codeobj__230;
static PyObject *__pyx_codeobj__232;
static PyObject *__pyx_codeobj__234;
static PyObject *__pyx_codeobj__236;
static PyObject *__pyx_codeobj__238;
static PyObject *__pyx_codeobj__240;
static PyObject *__pyx_codeobj__242;
static PyObject *__pyx_codeobj__244;
static PyObject *__pyx_codeobj__246;
static PyObject *__pyx_codeobj__248;
static PyObject *__pyx_codeobj__250;
static PyObject *__pyx_codeobj__252;
static PyObject *__pyx_codeobj__254;
static PyObject *__pyx_codeobj__256;
static PyObject *__pyx_codeobj__258;
static PyObject *__pyx_codeobj__260;
static PyObject *__pyx_codeobj__262;
static PyObject *__pyx_codeobj__264;
static PyObject *__pyx_codeobj__266;
static PyObject *__pyx_codeobj__268;
static PyObject *__pyx_codeobj__270;
static PyObject *__pyx_codeobj__272;
static PyObject *__pyx_codeobj__274;
static PyObject *__pyx_codeobj__276;
static PyObject *__pyx_codeobj__278;
static PyObject *__pyx_codeobj__280;
static PyObject *__pyx_codeobj__282;
static PyObject *__pyx_codeobj__284;
static PyObject *__pyx_codeobj__286;
static PyObject *__pyx_codeobj__288;
static PyObject *__pyx_codeobj__290;
static PyObject *__pyx_codeobj__292;
static PyObject *__pyx_codeobj__294;
static PyObject *__pyx_codeobj__296;
static PyObject *__pyx_codeobj__298;
static PyObject *__pyx_codeobj__300;
static PyObject *__pyx_codeobj__302;
static PyObject *__pyx_codeobj__304;
static PyObject *__pyx_codeobj__306;
static PyObject *__pyx_codeobj__308;
static PyObject *__pyx_codeobj__310;
static PyObject *__pyx_codeobj__312;
static PyObject *__pyx_codeobj__314;
static PyObject *__pyx_codeobj__316;
static PyObject *__pyx_codeobj__318;
static PyObject *__pyx_codeobj__320;
static PyObject *__pyx_codeobj__322;
static PyObject *__pyx_codeobj__324;
static PyObject *__pyx_codeobj__326;
static PyObject *__pyx_codeobj__328;
static PyObject *__pyx_codeobj__330;
static PyObject *__pyx_codeobj__332;
static PyObject *__pyx_codeobj__334;
static PyObject *__pyx_codeobj__336;
static PyObject *__pyx_codeobj__338;
static PyObject *__pyx_codeobj__340;
static PyObject *__pyx_codeobj__342;
static PyObject *__pyx_codeobj__344;
static PyObject *__pyx_codeobj__346;
static PyObject *__pyx_codeobj__348;
static PyObject *__pyx_codeobj__350;
static PyObject *__pyx_codeobj__352;
static PyObject *__pyx_codeobj__354;
static PyObject *__pyx_codeobj__356;
static PyObject *__pyx_codeobj__358;
static PyObject *__pyx_codeobj__360;
static PyObject *__pyx_codeobj__362;
static PyObject *__pyx_codeobj__364;
static PyObject *__pyx_codeobj__366;
static PyObject *__pyx_codeobj__368;
static PyObject *__pyx_codeobj__370;
static PyObject *__pyx_codeobj__372;
static PyObject *__pyx_codeobj__374;
static PyObject *__pyx_codeobj__376;
static PyObject *__pyx_codeobj__378;
static PyObject *__pyx_codeobj__380;
static PyObject *__pyx_codeobj__382;
static PyObject *__pyx_codeobj__384;
static PyObject *__pyx_codeobj__386;
static PyObject *__pyx_codeobj__388;
static PyObject *__pyx_codeobj__390;
static PyObject *__pyx_codeobj__392;
static PyObject *__pyx_codeobj__394;
static PyObject *__pyx_codeobj__396;
static PyObject *__pyx_codeobj__398;
static PyObject *__pyx_codeobj__400;
static PyObject *__pyx_codeobj__402;
static PyObject *__pyx_codeobj__404;
static PyObject *__pyx_codeobj__406;
static PyObject *__pyx_codeobj__408;
static PyObject *__pyx_codeobj__410;
static PyObject *__pyx_codeobj__412;
static PyObject *__pyx_codeobj__414;
static PyObject *__pyx_codeobj__416;
static PyObject *__pyx_codeobj__418;
static PyObject *__pyx_codeobj__420;
static PyObject *__pyx_codeobj__422;
static PyObject *__pyx_codeobj__424;
static PyObject *__pyx_codeobj__426;
static PyObject *__pyx_codeobj__428;
static PyObject *__pyx_codeobj__430;
static PyObject *__pyx_codeobj__432;
static PyObject *__pyx_codeobj__434;
static PyObject *__pyx_codeobj__436;
static PyObject *__pyx_codeobj__438;
static PyObject *__pyx_codeobj__440;
static PyObject *__pyx_codeobj__442;
static PyObject *__pyx_codeobj__444;
static PyObject *__pyx_codeobj__446;
static PyObject *__pyx_codeobj__448;
static PyObject *__pyx_codeobj__450;

/* "pymola/generated/ModelicaListener.pyx":12
 * 
 *     # Enter a parse tree produced by ModelicaParser#stored_definition.
 *     def enterStored_definition(self, ctx:ModelicaParser.Stored_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_1enterStored_definition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_1enterStored_definition = {"enterStored_definition", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_1enterStored_definition, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_1enterStored_definition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStored_definition (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterStored_definition", 1, 2, 2, 1); __PYX_ERR(0, 12, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterStored_definition") < 0)) __PYX_ERR(0, 12, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterStored_definition", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 12, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterStored_definition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_enterStored_definition(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_enterStored_definition(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStored_definition", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":16
 * 
 *     # Exit a parse tree produced by ModelicaParser#stored_definition.
 *     def exitStored_definition(self, ctx:ModelicaParser.Stored_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_3exitStored_definition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_3exitStored_definition = {"exitStored_definition", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_3exitStored_definition, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_3exitStored_definition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStored_definition (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitStored_definition", 1, 2, 2, 1); __PYX_ERR(0, 16, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitStored_definition") < 0)) __PYX_ERR(0, 16, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitStored_definition", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 16, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitStored_definition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_2exitStored_definition(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_2exitStored_definition(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStored_definition", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":21
 * 
 *     # Enter a parse tree produced by ModelicaParser#stored_definition_class.
 *     def enterStored_definition_class(self, ctx:ModelicaParser.Stored_definition_classContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_5enterStored_definition_class(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_5enterStored_definition_class = {"enterStored_definition_class", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_5enterStored_definition_class, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_5enterStored_definition_class(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStored_definition_class (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterStored_definition_class", 1, 2, 2, 1); __PYX_ERR(0, 21, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterStored_definition_class") < 0)) __PYX_ERR(0, 21, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterStored_definition_class", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 21, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterStored_definition_class", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_4enterStored_definition_class(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_4enterStored_definition_class(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStored_definition_class", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":25
 * 
 *     # Exit a parse tree produced by ModelicaParser#stored_definition_class.
 *     def exitStored_definition_class(self, ctx:ModelicaParser.Stored_definition_classContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_7exitStored_definition_class(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_7exitStored_definition_class = {"exitStored_definition_class", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_7exitStored_definition_class, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_7exitStored_definition_class(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStored_definition_class (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitStored_definition_class", 1, 2, 2, 1); __PYX_ERR(0, 25, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitStored_definition_class") < 0)) __PYX_ERR(0, 25, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitStored_definition_class", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 25, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitStored_definition_class", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_6exitStored_definition_class(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_6exitStored_definition_class(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStored_definition_class", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":30
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_definition.
 *     def enterClass_definition(self, ctx:ModelicaParser.Class_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_9enterClass_definition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_9enterClass_definition = {"enterClass_definition", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_9enterClass_definition, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_9enterClass_definition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_definition (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterClass_definition", 1, 2, 2, 1); __PYX_ERR(0, 30, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterClass_definition") < 0)) __PYX_ERR(0, 30, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterClass_definition", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 30, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterClass_definition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_8enterClass_definition(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_8enterClass_definition(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_definition", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":34
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_definition.
 *     def exitClass_definition(self, ctx:ModelicaParser.Class_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_11exitClass_definition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_11exitClass_definition = {"exitClass_definition", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_11exitClass_definition, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_11exitClass_definition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_definition (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitClass_definition", 1, 2, 2, 1); __PYX_ERR(0, 34, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitClass_definition") < 0)) __PYX_ERR(0, 34, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitClass_definition", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 34, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitClass_definition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_10exitClass_definition(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_10exitClass_definition(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_definition", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":39
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_prefixes.
 *     def enterClass_prefixes(self, ctx:ModelicaParser.Class_prefixesContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_13enterClass_prefixes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_13enterClass_prefixes = {"enterClass_prefixes", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_13enterClass_prefixes, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_13enterClass_prefixes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_prefixes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterClass_prefixes", 1, 2, 2, 1); __PYX_ERR(0, 39, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterClass_prefixes") < 0)) __PYX_ERR(0, 39, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterClass_prefixes", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 39, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterClass_prefixes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_12enterClass_prefixes(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_12enterClass_prefixes(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_prefixes", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":43
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_prefixes.
 *     def exitClass_prefixes(self, ctx:ModelicaParser.Class_prefixesContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_15exitClass_prefixes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_15exitClass_prefixes = {"exitClass_prefixes", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_15exitClass_prefixes, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_15exitClass_prefixes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_prefixes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitClass_prefixes", 1, 2, 2, 1); __PYX_ERR(0, 43, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitClass_prefixes") < 0)) __PYX_ERR(0, 43, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitClass_prefixes", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 43, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitClass_prefixes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_14exitClass_prefixes(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_14exitClass_prefixes(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_prefixes", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":48
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_type.
 *     def enterClass_type(self, ctx:ModelicaParser.Class_typeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_17enterClass_type(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_17enterClass_type = {"enterClass_type", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_17enterClass_type, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_17enterClass_type(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_type (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterClass_type", 1, 2, 2, 1); __PYX_ERR(0, 48, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterClass_type") < 0)) __PYX_ERR(0, 48, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterClass_type", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 48, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterClass_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_16enterClass_type(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_16enterClass_type(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_type", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":52
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_type.
 *     def exitClass_type(self, ctx:ModelicaParser.Class_typeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_19exitClass_type(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_19exitClass_type = {"exitClass_type", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_19exitClass_type, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_19exitClass_type(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_type (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitClass_type", 1, 2, 2, 1); __PYX_ERR(0, 52, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitClass_type") < 0)) __PYX_ERR(0, 52, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitClass_type", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 52, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitClass_type", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_18exitClass_type(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_18exitClass_type(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_type", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":57
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_spec_comp.
 *     def enterClass_spec_comp(self, ctx:ModelicaParser.Class_spec_compContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_21enterClass_spec_comp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_21enterClass_spec_comp = {"enterClass_spec_comp", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_21enterClass_spec_comp, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_21enterClass_spec_comp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_spec_comp (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterClass_spec_comp", 1, 2, 2, 1); __PYX_ERR(0, 57, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterClass_spec_comp") < 0)) __PYX_ERR(0, 57, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterClass_spec_comp", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 57, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterClass_spec_comp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_20enterClass_spec_comp(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_20enterClass_spec_comp(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_spec_comp", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":61
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_spec_comp.
 *     def exitClass_spec_comp(self, ctx:ModelicaParser.Class_spec_compContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_23exitClass_spec_comp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_23exitClass_spec_comp = {"exitClass_spec_comp", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_23exitClass_spec_comp, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_23exitClass_spec_comp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_spec_comp (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitClass_spec_comp", 1, 2, 2, 1); __PYX_ERR(0, 61, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitClass_spec_comp") < 0)) __PYX_ERR(0, 61, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitClass_spec_comp", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 61, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitClass_spec_comp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_22exitClass_spec_comp(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_22exitClass_spec_comp(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_spec_comp", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":66
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_spec_base.
 *     def enterClass_spec_base(self, ctx:ModelicaParser.Class_spec_baseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_25enterClass_spec_base(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_25enterClass_spec_base = {"enterClass_spec_base", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_25enterClass_spec_base, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_25enterClass_spec_base(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_spec_base (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterClass_spec_base", 1, 2, 2, 1); __PYX_ERR(0, 66, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterClass_spec_base") < 0)) __PYX_ERR(0, 66, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterClass_spec_base", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 66, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterClass_spec_base", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_24enterClass_spec_base(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_24enterClass_spec_base(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_spec_base", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":70
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_spec_base.
 *     def exitClass_spec_base(self, ctx:ModelicaParser.Class_spec_baseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_27exitClass_spec_base(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_27exitClass_spec_base = {"exitClass_spec_base", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_27exitClass_spec_base, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_27exitClass_spec_base(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_spec_base (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitClass_spec_base", 1, 2, 2, 1); __PYX_ERR(0, 70, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitClass_spec_base") < 0)) __PYX_ERR(0, 70, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitClass_spec_base", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 70, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitClass_spec_base", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_26exitClass_spec_base(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_26exitClass_spec_base(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_spec_base", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":75
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_spec_enum.
 *     def enterClass_spec_enum(self, ctx:ModelicaParser.Class_spec_enumContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_29enterClass_spec_enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_29enterClass_spec_enum = {"enterClass_spec_enum", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_29enterClass_spec_enum, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_29enterClass_spec_enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_spec_enum (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterClass_spec_enum", 1, 2, 2, 1); __PYX_ERR(0, 75, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterClass_spec_enum") < 0)) __PYX_ERR(0, 75, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterClass_spec_enum", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 75, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterClass_spec_enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_28enterClass_spec_enum(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_28enterClass_spec_enum(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_spec_enum", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":79
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_spec_enum.
 *     def exitClass_spec_enum(self, ctx:ModelicaParser.Class_spec_enumContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_31exitClass_spec_enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_31exitClass_spec_enum = {"exitClass_spec_enum", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_31exitClass_spec_enum, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_31exitClass_spec_enum(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_spec_enum (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitClass_spec_enum", 1, 2, 2, 1); __PYX_ERR(0, 79, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitClass_spec_enum") < 0)) __PYX_ERR(0, 79, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitClass_spec_enum", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 79, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitClass_spec_enum", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_30exitClass_spec_enum(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_30exitClass_spec_enum(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_spec_enum", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":84
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_spec_der.
 *     def enterClass_spec_der(self, ctx:ModelicaParser.Class_spec_derContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_33enterClass_spec_der(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_33enterClass_spec_der = {"enterClass_spec_der", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_33enterClass_spec_der, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_33enterClass_spec_der(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_spec_der (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterClass_spec_der", 1, 2, 2, 1); __PYX_ERR(0, 84, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterClass_spec_der") < 0)) __PYX_ERR(0, 84, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterClass_spec_der", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 84, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterClass_spec_der", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_32enterClass_spec_der(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_32enterClass_spec_der(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_spec_der", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":88
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_spec_der.
 *     def exitClass_spec_der(self, ctx:ModelicaParser.Class_spec_derContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_35exitClass_spec_der(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_35exitClass_spec_der = {"exitClass_spec_der", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_35exitClass_spec_der, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_35exitClass_spec_der(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_spec_der (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitClass_spec_der", 1, 2, 2, 1); __PYX_ERR(0, 88, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitClass_spec_der") < 0)) __PYX_ERR(0, 88, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitClass_spec_der", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 88, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitClass_spec_der", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_34exitClass_spec_der(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_34exitClass_spec_der(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_spec_der", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":93
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_spec_extends.
 *     def enterClass_spec_extends(self, ctx:ModelicaParser.Class_spec_extendsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_37enterClass_spec_extends(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_37enterClass_spec_extends = {"enterClass_spec_extends", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_37enterClass_spec_extends, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_37enterClass_spec_extends(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_spec_extends (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterClass_spec_extends", 1, 2, 2, 1); __PYX_ERR(0, 93, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterClass_spec_extends") < 0)) __PYX_ERR(0, 93, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterClass_spec_extends", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 93, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterClass_spec_extends", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_36enterClass_spec_extends(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_36enterClass_spec_extends(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_spec_extends", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":97
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_spec_extends.
 *     def exitClass_spec_extends(self, ctx:ModelicaParser.Class_spec_extendsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_39exitClass_spec_extends(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_39exitClass_spec_extends = {"exitClass_spec_extends", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_39exitClass_spec_extends, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_39exitClass_spec_extends(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_spec_extends (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitClass_spec_extends", 1, 2, 2, 1); __PYX_ERR(0, 97, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitClass_spec_extends") < 0)) __PYX_ERR(0, 97, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitClass_spec_extends", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 97, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitClass_spec_extends", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_38exitClass_spec_extends(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_38exitClass_spec_extends(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_spec_extends", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":102
 * 
 *     # Enter a parse tree produced by ModelicaParser#base_prefix.
 *     def enterBase_prefix(self, ctx:ModelicaParser.Base_prefixContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_41enterBase_prefix(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_41enterBase_prefix = {"enterBase_prefix", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_41enterBase_prefix, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_41enterBase_prefix(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterBase_prefix (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterBase_prefix", 1, 2, 2, 1); __PYX_ERR(0, 102, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterBase_prefix") < 0)) __PYX_ERR(0, 102, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterBase_prefix", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 102, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterBase_prefix", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_40enterBase_prefix(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_40enterBase_prefix(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterBase_prefix", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":106
 * 
 *     # Exit a parse tree produced by ModelicaParser#base_prefix.
 *     def exitBase_prefix(self, ctx:ModelicaParser.Base_prefixContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_43exitBase_prefix(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_43exitBase_prefix = {"exitBase_prefix", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_43exitBase_prefix, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_43exitBase_prefix(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitBase_prefix (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitBase_prefix", 1, 2, 2, 1); __PYX_ERR(0, 106, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitBase_prefix") < 0)) __PYX_ERR(0, 106, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitBase_prefix", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 106, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitBase_prefix", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_42exitBase_prefix(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_42exitBase_prefix(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitBase_prefix", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":111
 * 
 *     # Enter a parse tree produced by ModelicaParser#enum_list.
 *     def enterEnum_list(self, ctx:ModelicaParser.Enum_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_45enterEnum_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_45enterEnum_list = {"enterEnum_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_45enterEnum_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_45enterEnum_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEnum_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterEnum_list", 1, 2, 2, 1); __PYX_ERR(0, 111, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterEnum_list") < 0)) __PYX_ERR(0, 111, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterEnum_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 111, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterEnum_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_44enterEnum_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_44enterEnum_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEnum_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":115
 * 
 *     # Exit a parse tree produced by ModelicaParser#enum_list.
 *     def exitEnum_list(self, ctx:ModelicaParser.Enum_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_47exitEnum_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_47exitEnum_list = {"exitEnum_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_47exitEnum_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_47exitEnum_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEnum_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitEnum_list", 1, 2, 2, 1); __PYX_ERR(0, 115, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitEnum_list") < 0)) __PYX_ERR(0, 115, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitEnum_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 115, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitEnum_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_46exitEnum_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_46exitEnum_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEnum_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":120
 * 
 *     # Enter a parse tree produced by ModelicaParser#enumeration_literal.
 *     def enterEnumeration_literal(self, ctx:ModelicaParser.Enumeration_literalContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_49enterEnumeration_literal(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_49enterEnumeration_literal = {"enterEnumeration_literal", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_49enterEnumeration_literal, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_49enterEnumeration_literal(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEnumeration_literal (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterEnumeration_literal", 1, 2, 2, 1); __PYX_ERR(0, 120, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterEnumeration_literal") < 0)) __PYX_ERR(0, 120, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterEnumeration_literal", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 120, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterEnumeration_literal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_48enterEnumeration_literal(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_48enterEnumeration_literal(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEnumeration_literal", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":124
 * 
 *     # Exit a parse tree produced by ModelicaParser#enumeration_literal.
 *     def exitEnumeration_literal(self, ctx:ModelicaParser.Enumeration_literalContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_51exitEnumeration_literal(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_51exitEnumeration_literal = {"exitEnumeration_literal", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_51exitEnumeration_literal, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_51exitEnumeration_literal(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEnumeration_literal (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitEnumeration_literal", 1, 2, 2, 1); __PYX_ERR(0, 124, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitEnumeration_literal") < 0)) __PYX_ERR(0, 124, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitEnumeration_literal", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 124, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitEnumeration_literal", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_50exitEnumeration_literal(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_50exitEnumeration_literal(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEnumeration_literal", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":129
 * 
 *     # Enter a parse tree produced by ModelicaParser#composition.
 *     def enterComposition(self, ctx:ModelicaParser.CompositionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_53enterComposition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_53enterComposition = {"enterComposition", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_53enterComposition, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_53enterComposition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComposition (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterComposition", 1, 2, 2, 1); __PYX_ERR(0, 129, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterComposition") < 0)) __PYX_ERR(0, 129, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterComposition", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 129, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterComposition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_52enterComposition(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_52enterComposition(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComposition", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":133
 * 
 *     # Exit a parse tree produced by ModelicaParser#composition.
 *     def exitComposition(self, ctx:ModelicaParser.CompositionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_55exitComposition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_55exitComposition = {"exitComposition", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_55exitComposition, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_55exitComposition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComposition (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitComposition", 1, 2, 2, 1); __PYX_ERR(0, 133, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitComposition") < 0)) __PYX_ERR(0, 133, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitComposition", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 133, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitComposition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_54exitComposition(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_54exitComposition(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComposition", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":138
 * 
 *     # Enter a parse tree produced by ModelicaParser#language_specification.
 *     def enterLanguage_specification(self, ctx:ModelicaParser.Language_specificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_57enterLanguage_specification(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_57enterLanguage_specification = {"enterLanguage_specification", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_57enterLanguage_specification, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_57enterLanguage_specification(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterLanguage_specification (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterLanguage_specification", 1, 2, 2, 1); __PYX_ERR(0, 138, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterLanguage_specification") < 0)) __PYX_ERR(0, 138, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterLanguage_specification", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 138, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterLanguage_specification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_56enterLanguage_specification(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_56enterLanguage_specification(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterLanguage_specification", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":142
 * 
 *     # Exit a parse tree produced by ModelicaParser#language_specification.
 *     def exitLanguage_specification(self, ctx:ModelicaParser.Language_specificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_59exitLanguage_specification(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_59exitLanguage_specification = {"exitLanguage_specification", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_59exitLanguage_specification, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_59exitLanguage_specification(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitLanguage_specification (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitLanguage_specification", 1, 2, 2, 1); __PYX_ERR(0, 142, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitLanguage_specification") < 0)) __PYX_ERR(0, 142, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitLanguage_specification", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 142, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitLanguage_specification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_58exitLanguage_specification(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_58exitLanguage_specification(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitLanguage_specification", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":147
 * 
 *     # Enter a parse tree produced by ModelicaParser#external_function_call.
 *     def enterExternal_function_call(self, ctx:ModelicaParser.External_function_callContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_61enterExternal_function_call(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_61enterExternal_function_call = {"enterExternal_function_call", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_61enterExternal_function_call, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_61enterExternal_function_call(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExternal_function_call (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterExternal_function_call", 1, 2, 2, 1); __PYX_ERR(0, 147, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterExternal_function_call") < 0)) __PYX_ERR(0, 147, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterExternal_function_call", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 147, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterExternal_function_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_60enterExternal_function_call(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_60enterExternal_function_call(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExternal_function_call", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":151
 * 
 *     # Exit a parse tree produced by ModelicaParser#external_function_call.
 *     def exitExternal_function_call(self, ctx:ModelicaParser.External_function_callContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_63exitExternal_function_call(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_63exitExternal_function_call = {"exitExternal_function_call", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_63exitExternal_function_call, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_63exitExternal_function_call(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExternal_function_call (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitExternal_function_call", 1, 2, 2, 1); __PYX_ERR(0, 151, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitExternal_function_call") < 0)) __PYX_ERR(0, 151, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitExternal_function_call", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 151, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitExternal_function_call", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_62exitExternal_function_call(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_62exitExternal_function_call(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExternal_function_call", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":156
 * 
 *     # Enter a parse tree produced by ModelicaParser#element_list.
 *     def enterElement_list(self, ctx:ModelicaParser.Element_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_65enterElement_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_65enterElement_list = {"enterElement_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_65enterElement_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_65enterElement_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterElement_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterElement_list", 1, 2, 2, 1); __PYX_ERR(0, 156, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterElement_list") < 0)) __PYX_ERR(0, 156, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterElement_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 156, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterElement_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_64enterElement_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_64enterElement_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterElement_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":160
 * 
 *     # Exit a parse tree produced by ModelicaParser#element_list.
 *     def exitElement_list(self, ctx:ModelicaParser.Element_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_67exitElement_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_67exitElement_list = {"exitElement_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_67exitElement_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_67exitElement_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitElement_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitElement_list", 1, 2, 2, 1); __PYX_ERR(0, 160, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitElement_list") < 0)) __PYX_ERR(0, 160, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitElement_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 160, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitElement_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_66exitElement_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_66exitElement_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitElement_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":165
 * 
 *     # Enter a parse tree produced by ModelicaParser#element.
 *     def enterElement(self, ctx:ModelicaParser.ElementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_69enterElement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_69enterElement = {"enterElement", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_69enterElement, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_69enterElement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterElement (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterElement", 1, 2, 2, 1); __PYX_ERR(0, 165, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterElement") < 0)) __PYX_ERR(0, 165, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterElement", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 165, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterElement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_68enterElement(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_68enterElement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterElement", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":169
 * 
 *     # Exit a parse tree produced by ModelicaParser#element.
 *     def exitElement(self, ctx:ModelicaParser.ElementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_71exitElement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_71exitElement = {"exitElement", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_71exitElement, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_71exitElement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitElement (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitElement", 1, 2, 2, 1); __PYX_ERR(0, 169, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitElement") < 0)) __PYX_ERR(0, 169, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitElement", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 169, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitElement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_70exitElement(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_70exitElement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitElement", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":174
 * 
 *     # Enter a parse tree produced by ModelicaParser#regular_element.
 *     def enterRegular_element(self, ctx:ModelicaParser.Regular_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_73enterRegular_element(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_73enterRegular_element = {"enterRegular_element", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_73enterRegular_element, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_73enterRegular_element(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRegular_element (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterRegular_element", 1, 2, 2, 1); __PYX_ERR(0, 174, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterRegular_element") < 0)) __PYX_ERR(0, 174, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterRegular_element", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 174, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterRegular_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_72enterRegular_element(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_72enterRegular_element(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterRegular_element", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":178
 * 
 *     # Exit a parse tree produced by ModelicaParser#regular_element.
 *     def exitRegular_element(self, ctx:ModelicaParser.Regular_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_75exitRegular_element(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_75exitRegular_element = {"exitRegular_element", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_75exitRegular_element, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_75exitRegular_element(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRegular_element (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitRegular_element", 1, 2, 2, 1); __PYX_ERR(0, 178, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitRegular_element") < 0)) __PYX_ERR(0, 178, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitRegular_element", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 178, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitRegular_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_74exitRegular_element(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_74exitRegular_element(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitRegular_element", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":183
 * 
 *     # Enter a parse tree produced by ModelicaParser#replaceable_element.
 *     def enterReplaceable_element(self, ctx:ModelicaParser.Replaceable_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_77enterReplaceable_element(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_77enterReplaceable_element = {"enterReplaceable_element", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_77enterReplaceable_element, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_77enterReplaceable_element(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterReplaceable_element (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterReplaceable_element", 1, 2, 2, 1); __PYX_ERR(0, 183, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterReplaceable_element") < 0)) __PYX_ERR(0, 183, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterReplaceable_element", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 183, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterReplaceable_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_76enterReplaceable_element(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_76enterReplaceable_element(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterReplaceable_element", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":187
 * 
 *     # Exit a parse tree produced by ModelicaParser#replaceable_element.
 *     def exitReplaceable_element(self, ctx:ModelicaParser.Replaceable_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_79exitReplaceable_element(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_79exitReplaceable_element = {"exitReplaceable_element", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_79exitReplaceable_element, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_79exitReplaceable_element(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitReplaceable_element (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitReplaceable_element", 1, 2, 2, 1); __PYX_ERR(0, 187, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitReplaceable_element") < 0)) __PYX_ERR(0, 187, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitReplaceable_element", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 187, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitReplaceable_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_78exitReplaceable_element(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_78exitReplaceable_element(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitReplaceable_element", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":192
 * 
 *     # Enter a parse tree produced by ModelicaParser#import_clause.
 *     def enterImport_clause(self, ctx:ModelicaParser.Import_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_81enterImport_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_81enterImport_clause = {"enterImport_clause", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_81enterImport_clause, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_81enterImport_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterImport_clause (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterImport_clause", 1, 2, 2, 1); __PYX_ERR(0, 192, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterImport_clause") < 0)) __PYX_ERR(0, 192, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterImport_clause", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 192, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterImport_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_80enterImport_clause(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_80enterImport_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterImport_clause", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":196
 * 
 *     # Exit a parse tree produced by ModelicaParser#import_clause.
 *     def exitImport_clause(self, ctx:ModelicaParser.Import_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_83exitImport_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_83exitImport_clause = {"exitImport_clause", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_83exitImport_clause, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_83exitImport_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitImport_clause (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitImport_clause", 1, 2, 2, 1); __PYX_ERR(0, 196, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitImport_clause") < 0)) __PYX_ERR(0, 196, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitImport_clause", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 196, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitImport_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_82exitImport_clause(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_82exitImport_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitImport_clause", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":201
 * 
 *     # Enter a parse tree produced by ModelicaParser#import_list.
 *     def enterImport_list(self, ctx:ModelicaParser.Import_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_85enterImport_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_85enterImport_list = {"enterImport_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_85enterImport_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_85enterImport_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterImport_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterImport_list", 1, 2, 2, 1); __PYX_ERR(0, 201, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterImport_list") < 0)) __PYX_ERR(0, 201, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterImport_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 201, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterImport_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_84enterImport_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_84enterImport_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterImport_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":205
 * 
 *     # Exit a parse tree produced by ModelicaParser#import_list.
 *     def exitImport_list(self, ctx:ModelicaParser.Import_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_87exitImport_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_87exitImport_list = {"exitImport_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_87exitImport_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_87exitImport_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitImport_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitImport_list", 1, 2, 2, 1); __PYX_ERR(0, 205, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitImport_list") < 0)) __PYX_ERR(0, 205, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitImport_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 205, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitImport_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_86exitImport_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_86exitImport_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitImport_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":210
 * 
 *     # Enter a parse tree produced by ModelicaParser#extends_clause.
 *     def enterExtends_clause(self, ctx:ModelicaParser.Extends_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_89enterExtends_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_89enterExtends_clause = {"enterExtends_clause", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_89enterExtends_clause, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_89enterExtends_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExtends_clause (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterExtends_clause", 1, 2, 2, 1); __PYX_ERR(0, 210, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterExtends_clause") < 0)) __PYX_ERR(0, 210, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterExtends_clause", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 210, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterExtends_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_88enterExtends_clause(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_88enterExtends_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExtends_clause", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":214
 * 
 *     # Exit a parse tree produced by ModelicaParser#extends_clause.
 *     def exitExtends_clause(self, ctx:ModelicaParser.Extends_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_91exitExtends_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_91exitExtends_clause = {"exitExtends_clause", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_91exitExtends_clause, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_91exitExtends_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExtends_clause (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitExtends_clause", 1, 2, 2, 1); __PYX_ERR(0, 214, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitExtends_clause") < 0)) __PYX_ERR(0, 214, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitExtends_clause", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 214, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitExtends_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_90exitExtends_clause(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_90exitExtends_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExtends_clause", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":219
 * 
 *     # Enter a parse tree produced by ModelicaParser#constraining_clause.
 *     def enterConstraining_clause(self, ctx:ModelicaParser.Constraining_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_93enterConstraining_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_93enterConstraining_clause = {"enterConstraining_clause", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_93enterConstraining_clause, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_93enterConstraining_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterConstraining_clause (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterConstraining_clause", 1, 2, 2, 1); __PYX_ERR(0, 219, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterConstraining_clause") < 0)) __PYX_ERR(0, 219, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterConstraining_clause", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 219, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterConstraining_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_92enterConstraining_clause(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_92enterConstraining_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterConstraining_clause", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":223
 * 
 *     # Exit a parse tree produced by ModelicaParser#constraining_clause.
 *     def exitConstraining_clause(self, ctx:ModelicaParser.Constraining_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_95exitConstraining_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_95exitConstraining_clause = {"exitConstraining_clause", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_95exitConstraining_clause, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_95exitConstraining_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitConstraining_clause (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitConstraining_clause", 1, 2, 2, 1); __PYX_ERR(0, 223, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitConstraining_clause") < 0)) __PYX_ERR(0, 223, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitConstraining_clause", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 223, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitConstraining_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_94exitConstraining_clause(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_94exitConstraining_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitConstraining_clause", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":228
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_clause.
 *     def enterComponent_clause(self, ctx:ModelicaParser.Component_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_97enterComponent_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_97enterComponent_clause = {"enterComponent_clause", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_97enterComponent_clause, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_97enterComponent_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComponent_clause (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterComponent_clause", 1, 2, 2, 1); __PYX_ERR(0, 228, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterComponent_clause") < 0)) __PYX_ERR(0, 228, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterComponent_clause", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 228, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterComponent_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_96enterComponent_clause(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_96enterComponent_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComponent_clause", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":232
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_clause.
 *     def exitComponent_clause(self, ctx:ModelicaParser.Component_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_99exitComponent_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_99exitComponent_clause = {"exitComponent_clause", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_99exitComponent_clause, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_99exitComponent_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComponent_clause (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitComponent_clause", 1, 2, 2, 1); __PYX_ERR(0, 232, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitComponent_clause") < 0)) __PYX_ERR(0, 232, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitComponent_clause", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 232, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitComponent_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_98exitComponent_clause(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_98exitComponent_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComponent_clause", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":237
 * 
 *     # Enter a parse tree produced by ModelicaParser#type_prefix.
 *     def enterType_prefix(self, ctx:ModelicaParser.Type_prefixContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_101enterType_prefix(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_101enterType_prefix = {"enterType_prefix", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_101enterType_prefix, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_101enterType_prefix(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterType_prefix (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterType_prefix", 1, 2, 2, 1); __PYX_ERR(0, 237, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterType_prefix") < 0)) __PYX_ERR(0, 237, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterType_prefix", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 237, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterType_prefix", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_100enterType_prefix(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_100enterType_prefix(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterType_prefix", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":241
 * 
 *     # Exit a parse tree produced by ModelicaParser#type_prefix.
 *     def exitType_prefix(self, ctx:ModelicaParser.Type_prefixContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_103exitType_prefix(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_103exitType_prefix = {"exitType_prefix", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_103exitType_prefix, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_103exitType_prefix(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitType_prefix (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitType_prefix", 1, 2, 2, 1); __PYX_ERR(0, 241, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitType_prefix") < 0)) __PYX_ERR(0, 241, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitType_prefix", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 241, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitType_prefix", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_102exitType_prefix(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_102exitType_prefix(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitType_prefix", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":246
 * 
 *     # Enter a parse tree produced by ModelicaParser#type_specifier.
 *     def enterType_specifier(self, ctx:ModelicaParser.Type_specifierContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_105enterType_specifier(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_105enterType_specifier = {"enterType_specifier", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_105enterType_specifier, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_105enterType_specifier(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterType_specifier (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterType_specifier", 1, 2, 2, 1); __PYX_ERR(0, 246, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterType_specifier") < 0)) __PYX_ERR(0, 246, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterType_specifier", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 246, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterType_specifier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_104enterType_specifier(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_104enterType_specifier(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterType_specifier", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":250
 * 
 *     # Exit a parse tree produced by ModelicaParser#type_specifier.
 *     def exitType_specifier(self, ctx:ModelicaParser.Type_specifierContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_107exitType_specifier(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_107exitType_specifier = {"exitType_specifier", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_107exitType_specifier, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_107exitType_specifier(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitType_specifier (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitType_specifier", 1, 2, 2, 1); __PYX_ERR(0, 250, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitType_specifier") < 0)) __PYX_ERR(0, 250, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitType_specifier", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 250, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitType_specifier", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_106exitType_specifier(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_106exitType_specifier(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitType_specifier", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":255
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_list.
 *     def enterComponent_list(self, ctx:ModelicaParser.Component_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_109enterComponent_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_109enterComponent_list = {"enterComponent_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_109enterComponent_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_109enterComponent_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComponent_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterComponent_list", 1, 2, 2, 1); __PYX_ERR(0, 255, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterComponent_list") < 0)) __PYX_ERR(0, 255, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterComponent_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 255, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterComponent_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_108enterComponent_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_108enterComponent_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComponent_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":259
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_list.
 *     def exitComponent_list(self, ctx:ModelicaParser.Component_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_111exitComponent_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_111exitComponent_list = {"exitComponent_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_111exitComponent_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_111exitComponent_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComponent_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitComponent_list", 1, 2, 2, 1); __PYX_ERR(0, 259, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitComponent_list") < 0)) __PYX_ERR(0, 259, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitComponent_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 259, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitComponent_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_110exitComponent_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_110exitComponent_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComponent_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":264
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_declaration.
 *     def enterComponent_declaration(self, ctx:ModelicaParser.Component_declarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_113enterComponent_declaration(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_113enterComponent_declaration = {"enterComponent_declaration", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_113enterComponent_declaration, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_113enterComponent_declaration(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComponent_declaration (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterComponent_declaration", 1, 2, 2, 1); __PYX_ERR(0, 264, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterComponent_declaration") < 0)) __PYX_ERR(0, 264, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterComponent_declaration", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 264, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterComponent_declaration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_112enterComponent_declaration(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_112enterComponent_declaration(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComponent_declaration", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":268
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_declaration.
 *     def exitComponent_declaration(self, ctx:ModelicaParser.Component_declarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_115exitComponent_declaration(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_115exitComponent_declaration = {"exitComponent_declaration", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_115exitComponent_declaration, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_115exitComponent_declaration(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComponent_declaration (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitComponent_declaration", 1, 2, 2, 1); __PYX_ERR(0, 268, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitComponent_declaration") < 0)) __PYX_ERR(0, 268, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitComponent_declaration", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 268, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitComponent_declaration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_114exitComponent_declaration(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_114exitComponent_declaration(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComponent_declaration", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":273
 * 
 *     # Enter a parse tree produced by ModelicaParser#condition_attribute.
 *     def enterCondition_attribute(self, ctx:ModelicaParser.Condition_attributeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_117enterCondition_attribute(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_117enterCondition_attribute = {"enterCondition_attribute", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_117enterCondition_attribute, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_117enterCondition_attribute(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterCondition_attribute (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterCondition_attribute", 1, 2, 2, 1); __PYX_ERR(0, 273, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterCondition_attribute") < 0)) __PYX_ERR(0, 273, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterCondition_attribute", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 273, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterCondition_attribute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_116enterCondition_attribute(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_116enterCondition_attribute(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterCondition_attribute", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":277
 * 
 *     # Exit a parse tree produced by ModelicaParser#condition_attribute.
 *     def exitCondition_attribute(self, ctx:ModelicaParser.Condition_attributeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_119exitCondition_attribute(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_119exitCondition_attribute = {"exitCondition_attribute", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_119exitCondition_attribute, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_119exitCondition_attribute(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitCondition_attribute (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitCondition_attribute", 1, 2, 2, 1); __PYX_ERR(0, 277, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitCondition_attribute") < 0)) __PYX_ERR(0, 277, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitCondition_attribute", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 277, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitCondition_attribute", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_118exitCondition_attribute(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_118exitCondition_attribute(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitCondition_attribute", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":282
 * 
 *     # Enter a parse tree produced by ModelicaParser#declaration.
 *     def enterDeclaration(self, ctx:ModelicaParser.DeclarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_121enterDeclaration(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_121enterDeclaration = {"enterDeclaration", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_121enterDeclaration, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_121enterDeclaration(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterDeclaration (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterDeclaration", 1, 2, 2, 1); __PYX_ERR(0, 282, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterDeclaration") < 0)) __PYX_ERR(0, 282, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterDeclaration", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 282, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterDeclaration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_120enterDeclaration(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_120enterDeclaration(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterDeclaration", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":286
 * 
 *     # Exit a parse tree produced by ModelicaParser#declaration.
 *     def exitDeclaration(self, ctx:ModelicaParser.DeclarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_123exitDeclaration(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_123exitDeclaration = {"exitDeclaration", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_123exitDeclaration, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_123exitDeclaration(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitDeclaration (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitDeclaration", 1, 2, 2, 1); __PYX_ERR(0, 286, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitDeclaration") < 0)) __PYX_ERR(0, 286, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitDeclaration", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 286, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitDeclaration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_122exitDeclaration(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_122exitDeclaration(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitDeclaration", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":291
 * 
 *     # Enter a parse tree produced by ModelicaParser#modification_class.
 *     def enterModification_class(self, ctx:ModelicaParser.Modification_classContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_125enterModification_class(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_125enterModification_class = {"enterModification_class", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_125enterModification_class, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_125enterModification_class(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterModification_class (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterModification_class", 1, 2, 2, 1); __PYX_ERR(0, 291, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterModification_class") < 0)) __PYX_ERR(0, 291, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterModification_class", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 291, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterModification_class", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_124enterModification_class(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_124enterModification_class(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterModification_class", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":295
 * 
 *     # Exit a parse tree produced by ModelicaParser#modification_class.
 *     def exitModification_class(self, ctx:ModelicaParser.Modification_classContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_127exitModification_class(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_127exitModification_class = {"exitModification_class", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_127exitModification_class, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_127exitModification_class(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitModification_class (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitModification_class", 1, 2, 2, 1); __PYX_ERR(0, 295, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitModification_class") < 0)) __PYX_ERR(0, 295, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitModification_class", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 295, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitModification_class", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_126exitModification_class(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_126exitModification_class(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitModification_class", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":300
 * 
 *     # Enter a parse tree produced by ModelicaParser#modification_assignment.
 *     def enterModification_assignment(self, ctx:ModelicaParser.Modification_assignmentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_129enterModification_assignment(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_129enterModification_assignment = {"enterModification_assignment", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_129enterModification_assignment, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_129enterModification_assignment(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterModification_assignment (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterModification_assignment", 1, 2, 2, 1); __PYX_ERR(0, 300, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterModification_assignment") < 0)) __PYX_ERR(0, 300, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterModification_assignment", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 300, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterModification_assignment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_128enterModification_assignment(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_128enterModification_assignment(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterModification_assignment", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":304
 * 
 *     # Exit a parse tree produced by ModelicaParser#modification_assignment.
 *     def exitModification_assignment(self, ctx:ModelicaParser.Modification_assignmentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_131exitModification_assignment(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_131exitModification_assignment = {"exitModification_assignment", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_131exitModification_assignment, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_131exitModification_assignment(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitModification_assignment (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitModification_assignment", 1, 2, 2, 1); __PYX_ERR(0, 304, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitModification_assignment") < 0)) __PYX_ERR(0, 304, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitModification_assignment", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 304, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitModification_assignment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_130exitModification_assignment(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_130exitModification_assignment(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitModification_assignment", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":309
 * 
 *     # Enter a parse tree produced by ModelicaParser#modification_assignment2.
 *     def enterModification_assignment2(self, ctx:ModelicaParser.Modification_assignment2Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_133enterModification_assignment2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_133enterModification_assignment2 = {"enterModification_assignment2", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_133enterModification_assignment2, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_133enterModification_assignment2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterModification_assignment2 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterModification_assignment2", 1, 2, 2, 1); __PYX_ERR(0, 309, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterModification_assignment2") < 0)) __PYX_ERR(0, 309, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterModification_assignment2", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 309, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterModification_assignment2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_132enterModification_assignment2(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_132enterModification_assignment2(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterModification_assignment2", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":313
 * 
 *     # Exit a parse tree produced by ModelicaParser#modification_assignment2.
 *     def exitModification_assignment2(self, ctx:ModelicaParser.Modification_assignment2Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_135exitModification_assignment2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_135exitModification_assignment2 = {"exitModification_assignment2", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_135exitModification_assignment2, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_135exitModification_assignment2(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitModification_assignment2 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitModification_assignment2", 1, 2, 2, 1); __PYX_ERR(0, 313, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitModification_assignment2") < 0)) __PYX_ERR(0, 313, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitModification_assignment2", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 313, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitModification_assignment2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_134exitModification_assignment2(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_134exitModification_assignment2(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitModification_assignment2", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":318
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_modification.
 *     def enterClass_modification(self, ctx:ModelicaParser.Class_modificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_137enterClass_modification(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_137enterClass_modification = {"enterClass_modification", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_137enterClass_modification, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_137enterClass_modification(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_modification (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterClass_modification", 1, 2, 2, 1); __PYX_ERR(0, 318, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterClass_modification") < 0)) __PYX_ERR(0, 318, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterClass_modification", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 318, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterClass_modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_136enterClass_modification(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_136enterClass_modification(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterClass_modification", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":322
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_modification.
 *     def exitClass_modification(self, ctx:ModelicaParser.Class_modificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_139exitClass_modification(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_139exitClass_modification = {"exitClass_modification", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_139exitClass_modification, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_139exitClass_modification(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_modification (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitClass_modification", 1, 2, 2, 1); __PYX_ERR(0, 322, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitClass_modification") < 0)) __PYX_ERR(0, 322, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitClass_modification", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 322, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitClass_modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_138exitClass_modification(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_138exitClass_modification(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitClass_modification", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":327
 * 
 *     # Enter a parse tree produced by ModelicaParser#argument_list.
 *     def enterArgument_list(self, ctx:ModelicaParser.Argument_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_141enterArgument_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_141enterArgument_list = {"enterArgument_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_141enterArgument_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_141enterArgument_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterArgument_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterArgument_list", 1, 2, 2, 1); __PYX_ERR(0, 327, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterArgument_list") < 0)) __PYX_ERR(0, 327, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterArgument_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 327, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterArgument_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_140enterArgument_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_140enterArgument_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterArgument_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":331
 * 
 *     # Exit a parse tree produced by ModelicaParser#argument_list.
 *     def exitArgument_list(self, ctx:ModelicaParser.Argument_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_143exitArgument_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_143exitArgument_list = {"exitArgument_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_143exitArgument_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_143exitArgument_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitArgument_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitArgument_list", 1, 2, 2, 1); __PYX_ERR(0, 331, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitArgument_list") < 0)) __PYX_ERR(0, 331, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitArgument_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 331, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitArgument_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_142exitArgument_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_142exitArgument_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitArgument_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":336
 * 
 *     # Enter a parse tree produced by ModelicaParser#argument.
 *     def enterArgument(self, ctx:ModelicaParser.ArgumentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_145enterArgument(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_145enterArgument = {"enterArgument", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_145enterArgument, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_145enterArgument(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterArgument (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterArgument", 1, 2, 2, 1); __PYX_ERR(0, 336, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterArgument") < 0)) __PYX_ERR(0, 336, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterArgument", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 336, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterArgument", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_144enterArgument(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_144enterArgument(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterArgument", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":340
 * 
 *     # Exit a parse tree produced by ModelicaParser#argument.
 *     def exitArgument(self, ctx:ModelicaParser.ArgumentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_147exitArgument(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_147exitArgument = {"exitArgument", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_147exitArgument, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_147exitArgument(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitArgument (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitArgument", 1, 2, 2, 1); __PYX_ERR(0, 340, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitArgument") < 0)) __PYX_ERR(0, 340, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitArgument", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 340, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitArgument", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_146exitArgument(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_146exitArgument(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitArgument", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":345
 * 
 *     # Enter a parse tree produced by ModelicaParser#element_modification_or_replaceable.
 *     def enterElement_modification_or_replaceable(self, ctx:ModelicaParser.Element_modification_or_replaceableContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_149enterElement_modification_or_replaceable(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_149enterElement_modification_or_replaceable = {"enterElement_modification_or_replaceable", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_149enterElement_modification_or_replaceable, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_149enterElement_modification_or_replaceable(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterElement_modification_or_replaceable (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterElement_modification_or_replaceable", 1, 2, 2, 1); __PYX_ERR(0, 345, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterElement_modification_or_replaceable") < 0)) __PYX_ERR(0, 345, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterElement_modification_or_replaceable", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 345, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterElement_modification_or_replaceable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_148enterElement_modification_or_replaceable(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_148enterElement_modification_or_replaceable(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterElement_modification_or_replaceable", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":349
 * 
 *     # Exit a parse tree produced by ModelicaParser#element_modification_or_replaceable.
 *     def exitElement_modification_or_replaceable(self, ctx:ModelicaParser.Element_modification_or_replaceableContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_151exitElement_modification_or_replaceable(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_151exitElement_modification_or_replaceable = {"exitElement_modification_or_replaceable", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_151exitElement_modification_or_replaceable, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_151exitElement_modification_or_replaceable(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitElement_modification_or_replaceable (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitElement_modification_or_replaceable", 1, 2, 2, 1); __PYX_ERR(0, 349, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitElement_modification_or_replaceable") < 0)) __PYX_ERR(0, 349, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitElement_modification_or_replaceable", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 349, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitElement_modification_or_replaceable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_150exitElement_modification_or_replaceable(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_150exitElement_modification_or_replaceable(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitElement_modification_or_replaceable", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":354
 * 
 *     # Enter a parse tree produced by ModelicaParser#element_modification.
 *     def enterElement_modification(self, ctx:ModelicaParser.Element_modificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_153enterElement_modification(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_153enterElement_modification = {"enterElement_modification", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_153enterElement_modification, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_153enterElement_modification(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterElement_modification (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterElement_modification", 1, 2, 2, 1); __PYX_ERR(0, 354, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterElement_modification") < 0)) __PYX_ERR(0, 354, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterElement_modification", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 354, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterElement_modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_152enterElement_modification(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_152enterElement_modification(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterElement_modification", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":358
 * 
 *     # Exit a parse tree produced by ModelicaParser#element_modification.
 *     def exitElement_modification(self, ctx:ModelicaParser.Element_modificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_155exitElement_modification(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_155exitElement_modification = {"exitElement_modification", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_155exitElement_modification, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_155exitElement_modification(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitElement_modification (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitElement_modification", 1, 2, 2, 1); __PYX_ERR(0, 358, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitElement_modification") < 0)) __PYX_ERR(0, 358, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitElement_modification", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 358, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitElement_modification", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_154exitElement_modification(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_154exitElement_modification(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitElement_modification", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":363
 * 
 *     # Enter a parse tree produced by ModelicaParser#element_redeclaration.
 *     def enterElement_redeclaration(self, ctx:ModelicaParser.Element_redeclarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_157enterElement_redeclaration(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_157enterElement_redeclaration = {"enterElement_redeclaration", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_157enterElement_redeclaration, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_157enterElement_redeclaration(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterElement_redeclaration (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterElement_redeclaration", 1, 2, 2, 1); __PYX_ERR(0, 363, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterElement_redeclaration") < 0)) __PYX_ERR(0, 363, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterElement_redeclaration", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 363, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterElement_redeclaration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_156enterElement_redeclaration(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_156enterElement_redeclaration(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterElement_redeclaration", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":367
 * 
 *     # Exit a parse tree produced by ModelicaParser#element_redeclaration.
 *     def exitElement_redeclaration(self, ctx:ModelicaParser.Element_redeclarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_159exitElement_redeclaration(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_159exitElement_redeclaration = {"exitElement_redeclaration", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_159exitElement_redeclaration, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_159exitElement_redeclaration(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitElement_redeclaration (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitElement_redeclaration", 1, 2, 2, 1); __PYX_ERR(0, 367, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitElement_redeclaration") < 0)) __PYX_ERR(0, 367, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitElement_redeclaration", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 367, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitElement_redeclaration", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_158exitElement_redeclaration(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_158exitElement_redeclaration(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitElement_redeclaration", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":372
 * 
 *     # Enter a parse tree produced by ModelicaParser#element_replaceable.
 *     def enterElement_replaceable(self, ctx:ModelicaParser.Element_replaceableContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_161enterElement_replaceable(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_161enterElement_replaceable = {"enterElement_replaceable", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_161enterElement_replaceable, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_161enterElement_replaceable(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterElement_replaceable (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterElement_replaceable", 1, 2, 2, 1); __PYX_ERR(0, 372, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterElement_replaceable") < 0)) __PYX_ERR(0, 372, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterElement_replaceable", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 372, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterElement_replaceable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_160enterElement_replaceable(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_160enterElement_replaceable(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterElement_replaceable", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":376
 * 
 *     # Exit a parse tree produced by ModelicaParser#element_replaceable.
 *     def exitElement_replaceable(self, ctx:ModelicaParser.Element_replaceableContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_163exitElement_replaceable(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_163exitElement_replaceable = {"exitElement_replaceable", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_163exitElement_replaceable, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_163exitElement_replaceable(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitElement_replaceable (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitElement_replaceable", 1, 2, 2, 1); __PYX_ERR(0, 376, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitElement_replaceable") < 0)) __PYX_ERR(0, 376, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitElement_replaceable", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 376, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitElement_replaceable", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_162exitElement_replaceable(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_162exitElement_replaceable(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitElement_replaceable", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":381
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_clause1.
 *     def enterComponent_clause1(self, ctx:ModelicaParser.Component_clause1Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_165enterComponent_clause1(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_165enterComponent_clause1 = {"enterComponent_clause1", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_165enterComponent_clause1, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_165enterComponent_clause1(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComponent_clause1 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterComponent_clause1", 1, 2, 2, 1); __PYX_ERR(0, 381, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterComponent_clause1") < 0)) __PYX_ERR(0, 381, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterComponent_clause1", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 381, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterComponent_clause1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_164enterComponent_clause1(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_164enterComponent_clause1(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComponent_clause1", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":385
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_clause1.
 *     def exitComponent_clause1(self, ctx:ModelicaParser.Component_clause1Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_167exitComponent_clause1(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_167exitComponent_clause1 = {"exitComponent_clause1", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_167exitComponent_clause1, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_167exitComponent_clause1(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComponent_clause1 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitComponent_clause1", 1, 2, 2, 1); __PYX_ERR(0, 385, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitComponent_clause1") < 0)) __PYX_ERR(0, 385, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitComponent_clause1", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 385, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitComponent_clause1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_166exitComponent_clause1(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_166exitComponent_clause1(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComponent_clause1", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":390
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_declaration1.
 *     def enterComponent_declaration1(self, ctx:ModelicaParser.Component_declaration1Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_169enterComponent_declaration1(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_169enterComponent_declaration1 = {"enterComponent_declaration1", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_169enterComponent_declaration1, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_169enterComponent_declaration1(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComponent_declaration1 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterComponent_declaration1", 1, 2, 2, 1); __PYX_ERR(0, 390, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterComponent_declaration1") < 0)) __PYX_ERR(0, 390, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterComponent_declaration1", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 390, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterComponent_declaration1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_168enterComponent_declaration1(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_168enterComponent_declaration1(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComponent_declaration1", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":394
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_declaration1.
 *     def exitComponent_declaration1(self, ctx:ModelicaParser.Component_declaration1Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_171exitComponent_declaration1(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_171exitComponent_declaration1 = {"exitComponent_declaration1", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_171exitComponent_declaration1, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_171exitComponent_declaration1(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComponent_declaration1 (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitComponent_declaration1", 1, 2, 2, 1); __PYX_ERR(0, 394, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitComponent_declaration1") < 0)) __PYX_ERR(0, 394, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitComponent_declaration1", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 394, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitComponent_declaration1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_170exitComponent_declaration1(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_170exitComponent_declaration1(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComponent_declaration1", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":399
 * 
 *     # Enter a parse tree produced by ModelicaParser#short_class_definition.
 *     def enterShort_class_definition(self, ctx:ModelicaParser.Short_class_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_173enterShort_class_definition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_173enterShort_class_definition = {"enterShort_class_definition", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_173enterShort_class_definition, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_173enterShort_class_definition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterShort_class_definition (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterShort_class_definition", 1, 2, 2, 1); __PYX_ERR(0, 399, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterShort_class_definition") < 0)) __PYX_ERR(0, 399, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterShort_class_definition", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 399, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterShort_class_definition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_172enterShort_class_definition(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_172enterShort_class_definition(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterShort_class_definition", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":403
 * 
 *     # Exit a parse tree produced by ModelicaParser#short_class_definition.
 *     def exitShort_class_definition(self, ctx:ModelicaParser.Short_class_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_175exitShort_class_definition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_175exitShort_class_definition = {"exitShort_class_definition", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_175exitShort_class_definition, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_175exitShort_class_definition(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitShort_class_definition (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitShort_class_definition", 1, 2, 2, 1); __PYX_ERR(0, 403, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitShort_class_definition") < 0)) __PYX_ERR(0, 403, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitShort_class_definition", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 403, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitShort_class_definition", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_174exitShort_class_definition(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_174exitShort_class_definition(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitShort_class_definition", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":408
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_section.
 *     def enterEquation_section(self, ctx:ModelicaParser.Equation_sectionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_177enterEquation_section(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_177enterEquation_section = {"enterEquation_section", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_177enterEquation_section, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_177enterEquation_section(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEquation_section (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterEquation_section", 1, 2, 2, 1); __PYX_ERR(0, 408, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterEquation_section") < 0)) __PYX_ERR(0, 408, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterEquation_section", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 408, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterEquation_section", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_176enterEquation_section(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_176enterEquation_section(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEquation_section", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":412
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_section.
 *     def exitEquation_section(self, ctx:ModelicaParser.Equation_sectionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_179exitEquation_section(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_179exitEquation_section = {"exitEquation_section", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_179exitEquation_section, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_179exitEquation_section(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEquation_section (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitEquation_section", 1, 2, 2, 1); __PYX_ERR(0, 412, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitEquation_section") < 0)) __PYX_ERR(0, 412, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitEquation_section", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 412, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitEquation_section", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_178exitEquation_section(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_178exitEquation_section(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEquation_section", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":417
 * 
 *     # Enter a parse tree produced by ModelicaParser#algorithm_section.
 *     def enterAlgorithm_section(self, ctx:ModelicaParser.Algorithm_sectionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_181enterAlgorithm_section(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_181enterAlgorithm_section = {"enterAlgorithm_section", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_181enterAlgorithm_section, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_181enterAlgorithm_section(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterAlgorithm_section (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterAlgorithm_section", 1, 2, 2, 1); __PYX_ERR(0, 417, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterAlgorithm_section") < 0)) __PYX_ERR(0, 417, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterAlgorithm_section", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 417, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterAlgorithm_section", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_180enterAlgorithm_section(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_180enterAlgorithm_section(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterAlgorithm_section", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":421
 * 
 *     # Exit a parse tree produced by ModelicaParser#algorithm_section.
 *     def exitAlgorithm_section(self, ctx:ModelicaParser.Algorithm_sectionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_183exitAlgorithm_section(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_183exitAlgorithm_section = {"exitAlgorithm_section", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_183exitAlgorithm_section, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_183exitAlgorithm_section(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitAlgorithm_section (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitAlgorithm_section", 1, 2, 2, 1); __PYX_ERR(0, 421, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitAlgorithm_section") < 0)) __PYX_ERR(0, 421, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitAlgorithm_section", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 421, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitAlgorithm_section", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_182exitAlgorithm_section(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_182exitAlgorithm_section(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitAlgorithm_section", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":426
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_simple.
 *     def enterEquation_simple(self, ctx:ModelicaParser.Equation_simpleContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_185enterEquation_simple(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_185enterEquation_simple = {"enterEquation_simple", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_185enterEquation_simple, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_185enterEquation_simple(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEquation_simple (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterEquation_simple", 1, 2, 2, 1); __PYX_ERR(0, 426, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterEquation_simple") < 0)) __PYX_ERR(0, 426, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterEquation_simple", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 426, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterEquation_simple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_184enterEquation_simple(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_184enterEquation_simple(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEquation_simple", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":430
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_simple.
 *     def exitEquation_simple(self, ctx:ModelicaParser.Equation_simpleContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_187exitEquation_simple(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_187exitEquation_simple = {"exitEquation_simple", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_187exitEquation_simple, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_187exitEquation_simple(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEquation_simple (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitEquation_simple", 1, 2, 2, 1); __PYX_ERR(0, 430, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitEquation_simple") < 0)) __PYX_ERR(0, 430, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitEquation_simple", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 430, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitEquation_simple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_186exitEquation_simple(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_186exitEquation_simple(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEquation_simple", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":435
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_if.
 *     def enterEquation_if(self, ctx:ModelicaParser.Equation_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_189enterEquation_if(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_189enterEquation_if = {"enterEquation_if", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_189enterEquation_if, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_189enterEquation_if(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEquation_if (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterEquation_if", 1, 2, 2, 1); __PYX_ERR(0, 435, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterEquation_if") < 0)) __PYX_ERR(0, 435, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterEquation_if", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 435, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterEquation_if", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_188enterEquation_if(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_188enterEquation_if(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEquation_if", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":439
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_if.
 *     def exitEquation_if(self, ctx:ModelicaParser.Equation_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_191exitEquation_if(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_191exitEquation_if = {"exitEquation_if", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_191exitEquation_if, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_191exitEquation_if(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEquation_if (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitEquation_if", 1, 2, 2, 1); __PYX_ERR(0, 439, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitEquation_if") < 0)) __PYX_ERR(0, 439, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitEquation_if", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 439, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitEquation_if", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_190exitEquation_if(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_190exitEquation_if(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEquation_if", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":444
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_for.
 *     def enterEquation_for(self, ctx:ModelicaParser.Equation_forContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_193enterEquation_for(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_193enterEquation_for = {"enterEquation_for", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_193enterEquation_for, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_193enterEquation_for(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEquation_for (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterEquation_for", 1, 2, 2, 1); __PYX_ERR(0, 444, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterEquation_for") < 0)) __PYX_ERR(0, 444, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterEquation_for", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 444, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterEquation_for", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_192enterEquation_for(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_192enterEquation_for(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEquation_for", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":448
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_for.
 *     def exitEquation_for(self, ctx:ModelicaParser.Equation_forContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_195exitEquation_for(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_195exitEquation_for = {"exitEquation_for", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_195exitEquation_for, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_195exitEquation_for(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEquation_for (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitEquation_for", 1, 2, 2, 1); __PYX_ERR(0, 448, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitEquation_for") < 0)) __PYX_ERR(0, 448, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitEquation_for", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 448, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitEquation_for", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_194exitEquation_for(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_194exitEquation_for(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEquation_for", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":453
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_connect_clause.
 *     def enterEquation_connect_clause(self, ctx:ModelicaParser.Equation_connect_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_197enterEquation_connect_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_197enterEquation_connect_clause = {"enterEquation_connect_clause", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_197enterEquation_connect_clause, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_197enterEquation_connect_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEquation_connect_clause (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterEquation_connect_clause", 1, 2, 2, 1); __PYX_ERR(0, 453, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterEquation_connect_clause") < 0)) __PYX_ERR(0, 453, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterEquation_connect_clause", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 453, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterEquation_connect_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_196enterEquation_connect_clause(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_196enterEquation_connect_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEquation_connect_clause", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":457
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_connect_clause.
 *     def exitEquation_connect_clause(self, ctx:ModelicaParser.Equation_connect_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_199exitEquation_connect_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_199exitEquation_connect_clause = {"exitEquation_connect_clause", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_199exitEquation_connect_clause, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_199exitEquation_connect_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEquation_connect_clause (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitEquation_connect_clause", 1, 2, 2, 1); __PYX_ERR(0, 457, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitEquation_connect_clause") < 0)) __PYX_ERR(0, 457, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitEquation_connect_clause", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 457, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitEquation_connect_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_198exitEquation_connect_clause(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_198exitEquation_connect_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEquation_connect_clause", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":462
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_when.
 *     def enterEquation_when(self, ctx:ModelicaParser.Equation_whenContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_201enterEquation_when(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_201enterEquation_when = {"enterEquation_when", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_201enterEquation_when, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_201enterEquation_when(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEquation_when (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterEquation_when", 1, 2, 2, 1); __PYX_ERR(0, 462, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterEquation_when") < 0)) __PYX_ERR(0, 462, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterEquation_when", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 462, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterEquation_when", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_200enterEquation_when(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_200enterEquation_when(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEquation_when", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":466
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_when.
 *     def exitEquation_when(self, ctx:ModelicaParser.Equation_whenContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_203exitEquation_when(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_203exitEquation_when = {"exitEquation_when", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_203exitEquation_when, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_203exitEquation_when(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEquation_when (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitEquation_when", 1, 2, 2, 1); __PYX_ERR(0, 466, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitEquation_when") < 0)) __PYX_ERR(0, 466, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitEquation_when", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 466, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitEquation_when", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_202exitEquation_when(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_202exitEquation_when(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEquation_when", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":471
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_function.
 *     def enterEquation_function(self, ctx:ModelicaParser.Equation_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_205enterEquation_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_205enterEquation_function = {"enterEquation_function", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_205enterEquation_function, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_205enterEquation_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEquation_function (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterEquation_function", 1, 2, 2, 1); __PYX_ERR(0, 471, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterEquation_function") < 0)) __PYX_ERR(0, 471, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterEquation_function", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 471, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterEquation_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_204enterEquation_function(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_204enterEquation_function(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEquation_function", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":475
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_function.
 *     def exitEquation_function(self, ctx:ModelicaParser.Equation_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_207exitEquation_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_207exitEquation_function = {"exitEquation_function", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_207exitEquation_function, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_207exitEquation_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEquation_function (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitEquation_function", 1, 2, 2, 1); __PYX_ERR(0, 475, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitEquation_function") < 0)) __PYX_ERR(0, 475, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitEquation_function", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 475, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitEquation_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_206exitEquation_function(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_206exitEquation_function(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEquation_function", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":480
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation.
 *     def enterEquation(self, ctx:ModelicaParser.EquationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_209enterEquation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_209enterEquation = {"enterEquation", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_209enterEquation, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_209enterEquation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEquation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterEquation", 1, 2, 2, 1); __PYX_ERR(0, 480, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterEquation") < 0)) __PYX_ERR(0, 480, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterEquation", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 480, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterEquation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_208enterEquation(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_208enterEquation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterEquation", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":484
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation.
 *     def exitEquation(self, ctx:ModelicaParser.EquationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_211exitEquation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_211exitEquation = {"exitEquation", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_211exitEquation, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_211exitEquation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEquation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitEquation", 1, 2, 2, 1); __PYX_ERR(0, 484, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitEquation") < 0)) __PYX_ERR(0, 484, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitEquation", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 484, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitEquation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_210exitEquation(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_210exitEquation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitEquation", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":489
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_component_reference.
 *     def enterStatement_component_reference(self, ctx:ModelicaParser.Statement_component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_213enterStatement_component_reference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_213enterStatement_component_reference = {"enterStatement_component_reference", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_213enterStatement_component_reference, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_213enterStatement_component_reference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement_component_reference (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterStatement_component_reference", 1, 2, 2, 1); __PYX_ERR(0, 489, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterStatement_component_reference") < 0)) __PYX_ERR(0, 489, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterStatement_component_reference", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 489, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterStatement_component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_212enterStatement_component_reference(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_212enterStatement_component_reference(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement_component_reference", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":493
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_component_reference.
 *     def exitStatement_component_reference(self, ctx:ModelicaParser.Statement_component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_215exitStatement_component_reference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_215exitStatement_component_reference = {"exitStatement_component_reference", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_215exitStatement_component_reference, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_215exitStatement_component_reference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement_component_reference (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitStatement_component_reference", 1, 2, 2, 1); __PYX_ERR(0, 493, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitStatement_component_reference") < 0)) __PYX_ERR(0, 493, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitStatement_component_reference", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 493, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitStatement_component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_214exitStatement_component_reference(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_214exitStatement_component_reference(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement_component_reference", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":498
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_component_function.
 *     def enterStatement_component_function(self, ctx:ModelicaParser.Statement_component_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_217enterStatement_component_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_217enterStatement_component_function = {"enterStatement_component_function", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_217enterStatement_component_function, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_217enterStatement_component_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement_component_function (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterStatement_component_function", 1, 2, 2, 1); __PYX_ERR(0, 498, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterStatement_component_function") < 0)) __PYX_ERR(0, 498, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterStatement_component_function", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 498, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterStatement_component_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_216enterStatement_component_function(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_216enterStatement_component_function(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement_component_function", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":502
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_component_function.
 *     def exitStatement_component_function(self, ctx:ModelicaParser.Statement_component_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_219exitStatement_component_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_219exitStatement_component_function = {"exitStatement_component_function", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_219exitStatement_component_function, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_219exitStatement_component_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement_component_function (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitStatement_component_function", 1, 2, 2, 1); __PYX_ERR(0, 502, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitStatement_component_function") < 0)) __PYX_ERR(0, 502, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitStatement_component_function", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 502, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitStatement_component_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_218exitStatement_component_function(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_218exitStatement_component_function(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement_component_function", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":507
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_break.
 *     def enterStatement_break(self, ctx:ModelicaParser.Statement_breakContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_221enterStatement_break(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_221enterStatement_break = {"enterStatement_break", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_221enterStatement_break, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_221enterStatement_break(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement_break (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterStatement_break", 1, 2, 2, 1); __PYX_ERR(0, 507, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterStatement_break") < 0)) __PYX_ERR(0, 507, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterStatement_break", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 507, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterStatement_break", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_220enterStatement_break(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_220enterStatement_break(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement_break", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":511
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_break.
 *     def exitStatement_break(self, ctx:ModelicaParser.Statement_breakContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_223exitStatement_break(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_223exitStatement_break = {"exitStatement_break", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_223exitStatement_break, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_223exitStatement_break(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement_break (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitStatement_break", 1, 2, 2, 1); __PYX_ERR(0, 511, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitStatement_break") < 0)) __PYX_ERR(0, 511, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitStatement_break", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 511, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitStatement_break", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_222exitStatement_break(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_222exitStatement_break(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement_break", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":516
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_return.
 *     def enterStatement_return(self, ctx:ModelicaParser.Statement_returnContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_225enterStatement_return(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_225enterStatement_return = {"enterStatement_return", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_225enterStatement_return, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_225enterStatement_return(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement_return (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterStatement_return", 1, 2, 2, 1); __PYX_ERR(0, 516, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterStatement_return") < 0)) __PYX_ERR(0, 516, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterStatement_return", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 516, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterStatement_return", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_224enterStatement_return(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_224enterStatement_return(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement_return", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":520
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_return.
 *     def exitStatement_return(self, ctx:ModelicaParser.Statement_returnContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_227exitStatement_return(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_227exitStatement_return = {"exitStatement_return", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_227exitStatement_return, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_227exitStatement_return(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement_return (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitStatement_return", 1, 2, 2, 1); __PYX_ERR(0, 520, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitStatement_return") < 0)) __PYX_ERR(0, 520, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitStatement_return", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 520, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitStatement_return", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_226exitStatement_return(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_226exitStatement_return(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement_return", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":525
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_if.
 *     def enterStatement_if(self, ctx:ModelicaParser.Statement_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_229enterStatement_if(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_229enterStatement_if = {"enterStatement_if", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_229enterStatement_if, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_229enterStatement_if(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement_if (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterStatement_if", 1, 2, 2, 1); __PYX_ERR(0, 525, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterStatement_if") < 0)) __PYX_ERR(0, 525, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterStatement_if", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 525, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterStatement_if", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_228enterStatement_if(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_228enterStatement_if(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement_if", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":529
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_if.
 *     def exitStatement_if(self, ctx:ModelicaParser.Statement_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_231exitStatement_if(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_231exitStatement_if = {"exitStatement_if", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_231exitStatement_if, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_231exitStatement_if(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement_if (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitStatement_if", 1, 2, 2, 1); __PYX_ERR(0, 529, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitStatement_if") < 0)) __PYX_ERR(0, 529, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitStatement_if", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 529, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitStatement_if", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_230exitStatement_if(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_230exitStatement_if(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement_if", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":534
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_for.
 *     def enterStatement_for(self, ctx:ModelicaParser.Statement_forContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_233enterStatement_for(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_233enterStatement_for = {"enterStatement_for", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_233enterStatement_for, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_233enterStatement_for(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement_for (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterStatement_for", 1, 2, 2, 1); __PYX_ERR(0, 534, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterStatement_for") < 0)) __PYX_ERR(0, 534, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterStatement_for", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 534, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterStatement_for", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_232enterStatement_for(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_232enterStatement_for(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement_for", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":538
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_for.
 *     def exitStatement_for(self, ctx:ModelicaParser.Statement_forContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_235exitStatement_for(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_235exitStatement_for = {"exitStatement_for", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_235exitStatement_for, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_235exitStatement_for(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement_for (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitStatement_for", 1, 2, 2, 1); __PYX_ERR(0, 538, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitStatement_for") < 0)) __PYX_ERR(0, 538, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitStatement_for", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 538, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitStatement_for", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_234exitStatement_for(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_234exitStatement_for(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement_for", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":543
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_while.
 *     def enterStatement_while(self, ctx:ModelicaParser.Statement_whileContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_237enterStatement_while(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_237enterStatement_while = {"enterStatement_while", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_237enterStatement_while, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_237enterStatement_while(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement_while (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterStatement_while", 1, 2, 2, 1); __PYX_ERR(0, 543, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterStatement_while") < 0)) __PYX_ERR(0, 543, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterStatement_while", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 543, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterStatement_while", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_236enterStatement_while(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_236enterStatement_while(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement_while", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":547
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_while.
 *     def exitStatement_while(self, ctx:ModelicaParser.Statement_whileContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_239exitStatement_while(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_239exitStatement_while = {"exitStatement_while", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_239exitStatement_while, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_239exitStatement_while(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement_while (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitStatement_while", 1, 2, 2, 1); __PYX_ERR(0, 547, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitStatement_while") < 0)) __PYX_ERR(0, 547, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitStatement_while", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 547, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitStatement_while", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_238exitStatement_while(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_238exitStatement_while(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement_while", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":552
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_when.
 *     def enterStatement_when(self, ctx:ModelicaParser.Statement_whenContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_241enterStatement_when(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_241enterStatement_when = {"enterStatement_when", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_241enterStatement_when, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_241enterStatement_when(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement_when (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterStatement_when", 1, 2, 2, 1); __PYX_ERR(0, 552, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterStatement_when") < 0)) __PYX_ERR(0, 552, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterStatement_when", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 552, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterStatement_when", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_240enterStatement_when(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_240enterStatement_when(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement_when", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":556
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_when.
 *     def exitStatement_when(self, ctx:ModelicaParser.Statement_whenContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_243exitStatement_when(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_243exitStatement_when = {"exitStatement_when", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_243exitStatement_when, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_243exitStatement_when(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement_when (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitStatement_when", 1, 2, 2, 1); __PYX_ERR(0, 556, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitStatement_when") < 0)) __PYX_ERR(0, 556, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitStatement_when", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 556, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitStatement_when", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_242exitStatement_when(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_242exitStatement_when(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement_when", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":561
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement.
 *     def enterStatement(self, ctx:ModelicaParser.StatementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_245enterStatement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_245enterStatement = {"enterStatement", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_245enterStatement, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_245enterStatement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterStatement", 1, 2, 2, 1); __PYX_ERR(0, 561, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterStatement") < 0)) __PYX_ERR(0, 561, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterStatement", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 561, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterStatement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_244enterStatement(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_244enterStatement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterStatement", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":565
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement.
 *     def exitStatement(self, ctx:ModelicaParser.StatementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_247exitStatement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_247exitStatement = {"exitStatement", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_247exitStatement, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_247exitStatement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitStatement", 1, 2, 2, 1); __PYX_ERR(0, 565, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitStatement") < 0)) __PYX_ERR(0, 565, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitStatement", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 565, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitStatement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_246exitStatement(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_246exitStatement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitStatement", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":570
 * 
 *     # Enter a parse tree produced by ModelicaParser#if_equation.
 *     def enterIf_equation(self, ctx:ModelicaParser.If_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_249enterIf_equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_249enterIf_equation = {"enterIf_equation", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_249enterIf_equation, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_249enterIf_equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterIf_equation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterIf_equation", 1, 2, 2, 1); __PYX_ERR(0, 570, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterIf_equation") < 0)) __PYX_ERR(0, 570, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterIf_equation", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 570, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterIf_equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_248enterIf_equation(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_248enterIf_equation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterIf_equation", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":574
 * 
 *     # Exit a parse tree produced by ModelicaParser#if_equation.
 *     def exitIf_equation(self, ctx:ModelicaParser.If_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_251exitIf_equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_251exitIf_equation = {"exitIf_equation", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_251exitIf_equation, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_251exitIf_equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitIf_equation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitIf_equation", 1, 2, 2, 1); __PYX_ERR(0, 574, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitIf_equation") < 0)) __PYX_ERR(0, 574, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitIf_equation", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 574, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitIf_equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_250exitIf_equation(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_250exitIf_equation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitIf_equation", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":579
 * 
 *     # Enter a parse tree produced by ModelicaParser#if_statement.
 *     def enterIf_statement(self, ctx:ModelicaParser.If_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_253enterIf_statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_253enterIf_statement = {"enterIf_statement", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_253enterIf_statement, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_253enterIf_statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterIf_statement (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterIf_statement", 1, 2, 2, 1); __PYX_ERR(0, 579, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterIf_statement") < 0)) __PYX_ERR(0, 579, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterIf_statement", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 579, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterIf_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_252enterIf_statement(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_252enterIf_statement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterIf_statement", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":583
 * 
 *     # Exit a parse tree produced by ModelicaParser#if_statement.
 *     def exitIf_statement(self, ctx:ModelicaParser.If_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_255exitIf_statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_255exitIf_statement = {"exitIf_statement", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_255exitIf_statement, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_255exitIf_statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitIf_statement (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitIf_statement", 1, 2, 2, 1); __PYX_ERR(0, 583, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitIf_statement") < 0)) __PYX_ERR(0, 583, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitIf_statement", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 583, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitIf_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_254exitIf_statement(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_254exitIf_statement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitIf_statement", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":588
 * 
 *     # Enter a parse tree produced by ModelicaParser#for_equation.
 *     def enterFor_equation(self, ctx:ModelicaParser.For_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_257enterFor_equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_257enterFor_equation = {"enterFor_equation", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_257enterFor_equation, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_257enterFor_equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterFor_equation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterFor_equation", 1, 2, 2, 1); __PYX_ERR(0, 588, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterFor_equation") < 0)) __PYX_ERR(0, 588, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterFor_equation", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 588, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterFor_equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_256enterFor_equation(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_256enterFor_equation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterFor_equation", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":592
 * 
 *     # Exit a parse tree produced by ModelicaParser#for_equation.
 *     def exitFor_equation(self, ctx:ModelicaParser.For_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_259exitFor_equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_259exitFor_equation = {"exitFor_equation", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_259exitFor_equation, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_259exitFor_equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitFor_equation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitFor_equation", 1, 2, 2, 1); __PYX_ERR(0, 592, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitFor_equation") < 0)) __PYX_ERR(0, 592, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitFor_equation", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 592, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitFor_equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_258exitFor_equation(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_258exitFor_equation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitFor_equation", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":597
 * 
 *     # Enter a parse tree produced by ModelicaParser#for_statement.
 *     def enterFor_statement(self, ctx:ModelicaParser.For_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_261enterFor_statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_261enterFor_statement = {"enterFor_statement", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_261enterFor_statement, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_261enterFor_statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterFor_statement (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterFor_statement", 1, 2, 2, 1); __PYX_ERR(0, 597, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterFor_statement") < 0)) __PYX_ERR(0, 597, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterFor_statement", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 597, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterFor_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_260enterFor_statement(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_260enterFor_statement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterFor_statement", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":601
 * 
 *     # Exit a parse tree produced by ModelicaParser#for_statement.
 *     def exitFor_statement(self, ctx:ModelicaParser.For_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_263exitFor_statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_263exitFor_statement = {"exitFor_statement", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_263exitFor_statement, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_263exitFor_statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitFor_statement (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitFor_statement", 1, 2, 2, 1); __PYX_ERR(0, 601, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitFor_statement") < 0)) __PYX_ERR(0, 601, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitFor_statement", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 601, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitFor_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_262exitFor_statement(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_262exitFor_statement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitFor_statement", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":606
 * 
 *     # Enter a parse tree produced by ModelicaParser#for_indices.
 *     def enterFor_indices(self, ctx:ModelicaParser.For_indicesContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_265enterFor_indices(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_265enterFor_indices = {"enterFor_indices", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_265enterFor_indices, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_265enterFor_indices(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterFor_indices (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterFor_indices", 1, 2, 2, 1); __PYX_ERR(0, 606, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterFor_indices") < 0)) __PYX_ERR(0, 606, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterFor_indices", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 606, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterFor_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_264enterFor_indices(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_264enterFor_indices(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterFor_indices", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":610
 * 
 *     # Exit a parse tree produced by ModelicaParser#for_indices.
 *     def exitFor_indices(self, ctx:ModelicaParser.For_indicesContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_267exitFor_indices(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_267exitFor_indices = {"exitFor_indices", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_267exitFor_indices, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_267exitFor_indices(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitFor_indices (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitFor_indices", 1, 2, 2, 1); __PYX_ERR(0, 610, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitFor_indices") < 0)) __PYX_ERR(0, 610, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitFor_indices", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 610, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitFor_indices", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_266exitFor_indices(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_266exitFor_indices(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitFor_indices", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":615
 * 
 *     # Enter a parse tree produced by ModelicaParser#for_index.
 *     def enterFor_index(self, ctx:ModelicaParser.For_indexContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_269enterFor_index(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_269enterFor_index = {"enterFor_index", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_269enterFor_index, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_269enterFor_index(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterFor_index (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterFor_index", 1, 2, 2, 1); __PYX_ERR(0, 615, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterFor_index") < 0)) __PYX_ERR(0, 615, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterFor_index", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 615, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterFor_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_268enterFor_index(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_268enterFor_index(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterFor_index", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":619
 * 
 *     # Exit a parse tree produced by ModelicaParser#for_index.
 *     def exitFor_index(self, ctx:ModelicaParser.For_indexContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_271exitFor_index(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_271exitFor_index = {"exitFor_index", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_271exitFor_index, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_271exitFor_index(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitFor_index (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitFor_index", 1, 2, 2, 1); __PYX_ERR(0, 619, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitFor_index") < 0)) __PYX_ERR(0, 619, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitFor_index", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 619, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitFor_index", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_270exitFor_index(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_270exitFor_index(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitFor_index", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":624
 * 
 *     # Enter a parse tree produced by ModelicaParser#while_statement.
 *     def enterWhile_statement(self, ctx:ModelicaParser.While_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_273enterWhile_statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_273enterWhile_statement = {"enterWhile_statement", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_273enterWhile_statement, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_273enterWhile_statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterWhile_statement (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterWhile_statement", 1, 2, 2, 1); __PYX_ERR(0, 624, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterWhile_statement") < 0)) __PYX_ERR(0, 624, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterWhile_statement", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 624, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterWhile_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_272enterWhile_statement(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_272enterWhile_statement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterWhile_statement", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":628
 * 
 *     # Exit a parse tree produced by ModelicaParser#while_statement.
 *     def exitWhile_statement(self, ctx:ModelicaParser.While_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_275exitWhile_statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_275exitWhile_statement = {"exitWhile_statement", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_275exitWhile_statement, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_275exitWhile_statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitWhile_statement (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitWhile_statement", 1, 2, 2, 1); __PYX_ERR(0, 628, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitWhile_statement") < 0)) __PYX_ERR(0, 628, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitWhile_statement", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 628, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitWhile_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_274exitWhile_statement(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_274exitWhile_statement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitWhile_statement", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":633
 * 
 *     # Enter a parse tree produced by ModelicaParser#when_equation.
 *     def enterWhen_equation(self, ctx:ModelicaParser.When_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_277enterWhen_equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_277enterWhen_equation = {"enterWhen_equation", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_277enterWhen_equation, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_277enterWhen_equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterWhen_equation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterWhen_equation", 1, 2, 2, 1); __PYX_ERR(0, 633, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterWhen_equation") < 0)) __PYX_ERR(0, 633, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterWhen_equation", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 633, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterWhen_equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_276enterWhen_equation(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_276enterWhen_equation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterWhen_equation", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":637
 * 
 *     # Exit a parse tree produced by ModelicaParser#when_equation.
 *     def exitWhen_equation(self, ctx:ModelicaParser.When_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_279exitWhen_equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_279exitWhen_equation = {"exitWhen_equation", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_279exitWhen_equation, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_279exitWhen_equation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitWhen_equation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitWhen_equation", 1, 2, 2, 1); __PYX_ERR(0, 637, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitWhen_equation") < 0)) __PYX_ERR(0, 637, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitWhen_equation", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 637, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitWhen_equation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_278exitWhen_equation(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_278exitWhen_equation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitWhen_equation", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":642
 * 
 *     # Enter a parse tree produced by ModelicaParser#when_statement.
 *     def enterWhen_statement(self, ctx:ModelicaParser.When_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_281enterWhen_statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_281enterWhen_statement = {"enterWhen_statement", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_281enterWhen_statement, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_281enterWhen_statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterWhen_statement (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterWhen_statement", 1, 2, 2, 1); __PYX_ERR(0, 642, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterWhen_statement") < 0)) __PYX_ERR(0, 642, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterWhen_statement", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 642, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterWhen_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_280enterWhen_statement(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_280enterWhen_statement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterWhen_statement", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":646
 * 
 *     # Exit a parse tree produced by ModelicaParser#when_statement.
 *     def exitWhen_statement(self, ctx:ModelicaParser.When_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_283exitWhen_statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_283exitWhen_statement = {"exitWhen_statement", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_283exitWhen_statement, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_283exitWhen_statement(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitWhen_statement (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitWhen_statement", 1, 2, 2, 1); __PYX_ERR(0, 646, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitWhen_statement") < 0)) __PYX_ERR(0, 646, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitWhen_statement", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 646, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitWhen_statement", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_282exitWhen_statement(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_282exitWhen_statement(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitWhen_statement", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":651
 * 
 *     # Enter a parse tree produced by ModelicaParser#connect_clause.
 *     def enterConnect_clause(self, ctx:ModelicaParser.Connect_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_285enterConnect_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_285enterConnect_clause = {"enterConnect_clause", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_285enterConnect_clause, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_285enterConnect_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterConnect_clause (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterConnect_clause", 1, 2, 2, 1); __PYX_ERR(0, 651, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterConnect_clause") < 0)) __PYX_ERR(0, 651, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterConnect_clause", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 651, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterConnect_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_284enterConnect_clause(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_284enterConnect_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterConnect_clause", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":655
 * 
 *     # Exit a parse tree produced by ModelicaParser#connect_clause.
 *     def exitConnect_clause(self, ctx:ModelicaParser.Connect_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_287exitConnect_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_287exitConnect_clause = {"exitConnect_clause", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_287exitConnect_clause, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_287exitConnect_clause(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitConnect_clause (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitConnect_clause", 1, 2, 2, 1); __PYX_ERR(0, 655, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitConnect_clause") < 0)) __PYX_ERR(0, 655, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitConnect_clause", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 655, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitConnect_clause", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_286exitConnect_clause(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_286exitConnect_clause(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitConnect_clause", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":660
 * 
 *     # Enter a parse tree produced by ModelicaParser#expression_simple.
 *     def enterExpression_simple(self, ctx:ModelicaParser.Expression_simpleContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_289enterExpression_simple(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_289enterExpression_simple = {"enterExpression_simple", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_289enterExpression_simple, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_289enterExpression_simple(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpression_simple (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterExpression_simple", 1, 2, 2, 1); __PYX_ERR(0, 660, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterExpression_simple") < 0)) __PYX_ERR(0, 660, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterExpression_simple", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 660, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterExpression_simple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_288enterExpression_simple(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_288enterExpression_simple(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpression_simple", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":664
 * 
 *     # Exit a parse tree produced by ModelicaParser#expression_simple.
 *     def exitExpression_simple(self, ctx:ModelicaParser.Expression_simpleContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_291exitExpression_simple(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_291exitExpression_simple = {"exitExpression_simple", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_291exitExpression_simple, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_291exitExpression_simple(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpression_simple (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitExpression_simple", 1, 2, 2, 1); __PYX_ERR(0, 664, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitExpression_simple") < 0)) __PYX_ERR(0, 664, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitExpression_simple", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 664, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitExpression_simple", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_290exitExpression_simple(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_290exitExpression_simple(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpression_simple", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":669
 * 
 *     # Enter a parse tree produced by ModelicaParser#expression_if.
 *     def enterExpression_if(self, ctx:ModelicaParser.Expression_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_293enterExpression_if(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_293enterExpression_if = {"enterExpression_if", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_293enterExpression_if, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_293enterExpression_if(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpression_if (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterExpression_if", 1, 2, 2, 1); __PYX_ERR(0, 669, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterExpression_if") < 0)) __PYX_ERR(0, 669, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterExpression_if", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 669, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterExpression_if", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_292enterExpression_if(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_292enterExpression_if(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpression_if", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":673
 * 
 *     # Exit a parse tree produced by ModelicaParser#expression_if.
 *     def exitExpression_if(self, ctx:ModelicaParser.Expression_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_295exitExpression_if(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_295exitExpression_if = {"exitExpression_if", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_295exitExpression_if, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_295exitExpression_if(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpression_if (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitExpression_if", 1, 2, 2, 1); __PYX_ERR(0, 673, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitExpression_if") < 0)) __PYX_ERR(0, 673, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitExpression_if", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 673, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitExpression_if", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_294exitExpression_if(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_294exitExpression_if(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpression_if", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":678
 * 
 *     # Enter a parse tree produced by ModelicaParser#simple_expression.
 *     def enterSimple_expression(self, ctx:ModelicaParser.Simple_expressionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_297enterSimple_expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_297enterSimple_expression = {"enterSimple_expression", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_297enterSimple_expression, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_297enterSimple_expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterSimple_expression (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterSimple_expression", 1, 2, 2, 1); __PYX_ERR(0, 678, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterSimple_expression") < 0)) __PYX_ERR(0, 678, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterSimple_expression", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 678, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterSimple_expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_296enterSimple_expression(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_296enterSimple_expression(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterSimple_expression", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":682
 * 
 *     # Exit a parse tree produced by ModelicaParser#simple_expression.
 *     def exitSimple_expression(self, ctx:ModelicaParser.Simple_expressionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_299exitSimple_expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_299exitSimple_expression = {"exitSimple_expression", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_299exitSimple_expression, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_299exitSimple_expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitSimple_expression (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitSimple_expression", 1, 2, 2, 1); __PYX_ERR(0, 682, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitSimple_expression") < 0)) __PYX_ERR(0, 682, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitSimple_expression", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 682, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitSimple_expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_298exitSimple_expression(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_298exitSimple_expression(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitSimple_expression", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":687
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_neg.
 *     def enterExpr_neg(self, ctx:ModelicaParser.Expr_negContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_301enterExpr_neg(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_301enterExpr_neg = {"enterExpr_neg", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_301enterExpr_neg, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_301enterExpr_neg(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_neg (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterExpr_neg", 1, 2, 2, 1); __PYX_ERR(0, 687, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterExpr_neg") < 0)) __PYX_ERR(0, 687, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterExpr_neg", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 687, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterExpr_neg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_300enterExpr_neg(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_300enterExpr_neg(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_neg", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":691
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_neg.
 *     def exitExpr_neg(self, ctx:ModelicaParser.Expr_negContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_303exitExpr_neg(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_303exitExpr_neg = {"exitExpr_neg", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_303exitExpr_neg, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_303exitExpr_neg(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_neg (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitExpr_neg", 1, 2, 2, 1); __PYX_ERR(0, 691, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitExpr_neg") < 0)) __PYX_ERR(0, 691, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitExpr_neg", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 691, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitExpr_neg", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_302exitExpr_neg(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_302exitExpr_neg(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_neg", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":696
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_add.
 *     def enterExpr_add(self, ctx:ModelicaParser.Expr_addContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_305enterExpr_add(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_305enterExpr_add = {"enterExpr_add", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_305enterExpr_add, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_305enterExpr_add(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_add (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterExpr_add", 1, 2, 2, 1); __PYX_ERR(0, 696, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterExpr_add") < 0)) __PYX_ERR(0, 696, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterExpr_add", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 696, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterExpr_add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_304enterExpr_add(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_304enterExpr_add(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_add", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":700
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_add.
 *     def exitExpr_add(self, ctx:ModelicaParser.Expr_addContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_307exitExpr_add(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_307exitExpr_add = {"exitExpr_add", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_307exitExpr_add, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_307exitExpr_add(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_add (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitExpr_add", 1, 2, 2, 1); __PYX_ERR(0, 700, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitExpr_add") < 0)) __PYX_ERR(0, 700, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitExpr_add", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 700, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitExpr_add", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_306exitExpr_add(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_306exitExpr_add(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_add", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":705
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_exp.
 *     def enterExpr_exp(self, ctx:ModelicaParser.Expr_expContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_309enterExpr_exp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_309enterExpr_exp = {"enterExpr_exp", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_309enterExpr_exp, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_309enterExpr_exp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_exp (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterExpr_exp", 1, 2, 2, 1); __PYX_ERR(0, 705, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterExpr_exp") < 0)) __PYX_ERR(0, 705, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterExpr_exp", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 705, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterExpr_exp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_308enterExpr_exp(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_308enterExpr_exp(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_exp", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":709
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_exp.
 *     def exitExpr_exp(self, ctx:ModelicaParser.Expr_expContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_311exitExpr_exp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_311exitExpr_exp = {"exitExpr_exp", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_311exitExpr_exp, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_311exitExpr_exp(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_exp (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitExpr_exp", 1, 2, 2, 1); __PYX_ERR(0, 709, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitExpr_exp") < 0)) __PYX_ERR(0, 709, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitExpr_exp", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 709, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitExpr_exp", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_310exitExpr_exp(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_310exitExpr_exp(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_exp", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":714
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_or.
 *     def enterExpr_or(self, ctx:ModelicaParser.Expr_orContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_313enterExpr_or(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_313enterExpr_or = {"enterExpr_or", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_313enterExpr_or, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_313enterExpr_or(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_or (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterExpr_or", 1, 2, 2, 1); __PYX_ERR(0, 714, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterExpr_or") < 0)) __PYX_ERR(0, 714, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterExpr_or", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 714, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterExpr_or", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_312enterExpr_or(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_312enterExpr_or(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_or", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":718
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_or.
 *     def exitExpr_or(self, ctx:ModelicaParser.Expr_orContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_315exitExpr_or(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_315exitExpr_or = {"exitExpr_or", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_315exitExpr_or, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_315exitExpr_or(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_or (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitExpr_or", 1, 2, 2, 1); __PYX_ERR(0, 718, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitExpr_or") < 0)) __PYX_ERR(0, 718, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitExpr_or", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 718, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitExpr_or", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_314exitExpr_or(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_314exitExpr_or(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_or", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":723
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_primary.
 *     def enterExpr_primary(self, ctx:ModelicaParser.Expr_primaryContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_317enterExpr_primary(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_317enterExpr_primary = {"enterExpr_primary", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_317enterExpr_primary, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_317enterExpr_primary(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_primary (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterExpr_primary", 1, 2, 2, 1); __PYX_ERR(0, 723, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterExpr_primary") < 0)) __PYX_ERR(0, 723, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterExpr_primary", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 723, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterExpr_primary", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_316enterExpr_primary(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_316enterExpr_primary(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_primary", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":727
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_primary.
 *     def exitExpr_primary(self, ctx:ModelicaParser.Expr_primaryContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_319exitExpr_primary(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_319exitExpr_primary = {"exitExpr_primary", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_319exitExpr_primary, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_319exitExpr_primary(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_primary (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitExpr_primary", 1, 2, 2, 1); __PYX_ERR(0, 727, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitExpr_primary") < 0)) __PYX_ERR(0, 727, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitExpr_primary", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 727, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitExpr_primary", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_318exitExpr_primary(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_318exitExpr_primary(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_primary", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":732
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_and.
 *     def enterExpr_and(self, ctx:ModelicaParser.Expr_andContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_321enterExpr_and(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_321enterExpr_and = {"enterExpr_and", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_321enterExpr_and, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_321enterExpr_and(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_and (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterExpr_and", 1, 2, 2, 1); __PYX_ERR(0, 732, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterExpr_and") < 0)) __PYX_ERR(0, 732, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterExpr_and", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 732, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterExpr_and", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_320enterExpr_and(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_320enterExpr_and(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_and", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":736
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_and.
 *     def exitExpr_and(self, ctx:ModelicaParser.Expr_andContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_323exitExpr_and(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_323exitExpr_and = {"exitExpr_and", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_323exitExpr_and, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_323exitExpr_and(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_and (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitExpr_and", 1, 2, 2, 1); __PYX_ERR(0, 736, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitExpr_and") < 0)) __PYX_ERR(0, 736, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitExpr_and", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 736, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitExpr_and", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_322exitExpr_and(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_322exitExpr_and(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_and", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":741
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_rel.
 *     def enterExpr_rel(self, ctx:ModelicaParser.Expr_relContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_325enterExpr_rel(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_325enterExpr_rel = {"enterExpr_rel", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_325enterExpr_rel, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_325enterExpr_rel(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_rel (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterExpr_rel", 1, 2, 2, 1); __PYX_ERR(0, 741, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterExpr_rel") < 0)) __PYX_ERR(0, 741, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterExpr_rel", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 741, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterExpr_rel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_324enterExpr_rel(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_324enterExpr_rel(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_rel", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":745
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_rel.
 *     def exitExpr_rel(self, ctx:ModelicaParser.Expr_relContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_327exitExpr_rel(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_327exitExpr_rel = {"exitExpr_rel", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_327exitExpr_rel, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_327exitExpr_rel(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_rel (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitExpr_rel", 1, 2, 2, 1); __PYX_ERR(0, 745, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitExpr_rel") < 0)) __PYX_ERR(0, 745, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitExpr_rel", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 745, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitExpr_rel", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_326exitExpr_rel(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_326exitExpr_rel(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_rel", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":750
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_not.
 *     def enterExpr_not(self, ctx:ModelicaParser.Expr_notContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_329enterExpr_not(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_329enterExpr_not = {"enterExpr_not", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_329enterExpr_not, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_329enterExpr_not(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_not (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterExpr_not", 1, 2, 2, 1); __PYX_ERR(0, 750, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterExpr_not") < 0)) __PYX_ERR(0, 750, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterExpr_not", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 750, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterExpr_not", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_328enterExpr_not(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_328enterExpr_not(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_not", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":754
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_not.
 *     def exitExpr_not(self, ctx:ModelicaParser.Expr_notContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_331exitExpr_not(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_331exitExpr_not = {"exitExpr_not", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_331exitExpr_not, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_331exitExpr_not(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_not (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitExpr_not", 1, 2, 2, 1); __PYX_ERR(0, 754, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitExpr_not") < 0)) __PYX_ERR(0, 754, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitExpr_not", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 754, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitExpr_not", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_330exitExpr_not(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_330exitExpr_not(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_not", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":759
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_mul.
 *     def enterExpr_mul(self, ctx:ModelicaParser.Expr_mulContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_333enterExpr_mul(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_333enterExpr_mul = {"enterExpr_mul", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_333enterExpr_mul, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_333enterExpr_mul(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_mul (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterExpr_mul", 1, 2, 2, 1); __PYX_ERR(0, 759, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterExpr_mul") < 0)) __PYX_ERR(0, 759, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterExpr_mul", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 759, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterExpr_mul", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_332enterExpr_mul(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_332enterExpr_mul(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpr_mul", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":763
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_mul.
 *     def exitExpr_mul(self, ctx:ModelicaParser.Expr_mulContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_335exitExpr_mul(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_335exitExpr_mul = {"exitExpr_mul", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_335exitExpr_mul, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_335exitExpr_mul(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_mul (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitExpr_mul", 1, 2, 2, 1); __PYX_ERR(0, 763, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitExpr_mul") < 0)) __PYX_ERR(0, 763, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitExpr_mul", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 763, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitExpr_mul", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_334exitExpr_mul(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_334exitExpr_mul(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpr_mul", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":768
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_unsigned_number.
 *     def enterPrimary_unsigned_number(self, ctx:ModelicaParser.Primary_unsigned_numberContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_337enterPrimary_unsigned_number(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_337enterPrimary_unsigned_number = {"enterPrimary_unsigned_number", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_337enterPrimary_unsigned_number, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_337enterPrimary_unsigned_number(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_unsigned_number (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterPrimary_unsigned_number", 1, 2, 2, 1); __PYX_ERR(0, 768, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterPrimary_unsigned_number") < 0)) __PYX_ERR(0, 768, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterPrimary_unsigned_number", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 768, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterPrimary_unsigned_number", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_336enterPrimary_unsigned_number(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_336enterPrimary_unsigned_number(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_unsigned_number", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":772
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_unsigned_number.
 *     def exitPrimary_unsigned_number(self, ctx:ModelicaParser.Primary_unsigned_numberContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_339exitPrimary_unsigned_number(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_339exitPrimary_unsigned_number = {"exitPrimary_unsigned_number", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_339exitPrimary_unsigned_number, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_339exitPrimary_unsigned_number(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_unsigned_number (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitPrimary_unsigned_number", 1, 2, 2, 1); __PYX_ERR(0, 772, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitPrimary_unsigned_number") < 0)) __PYX_ERR(0, 772, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitPrimary_unsigned_number", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 772, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitPrimary_unsigned_number", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_338exitPrimary_unsigned_number(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_338exitPrimary_unsigned_number(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_unsigned_number", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":777
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_string.
 *     def enterPrimary_string(self, ctx:ModelicaParser.Primary_stringContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_341enterPrimary_string(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_341enterPrimary_string = {"enterPrimary_string", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_341enterPrimary_string, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_341enterPrimary_string(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_string (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterPrimary_string", 1, 2, 2, 1); __PYX_ERR(0, 777, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterPrimary_string") < 0)) __PYX_ERR(0, 777, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterPrimary_string", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 777, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterPrimary_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_340enterPrimary_string(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_340enterPrimary_string(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_string", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":781
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_string.
 *     def exitPrimary_string(self, ctx:ModelicaParser.Primary_stringContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_343exitPrimary_string(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_343exitPrimary_string = {"exitPrimary_string", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_343exitPrimary_string, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_343exitPrimary_string(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_string (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitPrimary_string", 1, 2, 2, 1); __PYX_ERR(0, 781, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitPrimary_string") < 0)) __PYX_ERR(0, 781, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitPrimary_string", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 781, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitPrimary_string", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_342exitPrimary_string(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_342exitPrimary_string(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_string", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":786
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_false.
 *     def enterPrimary_false(self, ctx:ModelicaParser.Primary_falseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_345enterPrimary_false(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_345enterPrimary_false = {"enterPrimary_false", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_345enterPrimary_false, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_345enterPrimary_false(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_false (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterPrimary_false", 1, 2, 2, 1); __PYX_ERR(0, 786, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterPrimary_false") < 0)) __PYX_ERR(0, 786, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterPrimary_false", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 786, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterPrimary_false", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_344enterPrimary_false(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_344enterPrimary_false(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_false", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":790
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_false.
 *     def exitPrimary_false(self, ctx:ModelicaParser.Primary_falseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_347exitPrimary_false(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_347exitPrimary_false = {"exitPrimary_false", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_347exitPrimary_false, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_347exitPrimary_false(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_false (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitPrimary_false", 1, 2, 2, 1); __PYX_ERR(0, 790, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitPrimary_false") < 0)) __PYX_ERR(0, 790, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitPrimary_false", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 790, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitPrimary_false", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_346exitPrimary_false(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_346exitPrimary_false(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_false", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":795
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_true.
 *     def enterPrimary_true(self, ctx:ModelicaParser.Primary_trueContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_349enterPrimary_true(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_349enterPrimary_true = {"enterPrimary_true", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_349enterPrimary_true, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_349enterPrimary_true(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_true (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterPrimary_true", 1, 2, 2, 1); __PYX_ERR(0, 795, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterPrimary_true") < 0)) __PYX_ERR(0, 795, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterPrimary_true", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 795, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterPrimary_true", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_348enterPrimary_true(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_348enterPrimary_true(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_true", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":799
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_true.
 *     def exitPrimary_true(self, ctx:ModelicaParser.Primary_trueContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_351exitPrimary_true(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_351exitPrimary_true = {"exitPrimary_true", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_351exitPrimary_true, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_351exitPrimary_true(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_true (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitPrimary_true", 1, 2, 2, 1); __PYX_ERR(0, 799, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitPrimary_true") < 0)) __PYX_ERR(0, 799, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitPrimary_true", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 799, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitPrimary_true", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_350exitPrimary_true(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_350exitPrimary_true(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_true", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":804
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_function.
 *     def enterPrimary_function(self, ctx:ModelicaParser.Primary_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_353enterPrimary_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_353enterPrimary_function = {"enterPrimary_function", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_353enterPrimary_function, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_353enterPrimary_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_function (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterPrimary_function", 1, 2, 2, 1); __PYX_ERR(0, 804, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterPrimary_function") < 0)) __PYX_ERR(0, 804, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterPrimary_function", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 804, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterPrimary_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_352enterPrimary_function(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_352enterPrimary_function(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_function", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":808
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_function.
 *     def exitPrimary_function(self, ctx:ModelicaParser.Primary_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_355exitPrimary_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_355exitPrimary_function = {"exitPrimary_function", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_355exitPrimary_function, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_355exitPrimary_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_function (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitPrimary_function", 1, 2, 2, 1); __PYX_ERR(0, 808, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitPrimary_function") < 0)) __PYX_ERR(0, 808, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitPrimary_function", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 808, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitPrimary_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_354exitPrimary_function(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_354exitPrimary_function(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_function", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":813
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_derivative.
 *     def enterPrimary_derivative(self, ctx:ModelicaParser.Primary_derivativeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_357enterPrimary_derivative(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_357enterPrimary_derivative = {"enterPrimary_derivative", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_357enterPrimary_derivative, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_357enterPrimary_derivative(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_derivative (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterPrimary_derivative", 1, 2, 2, 1); __PYX_ERR(0, 813, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterPrimary_derivative") < 0)) __PYX_ERR(0, 813, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterPrimary_derivative", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 813, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterPrimary_derivative", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_356enterPrimary_derivative(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_356enterPrimary_derivative(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_derivative", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":817
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_derivative.
 *     def exitPrimary_derivative(self, ctx:ModelicaParser.Primary_derivativeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_359exitPrimary_derivative(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_359exitPrimary_derivative = {"exitPrimary_derivative", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_359exitPrimary_derivative, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_359exitPrimary_derivative(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_derivative (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitPrimary_derivative", 1, 2, 2, 1); __PYX_ERR(0, 817, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitPrimary_derivative") < 0)) __PYX_ERR(0, 817, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitPrimary_derivative", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 817, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitPrimary_derivative", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_358exitPrimary_derivative(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_358exitPrimary_derivative(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_derivative", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":822
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_initial.
 *     def enterPrimary_initial(self, ctx:ModelicaParser.Primary_initialContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_361enterPrimary_initial(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_361enterPrimary_initial = {"enterPrimary_initial", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_361enterPrimary_initial, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_361enterPrimary_initial(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_initial (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterPrimary_initial", 1, 2, 2, 1); __PYX_ERR(0, 822, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterPrimary_initial") < 0)) __PYX_ERR(0, 822, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterPrimary_initial", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 822, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterPrimary_initial", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_360enterPrimary_initial(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_360enterPrimary_initial(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_initial", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":826
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_initial.
 *     def exitPrimary_initial(self, ctx:ModelicaParser.Primary_initialContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_363exitPrimary_initial(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_363exitPrimary_initial = {"exitPrimary_initial", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_363exitPrimary_initial, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_363exitPrimary_initial(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_initial (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitPrimary_initial", 1, 2, 2, 1); __PYX_ERR(0, 826, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitPrimary_initial") < 0)) __PYX_ERR(0, 826, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitPrimary_initial", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 826, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitPrimary_initial", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_362exitPrimary_initial(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_362exitPrimary_initial(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_initial", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":831
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_component_reference.
 *     def enterPrimary_component_reference(self, ctx:ModelicaParser.Primary_component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_365enterPrimary_component_reference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_365enterPrimary_component_reference = {"enterPrimary_component_reference", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_365enterPrimary_component_reference, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_365enterPrimary_component_reference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_component_reference (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterPrimary_component_reference", 1, 2, 2, 1); __PYX_ERR(0, 831, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterPrimary_component_reference") < 0)) __PYX_ERR(0, 831, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterPrimary_component_reference", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 831, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterPrimary_component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_364enterPrimary_component_reference(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_364enterPrimary_component_reference(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_component_reference", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":835
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_component_reference.
 *     def exitPrimary_component_reference(self, ctx:ModelicaParser.Primary_component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_367exitPrimary_component_reference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_367exitPrimary_component_reference = {"exitPrimary_component_reference", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_367exitPrimary_component_reference, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_367exitPrimary_component_reference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_component_reference (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitPrimary_component_reference", 1, 2, 2, 1); __PYX_ERR(0, 835, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitPrimary_component_reference") < 0)) __PYX_ERR(0, 835, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitPrimary_component_reference", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 835, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitPrimary_component_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_366exitPrimary_component_reference(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_366exitPrimary_component_reference(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_component_reference", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":840
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_output_expression_list.
 *     def enterPrimary_output_expression_list(self, ctx:ModelicaParser.Primary_output_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_369enterPrimary_output_expression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_369enterPrimary_output_expression_list = {"enterPrimary_output_expression_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_369enterPrimary_output_expression_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_369enterPrimary_output_expression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_output_expression_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterPrimary_output_expression_list", 1, 2, 2, 1); __PYX_ERR(0, 840, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterPrimary_output_expression_list") < 0)) __PYX_ERR(0, 840, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterPrimary_output_expression_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 840, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterPrimary_output_expression_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_368enterPrimary_output_expression_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_368enterPrimary_output_expression_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_output_expression_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":844
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_output_expression_list.
 *     def exitPrimary_output_expression_list(self, ctx:ModelicaParser.Primary_output_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_371exitPrimary_output_expression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_371exitPrimary_output_expression_list = {"exitPrimary_output_expression_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_371exitPrimary_output_expression_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_371exitPrimary_output_expression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_output_expression_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitPrimary_output_expression_list", 1, 2, 2, 1); __PYX_ERR(0, 844, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitPrimary_output_expression_list") < 0)) __PYX_ERR(0, 844, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitPrimary_output_expression_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 844, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitPrimary_output_expression_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_370exitPrimary_output_expression_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_370exitPrimary_output_expression_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_output_expression_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":849
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_expression_list.
 *     def enterPrimary_expression_list(self, ctx:ModelicaParser.Primary_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_373enterPrimary_expression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_373enterPrimary_expression_list = {"enterPrimary_expression_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_373enterPrimary_expression_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_373enterPrimary_expression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_expression_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterPrimary_expression_list", 1, 2, 2, 1); __PYX_ERR(0, 849, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterPrimary_expression_list") < 0)) __PYX_ERR(0, 849, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterPrimary_expression_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 849, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterPrimary_expression_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_372enterPrimary_expression_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_372enterPrimary_expression_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_expression_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":853
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_expression_list.
 *     def exitPrimary_expression_list(self, ctx:ModelicaParser.Primary_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_375exitPrimary_expression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_375exitPrimary_expression_list = {"exitPrimary_expression_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_375exitPrimary_expression_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_375exitPrimary_expression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_expression_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitPrimary_expression_list", 1, 2, 2, 1); __PYX_ERR(0, 853, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitPrimary_expression_list") < 0)) __PYX_ERR(0, 853, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitPrimary_expression_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 853, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitPrimary_expression_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_374exitPrimary_expression_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_374exitPrimary_expression_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_expression_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":858
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_function_arguments.
 *     def enterPrimary_function_arguments(self, ctx:ModelicaParser.Primary_function_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_377enterPrimary_function_arguments(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_377enterPrimary_function_arguments = {"enterPrimary_function_arguments", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_377enterPrimary_function_arguments, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_377enterPrimary_function_arguments(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_function_arguments (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterPrimary_function_arguments", 1, 2, 2, 1); __PYX_ERR(0, 858, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterPrimary_function_arguments") < 0)) __PYX_ERR(0, 858, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterPrimary_function_arguments", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 858, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterPrimary_function_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_376enterPrimary_function_arguments(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_376enterPrimary_function_arguments(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_function_arguments", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":862
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_function_arguments.
 *     def exitPrimary_function_arguments(self, ctx:ModelicaParser.Primary_function_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_379exitPrimary_function_arguments(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_379exitPrimary_function_arguments = {"exitPrimary_function_arguments", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_379exitPrimary_function_arguments, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_379exitPrimary_function_arguments(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_function_arguments (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitPrimary_function_arguments", 1, 2, 2, 1); __PYX_ERR(0, 862, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitPrimary_function_arguments") < 0)) __PYX_ERR(0, 862, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitPrimary_function_arguments", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 862, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitPrimary_function_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_378exitPrimary_function_arguments(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_378exitPrimary_function_arguments(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_function_arguments", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":867
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_end.
 *     def enterPrimary_end(self, ctx:ModelicaParser.Primary_endContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_381enterPrimary_end(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_381enterPrimary_end = {"enterPrimary_end", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_381enterPrimary_end, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_381enterPrimary_end(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_end (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterPrimary_end", 1, 2, 2, 1); __PYX_ERR(0, 867, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterPrimary_end") < 0)) __PYX_ERR(0, 867, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterPrimary_end", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 867, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterPrimary_end", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_380enterPrimary_end(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_380enterPrimary_end(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterPrimary_end", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":871
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_end.
 *     def exitPrimary_end(self, ctx:ModelicaParser.Primary_endContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_383exitPrimary_end(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_383exitPrimary_end = {"exitPrimary_end", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_383exitPrimary_end, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_383exitPrimary_end(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_end (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitPrimary_end", 1, 2, 2, 1); __PYX_ERR(0, 871, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitPrimary_end") < 0)) __PYX_ERR(0, 871, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitPrimary_end", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 871, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitPrimary_end", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_382exitPrimary_end(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_382exitPrimary_end(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitPrimary_end", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":876
 * 
 *     # Enter a parse tree produced by ModelicaParser#name.
 *     def enterName(self, ctx:ModelicaParser.NameContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_385enterName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_385enterName = {"enterName", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_385enterName, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_385enterName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterName (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterName", 1, 2, 2, 1); __PYX_ERR(0, 876, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterName") < 0)) __PYX_ERR(0, 876, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterName", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 876, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_384enterName(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_384enterName(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterName", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":880
 * 
 *     # Exit a parse tree produced by ModelicaParser#name.
 *     def exitName(self, ctx:ModelicaParser.NameContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_387exitName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_387exitName = {"exitName", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_387exitName, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_387exitName(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitName (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitName", 1, 2, 2, 1); __PYX_ERR(0, 880, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitName") < 0)) __PYX_ERR(0, 880, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitName", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 880, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitName", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_386exitName(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_386exitName(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitName", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":885
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_reference_element.
 *     def enterComponent_reference_element(self, ctx:ModelicaParser.Component_reference_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_389enterComponent_reference_element(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_389enterComponent_reference_element = {"enterComponent_reference_element", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_389enterComponent_reference_element, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_389enterComponent_reference_element(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComponent_reference_element (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterComponent_reference_element", 1, 2, 2, 1); __PYX_ERR(0, 885, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterComponent_reference_element") < 0)) __PYX_ERR(0, 885, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterComponent_reference_element", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 885, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterComponent_reference_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_388enterComponent_reference_element(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_388enterComponent_reference_element(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComponent_reference_element", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":889
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_reference_element.
 *     def exitComponent_reference_element(self, ctx:ModelicaParser.Component_reference_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_391exitComponent_reference_element(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_391exitComponent_reference_element = {"exitComponent_reference_element", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_391exitComponent_reference_element, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_391exitComponent_reference_element(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComponent_reference_element (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitComponent_reference_element", 1, 2, 2, 1); __PYX_ERR(0, 889, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitComponent_reference_element") < 0)) __PYX_ERR(0, 889, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitComponent_reference_element", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 889, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitComponent_reference_element", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_390exitComponent_reference_element(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_390exitComponent_reference_element(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComponent_reference_element", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":894
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_reference.
 *     def enterComponent_reference(self, ctx:ModelicaParser.Component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_393enterComponent_reference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_393enterComponent_reference = {"enterComponent_reference", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_393enterComponent_reference, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_393enterComponent_reference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComponent_reference (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterComponent_reference", 1, 2, 2, 1); __PYX_ERR(0, 894, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterComponent_reference") < 0)) __PYX_ERR(0, 894, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterComponent_reference", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 894, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterComponent_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_392enterComponent_reference(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_392enterComponent_reference(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComponent_reference", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":898
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_reference.
 *     def exitComponent_reference(self, ctx:ModelicaParser.Component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_395exitComponent_reference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_395exitComponent_reference = {"exitComponent_reference", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_395exitComponent_reference, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_395exitComponent_reference(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComponent_reference (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitComponent_reference", 1, 2, 2, 1); __PYX_ERR(0, 898, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitComponent_reference") < 0)) __PYX_ERR(0, 898, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitComponent_reference", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 898, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitComponent_reference", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_394exitComponent_reference(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_394exitComponent_reference(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComponent_reference", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":903
 * 
 *     # Enter a parse tree produced by ModelicaParser#function_call_args.
 *     def enterFunction_call_args(self, ctx:ModelicaParser.Function_call_argsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_397enterFunction_call_args(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_397enterFunction_call_args = {"enterFunction_call_args", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_397enterFunction_call_args, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_397enterFunction_call_args(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterFunction_call_args (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterFunction_call_args", 1, 2, 2, 1); __PYX_ERR(0, 903, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterFunction_call_args") < 0)) __PYX_ERR(0, 903, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterFunction_call_args", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 903, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterFunction_call_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_396enterFunction_call_args(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_396enterFunction_call_args(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterFunction_call_args", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":907
 * 
 *     # Exit a parse tree produced by ModelicaParser#function_call_args.
 *     def exitFunction_call_args(self, ctx:ModelicaParser.Function_call_argsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_399exitFunction_call_args(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_399exitFunction_call_args = {"exitFunction_call_args", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_399exitFunction_call_args, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_399exitFunction_call_args(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitFunction_call_args (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitFunction_call_args", 1, 2, 2, 1); __PYX_ERR(0, 907, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitFunction_call_args") < 0)) __PYX_ERR(0, 907, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitFunction_call_args", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 907, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitFunction_call_args", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_398exitFunction_call_args(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_398exitFunction_call_args(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitFunction_call_args", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":912
 * 
 *     # Enter a parse tree produced by ModelicaParser#function_arguments.
 *     def enterFunction_arguments(self, ctx:ModelicaParser.Function_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_401enterFunction_arguments(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_401enterFunction_arguments = {"enterFunction_arguments", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_401enterFunction_arguments, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_401enterFunction_arguments(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterFunction_arguments (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterFunction_arguments", 1, 2, 2, 1); __PYX_ERR(0, 912, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterFunction_arguments") < 0)) __PYX_ERR(0, 912, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterFunction_arguments", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 912, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterFunction_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_400enterFunction_arguments(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_400enterFunction_arguments(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterFunction_arguments", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":916
 * 
 *     # Exit a parse tree produced by ModelicaParser#function_arguments.
 *     def exitFunction_arguments(self, ctx:ModelicaParser.Function_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_403exitFunction_arguments(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_403exitFunction_arguments = {"exitFunction_arguments", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_403exitFunction_arguments, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_403exitFunction_arguments(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitFunction_arguments (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitFunction_arguments", 1, 2, 2, 1); __PYX_ERR(0, 916, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitFunction_arguments") < 0)) __PYX_ERR(0, 916, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitFunction_arguments", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 916, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitFunction_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_402exitFunction_arguments(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_402exitFunction_arguments(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitFunction_arguments", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":921
 * 
 *     # Enter a parse tree produced by ModelicaParser#named_arguments.
 *     def enterNamed_arguments(self, ctx:ModelicaParser.Named_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_405enterNamed_arguments(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_405enterNamed_arguments = {"enterNamed_arguments", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_405enterNamed_arguments, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_405enterNamed_arguments(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterNamed_arguments (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterNamed_arguments", 1, 2, 2, 1); __PYX_ERR(0, 921, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterNamed_arguments") < 0)) __PYX_ERR(0, 921, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterNamed_arguments", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 921, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterNamed_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_404enterNamed_arguments(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_404enterNamed_arguments(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterNamed_arguments", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":925
 * 
 *     # Exit a parse tree produced by ModelicaParser#named_arguments.
 *     def exitNamed_arguments(self, ctx:ModelicaParser.Named_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_407exitNamed_arguments(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_407exitNamed_arguments = {"exitNamed_arguments", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_407exitNamed_arguments, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_407exitNamed_arguments(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitNamed_arguments (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitNamed_arguments", 1, 2, 2, 1); __PYX_ERR(0, 925, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitNamed_arguments") < 0)) __PYX_ERR(0, 925, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitNamed_arguments", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 925, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitNamed_arguments", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_406exitNamed_arguments(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_406exitNamed_arguments(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitNamed_arguments", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":930
 * 
 *     # Enter a parse tree produced by ModelicaParser#named_argument.
 *     def enterNamed_argument(self, ctx:ModelicaParser.Named_argumentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_409enterNamed_argument(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_409enterNamed_argument = {"enterNamed_argument", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_409enterNamed_argument, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_409enterNamed_argument(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterNamed_argument (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterNamed_argument", 1, 2, 2, 1); __PYX_ERR(0, 930, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterNamed_argument") < 0)) __PYX_ERR(0, 930, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterNamed_argument", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 930, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterNamed_argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_408enterNamed_argument(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_408enterNamed_argument(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterNamed_argument", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":934
 * 
 *     # Exit a parse tree produced by ModelicaParser#named_argument.
 *     def exitNamed_argument(self, ctx:ModelicaParser.Named_argumentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_411exitNamed_argument(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_411exitNamed_argument = {"exitNamed_argument", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_411exitNamed_argument, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_411exitNamed_argument(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitNamed_argument (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitNamed_argument", 1, 2, 2, 1); __PYX_ERR(0, 934, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitNamed_argument") < 0)) __PYX_ERR(0, 934, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitNamed_argument", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 934, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitNamed_argument", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_410exitNamed_argument(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_410exitNamed_argument(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitNamed_argument", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":939
 * 
 *     # Enter a parse tree produced by ModelicaParser#argument_function.
 *     def enterArgument_function(self, ctx:ModelicaParser.Argument_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_413enterArgument_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_413enterArgument_function = {"enterArgument_function", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_413enterArgument_function, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_413enterArgument_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterArgument_function (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterArgument_function", 1, 2, 2, 1); __PYX_ERR(0, 939, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterArgument_function") < 0)) __PYX_ERR(0, 939, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterArgument_function", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 939, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterArgument_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_412enterArgument_function(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_412enterArgument_function(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterArgument_function", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":943
 * 
 *     # Exit a parse tree produced by ModelicaParser#argument_function.
 *     def exitArgument_function(self, ctx:ModelicaParser.Argument_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_415exitArgument_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_415exitArgument_function = {"exitArgument_function", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_415exitArgument_function, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_415exitArgument_function(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitArgument_function (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitArgument_function", 1, 2, 2, 1); __PYX_ERR(0, 943, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitArgument_function") < 0)) __PYX_ERR(0, 943, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitArgument_function", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 943, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitArgument_function", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_414exitArgument_function(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_414exitArgument_function(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitArgument_function", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":948
 * 
 *     # Enter a parse tree produced by ModelicaParser#argument_expression.
 *     def enterArgument_expression(self, ctx:ModelicaParser.Argument_expressionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_417enterArgument_expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_417enterArgument_expression = {"enterArgument_expression", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_417enterArgument_expression, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_417enterArgument_expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterArgument_expression (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterArgument_expression", 1, 2, 2, 1); __PYX_ERR(0, 948, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterArgument_expression") < 0)) __PYX_ERR(0, 948, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterArgument_expression", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 948, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterArgument_expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_416enterArgument_expression(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_416enterArgument_expression(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterArgument_expression", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":952
 * 
 *     # Exit a parse tree produced by ModelicaParser#argument_expression.
 *     def exitArgument_expression(self, ctx:ModelicaParser.Argument_expressionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_419exitArgument_expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_419exitArgument_expression = {"exitArgument_expression", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_419exitArgument_expression, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_419exitArgument_expression(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitArgument_expression (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitArgument_expression", 1, 2, 2, 1); __PYX_ERR(0, 952, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitArgument_expression") < 0)) __PYX_ERR(0, 952, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitArgument_expression", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 952, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitArgument_expression", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_418exitArgument_expression(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_418exitArgument_expression(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitArgument_expression", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":957
 * 
 *     # Enter a parse tree produced by ModelicaParser#output_expression_list.
 *     def enterOutput_expression_list(self, ctx:ModelicaParser.Output_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_421enterOutput_expression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_421enterOutput_expression_list = {"enterOutput_expression_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_421enterOutput_expression_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_421enterOutput_expression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterOutput_expression_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterOutput_expression_list", 1, 2, 2, 1); __PYX_ERR(0, 957, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterOutput_expression_list") < 0)) __PYX_ERR(0, 957, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterOutput_expression_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 957, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterOutput_expression_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_420enterOutput_expression_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_420enterOutput_expression_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterOutput_expression_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":961
 * 
 *     # Exit a parse tree produced by ModelicaParser#output_expression_list.
 *     def exitOutput_expression_list(self, ctx:ModelicaParser.Output_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_423exitOutput_expression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_423exitOutput_expression_list = {"exitOutput_expression_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_423exitOutput_expression_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_423exitOutput_expression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitOutput_expression_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitOutput_expression_list", 1, 2, 2, 1); __PYX_ERR(0, 961, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitOutput_expression_list") < 0)) __PYX_ERR(0, 961, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitOutput_expression_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 961, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitOutput_expression_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_422exitOutput_expression_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_422exitOutput_expression_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitOutput_expression_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":966
 * 
 *     # Enter a parse tree produced by ModelicaParser#expression_list.
 *     def enterExpression_list(self, ctx:ModelicaParser.Expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_425enterExpression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_425enterExpression_list = {"enterExpression_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_425enterExpression_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_425enterExpression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpression_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterExpression_list", 1, 2, 2, 1); __PYX_ERR(0, 966, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterExpression_list") < 0)) __PYX_ERR(0, 966, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterExpression_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 966, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterExpression_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_424enterExpression_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_424enterExpression_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterExpression_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":970
 * 
 *     # Exit a parse tree produced by ModelicaParser#expression_list.
 *     def exitExpression_list(self, ctx:ModelicaParser.Expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_427exitExpression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_427exitExpression_list = {"exitExpression_list", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_427exitExpression_list, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_427exitExpression_list(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpression_list (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitExpression_list", 1, 2, 2, 1); __PYX_ERR(0, 970, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitExpression_list") < 0)) __PYX_ERR(0, 970, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitExpression_list", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 970, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitExpression_list", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_426exitExpression_list(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_426exitExpression_list(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitExpression_list", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":975
 * 
 *     # Enter a parse tree produced by ModelicaParser#array_subscripts.
 *     def enterArray_subscripts(self, ctx:ModelicaParser.Array_subscriptsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_429enterArray_subscripts(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_429enterArray_subscripts = {"enterArray_subscripts", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_429enterArray_subscripts, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_429enterArray_subscripts(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterArray_subscripts (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterArray_subscripts", 1, 2, 2, 1); __PYX_ERR(0, 975, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterArray_subscripts") < 0)) __PYX_ERR(0, 975, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterArray_subscripts", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 975, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterArray_subscripts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_428enterArray_subscripts(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_428enterArray_subscripts(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterArray_subscripts", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":979
 * 
 *     # Exit a parse tree produced by ModelicaParser#array_subscripts.
 *     def exitArray_subscripts(self, ctx:ModelicaParser.Array_subscriptsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_431exitArray_subscripts(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_431exitArray_subscripts = {"exitArray_subscripts", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_431exitArray_subscripts, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_431exitArray_subscripts(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitArray_subscripts (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitArray_subscripts", 1, 2, 2, 1); __PYX_ERR(0, 979, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitArray_subscripts") < 0)) __PYX_ERR(0, 979, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitArray_subscripts", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 979, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitArray_subscripts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_430exitArray_subscripts(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_430exitArray_subscripts(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitArray_subscripts", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":984
 * 
 *     # Enter a parse tree produced by ModelicaParser#subscript.
 *     def enterSubscript(self, ctx:ModelicaParser.SubscriptContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_433enterSubscript(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_433enterSubscript = {"enterSubscript", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_433enterSubscript, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_433enterSubscript(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterSubscript (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterSubscript", 1, 2, 2, 1); __PYX_ERR(0, 984, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterSubscript") < 0)) __PYX_ERR(0, 984, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterSubscript", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 984, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterSubscript", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_432enterSubscript(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_432enterSubscript(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterSubscript", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":988
 * 
 *     # Exit a parse tree produced by ModelicaParser#subscript.
 *     def exitSubscript(self, ctx:ModelicaParser.SubscriptContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_435exitSubscript(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_435exitSubscript = {"exitSubscript", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_435exitSubscript, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_435exitSubscript(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitSubscript (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitSubscript", 1, 2, 2, 1); __PYX_ERR(0, 988, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitSubscript") < 0)) __PYX_ERR(0, 988, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitSubscript", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 988, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitSubscript", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_434exitSubscript(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_434exitSubscript(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitSubscript", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":993
 * 
 *     # Enter a parse tree produced by ModelicaParser#comment.
 *     def enterComment(self, ctx:ModelicaParser.CommentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_437enterComment(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_437enterComment = {"enterComment", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_437enterComment, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_437enterComment(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComment (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterComment", 1, 2, 2, 1); __PYX_ERR(0, 993, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterComment") < 0)) __PYX_ERR(0, 993, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterComment", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 993, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterComment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_436enterComment(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_436enterComment(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterComment", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":997
 * 
 *     # Exit a parse tree produced by ModelicaParser#comment.
 *     def exitComment(self, ctx:ModelicaParser.CommentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_439exitComment(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_439exitComment = {"exitComment", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_439exitComment, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_439exitComment(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComment (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitComment", 1, 2, 2, 1); __PYX_ERR(0, 997, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitComment") < 0)) __PYX_ERR(0, 997, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitComment", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 997, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitComment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_438exitComment(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_438exitComment(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitComment", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":1002
 * 
 *     # Enter a parse tree produced by ModelicaParser#string_comment.
 *     def enterString_comment(self, ctx:ModelicaParser.String_commentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_441enterString_comment(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_441enterString_comment = {"enterString_comment", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_441enterString_comment, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_441enterString_comment(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterString_comment (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterString_comment", 1, 2, 2, 1); __PYX_ERR(0, 1002, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterString_comment") < 0)) __PYX_ERR(0, 1002, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterString_comment", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1002, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterString_comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_440enterString_comment(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_440enterString_comment(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterString_comment", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":1006
 * 
 *     # Exit a parse tree produced by ModelicaParser#string_comment.
 *     def exitString_comment(self, ctx:ModelicaParser.String_commentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_443exitString_comment(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_443exitString_comment = {"exitString_comment", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_443exitString_comment, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_443exitString_comment(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitString_comment (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitString_comment", 1, 2, 2, 1); __PYX_ERR(0, 1006, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitString_comment") < 0)) __PYX_ERR(0, 1006, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitString_comment", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1006, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitString_comment", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_442exitString_comment(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_442exitString_comment(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitString_comment", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":1011
 * 
 *     # Enter a parse tree produced by ModelicaParser#annotation.
 *     def enterAnnotation(self, ctx:ModelicaParser.AnnotationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_445enterAnnotation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_445enterAnnotation = {"enterAnnotation", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_445enterAnnotation, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_445enterAnnotation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterAnnotation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("enterAnnotation", 1, 2, 2, 1); __PYX_ERR(0, 1011, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "enterAnnotation") < 0)) __PYX_ERR(0, 1011, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("enterAnnotation", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1011, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.enterAnnotation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_444enterAnnotation(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_444enterAnnotation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("enterAnnotation", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "pymola/generated/ModelicaListener.pyx":1015
 * 
 *     # Exit a parse tree produced by ModelicaParser#annotation.
 *     def exitAnnotation(self, ctx:ModelicaParser.AnnotationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */

/* Python wrapper */
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_447exitAnnotation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static PyMethodDef __pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_447exitAnnotation = {"exitAnnotation", (PyCFunction)__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_447exitAnnotation, METH_VARARGS|METH_KEYWORDS, 0};
static PyObject *__pyx_pw_6pymola_9generated_16ModelicaListener_16ModelicaListener_447exitAnnotation(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  CYTHON_UNUSED PyObject *__pyx_v_self = 0;
  CYTHON_UNUSED PyObject *__pyx_v_ctx = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitAnnotation (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_self,&__pyx_n_s_ctx,0};
    PyObject* values[2] = {0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_self)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_ctx)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("exitAnnotation", 1, 2, 2, 1); __PYX_ERR(0, 1015, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "exitAnnotation") < 0)) __PYX_ERR(0, 1015, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 2) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
    }
    __pyx_v_self = values[0];
    __pyx_v_ctx = values[1];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("exitAnnotation", 1, 2, 2, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 1015, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("pymola.generated.ModelicaListener.ModelicaListener.exitAnnotation", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_446exitAnnotation(__pyx_self, __pyx_v_self, __pyx_v_ctx);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_6pymola_9generated_16ModelicaListener_16ModelicaListener_446exitAnnotation(CYTHON_UNUSED PyObject *__pyx_self, CYTHON_UNUSED PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *__pyx_v_ctx) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("exitAnnotation", 0);

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

static int __pyx_import_star_set(PyObject *o, PyObject* py_name, char *name) {
  static const char* internal_type_names[] = {
    0
  };
  const char** type_name = internal_type_names;
  while (*type_name) {
    if (__Pyx_StrEq(name, *type_name)) {
      PyErr_Format(PyExc_TypeError, "Cannot overwrite C type %s", name);
      goto bad;
    }
    type_name++;
  }
  if (0);
  else {
    if (PyObject_SetAttr(__pyx_m, py_name, o) < 0) goto bad;
  }
  return 0;
  bad:
  return -1;
}

























































































































static int
__Pyx_import_all_from(PyObject *locals, PyObject *v)
{
    PyObject *all = PyObject_GetAttrString(v, "__all__");
    PyObject *dict, *name, *value;
    int skip_leading_underscores = 0;
    int pos, err;

    if (all == NULL) {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError))
            return -1;
        PyErr_Clear();
        dict = PyObject_GetAttrString(v, "__dict__");
        if (dict == NULL) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError))
                return -1;
            PyErr_SetString(PyExc_ImportError,
            "from-import-* object has no __dict__ and no __all__");
            return -1;
        }
#if PY_MAJOR_VERSION < 3
        all = PyObject_CallMethod(dict, (char *)"keys", NULL);
#else
        all = PyMapping_Keys(dict);
#endif
        Py_DECREF(dict);
        if (all == NULL)
            return -1;
        skip_leading_underscores = 1;
    }

    for (pos = 0, err = 0; ; pos++) {
        name = PySequence_GetItem(all, pos);
        if (name == NULL) {
            if (!PyErr_ExceptionMatches(PyExc_IndexError))
                err = -1;
            else
                PyErr_Clear();
            break;
        }
        if (skip_leading_underscores &&
#if PY_MAJOR_VERSION < 3
            PyString_Check(name) &&
            PyString_AS_STRING(name)[0] == '_')
#else
            PyUnicode_Check(name) &&
            PyUnicode_AS_UNICODE(name)[0] == '_')
#endif
        {
            Py_DECREF(name);
            continue;
        }
        value = PyObject_GetAttr(v, name);
        if (value == NULL)
            err = -1;
        else if (PyDict_CheckExact(locals))
            err = PyDict_SetItem(locals, name, value);
        else
            err = PyObject_SetItem(locals, name, value);
        Py_DECREF(name);
        Py_XDECREF(value);
        if (err != 0)
            break;
    }
    Py_DECREF(all);
    return err;
}


static int __pyx_import_star(PyObject* m) {

    int i;
    int ret = -1;
    char* s;
    PyObject *locals = 0;
    PyObject *list = 0;
#if PY_MAJOR_VERSION >= 3
    PyObject *utf8_name = 0;
#endif
    PyObject *name;
    PyObject *item;

    locals = PyDict_New();              if (!locals) goto bad;
    if (__Pyx_import_all_from(locals, m) < 0) goto bad;
    list = PyDict_Items(locals);        if (!list) goto bad;

    for(i=0; i<PyList_GET_SIZE(list); i++) {
        name = PyTuple_GET_ITEM(PyList_GET_ITEM(list, i), 0);
        item = PyTuple_GET_ITEM(PyList_GET_ITEM(list, i), 1);
#if PY_MAJOR_VERSION >= 3
        utf8_name = PyUnicode_AsUTF8String(name);
        if (!utf8_name) goto bad;
        s = PyBytes_AS_STRING(utf8_name);
        if (__pyx_import_star_set(item, name, s) < 0) goto bad;
        Py_DECREF(utf8_name); utf8_name = 0;
#else
        s = PyString_AsString(name);
        if (!s) goto bad;
        if (__pyx_import_star_set(item, name, s) < 0) goto bad;
#endif
    }
    ret = 0;

bad:
    Py_XDECREF(locals);
    Py_XDECREF(list);
#if PY_MAJOR_VERSION >= 3
    Py_XDECREF(utf8_name);
#endif
    return ret;
}



#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "ModelicaListener",
    0, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_n_s_, __pyx_k_, sizeof(__pyx_k_), 0, 0, 1, 1},
  {&__pyx_n_s_Algorithm_sectionContext, __pyx_k_Algorithm_sectionContext, sizeof(__pyx_k_Algorithm_sectionContext), 0, 0, 1, 1},
  {&__pyx_n_s_AnnotationContext, __pyx_k_AnnotationContext, sizeof(__pyx_k_AnnotationContext), 0, 0, 1, 1},
  {&__pyx_n_s_ArgumentContext, __pyx_k_ArgumentContext, sizeof(__pyx_k_ArgumentContext), 0, 0, 1, 1},
  {&__pyx_n_s_Argument_expressionContext, __pyx_k_Argument_expressionContext, sizeof(__pyx_k_Argument_expressionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Argument_functionContext, __pyx_k_Argument_functionContext, sizeof(__pyx_k_Argument_functionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Argument_listContext, __pyx_k_Argument_listContext, sizeof(__pyx_k_Argument_listContext), 0, 0, 1, 1},
  {&__pyx_n_s_Array_subscriptsContext, __pyx_k_Array_subscriptsContext, sizeof(__pyx_k_Array_subscriptsContext), 0, 0, 1, 1},
  {&__pyx_n_s_Base_prefixContext, __pyx_k_Base_prefixContext, sizeof(__pyx_k_Base_prefixContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_definitionContext, __pyx_k_Class_definitionContext, sizeof(__pyx_k_Class_definitionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_modificationContext, __pyx_k_Class_modificationContext, sizeof(__pyx_k_Class_modificationContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_prefixesContext, __pyx_k_Class_prefixesContext, sizeof(__pyx_k_Class_prefixesContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_spec_baseContext, __pyx_k_Class_spec_baseContext, sizeof(__pyx_k_Class_spec_baseContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_spec_compContext, __pyx_k_Class_spec_compContext, sizeof(__pyx_k_Class_spec_compContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_spec_derContext, __pyx_k_Class_spec_derContext, sizeof(__pyx_k_Class_spec_derContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_spec_enumContext, __pyx_k_Class_spec_enumContext, sizeof(__pyx_k_Class_spec_enumContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_spec_extendsContext, __pyx_k_Class_spec_extendsContext, sizeof(__pyx_k_Class_spec_extendsContext), 0, 0, 1, 1},
  {&__pyx_n_s_Class_typeContext, __pyx_k_Class_typeContext, sizeof(__pyx_k_Class_typeContext), 0, 0, 1, 1},
  {&__pyx_n_s_CommentContext, __pyx_k_CommentContext, sizeof(__pyx_k_CommentContext), 0, 0, 1, 1},
  {&__pyx_n_s_Component_clause1Context, __pyx_k_Component_clause1Context, sizeof(__pyx_k_Component_clause1Context), 0, 0, 1, 1},
  {&__pyx_n_s_Component_clauseContext, __pyx_k_Component_clauseContext, sizeof(__pyx_k_Component_clauseContext), 0, 0, 1, 1},
  {&__pyx_n_s_Component_declaration1Context, __pyx_k_Component_declaration1Context, sizeof(__pyx_k_Component_declaration1Context), 0, 0, 1, 1},
  {&__pyx_n_s_Component_declarationContext, __pyx_k_Component_declarationContext, sizeof(__pyx_k_Component_declarationContext), 0, 0, 1, 1},
  {&__pyx_n_s_Component_listContext, __pyx_k_Component_listContext, sizeof(__pyx_k_Component_listContext), 0, 0, 1, 1},
  {&__pyx_n_s_Component_referenceContext, __pyx_k_Component_referenceContext, sizeof(__pyx_k_Component_referenceContext), 0, 0, 1, 1},
  {&__pyx_n_s_Component_reference_elementConte, __pyx_k_Component_reference_elementConte, sizeof(__pyx_k_Component_reference_elementConte), 0, 0, 1, 1},
  {&__pyx_n_s_CompositionContext, __pyx_k_CompositionContext, sizeof(__pyx_k_CompositionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Condition_attributeContext, __pyx_k_Condition_attributeContext, sizeof(__pyx_k_Condition_attributeContext), 0, 0, 1, 1},
  {&__pyx_n_s_Connect_clauseContext, __pyx_k_Connect_clauseContext, sizeof(__pyx_k_Connect_clauseContext), 0, 0, 1, 1},
  {&__pyx_n_s_Constraining_clauseContext, __pyx_k_Constraining_clauseContext, sizeof(__pyx_k_Constraining_clauseContext), 0, 0, 1, 1},
  {&__pyx_n_s_DeclarationContext, __pyx_k_DeclarationContext, sizeof(__pyx_k_DeclarationContext), 0, 0, 1, 1},
  {&__pyx_n_s_ElementContext, __pyx_k_ElementContext, sizeof(__pyx_k_ElementContext), 0, 0, 1, 1},
  {&__pyx_n_s_Element_listContext, __pyx_k_Element_listContext, sizeof(__pyx_k_Element_listContext), 0, 0, 1, 1},
  {&__pyx_n_s_Element_modificationContext, __pyx_k_Element_modificationContext, sizeof(__pyx_k_Element_modificationContext), 0, 0, 1, 1},
  {&__pyx_n_s_Element_modification_or_replacea, __pyx_k_Element_modification_or_replacea, sizeof(__pyx_k_Element_modification_or_replacea), 0, 0, 1, 1},
  {&__pyx_n_s_Element_redeclarationContext, __pyx_k_Element_redeclarationContext, sizeof(__pyx_k_Element_redeclarationContext), 0, 0, 1, 1},
  {&__pyx_n_s_Element_replaceableContext, __pyx_k_Element_replaceableContext, sizeof(__pyx_k_Element_replaceableContext), 0, 0, 1, 1},
  {&__pyx_n_s_Enum_listContext, __pyx_k_Enum_listContext, sizeof(__pyx_k_Enum_listContext), 0, 0, 1, 1},
  {&__pyx_n_s_Enumeration_literalContext, __pyx_k_Enumeration_literalContext, sizeof(__pyx_k_Enumeration_literalContext), 0, 0, 1, 1},
  {&__pyx_n_s_EquationContext, __pyx_k_EquationContext, sizeof(__pyx_k_EquationContext), 0, 0, 1, 1},
  {&__pyx_n_s_Equation_connect_clauseContext, __pyx_k_Equation_connect_clauseContext, sizeof(__pyx_k_Equation_connect_clauseContext), 0, 0, 1, 1},
  {&__pyx_n_s_Equation_forContext, __pyx_k_Equation_forContext, sizeof(__pyx_k_Equation_forContext), 0, 0, 1, 1},
  {&__pyx_n_s_Equation_functionContext, __pyx_k_Equation_functionContext, sizeof(__pyx_k_Equation_functionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Equation_ifContext, __pyx_k_Equation_ifContext, sizeof(__pyx_k_Equation_ifContext), 0, 0, 1, 1},
  {&__pyx_n_s_Equation_sectionContext, __pyx_k_Equation_sectionContext, sizeof(__pyx_k_Equation_sectionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Equation_simpleContext, __pyx_k_Equation_simpleContext, sizeof(__pyx_k_Equation_simpleContext), 0, 0, 1, 1},
  {&__pyx_n_s_Equation_whenContext, __pyx_k_Equation_whenContext, sizeof(__pyx_k_Equation_whenContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_addContext, __pyx_k_Expr_addContext, sizeof(__pyx_k_Expr_addContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_andContext, __pyx_k_Expr_andContext, sizeof(__pyx_k_Expr_andContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_expContext, __pyx_k_Expr_expContext, sizeof(__pyx_k_Expr_expContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_mulContext, __pyx_k_Expr_mulContext, sizeof(__pyx_k_Expr_mulContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_negContext, __pyx_k_Expr_negContext, sizeof(__pyx_k_Expr_negContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_notContext, __pyx_k_Expr_notContext, sizeof(__pyx_k_Expr_notContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_orContext, __pyx_k_Expr_orContext, sizeof(__pyx_k_Expr_orContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_primaryContext, __pyx_k_Expr_primaryContext, sizeof(__pyx_k_Expr_primaryContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expr_relContext, __pyx_k_Expr_relContext, sizeof(__pyx_k_Expr_relContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expression_ifContext, __pyx_k_Expression_ifContext, sizeof(__pyx_k_Expression_ifContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expression_listContext, __pyx_k_Expression_listContext, sizeof(__pyx_k_Expression_listContext), 0, 0, 1, 1},
  {&__pyx_n_s_Expression_simpleContext, __pyx_k_Expression_simpleContext, sizeof(__pyx_k_Expression_simpleContext), 0, 0, 1, 1},
  {&__pyx_n_s_Extends_clauseContext, __pyx_k_Extends_clauseContext, sizeof(__pyx_k_Extends_clauseContext), 0, 0, 1, 1},
  {&__pyx_n_s_External_function_callContext, __pyx_k_External_function_callContext, sizeof(__pyx_k_External_function_callContext), 0, 0, 1, 1},
  {&__pyx_n_s_For_equationContext, __pyx_k_For_equationContext, sizeof(__pyx_k_For_equationContext), 0, 0, 1, 1},
  {&__pyx_n_s_For_indexContext, __pyx_k_For_indexContext, sizeof(__pyx_k_For_indexContext), 0, 0, 1, 1},
  {&__pyx_n_s_For_indicesContext, __pyx_k_For_indicesContext, sizeof(__pyx_k_For_indicesContext), 0, 0, 1, 1},
  {&__pyx_n_s_For_statementContext, __pyx_k_For_statementContext, sizeof(__pyx_k_For_statementContext), 0, 0, 1, 1},
  {&__pyx_n_s_Function_argumentsContext, __pyx_k_Function_argumentsContext, sizeof(__pyx_k_Function_argumentsContext), 0, 0, 1, 1},
  {&__pyx_n_s_Function_call_argsContext, __pyx_k_Function_call_argsContext, sizeof(__pyx_k_Function_call_argsContext), 0, 0, 1, 1},
  {&__pyx_n_s_If_equationContext, __pyx_k_If_equationContext, sizeof(__pyx_k_If_equationContext), 0, 0, 1, 1},
  {&__pyx_n_s_If_statementContext, __pyx_k_If_statementContext, sizeof(__pyx_k_If_statementContext), 0, 0, 1, 1},
  {&__pyx_n_s_Import_clauseContext, __pyx_k_Import_clauseContext, sizeof(__pyx_k_Import_clauseContext), 0, 0, 1, 1},
  {&__pyx_n_s_Import_listContext, __pyx_k_Import_listContext, sizeof(__pyx_k_Import_listContext), 0, 0, 1, 1},
  {&__pyx_n_s_Language_specificationContext, __pyx_k_Language_specificationContext, sizeof(__pyx_k_Language_specificationContext), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener, __pyx_k_ModelicaListener, sizeof(__pyx_k_ModelicaListener), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterAlgorithm, __pyx_k_ModelicaListener_enterAlgorithm, sizeof(__pyx_k_ModelicaListener_enterAlgorithm), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterAnnotation, __pyx_k_ModelicaListener_enterAnnotation, sizeof(__pyx_k_ModelicaListener_enterAnnotation), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterArgument, __pyx_k_ModelicaListener_enterArgument, sizeof(__pyx_k_ModelicaListener_enterArgument), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterArgument_e, __pyx_k_ModelicaListener_enterArgument_e, sizeof(__pyx_k_ModelicaListener_enterArgument_e), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterArgument_f, __pyx_k_ModelicaListener_enterArgument_f, sizeof(__pyx_k_ModelicaListener_enterArgument_f), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterArgument_l, __pyx_k_ModelicaListener_enterArgument_l, sizeof(__pyx_k_ModelicaListener_enterArgument_l), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterArray_subs, __pyx_k_ModelicaListener_enterArray_subs, sizeof(__pyx_k_ModelicaListener_enterArray_subs), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterBase_prefi, __pyx_k_ModelicaListener_enterBase_prefi, sizeof(__pyx_k_ModelicaListener_enterBase_prefi), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterClass_defi, __pyx_k_ModelicaListener_enterClass_defi, sizeof(__pyx_k_ModelicaListener_enterClass_defi), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterClass_modi, __pyx_k_ModelicaListener_enterClass_modi, sizeof(__pyx_k_ModelicaListener_enterClass_modi), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterClass_pref, __pyx_k_ModelicaListener_enterClass_pref, sizeof(__pyx_k_ModelicaListener_enterClass_pref), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterClass_spec, __pyx_k_ModelicaListener_enterClass_spec, sizeof(__pyx_k_ModelicaListener_enterClass_spec), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterClass_spec_2, __pyx_k_ModelicaListener_enterClass_spec_2, sizeof(__pyx_k_ModelicaListener_enterClass_spec_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterClass_spec_3, __pyx_k_ModelicaListener_enterClass_spec_3, sizeof(__pyx_k_ModelicaListener_enterClass_spec_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterClass_spec_4, __pyx_k_ModelicaListener_enterClass_spec_4, sizeof(__pyx_k_ModelicaListener_enterClass_spec_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterClass_spec_5, __pyx_k_ModelicaListener_enterClass_spec_5, sizeof(__pyx_k_ModelicaListener_enterClass_spec_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterClass_type, __pyx_k_ModelicaListener_enterClass_type, sizeof(__pyx_k_ModelicaListener_enterClass_type), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterComment, __pyx_k_ModelicaListener_enterComment, sizeof(__pyx_k_ModelicaListener_enterComment), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterComponent, __pyx_k_ModelicaListener_enterComponent, sizeof(__pyx_k_ModelicaListener_enterComponent), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterComponent_2, __pyx_k_ModelicaListener_enterComponent_2, sizeof(__pyx_k_ModelicaListener_enterComponent_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterComponent_3, __pyx_k_ModelicaListener_enterComponent_3, sizeof(__pyx_k_ModelicaListener_enterComponent_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterComponent_4, __pyx_k_ModelicaListener_enterComponent_4, sizeof(__pyx_k_ModelicaListener_enterComponent_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterComponent_5, __pyx_k_ModelicaListener_enterComponent_5, sizeof(__pyx_k_ModelicaListener_enterComponent_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterComponent_6, __pyx_k_ModelicaListener_enterComponent_6, sizeof(__pyx_k_ModelicaListener_enterComponent_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterComponent_7, __pyx_k_ModelicaListener_enterComponent_7, sizeof(__pyx_k_ModelicaListener_enterComponent_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterCompositio, __pyx_k_ModelicaListener_enterCompositio, sizeof(__pyx_k_ModelicaListener_enterCompositio), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterCondition, __pyx_k_ModelicaListener_enterCondition, sizeof(__pyx_k_ModelicaListener_enterCondition), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterConnect_cl, __pyx_k_ModelicaListener_enterConnect_cl, sizeof(__pyx_k_ModelicaListener_enterConnect_cl), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterConstraini, __pyx_k_ModelicaListener_enterConstraini, sizeof(__pyx_k_ModelicaListener_enterConstraini), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterDeclaratio, __pyx_k_ModelicaListener_enterDeclaratio, sizeof(__pyx_k_ModelicaListener_enterDeclaratio), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterElement, __pyx_k_ModelicaListener_enterElement, sizeof(__pyx_k_ModelicaListener_enterElement), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterElement_li, __pyx_k_ModelicaListener_enterElement_li, sizeof(__pyx_k_ModelicaListener_enterElement_li), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterElement_mo, __pyx_k_ModelicaListener_enterElement_mo, sizeof(__pyx_k_ModelicaListener_enterElement_mo), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterElement_mo_2, __pyx_k_ModelicaListener_enterElement_mo_2, sizeof(__pyx_k_ModelicaListener_enterElement_mo_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterElement_re, __pyx_k_ModelicaListener_enterElement_re, sizeof(__pyx_k_ModelicaListener_enterElement_re), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterElement_re_2, __pyx_k_ModelicaListener_enterElement_re_2, sizeof(__pyx_k_ModelicaListener_enterElement_re_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterEnum_list, __pyx_k_ModelicaListener_enterEnum_list, sizeof(__pyx_k_ModelicaListener_enterEnum_list), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterEnumeratio, __pyx_k_ModelicaListener_enterEnumeratio, sizeof(__pyx_k_ModelicaListener_enterEnumeratio), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterEquation, __pyx_k_ModelicaListener_enterEquation, sizeof(__pyx_k_ModelicaListener_enterEquation), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterEquation_c, __pyx_k_ModelicaListener_enterEquation_c, sizeof(__pyx_k_ModelicaListener_enterEquation_c), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterEquation_f, __pyx_k_ModelicaListener_enterEquation_f, sizeof(__pyx_k_ModelicaListener_enterEquation_f), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterEquation_f_2, __pyx_k_ModelicaListener_enterEquation_f_2, sizeof(__pyx_k_ModelicaListener_enterEquation_f_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterEquation_i, __pyx_k_ModelicaListener_enterEquation_i, sizeof(__pyx_k_ModelicaListener_enterEquation_i), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterEquation_s, __pyx_k_ModelicaListener_enterEquation_s, sizeof(__pyx_k_ModelicaListener_enterEquation_s), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterEquation_s_2, __pyx_k_ModelicaListener_enterEquation_s_2, sizeof(__pyx_k_ModelicaListener_enterEquation_s_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterEquation_w, __pyx_k_ModelicaListener_enterEquation_w, sizeof(__pyx_k_ModelicaListener_enterEquation_w), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterExpr_add, __pyx_k_ModelicaListener_enterExpr_add, sizeof(__pyx_k_ModelicaListener_enterExpr_add), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterExpr_and, __pyx_k_ModelicaListener_enterExpr_and, sizeof(__pyx_k_ModelicaListener_enterExpr_and), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterExpr_exp, __pyx_k_ModelicaListener_enterExpr_exp, sizeof(__pyx_k_ModelicaListener_enterExpr_exp), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterExpr_mul, __pyx_k_ModelicaListener_enterExpr_mul, sizeof(__pyx_k_ModelicaListener_enterExpr_mul), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterExpr_neg, __pyx_k_ModelicaListener_enterExpr_neg, sizeof(__pyx_k_ModelicaListener_enterExpr_neg), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterExpr_not, __pyx_k_ModelicaListener_enterExpr_not, sizeof(__pyx_k_ModelicaListener_enterExpr_not), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterExpr_or, __pyx_k_ModelicaListener_enterExpr_or, sizeof(__pyx_k_ModelicaListener_enterExpr_or), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterExpr_prima, __pyx_k_ModelicaListener_enterExpr_prima, sizeof(__pyx_k_ModelicaListener_enterExpr_prima), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterExpr_rel, __pyx_k_ModelicaListener_enterExpr_rel, sizeof(__pyx_k_ModelicaListener_enterExpr_rel), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterExpression, __pyx_k_ModelicaListener_enterExpression, sizeof(__pyx_k_ModelicaListener_enterExpression), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterExpression_2, __pyx_k_ModelicaListener_enterExpression_2, sizeof(__pyx_k_ModelicaListener_enterExpression_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterExpression_3, __pyx_k_ModelicaListener_enterExpression_3, sizeof(__pyx_k_ModelicaListener_enterExpression_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterExtends_cl, __pyx_k_ModelicaListener_enterExtends_cl, sizeof(__pyx_k_ModelicaListener_enterExtends_cl), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterExternal_f, __pyx_k_ModelicaListener_enterExternal_f, sizeof(__pyx_k_ModelicaListener_enterExternal_f), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterFor_equati, __pyx_k_ModelicaListener_enterFor_equati, sizeof(__pyx_k_ModelicaListener_enterFor_equati), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterFor_index, __pyx_k_ModelicaListener_enterFor_index, sizeof(__pyx_k_ModelicaListener_enterFor_index), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterFor_indice, __pyx_k_ModelicaListener_enterFor_indice, sizeof(__pyx_k_ModelicaListener_enterFor_indice), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterFor_statem, __pyx_k_ModelicaListener_enterFor_statem, sizeof(__pyx_k_ModelicaListener_enterFor_statem), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterFunction_a, __pyx_k_ModelicaListener_enterFunction_a, sizeof(__pyx_k_ModelicaListener_enterFunction_a), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterFunction_c, __pyx_k_ModelicaListener_enterFunction_c, sizeof(__pyx_k_ModelicaListener_enterFunction_c), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterIf_equatio, __pyx_k_ModelicaListener_enterIf_equatio, sizeof(__pyx_k_ModelicaListener_enterIf_equatio), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterIf_stateme, __pyx_k_ModelicaListener_enterIf_stateme, sizeof(__pyx_k_ModelicaListener_enterIf_stateme), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterImport_cla, __pyx_k_ModelicaListener_enterImport_cla, sizeof(__pyx_k_ModelicaListener_enterImport_cla), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterImport_lis, __pyx_k_ModelicaListener_enterImport_lis, sizeof(__pyx_k_ModelicaListener_enterImport_lis), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterLanguage_s, __pyx_k_ModelicaListener_enterLanguage_s, sizeof(__pyx_k_ModelicaListener_enterLanguage_s), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterModificati, __pyx_k_ModelicaListener_enterModificati, sizeof(__pyx_k_ModelicaListener_enterModificati), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterModificati_2, __pyx_k_ModelicaListener_enterModificati_2, sizeof(__pyx_k_ModelicaListener_enterModificati_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterModificati_3, __pyx_k_ModelicaListener_enterModificati_3, sizeof(__pyx_k_ModelicaListener_enterModificati_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterName, __pyx_k_ModelicaListener_enterName, sizeof(__pyx_k_ModelicaListener_enterName), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterNamed_argu, __pyx_k_ModelicaListener_enterNamed_argu, sizeof(__pyx_k_ModelicaListener_enterNamed_argu), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterNamed_argu_2, __pyx_k_ModelicaListener_enterNamed_argu_2, sizeof(__pyx_k_ModelicaListener_enterNamed_argu_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterOutput_exp, __pyx_k_ModelicaListener_enterOutput_exp, sizeof(__pyx_k_ModelicaListener_enterOutput_exp), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterPrimary_co, __pyx_k_ModelicaListener_enterPrimary_co, sizeof(__pyx_k_ModelicaListener_enterPrimary_co), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterPrimary_de, __pyx_k_ModelicaListener_enterPrimary_de, sizeof(__pyx_k_ModelicaListener_enterPrimary_de), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterPrimary_en, __pyx_k_ModelicaListener_enterPrimary_en, sizeof(__pyx_k_ModelicaListener_enterPrimary_en), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterPrimary_ex, __pyx_k_ModelicaListener_enterPrimary_ex, sizeof(__pyx_k_ModelicaListener_enterPrimary_ex), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterPrimary_fa, __pyx_k_ModelicaListener_enterPrimary_fa, sizeof(__pyx_k_ModelicaListener_enterPrimary_fa), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterPrimary_fu, __pyx_k_ModelicaListener_enterPrimary_fu, sizeof(__pyx_k_ModelicaListener_enterPrimary_fu), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterPrimary_fu_2, __pyx_k_ModelicaListener_enterPrimary_fu_2, sizeof(__pyx_k_ModelicaListener_enterPrimary_fu_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterPrimary_in, __pyx_k_ModelicaListener_enterPrimary_in, sizeof(__pyx_k_ModelicaListener_enterPrimary_in), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterPrimary_ou, __pyx_k_ModelicaListener_enterPrimary_ou, sizeof(__pyx_k_ModelicaListener_enterPrimary_ou), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterPrimary_st, __pyx_k_ModelicaListener_enterPrimary_st, sizeof(__pyx_k_ModelicaListener_enterPrimary_st), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterPrimary_tr, __pyx_k_ModelicaListener_enterPrimary_tr, sizeof(__pyx_k_ModelicaListener_enterPrimary_tr), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterPrimary_un, __pyx_k_ModelicaListener_enterPrimary_un, sizeof(__pyx_k_ModelicaListener_enterPrimary_un), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterRegular_el, __pyx_k_ModelicaListener_enterRegular_el, sizeof(__pyx_k_ModelicaListener_enterRegular_el), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterReplaceabl, __pyx_k_ModelicaListener_enterReplaceabl, sizeof(__pyx_k_ModelicaListener_enterReplaceabl), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterShort_clas, __pyx_k_ModelicaListener_enterShort_clas, sizeof(__pyx_k_ModelicaListener_enterShort_clas), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterSimple_exp, __pyx_k_ModelicaListener_enterSimple_exp, sizeof(__pyx_k_ModelicaListener_enterSimple_exp), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterStatement, __pyx_k_ModelicaListener_enterStatement, sizeof(__pyx_k_ModelicaListener_enterStatement), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterStatement_2, __pyx_k_ModelicaListener_enterStatement_2, sizeof(__pyx_k_ModelicaListener_enterStatement_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterStatement_3, __pyx_k_ModelicaListener_enterStatement_3, sizeof(__pyx_k_ModelicaListener_enterStatement_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterStatement_4, __pyx_k_ModelicaListener_enterStatement_4, sizeof(__pyx_k_ModelicaListener_enterStatement_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterStatement_5, __pyx_k_ModelicaListener_enterStatement_5, sizeof(__pyx_k_ModelicaListener_enterStatement_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterStatement_6, __pyx_k_ModelicaListener_enterStatement_6, sizeof(__pyx_k_ModelicaListener_enterStatement_6), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterStatement_7, __pyx_k_ModelicaListener_enterStatement_7, sizeof(__pyx_k_ModelicaListener_enterStatement_7), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterStatement_8, __pyx_k_ModelicaListener_enterStatement_8, sizeof(__pyx_k_ModelicaListener_enterStatement_8), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterStatement_9, __pyx_k_ModelicaListener_enterStatement_9, sizeof(__pyx_k_ModelicaListener_enterStatement_9), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterStored_def, __pyx_k_ModelicaListener_enterStored_def, sizeof(__pyx_k_ModelicaListener_enterStored_def), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterStored_def_2, __pyx_k_ModelicaListener_enterStored_def_2, sizeof(__pyx_k_ModelicaListener_enterStored_def_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterString_com, __pyx_k_ModelicaListener_enterString_com, sizeof(__pyx_k_ModelicaListener_enterString_com), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterSubscript, __pyx_k_ModelicaListener_enterSubscript, sizeof(__pyx_k_ModelicaListener_enterSubscript), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterType_prefi, __pyx_k_ModelicaListener_enterType_prefi, sizeof(__pyx_k_ModelicaListener_enterType_prefi), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterType_speci, __pyx_k_ModelicaListener_enterType_speci, sizeof(__pyx_k_ModelicaListener_enterType_speci), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterWhen_equat, __pyx_k_ModelicaListener_enterWhen_equat, sizeof(__pyx_k_ModelicaListener_enterWhen_equat), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterWhen_state, __pyx_k_ModelicaListener_enterWhen_state, sizeof(__pyx_k_ModelicaListener_enterWhen_state), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_enterWhile_stat, __pyx_k_ModelicaListener_enterWhile_stat, sizeof(__pyx_k_ModelicaListener_enterWhile_stat), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitAlgorithm_s, __pyx_k_ModelicaListener_exitAlgorithm_s, sizeof(__pyx_k_ModelicaListener_exitAlgorithm_s), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitAnnotation, __pyx_k_ModelicaListener_exitAnnotation, sizeof(__pyx_k_ModelicaListener_exitAnnotation), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitArgument, __pyx_k_ModelicaListener_exitArgument, sizeof(__pyx_k_ModelicaListener_exitArgument), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitArgument_ex, __pyx_k_ModelicaListener_exitArgument_ex, sizeof(__pyx_k_ModelicaListener_exitArgument_ex), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitArgument_fu, __pyx_k_ModelicaListener_exitArgument_fu, sizeof(__pyx_k_ModelicaListener_exitArgument_fu), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitArgument_li, __pyx_k_ModelicaListener_exitArgument_li, sizeof(__pyx_k_ModelicaListener_exitArgument_li), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitArray_subsc, __pyx_k_ModelicaListener_exitArray_subsc, sizeof(__pyx_k_ModelicaListener_exitArray_subsc), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitBase_prefix, __pyx_k_ModelicaListener_exitBase_prefix, sizeof(__pyx_k_ModelicaListener_exitBase_prefix), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitClass_defin, __pyx_k_ModelicaListener_exitClass_defin, sizeof(__pyx_k_ModelicaListener_exitClass_defin), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitClass_modif, __pyx_k_ModelicaListener_exitClass_modif, sizeof(__pyx_k_ModelicaListener_exitClass_modif), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitClass_prefi, __pyx_k_ModelicaListener_exitClass_prefi, sizeof(__pyx_k_ModelicaListener_exitClass_prefi), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitClass_spec, __pyx_k_ModelicaListener_exitClass_spec, sizeof(__pyx_k_ModelicaListener_exitClass_spec), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitClass_spec_2, __pyx_k_ModelicaListener_exitClass_spec_2, sizeof(__pyx_k_ModelicaListener_exitClass_spec_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitClass_spec_3, __pyx_k_ModelicaListener_exitClass_spec_3, sizeof(__pyx_k_ModelicaListener_exitClass_spec_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitClass_spec_4, __pyx_k_ModelicaListener_exitClass_spec_4, sizeof(__pyx_k_ModelicaListener_exitClass_spec_4), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitClass_spec_5, __pyx_k_ModelicaListener_exitClass_spec_5, sizeof(__pyx_k_ModelicaListener_exitClass_spec_5), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitClass_type, __pyx_k_ModelicaListener_exitClass_type, sizeof(__pyx_k_ModelicaListener_exitClass_type), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitComment, __pyx_k_ModelicaListener_exitComment, sizeof(__pyx_k_ModelicaListener_exitComment), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitComponent_c, __pyx_k_ModelicaListener_exitComponent_c, sizeof(__pyx_k_ModelicaListener_exitComponent_c), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitComponent_c_2, __pyx_k_ModelicaListener_exitComponent_c_2, sizeof(__pyx_k_ModelicaListener_exitComponent_c_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitComponent_d, __pyx_k_ModelicaListener_exitComponent_d, sizeof(__pyx_k_ModelicaListener_exitComponent_d), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitComponent_d_2, __pyx_k_ModelicaListener_exitComponent_d_2, sizeof(__pyx_k_ModelicaListener_exitComponent_d_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitComponent_l, __pyx_k_ModelicaListener_exitComponent_l, sizeof(__pyx_k_ModelicaListener_exitComponent_l), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitComponent_r, __pyx_k_ModelicaListener_exitComponent_r, sizeof(__pyx_k_ModelicaListener_exitComponent_r), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitComponent_r_2, __pyx_k_ModelicaListener_exitComponent_r_2, sizeof(__pyx_k_ModelicaListener_exitComponent_r_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitComposition, __pyx_k_ModelicaListener_exitComposition, sizeof(__pyx_k_ModelicaListener_exitComposition), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitCondition_a, __pyx_k_ModelicaListener_exitCondition_a, sizeof(__pyx_k_ModelicaListener_exitCondition_a), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitConnect_cla, __pyx_k_ModelicaListener_exitConnect_cla, sizeof(__pyx_k_ModelicaListener_exitConnect_cla), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitConstrainin, __pyx_k_ModelicaListener_exitConstrainin, sizeof(__pyx_k_ModelicaListener_exitConstrainin), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitDeclaration, __pyx_k_ModelicaListener_exitDeclaration, sizeof(__pyx_k_ModelicaListener_exitDeclaration), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitElement, __pyx_k_ModelicaListener_exitElement, sizeof(__pyx_k_ModelicaListener_exitElement), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitElement_lis, __pyx_k_ModelicaListener_exitElement_lis, sizeof(__pyx_k_ModelicaListener_exitElement_lis), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitElement_mod, __pyx_k_ModelicaListener_exitElement_mod, sizeof(__pyx_k_ModelicaListener_exitElement_mod), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitElement_mod_2, __pyx_k_ModelicaListener_exitElement_mod_2, sizeof(__pyx_k_ModelicaListener_exitElement_mod_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitElement_red, __pyx_k_ModelicaListener_exitElement_red, sizeof(__pyx_k_ModelicaListener_exitElement_red), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitElement_rep, __pyx_k_ModelicaListener_exitElement_rep, sizeof(__pyx_k_ModelicaListener_exitElement_rep), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitEnum_list, __pyx_k_ModelicaListener_exitEnum_list, sizeof(__pyx_k_ModelicaListener_exitEnum_list), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitEnumeration, __pyx_k_ModelicaListener_exitEnumeration, sizeof(__pyx_k_ModelicaListener_exitEnumeration), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitEquation, __pyx_k_ModelicaListener_exitEquation, sizeof(__pyx_k_ModelicaListener_exitEquation), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitEquation_co, __pyx_k_ModelicaListener_exitEquation_co, sizeof(__pyx_k_ModelicaListener_exitEquation_co), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitEquation_fo, __pyx_k_ModelicaListener_exitEquation_fo, sizeof(__pyx_k_ModelicaListener_exitEquation_fo), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitEquation_fu, __pyx_k_ModelicaListener_exitEquation_fu, sizeof(__pyx_k_ModelicaListener_exitEquation_fu), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitEquation_if, __pyx_k_ModelicaListener_exitEquation_if, sizeof(__pyx_k_ModelicaListener_exitEquation_if), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitEquation_se, __pyx_k_ModelicaListener_exitEquation_se, sizeof(__pyx_k_ModelicaListener_exitEquation_se), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitEquation_si, __pyx_k_ModelicaListener_exitEquation_si, sizeof(__pyx_k_ModelicaListener_exitEquation_si), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitEquation_wh, __pyx_k_ModelicaListener_exitEquation_wh, sizeof(__pyx_k_ModelicaListener_exitEquation_wh), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitExpr_add, __pyx_k_ModelicaListener_exitExpr_add, sizeof(__pyx_k_ModelicaListener_exitExpr_add), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitExpr_and, __pyx_k_ModelicaListener_exitExpr_and, sizeof(__pyx_k_ModelicaListener_exitExpr_and), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitExpr_exp, __pyx_k_ModelicaListener_exitExpr_exp, sizeof(__pyx_k_ModelicaListener_exitExpr_exp), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitExpr_mul, __pyx_k_ModelicaListener_exitExpr_mul, sizeof(__pyx_k_ModelicaListener_exitExpr_mul), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitExpr_neg, __pyx_k_ModelicaListener_exitExpr_neg, sizeof(__pyx_k_ModelicaListener_exitExpr_neg), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitExpr_not, __pyx_k_ModelicaListener_exitExpr_not, sizeof(__pyx_k_ModelicaListener_exitExpr_not), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitExpr_or, __pyx_k_ModelicaListener_exitExpr_or, sizeof(__pyx_k_ModelicaListener_exitExpr_or), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitExpr_primar, __pyx_k_ModelicaListener_exitExpr_primar, sizeof(__pyx_k_ModelicaListener_exitExpr_primar), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitExpr_rel, __pyx_k_ModelicaListener_exitExpr_rel, sizeof(__pyx_k_ModelicaListener_exitExpr_rel), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitExpression, __pyx_k_ModelicaListener_exitExpression, sizeof(__pyx_k_ModelicaListener_exitExpression), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitExpression_2, __pyx_k_ModelicaListener_exitExpression_2, sizeof(__pyx_k_ModelicaListener_exitExpression_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitExpression_3, __pyx_k_ModelicaListener_exitExpression_3, sizeof(__pyx_k_ModelicaListener_exitExpression_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitExtends_cla, __pyx_k_ModelicaListener_exitExtends_cla, sizeof(__pyx_k_ModelicaListener_exitExtends_cla), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitExternal_fu, __pyx_k_ModelicaListener_exitExternal_fu, sizeof(__pyx_k_ModelicaListener_exitExternal_fu), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitFor_equatio, __pyx_k_ModelicaListener_exitFor_equatio, sizeof(__pyx_k_ModelicaListener_exitFor_equatio), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitFor_index, __pyx_k_ModelicaListener_exitFor_index, sizeof(__pyx_k_ModelicaListener_exitFor_index), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitFor_indices, __pyx_k_ModelicaListener_exitFor_indices, sizeof(__pyx_k_ModelicaListener_exitFor_indices), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitFor_stateme, __pyx_k_ModelicaListener_exitFor_stateme, sizeof(__pyx_k_ModelicaListener_exitFor_stateme), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitFunction_ar, __pyx_k_ModelicaListener_exitFunction_ar, sizeof(__pyx_k_ModelicaListener_exitFunction_ar), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitFunction_ca, __pyx_k_ModelicaListener_exitFunction_ca, sizeof(__pyx_k_ModelicaListener_exitFunction_ca), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitIf_equation, __pyx_k_ModelicaListener_exitIf_equation, sizeof(__pyx_k_ModelicaListener_exitIf_equation), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitIf_statemen, __pyx_k_ModelicaListener_exitIf_statemen, sizeof(__pyx_k_ModelicaListener_exitIf_statemen), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitImport_clau, __pyx_k_ModelicaListener_exitImport_clau, sizeof(__pyx_k_ModelicaListener_exitImport_clau), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitImport_list, __pyx_k_ModelicaListener_exitImport_list, sizeof(__pyx_k_ModelicaListener_exitImport_list), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitLanguage_sp, __pyx_k_ModelicaListener_exitLanguage_sp, sizeof(__pyx_k_ModelicaListener_exitLanguage_sp), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitModificatio, __pyx_k_ModelicaListener_exitModificatio, sizeof(__pyx_k_ModelicaListener_exitModificatio), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitModificatio_2, __pyx_k_ModelicaListener_exitModificatio_2, sizeof(__pyx_k_ModelicaListener_exitModificatio_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitModificatio_3, __pyx_k_ModelicaListener_exitModificatio_3, sizeof(__pyx_k_ModelicaListener_exitModificatio_3), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitName, __pyx_k_ModelicaListener_exitName, sizeof(__pyx_k_ModelicaListener_exitName), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitNamed_argum, __pyx_k_ModelicaListener_exitNamed_argum, sizeof(__pyx_k_ModelicaListener_exitNamed_argum), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitNamed_argum_2, __pyx_k_ModelicaListener_exitNamed_argum_2, sizeof(__pyx_k_ModelicaListener_exitNamed_argum_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitOutput_expr, __pyx_k_ModelicaListener_exitOutput_expr, sizeof(__pyx_k_ModelicaListener_exitOutput_expr), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitPrimary_com, __pyx_k_ModelicaListener_exitPrimary_com, sizeof(__pyx_k_ModelicaListener_exitPrimary_com), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitPrimary_der, __pyx_k_ModelicaListener_exitPrimary_der, sizeof(__pyx_k_ModelicaListener_exitPrimary_der), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitPrimary_end, __pyx_k_ModelicaListener_exitPrimary_end, sizeof(__pyx_k_ModelicaListener_exitPrimary_end), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitPrimary_exp, __pyx_k_ModelicaListener_exitPrimary_exp, sizeof(__pyx_k_ModelicaListener_exitPrimary_exp), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitPrimary_fal, __pyx_k_ModelicaListener_exitPrimary_fal, sizeof(__pyx_k_ModelicaListener_exitPrimary_fal), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitPrimary_fun, __pyx_k_ModelicaListener_exitPrimary_fun, sizeof(__pyx_k_ModelicaListener_exitPrimary_fun), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitPrimary_fun_2, __pyx_k_ModelicaListener_exitPrimary_fun_2, sizeof(__pyx_k_ModelicaListener_exitPrimary_fun_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitPrimary_ini, __pyx_k_ModelicaListener_exitPrimary_ini, sizeof(__pyx_k_ModelicaListener_exitPrimary_ini), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitPrimary_out, __pyx_k_ModelicaListener_exitPrimary_out, sizeof(__pyx_k_ModelicaListener_exitPrimary_out), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitPrimary_str, __pyx_k_ModelicaListener_exitPrimary_str, sizeof(__pyx_k_ModelicaListener_exitPrimary_str), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitPrimary_tru, __pyx_k_ModelicaListener_exitPrimary_tru, sizeof(__pyx_k_ModelicaListener_exitPrimary_tru), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitPrimary_uns, __pyx_k_ModelicaListener_exitPrimary_uns, sizeof(__pyx_k_ModelicaListener_exitPrimary_uns), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitRegular_ele, __pyx_k_ModelicaListener_exitRegular_ele, sizeof(__pyx_k_ModelicaListener_exitRegular_ele), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitReplaceable, __pyx_k_ModelicaListener_exitReplaceable, sizeof(__pyx_k_ModelicaListener_exitReplaceable), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitShort_class, __pyx_k_ModelicaListener_exitShort_class, sizeof(__pyx_k_ModelicaListener_exitShort_class), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitSimple_expr, __pyx_k_ModelicaListener_exitSimple_expr, sizeof(__pyx_k_ModelicaListener_exitSimple_expr), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitStatement, __pyx_k_ModelicaListener_exitStatement, sizeof(__pyx_k_ModelicaListener_exitStatement), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitStatement_b, __pyx_k_ModelicaListener_exitStatement_b, sizeof(__pyx_k_ModelicaListener_exitStatement_b), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitStatement_c, __pyx_k_ModelicaListener_exitStatement_c, sizeof(__pyx_k_ModelicaListener_exitStatement_c), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitStatement_c_2, __pyx_k_ModelicaListener_exitStatement_c_2, sizeof(__pyx_k_ModelicaListener_exitStatement_c_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitStatement_f, __pyx_k_ModelicaListener_exitStatement_f, sizeof(__pyx_k_ModelicaListener_exitStatement_f), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitStatement_i, __pyx_k_ModelicaListener_exitStatement_i, sizeof(__pyx_k_ModelicaListener_exitStatement_i), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitStatement_r, __pyx_k_ModelicaListener_exitStatement_r, sizeof(__pyx_k_ModelicaListener_exitStatement_r), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitStatement_w, __pyx_k_ModelicaListener_exitStatement_w, sizeof(__pyx_k_ModelicaListener_exitStatement_w), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitStatement_w_2, __pyx_k_ModelicaListener_exitStatement_w_2, sizeof(__pyx_k_ModelicaListener_exitStatement_w_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitStored_defi, __pyx_k_ModelicaListener_exitStored_defi, sizeof(__pyx_k_ModelicaListener_exitStored_defi), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitStored_defi_2, __pyx_k_ModelicaListener_exitStored_defi_2, sizeof(__pyx_k_ModelicaListener_exitStored_defi_2), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitString_comm, __pyx_k_ModelicaListener_exitString_comm, sizeof(__pyx_k_ModelicaListener_exitString_comm), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitSubscript, __pyx_k_ModelicaListener_exitSubscript, sizeof(__pyx_k_ModelicaListener_exitSubscript), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitType_prefix, __pyx_k_ModelicaListener_exitType_prefix, sizeof(__pyx_k_ModelicaListener_exitType_prefix), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitType_specif, __pyx_k_ModelicaListener_exitType_specif, sizeof(__pyx_k_ModelicaListener_exitType_specif), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitWhen_equati, __pyx_k_ModelicaListener_exitWhen_equati, sizeof(__pyx_k_ModelicaListener_exitWhen_equati), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitWhen_statem, __pyx_k_ModelicaListener_exitWhen_statem, sizeof(__pyx_k_ModelicaListener_exitWhen_statem), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaListener_exitWhile_state, __pyx_k_ModelicaListener_exitWhile_state, sizeof(__pyx_k_ModelicaListener_exitWhile_state), 0, 0, 1, 1},
  {&__pyx_n_s_ModelicaParser, __pyx_k_ModelicaParser, sizeof(__pyx_k_ModelicaParser), 0, 0, 1, 1},
  {&__pyx_n_s_Modification_assignment2Context, __pyx_k_Modification_assignment2Context, sizeof(__pyx_k_Modification_assignment2Context), 0, 0, 1, 1},
  {&__pyx_n_s_Modification_assignmentContext, __pyx_k_Modification_assignmentContext, sizeof(__pyx_k_Modification_assignmentContext), 0, 0, 1, 1},
  {&__pyx_n_s_Modification_classContext, __pyx_k_Modification_classContext, sizeof(__pyx_k_Modification_classContext), 0, 0, 1, 1},
  {&__pyx_n_s_NameContext, __pyx_k_NameContext, sizeof(__pyx_k_NameContext), 0, 0, 1, 1},
  {&__pyx_n_s_Named_argumentContext, __pyx_k_Named_argumentContext, sizeof(__pyx_k_Named_argumentContext), 0, 0, 1, 1},
  {&__pyx_n_s_Named_argumentsContext, __pyx_k_Named_argumentsContext, sizeof(__pyx_k_Named_argumentsContext), 0, 0, 1, 1},
  {&__pyx_n_s_Output_expression_listContext, __pyx_k_Output_expression_listContext, sizeof(__pyx_k_Output_expression_listContext), 0, 0, 1, 1},
  {&__pyx_n_s_ParseTreeListener, __pyx_k_ParseTreeListener, sizeof(__pyx_k_ParseTreeListener), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_component_referenceConte, __pyx_k_Primary_component_referenceConte, sizeof(__pyx_k_Primary_component_referenceConte), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_derivativeContext, __pyx_k_Primary_derivativeContext, sizeof(__pyx_k_Primary_derivativeContext), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_endContext, __pyx_k_Primary_endContext, sizeof(__pyx_k_Primary_endContext), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_expression_listContext, __pyx_k_Primary_expression_listContext, sizeof(__pyx_k_Primary_expression_listContext), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_falseContext, __pyx_k_Primary_falseContext, sizeof(__pyx_k_Primary_falseContext), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_functionContext, __pyx_k_Primary_functionContext, sizeof(__pyx_k_Primary_functionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_function_argumentsContex, __pyx_k_Primary_function_argumentsContex, sizeof(__pyx_k_Primary_function_argumentsContex), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_initialContext, __pyx_k_Primary_initialContext, sizeof(__pyx_k_Primary_initialContext), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_output_expression_listCo, __pyx_k_Primary_output_expression_listCo, sizeof(__pyx_k_Primary_output_expression_listCo), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_stringContext, __pyx_k_Primary_stringContext, sizeof(__pyx_k_Primary_stringContext), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_trueContext, __pyx_k_Primary_trueContext, sizeof(__pyx_k_Primary_trueContext), 0, 0, 1, 1},
  {&__pyx_n_s_Primary_unsigned_numberContext, __pyx_k_Primary_unsigned_numberContext, sizeof(__pyx_k_Primary_unsigned_numberContext), 0, 0, 1, 1},
  {&__pyx_n_s_Regular_elementContext, __pyx_k_Regular_elementContext, sizeof(__pyx_k_Regular_elementContext), 0, 0, 1, 1},
  {&__pyx_n_s_Replaceable_elementContext, __pyx_k_Replaceable_elementContext, sizeof(__pyx_k_Replaceable_elementContext), 0, 0, 1, 1},
  {&__pyx_n_s_Short_class_definitionContext, __pyx_k_Short_class_definitionContext, sizeof(__pyx_k_Short_class_definitionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Simple_expressionContext, __pyx_k_Simple_expressionContext, sizeof(__pyx_k_Simple_expressionContext), 0, 0, 1, 1},
  {&__pyx_n_s_StatementContext, __pyx_k_StatementContext, sizeof(__pyx_k_StatementContext), 0, 0, 1, 1},
  {&__pyx_n_s_Statement_breakContext, __pyx_k_Statement_breakContext, sizeof(__pyx_k_Statement_breakContext), 0, 0, 1, 1},
  {&__pyx_n_s_Statement_component_functionCont, __pyx_k_Statement_component_functionCont, sizeof(__pyx_k_Statement_component_functionCont), 0, 0, 1, 1},
  {&__pyx_n_s_Statement_component_referenceCon, __pyx_k_Statement_component_referenceCon, sizeof(__pyx_k_Statement_component_referenceCon), 0, 0, 1, 1},
  {&__pyx_n_s_Statement_forContext, __pyx_k_Statement_forContext, sizeof(__pyx_k_Statement_forContext), 0, 0, 1, 1},
  {&__pyx_n_s_Statement_ifContext, __pyx_k_Statement_ifContext, sizeof(__pyx_k_Statement_ifContext), 0, 0, 1, 1},
  {&__pyx_n_s_Statement_returnContext, __pyx_k_Statement_returnContext, sizeof(__pyx_k_Statement_returnContext), 0, 0, 1, 1},
  {&__pyx_n_s_Statement_whenContext, __pyx_k_Statement_whenContext, sizeof(__pyx_k_Statement_whenContext), 0, 0, 1, 1},
  {&__pyx_n_s_Statement_whileContext, __pyx_k_Statement_whileContext, sizeof(__pyx_k_Statement_whileContext), 0, 0, 1, 1},
  {&__pyx_n_s_Stored_definitionContext, __pyx_k_Stored_definitionContext, sizeof(__pyx_k_Stored_definitionContext), 0, 0, 1, 1},
  {&__pyx_n_s_Stored_definition_classContext, __pyx_k_Stored_definition_classContext, sizeof(__pyx_k_Stored_definition_classContext), 0, 0, 1, 1},
  {&__pyx_n_s_String_commentContext, __pyx_k_String_commentContext, sizeof(__pyx_k_String_commentContext), 0, 0, 1, 1},
  {&__pyx_n_s_SubscriptContext, __pyx_k_SubscriptContext, sizeof(__pyx_k_SubscriptContext), 0, 0, 1, 1},
  {&__pyx_n_s_Type_prefixContext, __pyx_k_Type_prefixContext, sizeof(__pyx_k_Type_prefixContext), 0, 0, 1, 1},
  {&__pyx_n_s_Type_specifierContext, __pyx_k_Type_specifierContext, sizeof(__pyx_k_Type_specifierContext), 0, 0, 1, 1},
  {&__pyx_n_s_When_equationContext, __pyx_k_When_equationContext, sizeof(__pyx_k_When_equationContext), 0, 0, 1, 1},
  {&__pyx_n_s_When_statementContext, __pyx_k_When_statementContext, sizeof(__pyx_k_When_statementContext), 0, 0, 1, 1},
  {&__pyx_n_s_While_statementContext, __pyx_k_While_statementContext, sizeof(__pyx_k_While_statementContext), 0, 0, 1, 1},
  {&__pyx_kp_u__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 1, 0, 0},
  {&__pyx_n_s_antlr4, __pyx_k_antlr4, sizeof(__pyx_k_antlr4), 0, 0, 1, 1},
  {&__pyx_n_s_ctx, __pyx_k_ctx, sizeof(__pyx_k_ctx), 0, 0, 1, 1},
  {&__pyx_n_s_doc, __pyx_k_doc, sizeof(__pyx_k_doc), 0, 0, 1, 1},
  {&__pyx_n_s_enterAlgorithm_section, __pyx_k_enterAlgorithm_section, sizeof(__pyx_k_enterAlgorithm_section), 0, 0, 1, 1},
  {&__pyx_n_s_enterAnnotation, __pyx_k_enterAnnotation, sizeof(__pyx_k_enterAnnotation), 0, 0, 1, 1},
  {&__pyx_n_s_enterArgument, __pyx_k_enterArgument, sizeof(__pyx_k_enterArgument), 0, 0, 1, 1},
  {&__pyx_n_s_enterArgument_expression, __pyx_k_enterArgument_expression, sizeof(__pyx_k_enterArgument_expression), 0, 0, 1, 1},
  {&__pyx_n_s_enterArgument_function, __pyx_k_enterArgument_function, sizeof(__pyx_k_enterArgument_function), 0, 0, 1, 1},
  {&__pyx_n_s_enterArgument_list, __pyx_k_enterArgument_list, sizeof(__pyx_k_enterArgument_list), 0, 0, 1, 1},
  {&__pyx_n_s_enterArray_subscripts, __pyx_k_enterArray_subscripts, sizeof(__pyx_k_enterArray_subscripts), 0, 0, 1, 1},
  {&__pyx_n_s_enterBase_prefix, __pyx_k_enterBase_prefix, sizeof(__pyx_k_enterBase_prefix), 0, 0, 1, 1},
  {&__pyx_n_s_enterClass_definition, __pyx_k_enterClass_definition, sizeof(__pyx_k_enterClass_definition), 0, 0, 1, 1},
  {&__pyx_n_s_enterClass_modification, __pyx_k_enterClass_modification, sizeof(__pyx_k_enterClass_modification), 0, 0, 1, 1},
  {&__pyx_n_s_enterClass_prefixes, __pyx_k_enterClass_prefixes, sizeof(__pyx_k_enterClass_prefixes), 0, 0, 1, 1},
  {&__pyx_n_s_enterClass_spec_base, __pyx_k_enterClass_spec_base, sizeof(__pyx_k_enterClass_spec_base), 0, 0, 1, 1},
  {&__pyx_n_s_enterClass_spec_comp, __pyx_k_enterClass_spec_comp, sizeof(__pyx_k_enterClass_spec_comp), 0, 0, 1, 1},
  {&__pyx_n_s_enterClass_spec_der, __pyx_k_enterClass_spec_der, sizeof(__pyx_k_enterClass_spec_der), 0, 0, 1, 1},
  {&__pyx_n_s_enterClass_spec_enum, __pyx_k_enterClass_spec_enum, sizeof(__pyx_k_enterClass_spec_enum), 0, 0, 1, 1},
  {&__pyx_n_s_enterClass_spec_extends, __pyx_k_enterClass_spec_extends, sizeof(__pyx_k_enterClass_spec_extends), 0, 0, 1, 1},
  {&__pyx_n_s_enterClass_type, __pyx_k_enterClass_type, sizeof(__pyx_k_enterClass_type), 0, 0, 1, 1},
  {&__pyx_n_s_enterComment, __pyx_k_enterComment, sizeof(__pyx_k_enterComment), 0, 0, 1, 1},
  {&__pyx_n_s_enterComponent_clause, __pyx_k_enterComponent_clause, sizeof(__pyx_k_enterComponent_clause), 0, 0, 1, 1},
  {&__pyx_n_s_enterComponent_clause1, __pyx_k_enterComponent_clause1, sizeof(__pyx_k_enterComponent_clause1), 0, 0, 1, 1},
  {&__pyx_n_s_enterComponent_declaration, __pyx_k_enterComponent_declaration, sizeof(__pyx_k_enterComponent_declaration), 0, 0, 1, 1},
  {&__pyx_n_s_enterComponent_declaration1, __pyx_k_enterComponent_declaration1, sizeof(__pyx_k_enterComponent_declaration1), 0, 0, 1, 1},
  {&__pyx_n_s_enterComponent_list, __pyx_k_enterComponent_list, sizeof(__pyx_k_enterComponent_list), 0, 0, 1, 1},
  {&__pyx_n_s_enterComponent_reference, __pyx_k_enterComponent_reference, sizeof(__pyx_k_enterComponent_reference), 0, 0, 1, 1},
  {&__pyx_n_s_enterComponent_reference_element, __pyx_k_enterComponent_reference_element, sizeof(__pyx_k_enterComponent_reference_element), 0, 0, 1, 1},
  {&__pyx_n_s_enterComposition, __pyx_k_enterComposition, sizeof(__pyx_k_enterComposition), 0, 0, 1, 1},
  {&__pyx_n_s_enterCondition_attribute, __pyx_k_enterCondition_attribute, sizeof(__pyx_k_enterCondition_attribute), 0, 0, 1, 1},
  {&__pyx_n_s_enterConnect_clause, __pyx_k_enterConnect_clause, sizeof(__pyx_k_enterConnect_clause), 0, 0, 1, 1},
  {&__pyx_n_s_enterConstraining_clause, __pyx_k_enterConstraining_clause, sizeof(__pyx_k_enterConstraining_clause), 0, 0, 1, 1},
  {&__pyx_n_s_enterDeclaration, __pyx_k_enterDeclaration, sizeof(__pyx_k_enterDeclaration), 0, 0, 1, 1},
  {&__pyx_n_s_enterElement, __pyx_k_enterElement, sizeof(__pyx_k_enterElement), 0, 0, 1, 1},
  {&__pyx_n_s_enterElement_list, __pyx_k_enterElement_list, sizeof(__pyx_k_enterElement_list), 0, 0, 1, 1},
  {&__pyx_n_s_enterElement_modification, __pyx_k_enterElement_modification, sizeof(__pyx_k_enterElement_modification), 0, 0, 1, 1},
  {&__pyx_n_s_enterElement_modification_or_rep, __pyx_k_enterElement_modification_or_rep, sizeof(__pyx_k_enterElement_modification_or_rep), 0, 0, 1, 1},
  {&__pyx_n_s_enterElement_redeclaration, __pyx_k_enterElement_redeclaration, sizeof(__pyx_k_enterElement_redeclaration), 0, 0, 1, 1},
  {&__pyx_n_s_enterElement_replaceable, __pyx_k_enterElement_replaceable, sizeof(__pyx_k_enterElement_replaceable), 0, 0, 1, 1},
  {&__pyx_n_s_enterEnum_list, __pyx_k_enterEnum_list, sizeof(__pyx_k_enterEnum_list), 0, 0, 1, 1},
  {&__pyx_n_s_enterEnumeration_literal, __pyx_k_enterEnumeration_literal, sizeof(__pyx_k_enterEnumeration_literal), 0, 0, 1, 1},
  {&__pyx_n_s_enterEquation, __pyx_k_enterEquation, sizeof(__pyx_k_enterEquation), 0, 0, 1, 1},
  {&__pyx_n_s_enterEquation_connect_clause, __pyx_k_enterEquation_connect_clause, sizeof(__pyx_k_enterEquation_connect_clause), 0, 0, 1, 1},
  {&__pyx_n_s_enterEquation_for, __pyx_k_enterEquation_for, sizeof(__pyx_k_enterEquation_for), 0, 0, 1, 1},
  {&__pyx_n_s_enterEquation_function, __pyx_k_enterEquation_function, sizeof(__pyx_k_enterEquation_function), 0, 0, 1, 1},
  {&__pyx_n_s_enterEquation_if, __pyx_k_enterEquation_if, sizeof(__pyx_k_enterEquation_if), 0, 0, 1, 1},
  {&__pyx_n_s_enterEquation_section, __pyx_k_enterEquation_section, sizeof(__pyx_k_enterEquation_section), 0, 0, 1, 1},
  {&__pyx_n_s_enterEquation_simple, __pyx_k_enterEquation_simple, sizeof(__pyx_k_enterEquation_simple), 0, 0, 1, 1},
  {&__pyx_n_s_enterEquation_when, __pyx_k_enterEquation_when, sizeof(__pyx_k_enterEquation_when), 0, 0, 1, 1},
  {&__pyx_n_s_enterExpr_add, __pyx_k_enterExpr_add, sizeof(__pyx_k_enterExpr_add), 0, 0, 1, 1},
  {&__pyx_n_s_enterExpr_and, __pyx_k_enterExpr_and, sizeof(__pyx_k_enterExpr_and), 0, 0, 1, 1},
  {&__pyx_n_s_enterExpr_exp, __pyx_k_enterExpr_exp, sizeof(__pyx_k_enterExpr_exp), 0, 0, 1, 1},
  {&__pyx_n_s_enterExpr_mul, __pyx_k_enterExpr_mul, sizeof(__pyx_k_enterExpr_mul), 0, 0, 1, 1},
  {&__pyx_n_s_enterExpr_neg, __pyx_k_enterExpr_neg, sizeof(__pyx_k_enterExpr_neg), 0, 0, 1, 1},
  {&__pyx_n_s_enterExpr_not, __pyx_k_enterExpr_not, sizeof(__pyx_k_enterExpr_not), 0, 0, 1, 1},
  {&__pyx_n_s_enterExpr_or, __pyx_k_enterExpr_or, sizeof(__pyx_k_enterExpr_or), 0, 0, 1, 1},
  {&__pyx_n_s_enterExpr_primary, __pyx_k_enterExpr_primary, sizeof(__pyx_k_enterExpr_primary), 0, 0, 1, 1},
  {&__pyx_n_s_enterExpr_rel, __pyx_k_enterExpr_rel, sizeof(__pyx_k_enterExpr_rel), 0, 0, 1, 1},
  {&__pyx_n_s_enterExpression_if, __pyx_k_enterExpression_if, sizeof(__pyx_k_enterExpression_if), 0, 0, 1, 1},
  {&__pyx_n_s_enterExpression_list, __pyx_k_enterExpression_list, sizeof(__pyx_k_enterExpression_list), 0, 0, 1, 1},
  {&__pyx_n_s_enterExpression_simple, __pyx_k_enterExpression_simple, sizeof(__pyx_k_enterExpression_simple), 0, 0, 1, 1},
  {&__pyx_n_s_enterExtends_clause, __pyx_k_enterExtends_clause, sizeof(__pyx_k_enterExtends_clause), 0, 0, 1, 1},
  {&__pyx_n_s_enterExternal_function_call, __pyx_k_enterExternal_function_call, sizeof(__pyx_k_enterExternal_function_call), 0, 0, 1, 1},
  {&__pyx_n_s_enterFor_equation, __pyx_k_enterFor_equation, sizeof(__pyx_k_enterFor_equation), 0, 0, 1, 1},
  {&__pyx_n_s_enterFor_index, __pyx_k_enterFor_index, sizeof(__pyx_k_enterFor_index), 0, 0, 1, 1},
  {&__pyx_n_s_enterFor_indices, __pyx_k_enterFor_indices, sizeof(__pyx_k_enterFor_indices), 0, 0, 1, 1},
  {&__pyx_n_s_enterFor_statement, __pyx_k_enterFor_statement, sizeof(__pyx_k_enterFor_statement), 0, 0, 1, 1},
  {&__pyx_n_s_enterFunction_arguments, __pyx_k_enterFunction_arguments, sizeof(__pyx_k_enterFunction_arguments), 0, 0, 1, 1},
  {&__pyx_n_s_enterFunction_call_args, __pyx_k_enterFunction_call_args, sizeof(__pyx_k_enterFunction_call_args), 0, 0, 1, 1},
  {&__pyx_n_s_enterIf_equation, __pyx_k_enterIf_equation, sizeof(__pyx_k_enterIf_equation), 0, 0, 1, 1},
  {&__pyx_n_s_enterIf_statement, __pyx_k_enterIf_statement, sizeof(__pyx_k_enterIf_statement), 0, 0, 1, 1},
  {&__pyx_n_s_enterImport_clause, __pyx_k_enterImport_clause, sizeof(__pyx_k_enterImport_clause), 0, 0, 1, 1},
  {&__pyx_n_s_enterImport_list, __pyx_k_enterImport_list, sizeof(__pyx_k_enterImport_list), 0, 0, 1, 1},
  {&__pyx_n_s_enterLanguage_specification, __pyx_k_enterLanguage_specification, sizeof(__pyx_k_enterLanguage_specification), 0, 0, 1, 1},
  {&__pyx_n_s_enterModification_assignment, __pyx_k_enterModification_assignment, sizeof(__pyx_k_enterModification_assignment), 0, 0, 1, 1},
  {&__pyx_n_s_enterModification_assignment2, __pyx_k_enterModification_assignment2, sizeof(__pyx_k_enterModification_assignment2), 0, 0, 1, 1},
  {&__pyx_n_s_enterModification_class, __pyx_k_enterModification_class, sizeof(__pyx_k_enterModification_class), 0, 0, 1, 1},
  {&__pyx_n_s_enterName, __pyx_k_enterName, sizeof(__pyx_k_enterName), 0, 0, 1, 1},
  {&__pyx_n_s_enterNamed_argument, __pyx_k_enterNamed_argument, sizeof(__pyx_k_enterNamed_argument), 0, 0, 1, 1},
  {&__pyx_n_s_enterNamed_arguments, __pyx_k_enterNamed_arguments, sizeof(__pyx_k_enterNamed_arguments), 0, 0, 1, 1},
  {&__pyx_n_s_enterOutput_expression_list, __pyx_k_enterOutput_expression_list, sizeof(__pyx_k_enterOutput_expression_list), 0, 0, 1, 1},
  {&__pyx_n_s_enterPrimary_component_reference, __pyx_k_enterPrimary_component_reference, sizeof(__pyx_k_enterPrimary_component_reference), 0, 0, 1, 1},
  {&__pyx_n_s_enterPrimary_derivative, __pyx_k_enterPrimary_derivative, sizeof(__pyx_k_enterPrimary_derivative), 0, 0, 1, 1},
  {&__pyx_n_s_enterPrimary_end, __pyx_k_enterPrimary_end, sizeof(__pyx_k_enterPrimary_end), 0, 0, 1, 1},
  {&__pyx_n_s_enterPrimary_expression_list, __pyx_k_enterPrimary_expression_list, sizeof(__pyx_k_enterPrimary_expression_list), 0, 0, 1, 1},
  {&__pyx_n_s_enterPrimary_false, __pyx_k_enterPrimary_false, sizeof(__pyx_k_enterPrimary_false), 0, 0, 1, 1},
  {&__pyx_n_s_enterPrimary_function, __pyx_k_enterPrimary_function, sizeof(__pyx_k_enterPrimary_function), 0, 0, 1, 1},
  {&__pyx_n_s_enterPrimary_function_arguments, __pyx_k_enterPrimary_function_arguments, sizeof(__pyx_k_enterPrimary_function_arguments), 0, 0, 1, 1},
  {&__pyx_n_s_enterPrimary_initial, __pyx_k_enterPrimary_initial, sizeof(__pyx_k_enterPrimary_initial), 0, 0, 1, 1},
  {&__pyx_n_s_enterPrimary_output_expression_l, __pyx_k_enterPrimary_output_expression_l, sizeof(__pyx_k_enterPrimary_output_expression_l), 0, 0, 1, 1},
  {&__pyx_n_s_enterPrimary_string, __pyx_k_enterPrimary_string, sizeof(__pyx_k_enterPrimary_string), 0, 0, 1, 1},
  {&__pyx_n_s_enterPrimary_true, __pyx_k_enterPrimary_true, sizeof(__pyx_k_enterPrimary_true), 0, 0, 1, 1},
  {&__pyx_n_s_enterPrimary_unsigned_number, __pyx_k_enterPrimary_unsigned_number, sizeof(__pyx_k_enterPrimary_unsigned_number), 0, 0, 1, 1},
  {&__pyx_n_s_enterRegular_element, __pyx_k_enterRegular_element, sizeof(__pyx_k_enterRegular_element), 0, 0, 1, 1},
  {&__pyx_n_s_enterReplaceable_element, __pyx_k_enterReplaceable_element, sizeof(__pyx_k_enterReplaceable_element), 0, 0, 1, 1},
  {&__pyx_n_s_enterShort_class_definition, __pyx_k_enterShort_class_definition, sizeof(__pyx_k_enterShort_class_definition), 0, 0, 1, 1},
  {&__pyx_n_s_enterSimple_expression, __pyx_k_enterSimple_expression, sizeof(__pyx_k_enterSimple_expression), 0, 0, 1, 1},
  {&__pyx_n_s_enterStatement, __pyx_k_enterStatement, sizeof(__pyx_k_enterStatement), 0, 0, 1, 1},
  {&__pyx_n_s_enterStatement_break, __pyx_k_enterStatement_break, sizeof(__pyx_k_enterStatement_break), 0, 0, 1, 1},
  {&__pyx_n_s_enterStatement_component_functio, __pyx_k_enterStatement_component_functio, sizeof(__pyx_k_enterStatement_component_functio), 0, 0, 1, 1},
  {&__pyx_n_s_enterStatement_component_referen, __pyx_k_enterStatement_component_referen, sizeof(__pyx_k_enterStatement_component_referen), 0, 0, 1, 1},
  {&__pyx_n_s_enterStatement_for, __pyx_k_enterStatement_for, sizeof(__pyx_k_enterStatement_for), 0, 0, 1, 1},
  {&__pyx_n_s_enterStatement_if, __pyx_k_enterStatement_if, sizeof(__pyx_k_enterStatement_if), 0, 0, 1, 1},
  {&__pyx_n_s_enterStatement_return, __pyx_k_enterStatement_return, sizeof(__pyx_k_enterStatement_return), 0, 0, 1, 1},
  {&__pyx_n_s_enterStatement_when, __pyx_k_enterStatement_when, sizeof(__pyx_k_enterStatement_when), 0, 0, 1, 1},
  {&__pyx_n_s_enterStatement_while, __pyx_k_enterStatement_while, sizeof(__pyx_k_enterStatement_while), 0, 0, 1, 1},
  {&__pyx_n_s_enterStored_definition, __pyx_k_enterStored_definition, sizeof(__pyx_k_enterStored_definition), 0, 0, 1, 1},
  {&__pyx_n_s_enterStored_definition_class, __pyx_k_enterStored_definition_class, sizeof(__pyx_k_enterStored_definition_class), 0, 0, 1, 1},
  {&__pyx_n_s_enterString_comment, __pyx_k_enterString_comment, sizeof(__pyx_k_enterString_comment), 0, 0, 1, 1},
  {&__pyx_n_s_enterSubscript, __pyx_k_enterSubscript, sizeof(__pyx_k_enterSubscript), 0, 0, 1, 1},
  {&__pyx_n_s_enterType_prefix, __pyx_k_enterType_prefix, sizeof(__pyx_k_enterType_prefix), 0, 0, 1, 1},
  {&__pyx_n_s_enterType_specifier, __pyx_k_enterType_specifier, sizeof(__pyx_k_enterType_specifier), 0, 0, 1, 1},
  {&__pyx_n_s_enterWhen_equation, __pyx_k_enterWhen_equation, sizeof(__pyx_k_enterWhen_equation), 0, 0, 1, 1},
  {&__pyx_n_s_enterWhen_statement, __pyx_k_enterWhen_statement, sizeof(__pyx_k_enterWhen_statement), 0, 0, 1, 1},
  {&__pyx_n_s_enterWhile_statement, __pyx_k_enterWhile_statement, sizeof(__pyx_k_enterWhile_statement), 0, 0, 1, 1},
  {&__pyx_n_s_exitAlgorithm_section, __pyx_k_exitAlgorithm_section, sizeof(__pyx_k_exitAlgorithm_section), 0, 0, 1, 1},
  {&__pyx_n_s_exitAnnotation, __pyx_k_exitAnnotation, sizeof(__pyx_k_exitAnnotation), 0, 0, 1, 1},
  {&__pyx_n_s_exitArgument, __pyx_k_exitArgument, sizeof(__pyx_k_exitArgument), 0, 0, 1, 1},
  {&__pyx_n_s_exitArgument_expression, __pyx_k_exitArgument_expression, sizeof(__pyx_k_exitArgument_expression), 0, 0, 1, 1},
  {&__pyx_n_s_exitArgument_function, __pyx_k_exitArgument_function, sizeof(__pyx_k_exitArgument_function), 0, 0, 1, 1},
  {&__pyx_n_s_exitArgument_list, __pyx_k_exitArgument_list, sizeof(__pyx_k_exitArgument_list), 0, 0, 1, 1},
  {&__pyx_n_s_exitArray_subscripts, __pyx_k_exitArray_subscripts, sizeof(__pyx_k_exitArray_subscripts), 0, 0, 1, 1},
  {&__pyx_n_s_exitBase_prefix, __pyx_k_exitBase_prefix, sizeof(__pyx_k_exitBase_prefix), 0, 0, 1, 1},
  {&__pyx_n_s_exitClass_definition, __pyx_k_exitClass_definition, sizeof(__pyx_k_exitClass_definition), 0, 0, 1, 1},
  {&__pyx_n_s_exitClass_modification, __pyx_k_exitClass_modification, sizeof(__pyx_k_exitClass_modification), 0, 0, 1, 1},
  {&__pyx_n_s_exitClass_prefixes, __pyx_k_exitClass_prefixes, sizeof(__pyx_k_exitClass_prefixes), 0, 0, 1, 1},
  {&__pyx_n_s_exitClass_spec_base, __pyx_k_exitClass_spec_base, sizeof(__pyx_k_exitClass_spec_base), 0, 0, 1, 1},
  {&__pyx_n_s_exitClass_spec_comp, __pyx_k_exitClass_spec_comp, sizeof(__pyx_k_exitClass_spec_comp), 0, 0, 1, 1},
  {&__pyx_n_s_exitClass_spec_der, __pyx_k_exitClass_spec_der, sizeof(__pyx_k_exitClass_spec_der), 0, 0, 1, 1},
  {&__pyx_n_s_exitClass_spec_enum, __pyx_k_exitClass_spec_enum, sizeof(__pyx_k_exitClass_spec_enum), 0, 0, 1, 1},
  {&__pyx_n_s_exitClass_spec_extends, __pyx_k_exitClass_spec_extends, sizeof(__pyx_k_exitClass_spec_extends), 0, 0, 1, 1},
  {&__pyx_n_s_exitClass_type, __pyx_k_exitClass_type, sizeof(__pyx_k_exitClass_type), 0, 0, 1, 1},
  {&__pyx_n_s_exitComment, __pyx_k_exitComment, sizeof(__pyx_k_exitComment), 0, 0, 1, 1},
  {&__pyx_n_s_exitComponent_clause, __pyx_k_exitComponent_clause, sizeof(__pyx_k_exitComponent_clause), 0, 0, 1, 1},
  {&__pyx_n_s_exitComponent_clause1, __pyx_k_exitComponent_clause1, sizeof(__pyx_k_exitComponent_clause1), 0, 0, 1, 1},
  {&__pyx_n_s_exitComponent_declaration, __pyx_k_exitComponent_declaration, sizeof(__pyx_k_exitComponent_declaration), 0, 0, 1, 1},
  {&__pyx_n_s_exitComponent_declaration1, __pyx_k_exitComponent_declaration1, sizeof(__pyx_k_exitComponent_declaration1), 0, 0, 1, 1},
  {&__pyx_n_s_exitComponent_list, __pyx_k_exitComponent_list, sizeof(__pyx_k_exitComponent_list), 0, 0, 1, 1},
  {&__pyx_n_s_exitComponent_reference, __pyx_k_exitComponent_reference, sizeof(__pyx_k_exitComponent_reference), 0, 0, 1, 1},
  {&__pyx_n_s_exitComponent_reference_element, __pyx_k_exitComponent_reference_element, sizeof(__pyx_k_exitComponent_reference_element), 0, 0, 1, 1},
  {&__pyx_n_s_exitComposition, __pyx_k_exitComposition, sizeof(__pyx_k_exitComposition), 0, 0, 1, 1},
  {&__pyx_n_s_exitCondition_attribute, __pyx_k_exitCondition_attribute, sizeof(__pyx_k_exitCondition_attribute), 0, 0, 1, 1},
  {&__pyx_n_s_exitConnect_clause, __pyx_k_exitConnect_clause, sizeof(__pyx_k_exitConnect_clause), 0, 0, 1, 1},
  {&__pyx_n_s_exitConstraining_clause, __pyx_k_exitConstraining_clause, sizeof(__pyx_k_exitConstraining_clause), 0, 0, 1, 1},
  {&__pyx_n_s_exitDeclaration, __pyx_k_exitDeclaration, sizeof(__pyx_k_exitDeclaration), 0, 0, 1, 1},
  {&__pyx_n_s_exitElement, __pyx_k_exitElement, sizeof(__pyx_k_exitElement), 0, 0, 1, 1},
  {&__pyx_n_s_exitElement_list, __pyx_k_exitElement_list, sizeof(__pyx_k_exitElement_list), 0, 0, 1, 1},
  {&__pyx_n_s_exitElement_modification, __pyx_k_exitElement_modification, sizeof(__pyx_k_exitElement_modification), 0, 0, 1, 1},
  {&__pyx_n_s_exitElement_modification_or_repl, __pyx_k_exitElement_modification_or_repl, sizeof(__pyx_k_exitElement_modification_or_repl), 0, 0, 1, 1},
  {&__pyx_n_s_exitElement_redeclaration, __pyx_k_exitElement_redeclaration, sizeof(__pyx_k_exitElement_redeclaration), 0, 0, 1, 1},
  {&__pyx_n_s_exitElement_replaceable, __pyx_k_exitElement_replaceable, sizeof(__pyx_k_exitElement_replaceable), 0, 0, 1, 1},
  {&__pyx_n_s_exitEnum_list, __pyx_k_exitEnum_list, sizeof(__pyx_k_exitEnum_list), 0, 0, 1, 1},
  {&__pyx_n_s_exitEnumeration_literal, __pyx_k_exitEnumeration_literal, sizeof(__pyx_k_exitEnumeration_literal), 0, 0, 1, 1},
  {&__pyx_n_s_exitEquation, __pyx_k_exitEquation, sizeof(__pyx_k_exitEquation), 0, 0, 1, 1},
  {&__pyx_n_s_exitEquation_connect_clause, __pyx_k_exitEquation_connect_clause, sizeof(__pyx_k_exitEquation_connect_clause), 0, 0, 1, 1},
  {&__pyx_n_s_exitEquation_for, __pyx_k_exitEquation_for, sizeof(__pyx_k_exitEquation_for), 0, 0, 1, 1},
  {&__pyx_n_s_exitEquation_function, __pyx_k_exitEquation_function, sizeof(__pyx_k_exitEquation_function), 0, 0, 1, 1},
  {&__pyx_n_s_exitEquation_if, __pyx_k_exitEquation_if, sizeof(__pyx_k_exitEquation_if), 0, 0, 1, 1},
  {&__pyx_n_s_exitEquation_section, __pyx_k_exitEquation_section, sizeof(__pyx_k_exitEquation_section), 0, 0, 1, 1},
  {&__pyx_n_s_exitEquation_simple, __pyx_k_exitEquation_simple, sizeof(__pyx_k_exitEquation_simple), 0, 0, 1, 1},
  {&__pyx_n_s_exitEquation_when, __pyx_k_exitEquation_when, sizeof(__pyx_k_exitEquation_when), 0, 0, 1, 1},
  {&__pyx_n_s_exitExpr_add, __pyx_k_exitExpr_add, sizeof(__pyx_k_exitExpr_add), 0, 0, 1, 1},
  {&__pyx_n_s_exitExpr_and, __pyx_k_exitExpr_and, sizeof(__pyx_k_exitExpr_and), 0, 0, 1, 1},
  {&__pyx_n_s_exitExpr_exp, __pyx_k_exitExpr_exp, sizeof(__pyx_k_exitExpr_exp), 0, 0, 1, 1},
  {&__pyx_n_s_exitExpr_mul, __pyx_k_exitExpr_mul, sizeof(__pyx_k_exitExpr_mul), 0, 0, 1, 1},
  {&__pyx_n_s_exitExpr_neg, __pyx_k_exitExpr_neg, sizeof(__pyx_k_exitExpr_neg), 0, 0, 1, 1},
  {&__pyx_n_s_exitExpr_not, __pyx_k_exitExpr_not, sizeof(__pyx_k_exitExpr_not), 0, 0, 1, 1},
  {&__pyx_n_s_exitExpr_or, __pyx_k_exitExpr_or, sizeof(__pyx_k_exitExpr_or), 0, 0, 1, 1},
  {&__pyx_n_s_exitExpr_primary, __pyx_k_exitExpr_primary, sizeof(__pyx_k_exitExpr_primary), 0, 0, 1, 1},
  {&__pyx_n_s_exitExpr_rel, __pyx_k_exitExpr_rel, sizeof(__pyx_k_exitExpr_rel), 0, 0, 1, 1},
  {&__pyx_n_s_exitExpression_if, __pyx_k_exitExpression_if, sizeof(__pyx_k_exitExpression_if), 0, 0, 1, 1},
  {&__pyx_n_s_exitExpression_list, __pyx_k_exitExpression_list, sizeof(__pyx_k_exitExpression_list), 0, 0, 1, 1},
  {&__pyx_n_s_exitExpression_simple, __pyx_k_exitExpression_simple, sizeof(__pyx_k_exitExpression_simple), 0, 0, 1, 1},
  {&__pyx_n_s_exitExtends_clause, __pyx_k_exitExtends_clause, sizeof(__pyx_k_exitExtends_clause), 0, 0, 1, 1},
  {&__pyx_n_s_exitExternal_function_call, __pyx_k_exitExternal_function_call, sizeof(__pyx_k_exitExternal_function_call), 0, 0, 1, 1},
  {&__pyx_n_s_exitFor_equation, __pyx_k_exitFor_equation, sizeof(__pyx_k_exitFor_equation), 0, 0, 1, 1},
  {&__pyx_n_s_exitFor_index, __pyx_k_exitFor_index, sizeof(__pyx_k_exitFor_index), 0, 0, 1, 1},
  {&__pyx_n_s_exitFor_indices, __pyx_k_exitFor_indices, sizeof(__pyx_k_exitFor_indices), 0, 0, 1, 1},
  {&__pyx_n_s_exitFor_statement, __pyx_k_exitFor_statement, sizeof(__pyx_k_exitFor_statement), 0, 0, 1, 1},
  {&__pyx_n_s_exitFunction_arguments, __pyx_k_exitFunction_arguments, sizeof(__pyx_k_exitFunction_arguments), 0, 0, 1, 1},
  {&__pyx_n_s_exitFunction_call_args, __pyx_k_exitFunction_call_args, sizeof(__pyx_k_exitFunction_call_args), 0, 0, 1, 1},
  {&__pyx_n_s_exitIf_equation, __pyx_k_exitIf_equation, sizeof(__pyx_k_exitIf_equation), 0, 0, 1, 1},
  {&__pyx_n_s_exitIf_statement, __pyx_k_exitIf_statement, sizeof(__pyx_k_exitIf_statement), 0, 0, 1, 1},
  {&__pyx_n_s_exitImport_clause, __pyx_k_exitImport_clause, sizeof(__pyx_k_exitImport_clause), 0, 0, 1, 1},
  {&__pyx_n_s_exitImport_list, __pyx_k_exitImport_list, sizeof(__pyx_k_exitImport_list), 0, 0, 1, 1},
  {&__pyx_n_s_exitLanguage_specification, __pyx_k_exitLanguage_specification, sizeof(__pyx_k_exitLanguage_specification), 0, 0, 1, 1},
  {&__pyx_n_s_exitModification_assignment, __pyx_k_exitModification_assignment, sizeof(__pyx_k_exitModification_assignment), 0, 0, 1, 1},
  {&__pyx_n_s_exitModification_assignment2, __pyx_k_exitModification_assignment2, sizeof(__pyx_k_exitModification_assignment2), 0, 0, 1, 1},
  {&__pyx_n_s_exitModification_class, __pyx_k_exitModification_class, sizeof(__pyx_k_exitModification_class), 0, 0, 1, 1},
  {&__pyx_n_s_exitName, __pyx_k_exitName, sizeof(__pyx_k_exitName), 0, 0, 1, 1},
  {&__pyx_n_s_exitNamed_argument, __pyx_k_exitNamed_argument, sizeof(__pyx_k_exitNamed_argument), 0, 0, 1, 1},
  {&__pyx_n_s_exitNamed_arguments, __pyx_k_exitNamed_arguments, sizeof(__pyx_k_exitNamed_arguments), 0, 0, 1, 1},
  {&__pyx_n_s_exitOutput_expression_list, __pyx_k_exitOutput_expression_list, sizeof(__pyx_k_exitOutput_expression_list), 0, 0, 1, 1},
  {&__pyx_n_s_exitPrimary_component_reference, __pyx_k_exitPrimary_component_reference, sizeof(__pyx_k_exitPrimary_component_reference), 0, 0, 1, 1},
  {&__pyx_n_s_exitPrimary_derivative, __pyx_k_exitPrimary_derivative, sizeof(__pyx_k_exitPrimary_derivative), 0, 0, 1, 1},
  {&__pyx_n_s_exitPrimary_end, __pyx_k_exitPrimary_end, sizeof(__pyx_k_exitPrimary_end), 0, 0, 1, 1},
  {&__pyx_n_s_exitPrimary_expression_list, __pyx_k_exitPrimary_expression_list, sizeof(__pyx_k_exitPrimary_expression_list), 0, 0, 1, 1},
  {&__pyx_n_s_exitPrimary_false, __pyx_k_exitPrimary_false, sizeof(__pyx_k_exitPrimary_false), 0, 0, 1, 1},
  {&__pyx_n_s_exitPrimary_function, __pyx_k_exitPrimary_function, sizeof(__pyx_k_exitPrimary_function), 0, 0, 1, 1},
  {&__pyx_n_s_exitPrimary_function_arguments, __pyx_k_exitPrimary_function_arguments, sizeof(__pyx_k_exitPrimary_function_arguments), 0, 0, 1, 1},
  {&__pyx_n_s_exitPrimary_initial, __pyx_k_exitPrimary_initial, sizeof(__pyx_k_exitPrimary_initial), 0, 0, 1, 1},
  {&__pyx_n_s_exitPrimary_output_expression_li, __pyx_k_exitPrimary_output_expression_li, sizeof(__pyx_k_exitPrimary_output_expression_li), 0, 0, 1, 1},
  {&__pyx_n_s_exitPrimary_string, __pyx_k_exitPrimary_string, sizeof(__pyx_k_exitPrimary_string), 0, 0, 1, 1},
  {&__pyx_n_s_exitPrimary_true, __pyx_k_exitPrimary_true, sizeof(__pyx_k_exitPrimary_true), 0, 0, 1, 1},
  {&__pyx_n_s_exitPrimary_unsigned_number, __pyx_k_exitPrimary_unsigned_number, sizeof(__pyx_k_exitPrimary_unsigned_number), 0, 0, 1, 1},
  {&__pyx_n_s_exitRegular_element, __pyx_k_exitRegular_element, sizeof(__pyx_k_exitRegular_element), 0, 0, 1, 1},
  {&__pyx_n_s_exitReplaceable_element, __pyx_k_exitReplaceable_element, sizeof(__pyx_k_exitReplaceable_element), 0, 0, 1, 1},
  {&__pyx_n_s_exitShort_class_definition, __pyx_k_exitShort_class_definition, sizeof(__pyx_k_exitShort_class_definition), 0, 0, 1, 1},
  {&__pyx_n_s_exitSimple_expression, __pyx_k_exitSimple_expression, sizeof(__pyx_k_exitSimple_expression), 0, 0, 1, 1},
  {&__pyx_n_s_exitStatement, __pyx_k_exitStatement, sizeof(__pyx_k_exitStatement), 0, 0, 1, 1},
  {&__pyx_n_s_exitStatement_break, __pyx_k_exitStatement_break, sizeof(__pyx_k_exitStatement_break), 0, 0, 1, 1},
  {&__pyx_n_s_exitStatement_component_function, __pyx_k_exitStatement_component_function, sizeof(__pyx_k_exitStatement_component_function), 0, 0, 1, 1},
  {&__pyx_n_s_exitStatement_component_referenc, __pyx_k_exitStatement_component_referenc, sizeof(__pyx_k_exitStatement_component_referenc), 0, 0, 1, 1},
  {&__pyx_n_s_exitStatement_for, __pyx_k_exitStatement_for, sizeof(__pyx_k_exitStatement_for), 0, 0, 1, 1},
  {&__pyx_n_s_exitStatement_if, __pyx_k_exitStatement_if, sizeof(__pyx_k_exitStatement_if), 0, 0, 1, 1},
  {&__pyx_n_s_exitStatement_return, __pyx_k_exitStatement_return, sizeof(__pyx_k_exitStatement_return), 0, 0, 1, 1},
  {&__pyx_n_s_exitStatement_when, __pyx_k_exitStatement_when, sizeof(__pyx_k_exitStatement_when), 0, 0, 1, 1},
  {&__pyx_n_s_exitStatement_while, __pyx_k_exitStatement_while, sizeof(__pyx_k_exitStatement_while), 0, 0, 1, 1},
  {&__pyx_n_s_exitStored_definition, __pyx_k_exitStored_definition, sizeof(__pyx_k_exitStored_definition), 0, 0, 1, 1},
  {&__pyx_n_s_exitStored_definition_class, __pyx_k_exitStored_definition_class, sizeof(__pyx_k_exitStored_definition_class), 0, 0, 1, 1},
  {&__pyx_n_s_exitString_comment, __pyx_k_exitString_comment, sizeof(__pyx_k_exitString_comment), 0, 0, 1, 1},
  {&__pyx_n_s_exitSubscript, __pyx_k_exitSubscript, sizeof(__pyx_k_exitSubscript), 0, 0, 1, 1},
  {&__pyx_n_s_exitType_prefix, __pyx_k_exitType_prefix, sizeof(__pyx_k_exitType_prefix), 0, 0, 1, 1},
  {&__pyx_n_s_exitType_specifier, __pyx_k_exitType_specifier, sizeof(__pyx_k_exitType_specifier), 0, 0, 1, 1},
  {&__pyx_n_s_exitWhen_equation, __pyx_k_exitWhen_equation, sizeof(__pyx_k_exitWhen_equation), 0, 0, 1, 1},
  {&__pyx_n_s_exitWhen_statement, __pyx_k_exitWhen_statement, sizeof(__pyx_k_exitWhen_statement), 0, 0, 1, 1},
  {&__pyx_n_s_exitWhile_statement, __pyx_k_exitWhile_statement, sizeof(__pyx_k_exitWhile_statement), 0, 0, 1, 1},
  {&__pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_k_home_jgoppert_git_pymola_pymola, sizeof(__pyx_k_home_jgoppert_git_pymola_pymola), 0, 0, 1, 0},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_metaclass, __pyx_k_metaclass, sizeof(__pyx_k_metaclass), 0, 0, 1, 1},
  {&__pyx_n_s_module, __pyx_k_module, sizeof(__pyx_k_module), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_prepare, __pyx_k_prepare, sizeof(__pyx_k_prepare), 0, 0, 1, 1},
  {&__pyx_n_s_pymola_generated_ModelicaListene, __pyx_k_pymola_generated_ModelicaListene, sizeof(__pyx_k_pymola_generated_ModelicaListene), 0, 0, 1, 1},
  {&__pyx_n_s_qualname, __pyx_k_qualname, sizeof(__pyx_k_qualname), 0, 0, 1, 1},
  {&__pyx_n_s_self, __pyx_k_self, sizeof(__pyx_k_self), 0, 0, 1, 1},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  return 0;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "pymola/generated/ModelicaListener.pyx":12
 * 
 *     # Enter a parse tree produced by ModelicaParser#stored_definition.
 *     def enterStored_definition(self, ctx:ModelicaParser.Stored_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__3 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);
  __pyx_codeobj__4 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__3, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterStored_definition, 12, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__4)) __PYX_ERR(0, 12, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":16
 * 
 *     # Exit a parse tree produced by ModelicaParser#stored_definition.
 *     def exitStored_definition(self, ctx:ModelicaParser.Stored_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__5 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);
  __pyx_codeobj__6 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__5, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitStored_definition, 16, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__6)) __PYX_ERR(0, 16, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":21
 * 
 *     # Enter a parse tree produced by ModelicaParser#stored_definition_class.
 *     def enterStored_definition_class(self, ctx:ModelicaParser.Stored_definition_classContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__7 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);
  __pyx_codeobj__8 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__7, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterStored_definition_class, 21, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__8)) __PYX_ERR(0, 21, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":25
 * 
 *     # Exit a parse tree produced by ModelicaParser#stored_definition_class.
 *     def exitStored_definition_class(self, ctx:ModelicaParser.Stored_definition_classContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__9 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);
  __pyx_codeobj__10 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__9, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitStored_definition_class, 25, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__10)) __PYX_ERR(0, 25, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":30
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_definition.
 *     def enterClass_definition(self, ctx:ModelicaParser.Class_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__11 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);
  __pyx_codeobj__12 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__11, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterClass_definition, 30, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__12)) __PYX_ERR(0, 30, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":34
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_definition.
 *     def exitClass_definition(self, ctx:ModelicaParser.Class_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__13 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);
  __pyx_codeobj__14 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__13, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitClass_definition, 34, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__14)) __PYX_ERR(0, 34, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":39
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_prefixes.
 *     def enterClass_prefixes(self, ctx:ModelicaParser.Class_prefixesContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__15 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);
  __pyx_codeobj__16 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__15, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterClass_prefixes, 39, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__16)) __PYX_ERR(0, 39, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":43
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_prefixes.
 *     def exitClass_prefixes(self, ctx:ModelicaParser.Class_prefixesContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__17 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);
  __pyx_codeobj__18 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__17, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitClass_prefixes, 43, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__18)) __PYX_ERR(0, 43, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":48
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_type.
 *     def enterClass_type(self, ctx:ModelicaParser.Class_typeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__19 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);
  __pyx_codeobj__20 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__19, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterClass_type, 48, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__20)) __PYX_ERR(0, 48, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":52
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_type.
 *     def exitClass_type(self, ctx:ModelicaParser.Class_typeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__21 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);
  __pyx_codeobj__22 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__21, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitClass_type, 52, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__22)) __PYX_ERR(0, 52, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":57
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_spec_comp.
 *     def enterClass_spec_comp(self, ctx:ModelicaParser.Class_spec_compContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__23 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);
  __pyx_codeobj__24 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__23, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterClass_spec_comp, 57, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__24)) __PYX_ERR(0, 57, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":61
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_spec_comp.
 *     def exitClass_spec_comp(self, ctx:ModelicaParser.Class_spec_compContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__25 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);
  __pyx_codeobj__26 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__25, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitClass_spec_comp, 61, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__26)) __PYX_ERR(0, 61, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":66
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_spec_base.
 *     def enterClass_spec_base(self, ctx:ModelicaParser.Class_spec_baseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__27 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);
  __pyx_codeobj__28 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__27, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterClass_spec_base, 66, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__28)) __PYX_ERR(0, 66, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":70
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_spec_base.
 *     def exitClass_spec_base(self, ctx:ModelicaParser.Class_spec_baseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__29 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);
  __pyx_codeobj__30 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__29, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitClass_spec_base, 70, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__30)) __PYX_ERR(0, 70, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":75
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_spec_enum.
 *     def enterClass_spec_enum(self, ctx:ModelicaParser.Class_spec_enumContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__31 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);
  __pyx_codeobj__32 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__31, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterClass_spec_enum, 75, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__32)) __PYX_ERR(0, 75, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":79
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_spec_enum.
 *     def exitClass_spec_enum(self, ctx:ModelicaParser.Class_spec_enumContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__33 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);
  __pyx_codeobj__34 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__33, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitClass_spec_enum, 79, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__34)) __PYX_ERR(0, 79, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":84
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_spec_der.
 *     def enterClass_spec_der(self, ctx:ModelicaParser.Class_spec_derContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__35 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);
  __pyx_codeobj__36 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__35, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterClass_spec_der, 84, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__36)) __PYX_ERR(0, 84, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":88
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_spec_der.
 *     def exitClass_spec_der(self, ctx:ModelicaParser.Class_spec_derContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__37 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);
  __pyx_codeobj__38 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__37, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitClass_spec_der, 88, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__38)) __PYX_ERR(0, 88, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":93
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_spec_extends.
 *     def enterClass_spec_extends(self, ctx:ModelicaParser.Class_spec_extendsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__39 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);
  __pyx_codeobj__40 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__39, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterClass_spec_extends, 93, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__40)) __PYX_ERR(0, 93, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":97
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_spec_extends.
 *     def exitClass_spec_extends(self, ctx:ModelicaParser.Class_spec_extendsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__41 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitClass_spec_extends, 97, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 97, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":102
 * 
 *     # Enter a parse tree produced by ModelicaParser#base_prefix.
 *     def enterBase_prefix(self, ctx:ModelicaParser.Base_prefixContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__43 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterBase_prefix, 102, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 102, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":106
 * 
 *     # Exit a parse tree produced by ModelicaParser#base_prefix.
 *     def exitBase_prefix(self, ctx:ModelicaParser.Base_prefixContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__45 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__45, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitBase_prefix, 106, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(0, 106, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":111
 * 
 *     # Enter a parse tree produced by ModelicaParser#enum_list.
 *     def enterEnum_list(self, ctx:ModelicaParser.Enum_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__47 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterEnum_list, 111, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(0, 111, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":115
 * 
 *     # Exit a parse tree produced by ModelicaParser#enum_list.
 *     def exitEnum_list(self, ctx:ModelicaParser.Enum_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__49 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__49, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitEnum_list, 115, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(0, 115, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":120
 * 
 *     # Enter a parse tree produced by ModelicaParser#enumeration_literal.
 *     def enterEnumeration_literal(self, ctx:ModelicaParser.Enumeration_literalContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__51 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__51)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__51);
  __Pyx_GIVEREF(__pyx_tuple__51);
  __pyx_codeobj__52 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__51, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterEnumeration_literal, 120, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__52)) __PYX_ERR(0, 120, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":124
 * 
 *     # Exit a parse tree produced by ModelicaParser#enumeration_literal.
 *     def exitEnumeration_literal(self, ctx:ModelicaParser.Enumeration_literalContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__53 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__53)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__53);
  __Pyx_GIVEREF(__pyx_tuple__53);
  __pyx_codeobj__54 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__53, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitEnumeration_literal, 124, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__54)) __PYX_ERR(0, 124, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":129
 * 
 *     # Enter a parse tree produced by ModelicaParser#composition.
 *     def enterComposition(self, ctx:ModelicaParser.CompositionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__55 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__55)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__55);
  __Pyx_GIVEREF(__pyx_tuple__55);
  __pyx_codeobj__56 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__55, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterComposition, 129, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__56)) __PYX_ERR(0, 129, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":133
 * 
 *     # Exit a parse tree produced by ModelicaParser#composition.
 *     def exitComposition(self, ctx:ModelicaParser.CompositionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__57 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__57)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__57);
  __Pyx_GIVEREF(__pyx_tuple__57);
  __pyx_codeobj__58 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__57, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitComposition, 133, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__58)) __PYX_ERR(0, 133, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":138
 * 
 *     # Enter a parse tree produced by ModelicaParser#language_specification.
 *     def enterLanguage_specification(self, ctx:ModelicaParser.Language_specificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__59 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__59)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__59);
  __Pyx_GIVEREF(__pyx_tuple__59);
  __pyx_codeobj__60 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__59, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterLanguage_specification, 138, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__60)) __PYX_ERR(0, 138, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":142
 * 
 *     # Exit a parse tree produced by ModelicaParser#language_specification.
 *     def exitLanguage_specification(self, ctx:ModelicaParser.Language_specificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__61 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__61)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__61);
  __Pyx_GIVEREF(__pyx_tuple__61);
  __pyx_codeobj__62 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__61, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitLanguage_specification, 142, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__62)) __PYX_ERR(0, 142, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":147
 * 
 *     # Enter a parse tree produced by ModelicaParser#external_function_call.
 *     def enterExternal_function_call(self, ctx:ModelicaParser.External_function_callContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__63 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__63)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__63);
  __Pyx_GIVEREF(__pyx_tuple__63);
  __pyx_codeobj__64 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__63, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterExternal_function_call, 147, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__64)) __PYX_ERR(0, 147, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":151
 * 
 *     # Exit a parse tree produced by ModelicaParser#external_function_call.
 *     def exitExternal_function_call(self, ctx:ModelicaParser.External_function_callContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__65 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__65)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__65);
  __Pyx_GIVEREF(__pyx_tuple__65);
  __pyx_codeobj__66 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__65, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitExternal_function_call, 151, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__66)) __PYX_ERR(0, 151, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":156
 * 
 *     # Enter a parse tree produced by ModelicaParser#element_list.
 *     def enterElement_list(self, ctx:ModelicaParser.Element_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__67 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__67)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__67);
  __Pyx_GIVEREF(__pyx_tuple__67);
  __pyx_codeobj__68 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__67, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterElement_list, 156, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__68)) __PYX_ERR(0, 156, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":160
 * 
 *     # Exit a parse tree produced by ModelicaParser#element_list.
 *     def exitElement_list(self, ctx:ModelicaParser.Element_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__69 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__69)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__69);
  __Pyx_GIVEREF(__pyx_tuple__69);
  __pyx_codeobj__70 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__69, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitElement_list, 160, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__70)) __PYX_ERR(0, 160, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":165
 * 
 *     # Enter a parse tree produced by ModelicaParser#element.
 *     def enterElement(self, ctx:ModelicaParser.ElementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__71 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__71)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__71);
  __Pyx_GIVEREF(__pyx_tuple__71);
  __pyx_codeobj__72 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__71, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterElement, 165, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__72)) __PYX_ERR(0, 165, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":169
 * 
 *     # Exit a parse tree produced by ModelicaParser#element.
 *     def exitElement(self, ctx:ModelicaParser.ElementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__73 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__73)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__73);
  __Pyx_GIVEREF(__pyx_tuple__73);
  __pyx_codeobj__74 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__73, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitElement, 169, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__74)) __PYX_ERR(0, 169, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":174
 * 
 *     # Enter a parse tree produced by ModelicaParser#regular_element.
 *     def enterRegular_element(self, ctx:ModelicaParser.Regular_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__75 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__75)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__75);
  __Pyx_GIVEREF(__pyx_tuple__75);
  __pyx_codeobj__76 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__75, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterRegular_element, 174, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__76)) __PYX_ERR(0, 174, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":178
 * 
 *     # Exit a parse tree produced by ModelicaParser#regular_element.
 *     def exitRegular_element(self, ctx:ModelicaParser.Regular_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__77 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__77)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__77);
  __Pyx_GIVEREF(__pyx_tuple__77);
  __pyx_codeobj__78 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__77, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitRegular_element, 178, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__78)) __PYX_ERR(0, 178, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":183
 * 
 *     # Enter a parse tree produced by ModelicaParser#replaceable_element.
 *     def enterReplaceable_element(self, ctx:ModelicaParser.Replaceable_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__79 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__79)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__79);
  __Pyx_GIVEREF(__pyx_tuple__79);
  __pyx_codeobj__80 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__79, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterReplaceable_element, 183, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__80)) __PYX_ERR(0, 183, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":187
 * 
 *     # Exit a parse tree produced by ModelicaParser#replaceable_element.
 *     def exitReplaceable_element(self, ctx:ModelicaParser.Replaceable_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__81 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__81)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__81);
  __Pyx_GIVEREF(__pyx_tuple__81);
  __pyx_codeobj__82 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__81, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitReplaceable_element, 187, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__82)) __PYX_ERR(0, 187, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":192
 * 
 *     # Enter a parse tree produced by ModelicaParser#import_clause.
 *     def enterImport_clause(self, ctx:ModelicaParser.Import_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__83 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__83)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__83);
  __Pyx_GIVEREF(__pyx_tuple__83);
  __pyx_codeobj__84 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__83, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterImport_clause, 192, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__84)) __PYX_ERR(0, 192, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":196
 * 
 *     # Exit a parse tree produced by ModelicaParser#import_clause.
 *     def exitImport_clause(self, ctx:ModelicaParser.Import_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__85 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__85)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__85);
  __Pyx_GIVEREF(__pyx_tuple__85);
  __pyx_codeobj__86 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__85, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitImport_clause, 196, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__86)) __PYX_ERR(0, 196, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":201
 * 
 *     # Enter a parse tree produced by ModelicaParser#import_list.
 *     def enterImport_list(self, ctx:ModelicaParser.Import_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__87 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__87)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__87);
  __Pyx_GIVEREF(__pyx_tuple__87);
  __pyx_codeobj__88 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__87, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterImport_list, 201, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__88)) __PYX_ERR(0, 201, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":205
 * 
 *     # Exit a parse tree produced by ModelicaParser#import_list.
 *     def exitImport_list(self, ctx:ModelicaParser.Import_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__89 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__89)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__89);
  __Pyx_GIVEREF(__pyx_tuple__89);
  __pyx_codeobj__90 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__89, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitImport_list, 205, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__90)) __PYX_ERR(0, 205, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":210
 * 
 *     # Enter a parse tree produced by ModelicaParser#extends_clause.
 *     def enterExtends_clause(self, ctx:ModelicaParser.Extends_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__91 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__91)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__91);
  __Pyx_GIVEREF(__pyx_tuple__91);
  __pyx_codeobj__92 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__91, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterExtends_clause, 210, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__92)) __PYX_ERR(0, 210, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":214
 * 
 *     # Exit a parse tree produced by ModelicaParser#extends_clause.
 *     def exitExtends_clause(self, ctx:ModelicaParser.Extends_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__93 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__93)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__93);
  __Pyx_GIVEREF(__pyx_tuple__93);
  __pyx_codeobj__94 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__93, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitExtends_clause, 214, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__94)) __PYX_ERR(0, 214, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":219
 * 
 *     # Enter a parse tree produced by ModelicaParser#constraining_clause.
 *     def enterConstraining_clause(self, ctx:ModelicaParser.Constraining_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__95 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__95)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__95);
  __Pyx_GIVEREF(__pyx_tuple__95);
  __pyx_codeobj__96 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__95, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterConstraining_clause, 219, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__96)) __PYX_ERR(0, 219, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":223
 * 
 *     # Exit a parse tree produced by ModelicaParser#constraining_clause.
 *     def exitConstraining_clause(self, ctx:ModelicaParser.Constraining_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__97 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__97)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__97);
  __Pyx_GIVEREF(__pyx_tuple__97);
  __pyx_codeobj__98 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__97, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitConstraining_clause, 223, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__98)) __PYX_ERR(0, 223, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":228
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_clause.
 *     def enterComponent_clause(self, ctx:ModelicaParser.Component_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__99 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__99)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__99);
  __Pyx_GIVEREF(__pyx_tuple__99);
  __pyx_codeobj__100 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__99, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterComponent_clause, 228, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__100)) __PYX_ERR(0, 228, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":232
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_clause.
 *     def exitComponent_clause(self, ctx:ModelicaParser.Component_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__101 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__101)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__101);
  __Pyx_GIVEREF(__pyx_tuple__101);
  __pyx_codeobj__102 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__101, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitComponent_clause, 232, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__102)) __PYX_ERR(0, 232, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":237
 * 
 *     # Enter a parse tree produced by ModelicaParser#type_prefix.
 *     def enterType_prefix(self, ctx:ModelicaParser.Type_prefixContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__103 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__103)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__103);
  __Pyx_GIVEREF(__pyx_tuple__103);
  __pyx_codeobj__104 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__103, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterType_prefix, 237, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__104)) __PYX_ERR(0, 237, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":241
 * 
 *     # Exit a parse tree produced by ModelicaParser#type_prefix.
 *     def exitType_prefix(self, ctx:ModelicaParser.Type_prefixContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__105 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__105)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__105);
  __Pyx_GIVEREF(__pyx_tuple__105);
  __pyx_codeobj__106 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__105, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitType_prefix, 241, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__106)) __PYX_ERR(0, 241, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":246
 * 
 *     # Enter a parse tree produced by ModelicaParser#type_specifier.
 *     def enterType_specifier(self, ctx:ModelicaParser.Type_specifierContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__107 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__107)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__107);
  __Pyx_GIVEREF(__pyx_tuple__107);
  __pyx_codeobj__108 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__107, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterType_specifier, 246, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__108)) __PYX_ERR(0, 246, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":250
 * 
 *     # Exit a parse tree produced by ModelicaParser#type_specifier.
 *     def exitType_specifier(self, ctx:ModelicaParser.Type_specifierContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__109 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__109)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__109);
  __Pyx_GIVEREF(__pyx_tuple__109);
  __pyx_codeobj__110 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__109, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitType_specifier, 250, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__110)) __PYX_ERR(0, 250, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":255
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_list.
 *     def enterComponent_list(self, ctx:ModelicaParser.Component_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__111 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__111)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__111);
  __Pyx_GIVEREF(__pyx_tuple__111);
  __pyx_codeobj__112 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__111, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterComponent_list, 255, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__112)) __PYX_ERR(0, 255, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":259
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_list.
 *     def exitComponent_list(self, ctx:ModelicaParser.Component_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__113 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__113)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__113);
  __Pyx_GIVEREF(__pyx_tuple__113);
  __pyx_codeobj__114 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__113, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitComponent_list, 259, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__114)) __PYX_ERR(0, 259, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":264
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_declaration.
 *     def enterComponent_declaration(self, ctx:ModelicaParser.Component_declarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__115 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__115)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__115);
  __Pyx_GIVEREF(__pyx_tuple__115);
  __pyx_codeobj__116 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__115, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterComponent_declaration, 264, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__116)) __PYX_ERR(0, 264, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":268
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_declaration.
 *     def exitComponent_declaration(self, ctx:ModelicaParser.Component_declarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__117 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__117)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__117);
  __Pyx_GIVEREF(__pyx_tuple__117);
  __pyx_codeobj__118 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__117, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitComponent_declaration, 268, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__118)) __PYX_ERR(0, 268, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":273
 * 
 *     # Enter a parse tree produced by ModelicaParser#condition_attribute.
 *     def enterCondition_attribute(self, ctx:ModelicaParser.Condition_attributeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__119 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__119)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__119);
  __Pyx_GIVEREF(__pyx_tuple__119);
  __pyx_codeobj__120 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__119, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterCondition_attribute, 273, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__120)) __PYX_ERR(0, 273, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":277
 * 
 *     # Exit a parse tree produced by ModelicaParser#condition_attribute.
 *     def exitCondition_attribute(self, ctx:ModelicaParser.Condition_attributeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__121 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__121)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__121);
  __Pyx_GIVEREF(__pyx_tuple__121);
  __pyx_codeobj__122 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__121, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitCondition_attribute, 277, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__122)) __PYX_ERR(0, 277, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":282
 * 
 *     # Enter a parse tree produced by ModelicaParser#declaration.
 *     def enterDeclaration(self, ctx:ModelicaParser.DeclarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__123 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__123)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__123);
  __Pyx_GIVEREF(__pyx_tuple__123);
  __pyx_codeobj__124 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__123, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterDeclaration, 282, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__124)) __PYX_ERR(0, 282, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":286
 * 
 *     # Exit a parse tree produced by ModelicaParser#declaration.
 *     def exitDeclaration(self, ctx:ModelicaParser.DeclarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__125 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__125)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__125);
  __Pyx_GIVEREF(__pyx_tuple__125);
  __pyx_codeobj__126 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__125, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitDeclaration, 286, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__126)) __PYX_ERR(0, 286, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":291
 * 
 *     # Enter a parse tree produced by ModelicaParser#modification_class.
 *     def enterModification_class(self, ctx:ModelicaParser.Modification_classContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__127 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__127)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__127);
  __Pyx_GIVEREF(__pyx_tuple__127);
  __pyx_codeobj__128 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__127, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterModification_class, 291, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__128)) __PYX_ERR(0, 291, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":295
 * 
 *     # Exit a parse tree produced by ModelicaParser#modification_class.
 *     def exitModification_class(self, ctx:ModelicaParser.Modification_classContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__129 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__129)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__129);
  __Pyx_GIVEREF(__pyx_tuple__129);
  __pyx_codeobj__130 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__129, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitModification_class, 295, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__130)) __PYX_ERR(0, 295, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":300
 * 
 *     # Enter a parse tree produced by ModelicaParser#modification_assignment.
 *     def enterModification_assignment(self, ctx:ModelicaParser.Modification_assignmentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__131 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__131)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__131);
  __Pyx_GIVEREF(__pyx_tuple__131);
  __pyx_codeobj__132 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__131, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterModification_assignment, 300, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__132)) __PYX_ERR(0, 300, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":304
 * 
 *     # Exit a parse tree produced by ModelicaParser#modification_assignment.
 *     def exitModification_assignment(self, ctx:ModelicaParser.Modification_assignmentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__133 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__133)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__133);
  __Pyx_GIVEREF(__pyx_tuple__133);
  __pyx_codeobj__134 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__133, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitModification_assignment, 304, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__134)) __PYX_ERR(0, 304, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":309
 * 
 *     # Enter a parse tree produced by ModelicaParser#modification_assignment2.
 *     def enterModification_assignment2(self, ctx:ModelicaParser.Modification_assignment2Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__135 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__135)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__135);
  __Pyx_GIVEREF(__pyx_tuple__135);
  __pyx_codeobj__136 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__135, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterModification_assignment2, 309, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__136)) __PYX_ERR(0, 309, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":313
 * 
 *     # Exit a parse tree produced by ModelicaParser#modification_assignment2.
 *     def exitModification_assignment2(self, ctx:ModelicaParser.Modification_assignment2Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__137 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__137)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__137);
  __Pyx_GIVEREF(__pyx_tuple__137);
  __pyx_codeobj__138 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__137, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitModification_assignment2, 313, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__138)) __PYX_ERR(0, 313, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":318
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_modification.
 *     def enterClass_modification(self, ctx:ModelicaParser.Class_modificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__139 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__139)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__139);
  __Pyx_GIVEREF(__pyx_tuple__139);
  __pyx_codeobj__140 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__139, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterClass_modification, 318, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__140)) __PYX_ERR(0, 318, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":322
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_modification.
 *     def exitClass_modification(self, ctx:ModelicaParser.Class_modificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__141 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__141)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__141);
  __Pyx_GIVEREF(__pyx_tuple__141);
  __pyx_codeobj__142 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__141, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitClass_modification, 322, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__142)) __PYX_ERR(0, 322, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":327
 * 
 *     # Enter a parse tree produced by ModelicaParser#argument_list.
 *     def enterArgument_list(self, ctx:ModelicaParser.Argument_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__143 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__143)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__143);
  __Pyx_GIVEREF(__pyx_tuple__143);
  __pyx_codeobj__144 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__143, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterArgument_list, 327, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__144)) __PYX_ERR(0, 327, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":331
 * 
 *     # Exit a parse tree produced by ModelicaParser#argument_list.
 *     def exitArgument_list(self, ctx:ModelicaParser.Argument_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__145 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__145)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__145);
  __Pyx_GIVEREF(__pyx_tuple__145);
  __pyx_codeobj__146 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__145, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitArgument_list, 331, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__146)) __PYX_ERR(0, 331, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":336
 * 
 *     # Enter a parse tree produced by ModelicaParser#argument.
 *     def enterArgument(self, ctx:ModelicaParser.ArgumentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__147 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__147)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__147);
  __Pyx_GIVEREF(__pyx_tuple__147);
  __pyx_codeobj__148 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__147, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterArgument, 336, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__148)) __PYX_ERR(0, 336, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":340
 * 
 *     # Exit a parse tree produced by ModelicaParser#argument.
 *     def exitArgument(self, ctx:ModelicaParser.ArgumentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__149 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__149)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__149);
  __Pyx_GIVEREF(__pyx_tuple__149);
  __pyx_codeobj__150 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__149, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitArgument, 340, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__150)) __PYX_ERR(0, 340, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":345
 * 
 *     # Enter a parse tree produced by ModelicaParser#element_modification_or_replaceable.
 *     def enterElement_modification_or_replaceable(self, ctx:ModelicaParser.Element_modification_or_replaceableContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__151 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__151)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__151);
  __Pyx_GIVEREF(__pyx_tuple__151);
  __pyx_codeobj__152 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__151, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterElement_modification_or_rep, 345, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__152)) __PYX_ERR(0, 345, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":349
 * 
 *     # Exit a parse tree produced by ModelicaParser#element_modification_or_replaceable.
 *     def exitElement_modification_or_replaceable(self, ctx:ModelicaParser.Element_modification_or_replaceableContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__153 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__153)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__153);
  __Pyx_GIVEREF(__pyx_tuple__153);
  __pyx_codeobj__154 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__153, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitElement_modification_or_repl, 349, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__154)) __PYX_ERR(0, 349, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":354
 * 
 *     # Enter a parse tree produced by ModelicaParser#element_modification.
 *     def enterElement_modification(self, ctx:ModelicaParser.Element_modificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__155 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__155)) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__155);
  __Pyx_GIVEREF(__pyx_tuple__155);
  __pyx_codeobj__156 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__155, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterElement_modification, 354, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__156)) __PYX_ERR(0, 354, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":358
 * 
 *     # Exit a parse tree produced by ModelicaParser#element_modification.
 *     def exitElement_modification(self, ctx:ModelicaParser.Element_modificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__157 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__157)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__157);
  __Pyx_GIVEREF(__pyx_tuple__157);
  __pyx_codeobj__158 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__157, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitElement_modification, 358, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__158)) __PYX_ERR(0, 358, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":363
 * 
 *     # Enter a parse tree produced by ModelicaParser#element_redeclaration.
 *     def enterElement_redeclaration(self, ctx:ModelicaParser.Element_redeclarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__159 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__159)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__159);
  __Pyx_GIVEREF(__pyx_tuple__159);
  __pyx_codeobj__160 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__159, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterElement_redeclaration, 363, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__160)) __PYX_ERR(0, 363, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":367
 * 
 *     # Exit a parse tree produced by ModelicaParser#element_redeclaration.
 *     def exitElement_redeclaration(self, ctx:ModelicaParser.Element_redeclarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__161 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__161)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__161);
  __Pyx_GIVEREF(__pyx_tuple__161);
  __pyx_codeobj__162 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__161, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitElement_redeclaration, 367, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__162)) __PYX_ERR(0, 367, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":372
 * 
 *     # Enter a parse tree produced by ModelicaParser#element_replaceable.
 *     def enterElement_replaceable(self, ctx:ModelicaParser.Element_replaceableContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__163 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__163)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__163);
  __Pyx_GIVEREF(__pyx_tuple__163);
  __pyx_codeobj__164 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__163, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterElement_replaceable, 372, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__164)) __PYX_ERR(0, 372, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":376
 * 
 *     # Exit a parse tree produced by ModelicaParser#element_replaceable.
 *     def exitElement_replaceable(self, ctx:ModelicaParser.Element_replaceableContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__165 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__165)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__165);
  __Pyx_GIVEREF(__pyx_tuple__165);
  __pyx_codeobj__166 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__165, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitElement_replaceable, 376, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__166)) __PYX_ERR(0, 376, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":381
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_clause1.
 *     def enterComponent_clause1(self, ctx:ModelicaParser.Component_clause1Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__167 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__167)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__167);
  __Pyx_GIVEREF(__pyx_tuple__167);
  __pyx_codeobj__168 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__167, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterComponent_clause1, 381, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__168)) __PYX_ERR(0, 381, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":385
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_clause1.
 *     def exitComponent_clause1(self, ctx:ModelicaParser.Component_clause1Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__169 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__169)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__169);
  __Pyx_GIVEREF(__pyx_tuple__169);
  __pyx_codeobj__170 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__169, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitComponent_clause1, 385, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__170)) __PYX_ERR(0, 385, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":390
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_declaration1.
 *     def enterComponent_declaration1(self, ctx:ModelicaParser.Component_declaration1Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__171 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__171)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__171);
  __Pyx_GIVEREF(__pyx_tuple__171);
  __pyx_codeobj__172 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__171, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterComponent_declaration1, 390, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__172)) __PYX_ERR(0, 390, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":394
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_declaration1.
 *     def exitComponent_declaration1(self, ctx:ModelicaParser.Component_declaration1Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__173 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__173)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__173);
  __Pyx_GIVEREF(__pyx_tuple__173);
  __pyx_codeobj__174 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__173, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitComponent_declaration1, 394, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__174)) __PYX_ERR(0, 394, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":399
 * 
 *     # Enter a parse tree produced by ModelicaParser#short_class_definition.
 *     def enterShort_class_definition(self, ctx:ModelicaParser.Short_class_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__175 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__175)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__175);
  __Pyx_GIVEREF(__pyx_tuple__175);
  __pyx_codeobj__176 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__175, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterShort_class_definition, 399, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__176)) __PYX_ERR(0, 399, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":403
 * 
 *     # Exit a parse tree produced by ModelicaParser#short_class_definition.
 *     def exitShort_class_definition(self, ctx:ModelicaParser.Short_class_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__177 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__177)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__177);
  __Pyx_GIVEREF(__pyx_tuple__177);
  __pyx_codeobj__178 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__177, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitShort_class_definition, 403, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__178)) __PYX_ERR(0, 403, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":408
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_section.
 *     def enterEquation_section(self, ctx:ModelicaParser.Equation_sectionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__179 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__179)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__179);
  __Pyx_GIVEREF(__pyx_tuple__179);
  __pyx_codeobj__180 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__179, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterEquation_section, 408, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__180)) __PYX_ERR(0, 408, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":412
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_section.
 *     def exitEquation_section(self, ctx:ModelicaParser.Equation_sectionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__181 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__181)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__181);
  __Pyx_GIVEREF(__pyx_tuple__181);
  __pyx_codeobj__182 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__181, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitEquation_section, 412, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__182)) __PYX_ERR(0, 412, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":417
 * 
 *     # Enter a parse tree produced by ModelicaParser#algorithm_section.
 *     def enterAlgorithm_section(self, ctx:ModelicaParser.Algorithm_sectionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__183 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__183)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__183);
  __Pyx_GIVEREF(__pyx_tuple__183);
  __pyx_codeobj__184 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__183, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterAlgorithm_section, 417, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__184)) __PYX_ERR(0, 417, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":421
 * 
 *     # Exit a parse tree produced by ModelicaParser#algorithm_section.
 *     def exitAlgorithm_section(self, ctx:ModelicaParser.Algorithm_sectionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__185 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__185)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__185);
  __Pyx_GIVEREF(__pyx_tuple__185);
  __pyx_codeobj__186 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__185, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitAlgorithm_section, 421, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__186)) __PYX_ERR(0, 421, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":426
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_simple.
 *     def enterEquation_simple(self, ctx:ModelicaParser.Equation_simpleContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__187 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__187)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__187);
  __Pyx_GIVEREF(__pyx_tuple__187);
  __pyx_codeobj__188 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__187, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterEquation_simple, 426, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__188)) __PYX_ERR(0, 426, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":430
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_simple.
 *     def exitEquation_simple(self, ctx:ModelicaParser.Equation_simpleContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__189 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__189)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__189);
  __Pyx_GIVEREF(__pyx_tuple__189);
  __pyx_codeobj__190 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__189, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitEquation_simple, 430, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__190)) __PYX_ERR(0, 430, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":435
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_if.
 *     def enterEquation_if(self, ctx:ModelicaParser.Equation_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__191 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__191)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__191);
  __Pyx_GIVEREF(__pyx_tuple__191);
  __pyx_codeobj__192 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__191, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterEquation_if, 435, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__192)) __PYX_ERR(0, 435, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":439
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_if.
 *     def exitEquation_if(self, ctx:ModelicaParser.Equation_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__193 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__193)) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__193);
  __Pyx_GIVEREF(__pyx_tuple__193);
  __pyx_codeobj__194 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__193, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitEquation_if, 439, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__194)) __PYX_ERR(0, 439, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":444
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_for.
 *     def enterEquation_for(self, ctx:ModelicaParser.Equation_forContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__195 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__195)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__195);
  __Pyx_GIVEREF(__pyx_tuple__195);
  __pyx_codeobj__196 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__195, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterEquation_for, 444, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__196)) __PYX_ERR(0, 444, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":448
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_for.
 *     def exitEquation_for(self, ctx:ModelicaParser.Equation_forContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__197 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__197)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__197);
  __Pyx_GIVEREF(__pyx_tuple__197);
  __pyx_codeobj__198 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__197, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitEquation_for, 448, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__198)) __PYX_ERR(0, 448, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":453
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_connect_clause.
 *     def enterEquation_connect_clause(self, ctx:ModelicaParser.Equation_connect_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__199 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__199)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__199);
  __Pyx_GIVEREF(__pyx_tuple__199);
  __pyx_codeobj__200 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__199, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterEquation_connect_clause, 453, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__200)) __PYX_ERR(0, 453, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":457
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_connect_clause.
 *     def exitEquation_connect_clause(self, ctx:ModelicaParser.Equation_connect_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__201 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__201)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__201);
  __Pyx_GIVEREF(__pyx_tuple__201);
  __pyx_codeobj__202 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__201, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitEquation_connect_clause, 457, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__202)) __PYX_ERR(0, 457, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":462
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_when.
 *     def enterEquation_when(self, ctx:ModelicaParser.Equation_whenContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__203 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__203)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__203);
  __Pyx_GIVEREF(__pyx_tuple__203);
  __pyx_codeobj__204 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__203, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterEquation_when, 462, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__204)) __PYX_ERR(0, 462, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":466
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_when.
 *     def exitEquation_when(self, ctx:ModelicaParser.Equation_whenContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__205 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__205)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__205);
  __Pyx_GIVEREF(__pyx_tuple__205);
  __pyx_codeobj__206 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__205, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitEquation_when, 466, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__206)) __PYX_ERR(0, 466, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":471
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_function.
 *     def enterEquation_function(self, ctx:ModelicaParser.Equation_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__207 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__207)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__207);
  __Pyx_GIVEREF(__pyx_tuple__207);
  __pyx_codeobj__208 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__207, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterEquation_function, 471, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__208)) __PYX_ERR(0, 471, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":475
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_function.
 *     def exitEquation_function(self, ctx:ModelicaParser.Equation_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__209 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__209)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__209);
  __Pyx_GIVEREF(__pyx_tuple__209);
  __pyx_codeobj__210 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__209, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitEquation_function, 475, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__210)) __PYX_ERR(0, 475, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":480
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation.
 *     def enterEquation(self, ctx:ModelicaParser.EquationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__211 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__211)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__211);
  __Pyx_GIVEREF(__pyx_tuple__211);
  __pyx_codeobj__212 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__211, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterEquation, 480, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__212)) __PYX_ERR(0, 480, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":484
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation.
 *     def exitEquation(self, ctx:ModelicaParser.EquationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__213 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__213)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__213);
  __Pyx_GIVEREF(__pyx_tuple__213);
  __pyx_codeobj__214 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__213, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitEquation, 484, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__214)) __PYX_ERR(0, 484, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":489
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_component_reference.
 *     def enterStatement_component_reference(self, ctx:ModelicaParser.Statement_component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__215 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__215)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__215);
  __Pyx_GIVEREF(__pyx_tuple__215);
  __pyx_codeobj__216 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__215, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterStatement_component_referen, 489, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__216)) __PYX_ERR(0, 489, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":493
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_component_reference.
 *     def exitStatement_component_reference(self, ctx:ModelicaParser.Statement_component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__217 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__217)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__217);
  __Pyx_GIVEREF(__pyx_tuple__217);
  __pyx_codeobj__218 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__217, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitStatement_component_referenc, 493, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__218)) __PYX_ERR(0, 493, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":498
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_component_function.
 *     def enterStatement_component_function(self, ctx:ModelicaParser.Statement_component_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__219 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__219)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__219);
  __Pyx_GIVEREF(__pyx_tuple__219);
  __pyx_codeobj__220 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__219, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterStatement_component_functio, 498, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__220)) __PYX_ERR(0, 498, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":502
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_component_function.
 *     def exitStatement_component_function(self, ctx:ModelicaParser.Statement_component_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__221 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__221)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__221);
  __Pyx_GIVEREF(__pyx_tuple__221);
  __pyx_codeobj__222 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__221, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitStatement_component_function, 502, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__222)) __PYX_ERR(0, 502, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":507
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_break.
 *     def enterStatement_break(self, ctx:ModelicaParser.Statement_breakContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__223 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__223)) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__223);
  __Pyx_GIVEREF(__pyx_tuple__223);
  __pyx_codeobj__224 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__223, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterStatement_break, 507, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__224)) __PYX_ERR(0, 507, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":511
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_break.
 *     def exitStatement_break(self, ctx:ModelicaParser.Statement_breakContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__225 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__225)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__225);
  __Pyx_GIVEREF(__pyx_tuple__225);
  __pyx_codeobj__226 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__225, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitStatement_break, 511, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__226)) __PYX_ERR(0, 511, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":516
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_return.
 *     def enterStatement_return(self, ctx:ModelicaParser.Statement_returnContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__227 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__227)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__227);
  __Pyx_GIVEREF(__pyx_tuple__227);
  __pyx_codeobj__228 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__227, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterStatement_return, 516, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__228)) __PYX_ERR(0, 516, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":520
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_return.
 *     def exitStatement_return(self, ctx:ModelicaParser.Statement_returnContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__229 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__229)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__229);
  __Pyx_GIVEREF(__pyx_tuple__229);
  __pyx_codeobj__230 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__229, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitStatement_return, 520, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__230)) __PYX_ERR(0, 520, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":525
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_if.
 *     def enterStatement_if(self, ctx:ModelicaParser.Statement_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__231 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__231)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__231);
  __Pyx_GIVEREF(__pyx_tuple__231);
  __pyx_codeobj__232 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__231, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterStatement_if, 525, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__232)) __PYX_ERR(0, 525, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":529
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_if.
 *     def exitStatement_if(self, ctx:ModelicaParser.Statement_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__233 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__233)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__233);
  __Pyx_GIVEREF(__pyx_tuple__233);
  __pyx_codeobj__234 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__233, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitStatement_if, 529, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__234)) __PYX_ERR(0, 529, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":534
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_for.
 *     def enterStatement_for(self, ctx:ModelicaParser.Statement_forContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__235 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__235)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__235);
  __Pyx_GIVEREF(__pyx_tuple__235);
  __pyx_codeobj__236 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__235, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterStatement_for, 534, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__236)) __PYX_ERR(0, 534, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":538
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_for.
 *     def exitStatement_for(self, ctx:ModelicaParser.Statement_forContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__237 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__237)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__237);
  __Pyx_GIVEREF(__pyx_tuple__237);
  __pyx_codeobj__238 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__237, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitStatement_for, 538, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__238)) __PYX_ERR(0, 538, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":543
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_while.
 *     def enterStatement_while(self, ctx:ModelicaParser.Statement_whileContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__239 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__239)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__239);
  __Pyx_GIVEREF(__pyx_tuple__239);
  __pyx_codeobj__240 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__239, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterStatement_while, 543, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__240)) __PYX_ERR(0, 543, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":547
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_while.
 *     def exitStatement_while(self, ctx:ModelicaParser.Statement_whileContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__241 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__241)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__241);
  __Pyx_GIVEREF(__pyx_tuple__241);
  __pyx_codeobj__242 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__241, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitStatement_while, 547, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__242)) __PYX_ERR(0, 547, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":552
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_when.
 *     def enterStatement_when(self, ctx:ModelicaParser.Statement_whenContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__243 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__243)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__243);
  __Pyx_GIVEREF(__pyx_tuple__243);
  __pyx_codeobj__244 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__243, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterStatement_when, 552, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__244)) __PYX_ERR(0, 552, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":556
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_when.
 *     def exitStatement_when(self, ctx:ModelicaParser.Statement_whenContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__245 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__245)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__245);
  __Pyx_GIVEREF(__pyx_tuple__245);
  __pyx_codeobj__246 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__245, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitStatement_when, 556, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__246)) __PYX_ERR(0, 556, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":561
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement.
 *     def enterStatement(self, ctx:ModelicaParser.StatementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__247 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__247)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__247);
  __Pyx_GIVEREF(__pyx_tuple__247);
  __pyx_codeobj__248 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__247, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterStatement, 561, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__248)) __PYX_ERR(0, 561, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":565
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement.
 *     def exitStatement(self, ctx:ModelicaParser.StatementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__249 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__249)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__249);
  __Pyx_GIVEREF(__pyx_tuple__249);
  __pyx_codeobj__250 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__249, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitStatement, 565, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__250)) __PYX_ERR(0, 565, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":570
 * 
 *     # Enter a parse tree produced by ModelicaParser#if_equation.
 *     def enterIf_equation(self, ctx:ModelicaParser.If_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__251 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__251)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__251);
  __Pyx_GIVEREF(__pyx_tuple__251);
  __pyx_codeobj__252 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__251, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterIf_equation, 570, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__252)) __PYX_ERR(0, 570, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":574
 * 
 *     # Exit a parse tree produced by ModelicaParser#if_equation.
 *     def exitIf_equation(self, ctx:ModelicaParser.If_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__253 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__253)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__253);
  __Pyx_GIVEREF(__pyx_tuple__253);
  __pyx_codeobj__254 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__253, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitIf_equation, 574, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__254)) __PYX_ERR(0, 574, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":579
 * 
 *     # Enter a parse tree produced by ModelicaParser#if_statement.
 *     def enterIf_statement(self, ctx:ModelicaParser.If_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__255 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__255)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__255);
  __Pyx_GIVEREF(__pyx_tuple__255);
  __pyx_codeobj__256 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__255, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterIf_statement, 579, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__256)) __PYX_ERR(0, 579, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":583
 * 
 *     # Exit a parse tree produced by ModelicaParser#if_statement.
 *     def exitIf_statement(self, ctx:ModelicaParser.If_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__257 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__257)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__257);
  __Pyx_GIVEREF(__pyx_tuple__257);
  __pyx_codeobj__258 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__257, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitIf_statement, 583, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__258)) __PYX_ERR(0, 583, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":588
 * 
 *     # Enter a parse tree produced by ModelicaParser#for_equation.
 *     def enterFor_equation(self, ctx:ModelicaParser.For_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__259 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__259)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__259);
  __Pyx_GIVEREF(__pyx_tuple__259);
  __pyx_codeobj__260 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__259, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterFor_equation, 588, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__260)) __PYX_ERR(0, 588, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":592
 * 
 *     # Exit a parse tree produced by ModelicaParser#for_equation.
 *     def exitFor_equation(self, ctx:ModelicaParser.For_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__261 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__261)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__261);
  __Pyx_GIVEREF(__pyx_tuple__261);
  __pyx_codeobj__262 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__261, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitFor_equation, 592, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__262)) __PYX_ERR(0, 592, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":597
 * 
 *     # Enter a parse tree produced by ModelicaParser#for_statement.
 *     def enterFor_statement(self, ctx:ModelicaParser.For_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__263 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__263)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__263);
  __Pyx_GIVEREF(__pyx_tuple__263);
  __pyx_codeobj__264 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__263, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterFor_statement, 597, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__264)) __PYX_ERR(0, 597, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":601
 * 
 *     # Exit a parse tree produced by ModelicaParser#for_statement.
 *     def exitFor_statement(self, ctx:ModelicaParser.For_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__265 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__265)) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__265);
  __Pyx_GIVEREF(__pyx_tuple__265);
  __pyx_codeobj__266 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__265, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitFor_statement, 601, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__266)) __PYX_ERR(0, 601, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":606
 * 
 *     # Enter a parse tree produced by ModelicaParser#for_indices.
 *     def enterFor_indices(self, ctx:ModelicaParser.For_indicesContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__267 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__267)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__267);
  __Pyx_GIVEREF(__pyx_tuple__267);
  __pyx_codeobj__268 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__267, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterFor_indices, 606, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__268)) __PYX_ERR(0, 606, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":610
 * 
 *     # Exit a parse tree produced by ModelicaParser#for_indices.
 *     def exitFor_indices(self, ctx:ModelicaParser.For_indicesContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__269 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__269)) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__269);
  __Pyx_GIVEREF(__pyx_tuple__269);
  __pyx_codeobj__270 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__269, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitFor_indices, 610, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__270)) __PYX_ERR(0, 610, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":615
 * 
 *     # Enter a parse tree produced by ModelicaParser#for_index.
 *     def enterFor_index(self, ctx:ModelicaParser.For_indexContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__271 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__271)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__271);
  __Pyx_GIVEREF(__pyx_tuple__271);
  __pyx_codeobj__272 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__271, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterFor_index, 615, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__272)) __PYX_ERR(0, 615, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":619
 * 
 *     # Exit a parse tree produced by ModelicaParser#for_index.
 *     def exitFor_index(self, ctx:ModelicaParser.For_indexContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__273 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__273)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__273);
  __Pyx_GIVEREF(__pyx_tuple__273);
  __pyx_codeobj__274 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__273, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitFor_index, 619, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__274)) __PYX_ERR(0, 619, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":624
 * 
 *     # Enter a parse tree produced by ModelicaParser#while_statement.
 *     def enterWhile_statement(self, ctx:ModelicaParser.While_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__275 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__275)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__275);
  __Pyx_GIVEREF(__pyx_tuple__275);
  __pyx_codeobj__276 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__275, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterWhile_statement, 624, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__276)) __PYX_ERR(0, 624, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":628
 * 
 *     # Exit a parse tree produced by ModelicaParser#while_statement.
 *     def exitWhile_statement(self, ctx:ModelicaParser.While_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__277 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__277)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__277);
  __Pyx_GIVEREF(__pyx_tuple__277);
  __pyx_codeobj__278 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__277, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitWhile_statement, 628, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__278)) __PYX_ERR(0, 628, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":633
 * 
 *     # Enter a parse tree produced by ModelicaParser#when_equation.
 *     def enterWhen_equation(self, ctx:ModelicaParser.When_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__279 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__279)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__279);
  __Pyx_GIVEREF(__pyx_tuple__279);
  __pyx_codeobj__280 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__279, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterWhen_equation, 633, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__280)) __PYX_ERR(0, 633, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":637
 * 
 *     # Exit a parse tree produced by ModelicaParser#when_equation.
 *     def exitWhen_equation(self, ctx:ModelicaParser.When_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__281 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__281)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__281);
  __Pyx_GIVEREF(__pyx_tuple__281);
  __pyx_codeobj__282 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__281, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitWhen_equation, 637, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__282)) __PYX_ERR(0, 637, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":642
 * 
 *     # Enter a parse tree produced by ModelicaParser#when_statement.
 *     def enterWhen_statement(self, ctx:ModelicaParser.When_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__283 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__283)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__283);
  __Pyx_GIVEREF(__pyx_tuple__283);
  __pyx_codeobj__284 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__283, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterWhen_statement, 642, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__284)) __PYX_ERR(0, 642, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":646
 * 
 *     # Exit a parse tree produced by ModelicaParser#when_statement.
 *     def exitWhen_statement(self, ctx:ModelicaParser.When_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__285 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__285)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__285);
  __Pyx_GIVEREF(__pyx_tuple__285);
  __pyx_codeobj__286 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__285, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitWhen_statement, 646, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__286)) __PYX_ERR(0, 646, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":651
 * 
 *     # Enter a parse tree produced by ModelicaParser#connect_clause.
 *     def enterConnect_clause(self, ctx:ModelicaParser.Connect_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__287 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__287)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__287);
  __Pyx_GIVEREF(__pyx_tuple__287);
  __pyx_codeobj__288 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__287, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterConnect_clause, 651, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__288)) __PYX_ERR(0, 651, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":655
 * 
 *     # Exit a parse tree produced by ModelicaParser#connect_clause.
 *     def exitConnect_clause(self, ctx:ModelicaParser.Connect_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__289 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__289)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__289);
  __Pyx_GIVEREF(__pyx_tuple__289);
  __pyx_codeobj__290 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__289, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitConnect_clause, 655, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__290)) __PYX_ERR(0, 655, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":660
 * 
 *     # Enter a parse tree produced by ModelicaParser#expression_simple.
 *     def enterExpression_simple(self, ctx:ModelicaParser.Expression_simpleContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__291 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__291)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__291);
  __Pyx_GIVEREF(__pyx_tuple__291);
  __pyx_codeobj__292 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__291, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterExpression_simple, 660, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__292)) __PYX_ERR(0, 660, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":664
 * 
 *     # Exit a parse tree produced by ModelicaParser#expression_simple.
 *     def exitExpression_simple(self, ctx:ModelicaParser.Expression_simpleContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__293 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__293)) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__293);
  __Pyx_GIVEREF(__pyx_tuple__293);
  __pyx_codeobj__294 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__293, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitExpression_simple, 664, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__294)) __PYX_ERR(0, 664, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":669
 * 
 *     # Enter a parse tree produced by ModelicaParser#expression_if.
 *     def enterExpression_if(self, ctx:ModelicaParser.Expression_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__295 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__295)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__295);
  __Pyx_GIVEREF(__pyx_tuple__295);
  __pyx_codeobj__296 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__295, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterExpression_if, 669, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__296)) __PYX_ERR(0, 669, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":673
 * 
 *     # Exit a parse tree produced by ModelicaParser#expression_if.
 *     def exitExpression_if(self, ctx:ModelicaParser.Expression_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__297 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__297)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__297);
  __Pyx_GIVEREF(__pyx_tuple__297);
  __pyx_codeobj__298 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__297, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitExpression_if, 673, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__298)) __PYX_ERR(0, 673, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":678
 * 
 *     # Enter a parse tree produced by ModelicaParser#simple_expression.
 *     def enterSimple_expression(self, ctx:ModelicaParser.Simple_expressionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__299 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__299)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__299);
  __Pyx_GIVEREF(__pyx_tuple__299);
  __pyx_codeobj__300 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__299, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterSimple_expression, 678, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__300)) __PYX_ERR(0, 678, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":682
 * 
 *     # Exit a parse tree produced by ModelicaParser#simple_expression.
 *     def exitSimple_expression(self, ctx:ModelicaParser.Simple_expressionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__301 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__301)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__301);
  __Pyx_GIVEREF(__pyx_tuple__301);
  __pyx_codeobj__302 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__301, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitSimple_expression, 682, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__302)) __PYX_ERR(0, 682, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":687
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_neg.
 *     def enterExpr_neg(self, ctx:ModelicaParser.Expr_negContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__303 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__303)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__303);
  __Pyx_GIVEREF(__pyx_tuple__303);
  __pyx_codeobj__304 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__303, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterExpr_neg, 687, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__304)) __PYX_ERR(0, 687, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":691
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_neg.
 *     def exitExpr_neg(self, ctx:ModelicaParser.Expr_negContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__305 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__305)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__305);
  __Pyx_GIVEREF(__pyx_tuple__305);
  __pyx_codeobj__306 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__305, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitExpr_neg, 691, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__306)) __PYX_ERR(0, 691, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":696
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_add.
 *     def enterExpr_add(self, ctx:ModelicaParser.Expr_addContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__307 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__307)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__307);
  __Pyx_GIVEREF(__pyx_tuple__307);
  __pyx_codeobj__308 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__307, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterExpr_add, 696, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__308)) __PYX_ERR(0, 696, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":700
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_add.
 *     def exitExpr_add(self, ctx:ModelicaParser.Expr_addContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__309 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__309)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__309);
  __Pyx_GIVEREF(__pyx_tuple__309);
  __pyx_codeobj__310 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__309, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitExpr_add, 700, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__310)) __PYX_ERR(0, 700, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":705
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_exp.
 *     def enterExpr_exp(self, ctx:ModelicaParser.Expr_expContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__311 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__311)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__311);
  __Pyx_GIVEREF(__pyx_tuple__311);
  __pyx_codeobj__312 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__311, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterExpr_exp, 705, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__312)) __PYX_ERR(0, 705, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":709
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_exp.
 *     def exitExpr_exp(self, ctx:ModelicaParser.Expr_expContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__313 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__313)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__313);
  __Pyx_GIVEREF(__pyx_tuple__313);
  __pyx_codeobj__314 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__313, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitExpr_exp, 709, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__314)) __PYX_ERR(0, 709, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":714
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_or.
 *     def enterExpr_or(self, ctx:ModelicaParser.Expr_orContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__315 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__315)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__315);
  __Pyx_GIVEREF(__pyx_tuple__315);
  __pyx_codeobj__316 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__315, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterExpr_or, 714, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__316)) __PYX_ERR(0, 714, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":718
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_or.
 *     def exitExpr_or(self, ctx:ModelicaParser.Expr_orContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__317 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__317)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__317);
  __Pyx_GIVEREF(__pyx_tuple__317);
  __pyx_codeobj__318 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__317, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitExpr_or, 718, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__318)) __PYX_ERR(0, 718, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":723
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_primary.
 *     def enterExpr_primary(self, ctx:ModelicaParser.Expr_primaryContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__319 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__319)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__319);
  __Pyx_GIVEREF(__pyx_tuple__319);
  __pyx_codeobj__320 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__319, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterExpr_primary, 723, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__320)) __PYX_ERR(0, 723, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":727
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_primary.
 *     def exitExpr_primary(self, ctx:ModelicaParser.Expr_primaryContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__321 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__321)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__321);
  __Pyx_GIVEREF(__pyx_tuple__321);
  __pyx_codeobj__322 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__321, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitExpr_primary, 727, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__322)) __PYX_ERR(0, 727, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":732
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_and.
 *     def enterExpr_and(self, ctx:ModelicaParser.Expr_andContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__323 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__323)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__323);
  __Pyx_GIVEREF(__pyx_tuple__323);
  __pyx_codeobj__324 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__323, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterExpr_and, 732, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__324)) __PYX_ERR(0, 732, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":736
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_and.
 *     def exitExpr_and(self, ctx:ModelicaParser.Expr_andContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__325 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__325)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__325);
  __Pyx_GIVEREF(__pyx_tuple__325);
  __pyx_codeobj__326 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__325, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitExpr_and, 736, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__326)) __PYX_ERR(0, 736, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":741
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_rel.
 *     def enterExpr_rel(self, ctx:ModelicaParser.Expr_relContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__327 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__327)) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__327);
  __Pyx_GIVEREF(__pyx_tuple__327);
  __pyx_codeobj__328 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__327, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterExpr_rel, 741, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__328)) __PYX_ERR(0, 741, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":745
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_rel.
 *     def exitExpr_rel(self, ctx:ModelicaParser.Expr_relContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__329 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__329)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__329);
  __Pyx_GIVEREF(__pyx_tuple__329);
  __pyx_codeobj__330 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__329, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitExpr_rel, 745, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__330)) __PYX_ERR(0, 745, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":750
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_not.
 *     def enterExpr_not(self, ctx:ModelicaParser.Expr_notContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__331 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__331)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__331);
  __Pyx_GIVEREF(__pyx_tuple__331);
  __pyx_codeobj__332 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__331, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterExpr_not, 750, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__332)) __PYX_ERR(0, 750, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":754
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_not.
 *     def exitExpr_not(self, ctx:ModelicaParser.Expr_notContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__333 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__333)) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__333);
  __Pyx_GIVEREF(__pyx_tuple__333);
  __pyx_codeobj__334 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__333, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitExpr_not, 754, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__334)) __PYX_ERR(0, 754, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":759
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_mul.
 *     def enterExpr_mul(self, ctx:ModelicaParser.Expr_mulContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__335 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__335)) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__335);
  __Pyx_GIVEREF(__pyx_tuple__335);
  __pyx_codeobj__336 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__335, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterExpr_mul, 759, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__336)) __PYX_ERR(0, 759, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":763
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_mul.
 *     def exitExpr_mul(self, ctx:ModelicaParser.Expr_mulContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__337 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__337)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__337);
  __Pyx_GIVEREF(__pyx_tuple__337);
  __pyx_codeobj__338 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__337, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitExpr_mul, 763, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__338)) __PYX_ERR(0, 763, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":768
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_unsigned_number.
 *     def enterPrimary_unsigned_number(self, ctx:ModelicaParser.Primary_unsigned_numberContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__339 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__339)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__339);
  __Pyx_GIVEREF(__pyx_tuple__339);
  __pyx_codeobj__340 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__339, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterPrimary_unsigned_number, 768, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__340)) __PYX_ERR(0, 768, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":772
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_unsigned_number.
 *     def exitPrimary_unsigned_number(self, ctx:ModelicaParser.Primary_unsigned_numberContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__341 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__341)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__341);
  __Pyx_GIVEREF(__pyx_tuple__341);
  __pyx_codeobj__342 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__341, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitPrimary_unsigned_number, 772, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__342)) __PYX_ERR(0, 772, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":777
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_string.
 *     def enterPrimary_string(self, ctx:ModelicaParser.Primary_stringContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__343 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__343)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__343);
  __Pyx_GIVEREF(__pyx_tuple__343);
  __pyx_codeobj__344 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__343, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterPrimary_string, 777, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__344)) __PYX_ERR(0, 777, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":781
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_string.
 *     def exitPrimary_string(self, ctx:ModelicaParser.Primary_stringContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__345 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__345)) __PYX_ERR(0, 781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__345);
  __Pyx_GIVEREF(__pyx_tuple__345);
  __pyx_codeobj__346 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__345, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitPrimary_string, 781, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__346)) __PYX_ERR(0, 781, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":786
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_false.
 *     def enterPrimary_false(self, ctx:ModelicaParser.Primary_falseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__347 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__347)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__347);
  __Pyx_GIVEREF(__pyx_tuple__347);
  __pyx_codeobj__348 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__347, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterPrimary_false, 786, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__348)) __PYX_ERR(0, 786, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":790
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_false.
 *     def exitPrimary_false(self, ctx:ModelicaParser.Primary_falseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__349 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__349)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__349);
  __Pyx_GIVEREF(__pyx_tuple__349);
  __pyx_codeobj__350 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__349, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitPrimary_false, 790, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__350)) __PYX_ERR(0, 790, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":795
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_true.
 *     def enterPrimary_true(self, ctx:ModelicaParser.Primary_trueContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__351 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__351)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__351);
  __Pyx_GIVEREF(__pyx_tuple__351);
  __pyx_codeobj__352 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__351, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterPrimary_true, 795, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__352)) __PYX_ERR(0, 795, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":799
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_true.
 *     def exitPrimary_true(self, ctx:ModelicaParser.Primary_trueContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__353 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__353)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__353);
  __Pyx_GIVEREF(__pyx_tuple__353);
  __pyx_codeobj__354 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__353, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitPrimary_true, 799, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__354)) __PYX_ERR(0, 799, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":804
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_function.
 *     def enterPrimary_function(self, ctx:ModelicaParser.Primary_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__355 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__355)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__355);
  __Pyx_GIVEREF(__pyx_tuple__355);
  __pyx_codeobj__356 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__355, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterPrimary_function, 804, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__356)) __PYX_ERR(0, 804, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":808
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_function.
 *     def exitPrimary_function(self, ctx:ModelicaParser.Primary_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__357 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__357)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__357);
  __Pyx_GIVEREF(__pyx_tuple__357);
  __pyx_codeobj__358 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__357, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitPrimary_function, 808, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__358)) __PYX_ERR(0, 808, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":813
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_derivative.
 *     def enterPrimary_derivative(self, ctx:ModelicaParser.Primary_derivativeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__359 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__359)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__359);
  __Pyx_GIVEREF(__pyx_tuple__359);
  __pyx_codeobj__360 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__359, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterPrimary_derivative, 813, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__360)) __PYX_ERR(0, 813, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":817
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_derivative.
 *     def exitPrimary_derivative(self, ctx:ModelicaParser.Primary_derivativeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__361 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__361)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__361);
  __Pyx_GIVEREF(__pyx_tuple__361);
  __pyx_codeobj__362 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__361, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitPrimary_derivative, 817, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__362)) __PYX_ERR(0, 817, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":822
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_initial.
 *     def enterPrimary_initial(self, ctx:ModelicaParser.Primary_initialContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__363 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__363)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__363);
  __Pyx_GIVEREF(__pyx_tuple__363);
  __pyx_codeobj__364 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__363, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterPrimary_initial, 822, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__364)) __PYX_ERR(0, 822, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":826
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_initial.
 *     def exitPrimary_initial(self, ctx:ModelicaParser.Primary_initialContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__365 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__365)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__365);
  __Pyx_GIVEREF(__pyx_tuple__365);
  __pyx_codeobj__366 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__365, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitPrimary_initial, 826, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__366)) __PYX_ERR(0, 826, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":831
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_component_reference.
 *     def enterPrimary_component_reference(self, ctx:ModelicaParser.Primary_component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__367 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__367)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__367);
  __Pyx_GIVEREF(__pyx_tuple__367);
  __pyx_codeobj__368 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__367, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterPrimary_component_reference, 831, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__368)) __PYX_ERR(0, 831, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":835
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_component_reference.
 *     def exitPrimary_component_reference(self, ctx:ModelicaParser.Primary_component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__369 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__369)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__369);
  __Pyx_GIVEREF(__pyx_tuple__369);
  __pyx_codeobj__370 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__369, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitPrimary_component_reference, 835, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__370)) __PYX_ERR(0, 835, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":840
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_output_expression_list.
 *     def enterPrimary_output_expression_list(self, ctx:ModelicaParser.Primary_output_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__371 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__371)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__371);
  __Pyx_GIVEREF(__pyx_tuple__371);
  __pyx_codeobj__372 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__371, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterPrimary_output_expression_l, 840, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__372)) __PYX_ERR(0, 840, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":844
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_output_expression_list.
 *     def exitPrimary_output_expression_list(self, ctx:ModelicaParser.Primary_output_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__373 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__373)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__373);
  __Pyx_GIVEREF(__pyx_tuple__373);
  __pyx_codeobj__374 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__373, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitPrimary_output_expression_li, 844, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__374)) __PYX_ERR(0, 844, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":849
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_expression_list.
 *     def enterPrimary_expression_list(self, ctx:ModelicaParser.Primary_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__375 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__375)) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__375);
  __Pyx_GIVEREF(__pyx_tuple__375);
  __pyx_codeobj__376 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__375, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterPrimary_expression_list, 849, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__376)) __PYX_ERR(0, 849, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":853
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_expression_list.
 *     def exitPrimary_expression_list(self, ctx:ModelicaParser.Primary_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__377 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__377)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__377);
  __Pyx_GIVEREF(__pyx_tuple__377);
  __pyx_codeobj__378 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__377, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitPrimary_expression_list, 853, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__378)) __PYX_ERR(0, 853, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":858
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_function_arguments.
 *     def enterPrimary_function_arguments(self, ctx:ModelicaParser.Primary_function_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__379 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__379)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__379);
  __Pyx_GIVEREF(__pyx_tuple__379);
  __pyx_codeobj__380 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__379, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterPrimary_function_arguments, 858, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__380)) __PYX_ERR(0, 858, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":862
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_function_arguments.
 *     def exitPrimary_function_arguments(self, ctx:ModelicaParser.Primary_function_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__381 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__381)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__381);
  __Pyx_GIVEREF(__pyx_tuple__381);
  __pyx_codeobj__382 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__381, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitPrimary_function_arguments, 862, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__382)) __PYX_ERR(0, 862, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":867
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_end.
 *     def enterPrimary_end(self, ctx:ModelicaParser.Primary_endContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__383 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__383)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__383);
  __Pyx_GIVEREF(__pyx_tuple__383);
  __pyx_codeobj__384 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__383, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterPrimary_end, 867, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__384)) __PYX_ERR(0, 867, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":871
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_end.
 *     def exitPrimary_end(self, ctx:ModelicaParser.Primary_endContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__385 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__385)) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__385);
  __Pyx_GIVEREF(__pyx_tuple__385);
  __pyx_codeobj__386 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__385, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitPrimary_end, 871, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__386)) __PYX_ERR(0, 871, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":876
 * 
 *     # Enter a parse tree produced by ModelicaParser#name.
 *     def enterName(self, ctx:ModelicaParser.NameContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__387 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__387)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__387);
  __Pyx_GIVEREF(__pyx_tuple__387);
  __pyx_codeobj__388 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__387, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterName, 876, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__388)) __PYX_ERR(0, 876, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":880
 * 
 *     # Exit a parse tree produced by ModelicaParser#name.
 *     def exitName(self, ctx:ModelicaParser.NameContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__389 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__389)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__389);
  __Pyx_GIVEREF(__pyx_tuple__389);
  __pyx_codeobj__390 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__389, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitName, 880, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__390)) __PYX_ERR(0, 880, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":885
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_reference_element.
 *     def enterComponent_reference_element(self, ctx:ModelicaParser.Component_reference_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__391 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__391)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__391);
  __Pyx_GIVEREF(__pyx_tuple__391);
  __pyx_codeobj__392 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__391, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterComponent_reference_element, 885, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__392)) __PYX_ERR(0, 885, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":889
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_reference_element.
 *     def exitComponent_reference_element(self, ctx:ModelicaParser.Component_reference_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__393 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__393)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__393);
  __Pyx_GIVEREF(__pyx_tuple__393);
  __pyx_codeobj__394 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__393, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitComponent_reference_element, 889, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__394)) __PYX_ERR(0, 889, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":894
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_reference.
 *     def enterComponent_reference(self, ctx:ModelicaParser.Component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__395 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__395)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__395);
  __Pyx_GIVEREF(__pyx_tuple__395);
  __pyx_codeobj__396 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__395, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterComponent_reference, 894, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__396)) __PYX_ERR(0, 894, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":898
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_reference.
 *     def exitComponent_reference(self, ctx:ModelicaParser.Component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__397 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__397)) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__397);
  __Pyx_GIVEREF(__pyx_tuple__397);
  __pyx_codeobj__398 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__397, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitComponent_reference, 898, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__398)) __PYX_ERR(0, 898, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":903
 * 
 *     # Enter a parse tree produced by ModelicaParser#function_call_args.
 *     def enterFunction_call_args(self, ctx:ModelicaParser.Function_call_argsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__399 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__399)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__399);
  __Pyx_GIVEREF(__pyx_tuple__399);
  __pyx_codeobj__400 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__399, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterFunction_call_args, 903, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__400)) __PYX_ERR(0, 903, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":907
 * 
 *     # Exit a parse tree produced by ModelicaParser#function_call_args.
 *     def exitFunction_call_args(self, ctx:ModelicaParser.Function_call_argsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__401 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__401)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__401);
  __Pyx_GIVEREF(__pyx_tuple__401);
  __pyx_codeobj__402 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__401, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitFunction_call_args, 907, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__402)) __PYX_ERR(0, 907, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":912
 * 
 *     # Enter a parse tree produced by ModelicaParser#function_arguments.
 *     def enterFunction_arguments(self, ctx:ModelicaParser.Function_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__403 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__403)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__403);
  __Pyx_GIVEREF(__pyx_tuple__403);
  __pyx_codeobj__404 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__403, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterFunction_arguments, 912, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__404)) __PYX_ERR(0, 912, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":916
 * 
 *     # Exit a parse tree produced by ModelicaParser#function_arguments.
 *     def exitFunction_arguments(self, ctx:ModelicaParser.Function_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__405 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__405)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__405);
  __Pyx_GIVEREF(__pyx_tuple__405);
  __pyx_codeobj__406 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__405, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitFunction_arguments, 916, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__406)) __PYX_ERR(0, 916, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":921
 * 
 *     # Enter a parse tree produced by ModelicaParser#named_arguments.
 *     def enterNamed_arguments(self, ctx:ModelicaParser.Named_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__407 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__407)) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__407);
  __Pyx_GIVEREF(__pyx_tuple__407);
  __pyx_codeobj__408 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__407, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterNamed_arguments, 921, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__408)) __PYX_ERR(0, 921, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":925
 * 
 *     # Exit a parse tree produced by ModelicaParser#named_arguments.
 *     def exitNamed_arguments(self, ctx:ModelicaParser.Named_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__409 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__409)) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__409);
  __Pyx_GIVEREF(__pyx_tuple__409);
  __pyx_codeobj__410 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__409, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitNamed_arguments, 925, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__410)) __PYX_ERR(0, 925, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":930
 * 
 *     # Enter a parse tree produced by ModelicaParser#named_argument.
 *     def enterNamed_argument(self, ctx:ModelicaParser.Named_argumentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__411 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__411)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__411);
  __Pyx_GIVEREF(__pyx_tuple__411);
  __pyx_codeobj__412 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__411, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterNamed_argument, 930, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__412)) __PYX_ERR(0, 930, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":934
 * 
 *     # Exit a parse tree produced by ModelicaParser#named_argument.
 *     def exitNamed_argument(self, ctx:ModelicaParser.Named_argumentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__413 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__413)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__413);
  __Pyx_GIVEREF(__pyx_tuple__413);
  __pyx_codeobj__414 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__413, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitNamed_argument, 934, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__414)) __PYX_ERR(0, 934, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":939
 * 
 *     # Enter a parse tree produced by ModelicaParser#argument_function.
 *     def enterArgument_function(self, ctx:ModelicaParser.Argument_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__415 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__415)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__415);
  __Pyx_GIVEREF(__pyx_tuple__415);
  __pyx_codeobj__416 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__415, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterArgument_function, 939, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__416)) __PYX_ERR(0, 939, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":943
 * 
 *     # Exit a parse tree produced by ModelicaParser#argument_function.
 *     def exitArgument_function(self, ctx:ModelicaParser.Argument_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__417 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__417)) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__417);
  __Pyx_GIVEREF(__pyx_tuple__417);
  __pyx_codeobj__418 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__417, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitArgument_function, 943, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__418)) __PYX_ERR(0, 943, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":948
 * 
 *     # Enter a parse tree produced by ModelicaParser#argument_expression.
 *     def enterArgument_expression(self, ctx:ModelicaParser.Argument_expressionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__419 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__419)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__419);
  __Pyx_GIVEREF(__pyx_tuple__419);
  __pyx_codeobj__420 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__419, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterArgument_expression, 948, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__420)) __PYX_ERR(0, 948, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":952
 * 
 *     # Exit a parse tree produced by ModelicaParser#argument_expression.
 *     def exitArgument_expression(self, ctx:ModelicaParser.Argument_expressionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__421 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__421)) __PYX_ERR(0, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__421);
  __Pyx_GIVEREF(__pyx_tuple__421);
  __pyx_codeobj__422 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__421, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitArgument_expression, 952, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__422)) __PYX_ERR(0, 952, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":957
 * 
 *     # Enter a parse tree produced by ModelicaParser#output_expression_list.
 *     def enterOutput_expression_list(self, ctx:ModelicaParser.Output_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__423 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__423)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__423);
  __Pyx_GIVEREF(__pyx_tuple__423);
  __pyx_codeobj__424 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__423, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterOutput_expression_list, 957, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__424)) __PYX_ERR(0, 957, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":961
 * 
 *     # Exit a parse tree produced by ModelicaParser#output_expression_list.
 *     def exitOutput_expression_list(self, ctx:ModelicaParser.Output_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__425 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__425)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__425);
  __Pyx_GIVEREF(__pyx_tuple__425);
  __pyx_codeobj__426 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__425, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitOutput_expression_list, 961, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__426)) __PYX_ERR(0, 961, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":966
 * 
 *     # Enter a parse tree produced by ModelicaParser#expression_list.
 *     def enterExpression_list(self, ctx:ModelicaParser.Expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__427 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__427)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__427);
  __Pyx_GIVEREF(__pyx_tuple__427);
  __pyx_codeobj__428 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__427, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterExpression_list, 966, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__428)) __PYX_ERR(0, 966, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":970
 * 
 *     # Exit a parse tree produced by ModelicaParser#expression_list.
 *     def exitExpression_list(self, ctx:ModelicaParser.Expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__429 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__429)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__429);
  __Pyx_GIVEREF(__pyx_tuple__429);
  __pyx_codeobj__430 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__429, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitExpression_list, 970, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__430)) __PYX_ERR(0, 970, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":975
 * 
 *     # Enter a parse tree produced by ModelicaParser#array_subscripts.
 *     def enterArray_subscripts(self, ctx:ModelicaParser.Array_subscriptsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__431 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__431)) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__431);
  __Pyx_GIVEREF(__pyx_tuple__431);
  __pyx_codeobj__432 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__431, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterArray_subscripts, 975, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__432)) __PYX_ERR(0, 975, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":979
 * 
 *     # Exit a parse tree produced by ModelicaParser#array_subscripts.
 *     def exitArray_subscripts(self, ctx:ModelicaParser.Array_subscriptsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__433 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__433)) __PYX_ERR(0, 979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__433);
  __Pyx_GIVEREF(__pyx_tuple__433);
  __pyx_codeobj__434 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__433, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitArray_subscripts, 979, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__434)) __PYX_ERR(0, 979, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":984
 * 
 *     # Enter a parse tree produced by ModelicaParser#subscript.
 *     def enterSubscript(self, ctx:ModelicaParser.SubscriptContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__435 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__435)) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__435);
  __Pyx_GIVEREF(__pyx_tuple__435);
  __pyx_codeobj__436 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__435, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterSubscript, 984, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__436)) __PYX_ERR(0, 984, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":988
 * 
 *     # Exit a parse tree produced by ModelicaParser#subscript.
 *     def exitSubscript(self, ctx:ModelicaParser.SubscriptContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__437 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__437)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__437);
  __Pyx_GIVEREF(__pyx_tuple__437);
  __pyx_codeobj__438 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__437, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitSubscript, 988, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__438)) __PYX_ERR(0, 988, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":993
 * 
 *     # Enter a parse tree produced by ModelicaParser#comment.
 *     def enterComment(self, ctx:ModelicaParser.CommentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__439 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__439)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__439);
  __Pyx_GIVEREF(__pyx_tuple__439);
  __pyx_codeobj__440 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__439, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterComment, 993, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__440)) __PYX_ERR(0, 993, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":997
 * 
 *     # Exit a parse tree produced by ModelicaParser#comment.
 *     def exitComment(self, ctx:ModelicaParser.CommentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__441 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__441)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__441);
  __Pyx_GIVEREF(__pyx_tuple__441);
  __pyx_codeobj__442 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__441, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitComment, 997, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__442)) __PYX_ERR(0, 997, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":1002
 * 
 *     # Enter a parse tree produced by ModelicaParser#string_comment.
 *     def enterString_comment(self, ctx:ModelicaParser.String_commentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__443 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__443)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__443);
  __Pyx_GIVEREF(__pyx_tuple__443);
  __pyx_codeobj__444 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__443, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterString_comment, 1002, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__444)) __PYX_ERR(0, 1002, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":1006
 * 
 *     # Exit a parse tree produced by ModelicaParser#string_comment.
 *     def exitString_comment(self, ctx:ModelicaParser.String_commentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__445 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__445)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__445);
  __Pyx_GIVEREF(__pyx_tuple__445);
  __pyx_codeobj__446 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__445, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitString_comment, 1006, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__446)) __PYX_ERR(0, 1006, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":1011
 * 
 *     # Enter a parse tree produced by ModelicaParser#annotation.
 *     def enterAnnotation(self, ctx:ModelicaParser.AnnotationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__447 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__447)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__447);
  __Pyx_GIVEREF(__pyx_tuple__447);
  __pyx_codeobj__448 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__447, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_enterAnnotation, 1011, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__448)) __PYX_ERR(0, 1011, __pyx_L1_error)

  /* "pymola/generated/ModelicaListener.pyx":1015
 * 
 *     # Exit a parse tree produced by ModelicaParser#annotation.
 *     def exitAnnotation(self, ctx:ModelicaParser.AnnotationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_tuple__449 = PyTuple_Pack(2, __pyx_n_s_self, __pyx_n_s_ctx); if (unlikely(!__pyx_tuple__449)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__449);
  __Pyx_GIVEREF(__pyx_tuple__449);
  __pyx_codeobj__450 = (PyObject*)__Pyx_PyCode_New(2, 0, 2, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__449, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_home_jgoppert_git_pymola_pymola, __pyx_n_s_exitAnnotation, 1015, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__450)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC initModelicaListener(void); /*proto*/
PyMODINIT_FUNC initModelicaListener(void)
#else
PyMODINIT_FUNC PyInit_ModelicaListener(void); /*proto*/
PyMODINIT_FUNC PyInit_ModelicaListener(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit_ModelicaListener(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("ModelicaListener", __pyx_methods, 0, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_pymola__generated__ModelicaListener) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "pymola.generated.ModelicaListener")) {
      if (unlikely(PyDict_SetItemString(modules, "pymola.generated.ModelicaListener", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  /*--- Type import code ---*/
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "pymola/generated/ModelicaListener.pyx":2
 * # Generated from Modelica.g4 by ANTLR 4.7
 * from antlr4 import *             # <<<<<<<<<<<<<<
 * if __name__ is not None and "." in __name__:
 *     from .ModelicaParser import ModelicaParser
 */
  __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_INCREF(__pyx_n_s_);
  __Pyx_GIVEREF(__pyx_n_s_);
  PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_);
  __pyx_t_2 = __Pyx_Import(__pyx_n_s_antlr4, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 2, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_import_star(__pyx_t_2) < 0) __PYX_ERR(0, 2, __pyx_L1_error);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "pymola/generated/ModelicaListener.pyx":3
 * # Generated from Modelica.g4 by ANTLR 4.7
 * from antlr4 import *
 * if __name__ is not None and "." in __name__:             # <<<<<<<<<<<<<<
 *     from .ModelicaParser import ModelicaParser
 * else:
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_4 = (__pyx_t_2 != Py_None);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L3_bool_binop_done;
  }
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = (__Pyx_PySequence_ContainsTF(__pyx_kp_u__2, __pyx_t_2, Py_EQ)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 3, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_5 != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L3_bool_binop_done:;
  if (__pyx_t_3) {

    /* "pymola/generated/ModelicaListener.pyx":4
 * from antlr4 import *
 * if __name__ is not None and "." in __name__:
 *     from .ModelicaParser import ModelicaParser             # <<<<<<<<<<<<<<
 * else:
 *     from ModelicaParser import ModelicaParser
 */
    __pyx_t_2 = PyList_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_n_s_ModelicaParser);
    __Pyx_GIVEREF(__pyx_n_s_ModelicaParser);
    PyList_SET_ITEM(__pyx_t_2, 0, __pyx_n_s_ModelicaParser);
    __pyx_t_1 = __Pyx_Import(__pyx_n_s_ModelicaParser, __pyx_t_2, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 4, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_ImportFrom(__pyx_t_1, __pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 4, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_ModelicaParser, __pyx_t_2) < 0) __PYX_ERR(0, 4, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "pymola/generated/ModelicaListener.pyx":3
 * # Generated from Modelica.g4 by ANTLR 4.7
 * from antlr4 import *
 * if __name__ is not None and "." in __name__:             # <<<<<<<<<<<<<<
 *     from .ModelicaParser import ModelicaParser
 * else:
 */
    goto __pyx_L2;
  }

  /* "pymola/generated/ModelicaListener.pyx":6
 *     from .ModelicaParser import ModelicaParser
 * else:
 *     from ModelicaParser import ModelicaParser             # <<<<<<<<<<<<<<
 * 
 * # This class defines a complete listener for a parse tree produced by ModelicaParser.
 */
  /*else*/ {
    __pyx_t_1 = PyList_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_INCREF(__pyx_n_s_ModelicaParser);
    __Pyx_GIVEREF(__pyx_n_s_ModelicaParser);
    PyList_SET_ITEM(__pyx_t_1, 0, __pyx_n_s_ModelicaParser);
    __pyx_t_2 = __Pyx_Import(__pyx_n_s_ModelicaParser, __pyx_t_1, 0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_ImportFrom(__pyx_t_2, __pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_d, __pyx_n_s_ModelicaParser, __pyx_t_1) < 0) __PYX_ERR(0, 6, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __pyx_L2:;

  /* "pymola/generated/ModelicaListener.pyx":9
 * 
 * # This class defines a complete listener for a parse tree produced by ModelicaParser.
 * class ModelicaListener(ParseTreeListener):             # <<<<<<<<<<<<<<
 * 
 *     # Enter a parse tree produced by ModelicaParser#stored_definition.
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_ParseTreeListener); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_CalculateMetaclass(NULL, __pyx_t_1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = __Pyx_Py3MetaclassPrepare(__pyx_t_2, __pyx_t_1, __pyx_n_s_ModelicaListener, __pyx_n_s_ModelicaListener, (PyObject *) NULL, __pyx_n_s_pymola_generated_ModelicaListene, (PyObject *) NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);

  /* "pymola/generated/ModelicaListener.pyx":12
 * 
 *     # Enter a parse tree produced by ModelicaParser#stored_definition.
 *     def enterStored_definition(self, ctx:ModelicaParser.Stored_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Stored_definitionContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_1enterStored_definition, 0, __pyx_n_s_ModelicaListener_enterStored_def, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__4)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterStored_definition, __pyx_t_9) < 0) __PYX_ERR(0, 12, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":16
 * 
 *     # Exit a parse tree produced by ModelicaParser#stored_definition.
 *     def exitStored_definition(self, ctx:ModelicaParser.Stored_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Stored_definitionContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_3exitStored_definition, 0, __pyx_n_s_ModelicaListener_exitStored_defi, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__6)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitStored_definition, __pyx_t_8) < 0) __PYX_ERR(0, 16, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":21
 * 
 *     # Enter a parse tree produced by ModelicaParser#stored_definition_class.
 *     def enterStored_definition_class(self, ctx:ModelicaParser.Stored_definition_classContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Stored_definition_classContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_5enterStored_definition_class, 0, __pyx_n_s_ModelicaListener_enterStored_def_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__8)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterStored_definition_class, __pyx_t_7) < 0) __PYX_ERR(0, 21, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":25
 * 
 *     # Exit a parse tree produced by ModelicaParser#stored_definition_class.
 *     def exitStored_definition_class(self, ctx:ModelicaParser.Stored_definition_classContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Stored_definition_classContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_7exitStored_definition_class, 0, __pyx_n_s_ModelicaListener_exitStored_defi_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__10)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitStored_definition_class, __pyx_t_9) < 0) __PYX_ERR(0, 25, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":30
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_definition.
 *     def enterClass_definition(self, ctx:ModelicaParser.Class_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Class_definitionContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_9enterClass_definition, 0, __pyx_n_s_ModelicaListener_enterClass_defi, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__12)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterClass_definition, __pyx_t_8) < 0) __PYX_ERR(0, 30, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":34
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_definition.
 *     def exitClass_definition(self, ctx:ModelicaParser.Class_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Class_definitionContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_11exitClass_definition, 0, __pyx_n_s_ModelicaListener_exitClass_defin, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__14)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitClass_definition, __pyx_t_7) < 0) __PYX_ERR(0, 34, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":39
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_prefixes.
 *     def enterClass_prefixes(self, ctx:ModelicaParser.Class_prefixesContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Class_prefixesContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_13enterClass_prefixes, 0, __pyx_n_s_ModelicaListener_enterClass_pref, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__16)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterClass_prefixes, __pyx_t_9) < 0) __PYX_ERR(0, 39, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":43
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_prefixes.
 *     def exitClass_prefixes(self, ctx:ModelicaParser.Class_prefixesContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Class_prefixesContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_15exitClass_prefixes, 0, __pyx_n_s_ModelicaListener_exitClass_prefi, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__18)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitClass_prefixes, __pyx_t_8) < 0) __PYX_ERR(0, 43, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":48
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_type.
 *     def enterClass_type(self, ctx:ModelicaParser.Class_typeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Class_typeContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_17enterClass_type, 0, __pyx_n_s_ModelicaListener_enterClass_type, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__20)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterClass_type, __pyx_t_7) < 0) __PYX_ERR(0, 48, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":52
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_type.
 *     def exitClass_type(self, ctx:ModelicaParser.Class_typeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Class_typeContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_19exitClass_type, 0, __pyx_n_s_ModelicaListener_exitClass_type, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__22)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitClass_type, __pyx_t_9) < 0) __PYX_ERR(0, 52, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":57
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_spec_comp.
 *     def enterClass_spec_comp(self, ctx:ModelicaParser.Class_spec_compContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Class_spec_compContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_21enterClass_spec_comp, 0, __pyx_n_s_ModelicaListener_enterClass_spec, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__24)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterClass_spec_comp, __pyx_t_8) < 0) __PYX_ERR(0, 57, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":61
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_spec_comp.
 *     def exitClass_spec_comp(self, ctx:ModelicaParser.Class_spec_compContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Class_spec_compContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_23exitClass_spec_comp, 0, __pyx_n_s_ModelicaListener_exitClass_spec, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__26)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitClass_spec_comp, __pyx_t_7) < 0) __PYX_ERR(0, 61, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":66
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_spec_base.
 *     def enterClass_spec_base(self, ctx:ModelicaParser.Class_spec_baseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Class_spec_baseContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_25enterClass_spec_base, 0, __pyx_n_s_ModelicaListener_enterClass_spec_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__28)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterClass_spec_base, __pyx_t_9) < 0) __PYX_ERR(0, 66, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":70
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_spec_base.
 *     def exitClass_spec_base(self, ctx:ModelicaParser.Class_spec_baseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Class_spec_baseContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_27exitClass_spec_base, 0, __pyx_n_s_ModelicaListener_exitClass_spec_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__30)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitClass_spec_base, __pyx_t_8) < 0) __PYX_ERR(0, 70, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":75
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_spec_enum.
 *     def enterClass_spec_enum(self, ctx:ModelicaParser.Class_spec_enumContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Class_spec_enumContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_29enterClass_spec_enum, 0, __pyx_n_s_ModelicaListener_enterClass_spec_3, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__32)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterClass_spec_enum, __pyx_t_7) < 0) __PYX_ERR(0, 75, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":79
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_spec_enum.
 *     def exitClass_spec_enum(self, ctx:ModelicaParser.Class_spec_enumContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Class_spec_enumContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_31exitClass_spec_enum, 0, __pyx_n_s_ModelicaListener_exitClass_spec_3, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__34)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitClass_spec_enum, __pyx_t_9) < 0) __PYX_ERR(0, 79, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":84
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_spec_der.
 *     def enterClass_spec_der(self, ctx:ModelicaParser.Class_spec_derContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Class_spec_derContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_33enterClass_spec_der, 0, __pyx_n_s_ModelicaListener_enterClass_spec_4, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__36)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterClass_spec_der, __pyx_t_8) < 0) __PYX_ERR(0, 84, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":88
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_spec_der.
 *     def exitClass_spec_der(self, ctx:ModelicaParser.Class_spec_derContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Class_spec_derContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_35exitClass_spec_der, 0, __pyx_n_s_ModelicaListener_exitClass_spec_4, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__38)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitClass_spec_der, __pyx_t_7) < 0) __PYX_ERR(0, 88, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":93
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_spec_extends.
 *     def enterClass_spec_extends(self, ctx:ModelicaParser.Class_spec_extendsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Class_spec_extendsContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_37enterClass_spec_extends, 0, __pyx_n_s_ModelicaListener_enterClass_spec_5, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__40)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterClass_spec_extends, __pyx_t_9) < 0) __PYX_ERR(0, 93, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":97
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_spec_extends.
 *     def exitClass_spec_extends(self, ctx:ModelicaParser.Class_spec_extendsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Class_spec_extendsContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_39exitClass_spec_extends, 0, __pyx_n_s_ModelicaListener_exitClass_spec_5, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__42)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitClass_spec_extends, __pyx_t_8) < 0) __PYX_ERR(0, 97, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":102
 * 
 *     # Enter a parse tree produced by ModelicaParser#base_prefix.
 *     def enterBase_prefix(self, ctx:ModelicaParser.Base_prefixContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Base_prefixContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_41enterBase_prefix, 0, __pyx_n_s_ModelicaListener_enterBase_prefi, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__44)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterBase_prefix, __pyx_t_7) < 0) __PYX_ERR(0, 102, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":106
 * 
 *     # Exit a parse tree produced by ModelicaParser#base_prefix.
 *     def exitBase_prefix(self, ctx:ModelicaParser.Base_prefixContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Base_prefixContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_43exitBase_prefix, 0, __pyx_n_s_ModelicaListener_exitBase_prefix, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__46)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitBase_prefix, __pyx_t_9) < 0) __PYX_ERR(0, 106, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":111
 * 
 *     # Enter a parse tree produced by ModelicaParser#enum_list.
 *     def enterEnum_list(self, ctx:ModelicaParser.Enum_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Enum_listContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_45enterEnum_list, 0, __pyx_n_s_ModelicaListener_enterEnum_list, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__48)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterEnum_list, __pyx_t_8) < 0) __PYX_ERR(0, 111, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":115
 * 
 *     # Exit a parse tree produced by ModelicaParser#enum_list.
 *     def exitEnum_list(self, ctx:ModelicaParser.Enum_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Enum_listContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_47exitEnum_list, 0, __pyx_n_s_ModelicaListener_exitEnum_list, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__50)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitEnum_list, __pyx_t_7) < 0) __PYX_ERR(0, 115, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":120
 * 
 *     # Enter a parse tree produced by ModelicaParser#enumeration_literal.
 *     def enterEnumeration_literal(self, ctx:ModelicaParser.Enumeration_literalContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Enumeration_literalContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_49enterEnumeration_literal, 0, __pyx_n_s_ModelicaListener_enterEnumeratio, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__52)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterEnumeration_literal, __pyx_t_9) < 0) __PYX_ERR(0, 120, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":124
 * 
 *     # Exit a parse tree produced by ModelicaParser#enumeration_literal.
 *     def exitEnumeration_literal(self, ctx:ModelicaParser.Enumeration_literalContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Enumeration_literalContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_51exitEnumeration_literal, 0, __pyx_n_s_ModelicaListener_exitEnumeration, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__54)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitEnumeration_literal, __pyx_t_8) < 0) __PYX_ERR(0, 124, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":129
 * 
 *     # Enter a parse tree produced by ModelicaParser#composition.
 *     def enterComposition(self, ctx:ModelicaParser.CompositionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_CompositionContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_53enterComposition, 0, __pyx_n_s_ModelicaListener_enterCompositio, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__56)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterComposition, __pyx_t_7) < 0) __PYX_ERR(0, 129, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":133
 * 
 *     # Exit a parse tree produced by ModelicaParser#composition.
 *     def exitComposition(self, ctx:ModelicaParser.CompositionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_CompositionContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_55exitComposition, 0, __pyx_n_s_ModelicaListener_exitComposition, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__58)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitComposition, __pyx_t_9) < 0) __PYX_ERR(0, 133, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":138
 * 
 *     # Enter a parse tree produced by ModelicaParser#language_specification.
 *     def enterLanguage_specification(self, ctx:ModelicaParser.Language_specificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Language_specificationContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_57enterLanguage_specification, 0, __pyx_n_s_ModelicaListener_enterLanguage_s, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__60)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterLanguage_specification, __pyx_t_8) < 0) __PYX_ERR(0, 138, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":142
 * 
 *     # Exit a parse tree produced by ModelicaParser#language_specification.
 *     def exitLanguage_specification(self, ctx:ModelicaParser.Language_specificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Language_specificationContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_59exitLanguage_specification, 0, __pyx_n_s_ModelicaListener_exitLanguage_sp, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__62)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitLanguage_specification, __pyx_t_7) < 0) __PYX_ERR(0, 142, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":147
 * 
 *     # Enter a parse tree produced by ModelicaParser#external_function_call.
 *     def enterExternal_function_call(self, ctx:ModelicaParser.External_function_callContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_External_function_callContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_61enterExternal_function_call, 0, __pyx_n_s_ModelicaListener_enterExternal_f, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__64)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterExternal_function_call, __pyx_t_9) < 0) __PYX_ERR(0, 147, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":151
 * 
 *     # Exit a parse tree produced by ModelicaParser#external_function_call.
 *     def exitExternal_function_call(self, ctx:ModelicaParser.External_function_callContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_External_function_callContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_63exitExternal_function_call, 0, __pyx_n_s_ModelicaListener_exitExternal_fu, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__66)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitExternal_function_call, __pyx_t_8) < 0) __PYX_ERR(0, 151, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":156
 * 
 *     # Enter a parse tree produced by ModelicaParser#element_list.
 *     def enterElement_list(self, ctx:ModelicaParser.Element_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Element_listContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_65enterElement_list, 0, __pyx_n_s_ModelicaListener_enterElement_li, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__68)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterElement_list, __pyx_t_7) < 0) __PYX_ERR(0, 156, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":160
 * 
 *     # Exit a parse tree produced by ModelicaParser#element_list.
 *     def exitElement_list(self, ctx:ModelicaParser.Element_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Element_listContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_67exitElement_list, 0, __pyx_n_s_ModelicaListener_exitElement_lis, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__70)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitElement_list, __pyx_t_9) < 0) __PYX_ERR(0, 160, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":165
 * 
 *     # Enter a parse tree produced by ModelicaParser#element.
 *     def enterElement(self, ctx:ModelicaParser.ElementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ElementContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_69enterElement, 0, __pyx_n_s_ModelicaListener_enterElement, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__72)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterElement, __pyx_t_8) < 0) __PYX_ERR(0, 165, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":169
 * 
 *     # Exit a parse tree produced by ModelicaParser#element.
 *     def exitElement(self, ctx:ModelicaParser.ElementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_ElementContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_71exitElement, 0, __pyx_n_s_ModelicaListener_exitElement, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__74)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitElement, __pyx_t_7) < 0) __PYX_ERR(0, 169, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":174
 * 
 *     # Enter a parse tree produced by ModelicaParser#regular_element.
 *     def enterRegular_element(self, ctx:ModelicaParser.Regular_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Regular_elementContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_73enterRegular_element, 0, __pyx_n_s_ModelicaListener_enterRegular_el, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__76)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterRegular_element, __pyx_t_9) < 0) __PYX_ERR(0, 174, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":178
 * 
 *     # Exit a parse tree produced by ModelicaParser#regular_element.
 *     def exitRegular_element(self, ctx:ModelicaParser.Regular_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Regular_elementContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_75exitRegular_element, 0, __pyx_n_s_ModelicaListener_exitRegular_ele, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__78)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitRegular_element, __pyx_t_8) < 0) __PYX_ERR(0, 178, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":183
 * 
 *     # Enter a parse tree produced by ModelicaParser#replaceable_element.
 *     def enterReplaceable_element(self, ctx:ModelicaParser.Replaceable_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Replaceable_elementContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_77enterReplaceable_element, 0, __pyx_n_s_ModelicaListener_enterReplaceabl, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__80)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterReplaceable_element, __pyx_t_7) < 0) __PYX_ERR(0, 183, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":187
 * 
 *     # Exit a parse tree produced by ModelicaParser#replaceable_element.
 *     def exitReplaceable_element(self, ctx:ModelicaParser.Replaceable_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Replaceable_elementContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_79exitReplaceable_element, 0, __pyx_n_s_ModelicaListener_exitReplaceable, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__82)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitReplaceable_element, __pyx_t_9) < 0) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":192
 * 
 *     # Enter a parse tree produced by ModelicaParser#import_clause.
 *     def enterImport_clause(self, ctx:ModelicaParser.Import_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Import_clauseContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_81enterImport_clause, 0, __pyx_n_s_ModelicaListener_enterImport_cla, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__84)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterImport_clause, __pyx_t_8) < 0) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":196
 * 
 *     # Exit a parse tree produced by ModelicaParser#import_clause.
 *     def exitImport_clause(self, ctx:ModelicaParser.Import_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Import_clauseContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_83exitImport_clause, 0, __pyx_n_s_ModelicaListener_exitImport_clau, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__86)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitImport_clause, __pyx_t_7) < 0) __PYX_ERR(0, 196, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":201
 * 
 *     # Enter a parse tree produced by ModelicaParser#import_list.
 *     def enterImport_list(self, ctx:ModelicaParser.Import_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Import_listContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_85enterImport_list, 0, __pyx_n_s_ModelicaListener_enterImport_lis, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__88)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterImport_list, __pyx_t_9) < 0) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":205
 * 
 *     # Exit a parse tree produced by ModelicaParser#import_list.
 *     def exitImport_list(self, ctx:ModelicaParser.Import_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Import_listContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_87exitImport_list, 0, __pyx_n_s_ModelicaListener_exitImport_list, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__90)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitImport_list, __pyx_t_8) < 0) __PYX_ERR(0, 205, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":210
 * 
 *     # Enter a parse tree produced by ModelicaParser#extends_clause.
 *     def enterExtends_clause(self, ctx:ModelicaParser.Extends_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Extends_clauseContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_89enterExtends_clause, 0, __pyx_n_s_ModelicaListener_enterExtends_cl, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__92)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterExtends_clause, __pyx_t_7) < 0) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":214
 * 
 *     # Exit a parse tree produced by ModelicaParser#extends_clause.
 *     def exitExtends_clause(self, ctx:ModelicaParser.Extends_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Extends_clauseContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_91exitExtends_clause, 0, __pyx_n_s_ModelicaListener_exitExtends_cla, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__94)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitExtends_clause, __pyx_t_9) < 0) __PYX_ERR(0, 214, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":219
 * 
 *     # Enter a parse tree produced by ModelicaParser#constraining_clause.
 *     def enterConstraining_clause(self, ctx:ModelicaParser.Constraining_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Constraining_clauseContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_93enterConstraining_clause, 0, __pyx_n_s_ModelicaListener_enterConstraini, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__96)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterConstraining_clause, __pyx_t_8) < 0) __PYX_ERR(0, 219, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":223
 * 
 *     # Exit a parse tree produced by ModelicaParser#constraining_clause.
 *     def exitConstraining_clause(self, ctx:ModelicaParser.Constraining_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Constraining_clauseContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_95exitConstraining_clause, 0, __pyx_n_s_ModelicaListener_exitConstrainin, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__98)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitConstraining_clause, __pyx_t_7) < 0) __PYX_ERR(0, 223, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":228
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_clause.
 *     def enterComponent_clause(self, ctx:ModelicaParser.Component_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Component_clauseContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_97enterComponent_clause, 0, __pyx_n_s_ModelicaListener_enterComponent, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__100)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterComponent_clause, __pyx_t_9) < 0) __PYX_ERR(0, 228, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":232
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_clause.
 *     def exitComponent_clause(self, ctx:ModelicaParser.Component_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Component_clauseContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_99exitComponent_clause, 0, __pyx_n_s_ModelicaListener_exitComponent_c, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__102)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitComponent_clause, __pyx_t_8) < 0) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":237
 * 
 *     # Enter a parse tree produced by ModelicaParser#type_prefix.
 *     def enterType_prefix(self, ctx:ModelicaParser.Type_prefixContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Type_prefixContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_101enterType_prefix, 0, __pyx_n_s_ModelicaListener_enterType_prefi, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__104)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterType_prefix, __pyx_t_7) < 0) __PYX_ERR(0, 237, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":241
 * 
 *     # Exit a parse tree produced by ModelicaParser#type_prefix.
 *     def exitType_prefix(self, ctx:ModelicaParser.Type_prefixContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Type_prefixContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_103exitType_prefix, 0, __pyx_n_s_ModelicaListener_exitType_prefix, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__106)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitType_prefix, __pyx_t_9) < 0) __PYX_ERR(0, 241, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":246
 * 
 *     # Enter a parse tree produced by ModelicaParser#type_specifier.
 *     def enterType_specifier(self, ctx:ModelicaParser.Type_specifierContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Type_specifierContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_105enterType_specifier, 0, __pyx_n_s_ModelicaListener_enterType_speci, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__108)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterType_specifier, __pyx_t_8) < 0) __PYX_ERR(0, 246, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":250
 * 
 *     # Exit a parse tree produced by ModelicaParser#type_specifier.
 *     def exitType_specifier(self, ctx:ModelicaParser.Type_specifierContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Type_specifierContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_107exitType_specifier, 0, __pyx_n_s_ModelicaListener_exitType_specif, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__110)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitType_specifier, __pyx_t_7) < 0) __PYX_ERR(0, 250, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":255
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_list.
 *     def enterComponent_list(self, ctx:ModelicaParser.Component_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Component_listContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_109enterComponent_list, 0, __pyx_n_s_ModelicaListener_enterComponent_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__112)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterComponent_list, __pyx_t_9) < 0) __PYX_ERR(0, 255, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":259
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_list.
 *     def exitComponent_list(self, ctx:ModelicaParser.Component_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Component_listContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_111exitComponent_list, 0, __pyx_n_s_ModelicaListener_exitComponent_l, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__114)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitComponent_list, __pyx_t_8) < 0) __PYX_ERR(0, 259, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":264
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_declaration.
 *     def enterComponent_declaration(self, ctx:ModelicaParser.Component_declarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Component_declarationContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_113enterComponent_declaration, 0, __pyx_n_s_ModelicaListener_enterComponent_3, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__116)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterComponent_declaration, __pyx_t_7) < 0) __PYX_ERR(0, 264, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":268
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_declaration.
 *     def exitComponent_declaration(self, ctx:ModelicaParser.Component_declarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Component_declarationContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_115exitComponent_declaration, 0, __pyx_n_s_ModelicaListener_exitComponent_d, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__118)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitComponent_declaration, __pyx_t_9) < 0) __PYX_ERR(0, 268, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":273
 * 
 *     # Enter a parse tree produced by ModelicaParser#condition_attribute.
 *     def enterCondition_attribute(self, ctx:ModelicaParser.Condition_attributeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Condition_attributeContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_117enterCondition_attribute, 0, __pyx_n_s_ModelicaListener_enterCondition, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__120)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterCondition_attribute, __pyx_t_8) < 0) __PYX_ERR(0, 273, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":277
 * 
 *     # Exit a parse tree produced by ModelicaParser#condition_attribute.
 *     def exitCondition_attribute(self, ctx:ModelicaParser.Condition_attributeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Condition_attributeContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_119exitCondition_attribute, 0, __pyx_n_s_ModelicaListener_exitCondition_a, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__122)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitCondition_attribute, __pyx_t_7) < 0) __PYX_ERR(0, 277, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":282
 * 
 *     # Enter a parse tree produced by ModelicaParser#declaration.
 *     def enterDeclaration(self, ctx:ModelicaParser.DeclarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_DeclarationContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_121enterDeclaration, 0, __pyx_n_s_ModelicaListener_enterDeclaratio, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__124)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterDeclaration, __pyx_t_9) < 0) __PYX_ERR(0, 282, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":286
 * 
 *     # Exit a parse tree produced by ModelicaParser#declaration.
 *     def exitDeclaration(self, ctx:ModelicaParser.DeclarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_DeclarationContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_123exitDeclaration, 0, __pyx_n_s_ModelicaListener_exitDeclaration, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__126)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitDeclaration, __pyx_t_8) < 0) __PYX_ERR(0, 286, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":291
 * 
 *     # Enter a parse tree produced by ModelicaParser#modification_class.
 *     def enterModification_class(self, ctx:ModelicaParser.Modification_classContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Modification_classContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_125enterModification_class, 0, __pyx_n_s_ModelicaListener_enterModificati, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__128)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterModification_class, __pyx_t_7) < 0) __PYX_ERR(0, 291, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":295
 * 
 *     # Exit a parse tree produced by ModelicaParser#modification_class.
 *     def exitModification_class(self, ctx:ModelicaParser.Modification_classContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Modification_classContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_127exitModification_class, 0, __pyx_n_s_ModelicaListener_exitModificatio, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__130)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitModification_class, __pyx_t_9) < 0) __PYX_ERR(0, 295, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":300
 * 
 *     # Enter a parse tree produced by ModelicaParser#modification_assignment.
 *     def enterModification_assignment(self, ctx:ModelicaParser.Modification_assignmentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Modification_assignmentContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_129enterModification_assignment, 0, __pyx_n_s_ModelicaListener_enterModificati_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__132)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterModification_assignment, __pyx_t_8) < 0) __PYX_ERR(0, 300, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":304
 * 
 *     # Exit a parse tree produced by ModelicaParser#modification_assignment.
 *     def exitModification_assignment(self, ctx:ModelicaParser.Modification_assignmentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Modification_assignmentContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_131exitModification_assignment, 0, __pyx_n_s_ModelicaListener_exitModificatio_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__134)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitModification_assignment, __pyx_t_7) < 0) __PYX_ERR(0, 304, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":309
 * 
 *     # Enter a parse tree produced by ModelicaParser#modification_assignment2.
 *     def enterModification_assignment2(self, ctx:ModelicaParser.Modification_assignment2Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Modification_assignment2Context); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_133enterModification_assignment2, 0, __pyx_n_s_ModelicaListener_enterModificati_3, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__136)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterModification_assignment2, __pyx_t_9) < 0) __PYX_ERR(0, 309, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":313
 * 
 *     # Exit a parse tree produced by ModelicaParser#modification_assignment2.
 *     def exitModification_assignment2(self, ctx:ModelicaParser.Modification_assignment2Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Modification_assignment2Context); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_135exitModification_assignment2, 0, __pyx_n_s_ModelicaListener_exitModificatio_3, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__138)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitModification_assignment2, __pyx_t_8) < 0) __PYX_ERR(0, 313, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":318
 * 
 *     # Enter a parse tree produced by ModelicaParser#class_modification.
 *     def enterClass_modification(self, ctx:ModelicaParser.Class_modificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Class_modificationContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_137enterClass_modification, 0, __pyx_n_s_ModelicaListener_enterClass_modi, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__140)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterClass_modification, __pyx_t_7) < 0) __PYX_ERR(0, 318, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":322
 * 
 *     # Exit a parse tree produced by ModelicaParser#class_modification.
 *     def exitClass_modification(self, ctx:ModelicaParser.Class_modificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Class_modificationContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_139exitClass_modification, 0, __pyx_n_s_ModelicaListener_exitClass_modif, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__142)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitClass_modification, __pyx_t_9) < 0) __PYX_ERR(0, 322, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":327
 * 
 *     # Enter a parse tree produced by ModelicaParser#argument_list.
 *     def enterArgument_list(self, ctx:ModelicaParser.Argument_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Argument_listContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_141enterArgument_list, 0, __pyx_n_s_ModelicaListener_enterArgument_l, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__144)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterArgument_list, __pyx_t_8) < 0) __PYX_ERR(0, 327, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":331
 * 
 *     # Exit a parse tree produced by ModelicaParser#argument_list.
 *     def exitArgument_list(self, ctx:ModelicaParser.Argument_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Argument_listContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_143exitArgument_list, 0, __pyx_n_s_ModelicaListener_exitArgument_li, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__146)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitArgument_list, __pyx_t_7) < 0) __PYX_ERR(0, 331, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":336
 * 
 *     # Enter a parse tree produced by ModelicaParser#argument.
 *     def enterArgument(self, ctx:ModelicaParser.ArgumentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_ArgumentContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_145enterArgument, 0, __pyx_n_s_ModelicaListener_enterArgument, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__148)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterArgument, __pyx_t_9) < 0) __PYX_ERR(0, 336, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":340
 * 
 *     # Exit a parse tree produced by ModelicaParser#argument.
 *     def exitArgument(self, ctx:ModelicaParser.ArgumentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_ArgumentContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_147exitArgument, 0, __pyx_n_s_ModelicaListener_exitArgument, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__150)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitArgument, __pyx_t_8) < 0) __PYX_ERR(0, 340, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":345
 * 
 *     # Enter a parse tree produced by ModelicaParser#element_modification_or_replaceable.
 *     def enterElement_modification_or_replaceable(self, ctx:ModelicaParser.Element_modification_or_replaceableContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Element_modification_or_replacea); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_149enterElement_modification_or_replaceable, 0, __pyx_n_s_ModelicaListener_enterElement_mo, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__152)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterElement_modification_or_rep, __pyx_t_7) < 0) __PYX_ERR(0, 345, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":349
 * 
 *     # Exit a parse tree produced by ModelicaParser#element_modification_or_replaceable.
 *     def exitElement_modification_or_replaceable(self, ctx:ModelicaParser.Element_modification_or_replaceableContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Element_modification_or_replacea); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_151exitElement_modification_or_replaceable, 0, __pyx_n_s_ModelicaListener_exitElement_mod, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__154)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitElement_modification_or_repl, __pyx_t_9) < 0) __PYX_ERR(0, 349, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":354
 * 
 *     # Enter a parse tree produced by ModelicaParser#element_modification.
 *     def enterElement_modification(self, ctx:ModelicaParser.Element_modificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Element_modificationContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_153enterElement_modification, 0, __pyx_n_s_ModelicaListener_enterElement_mo_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__156)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterElement_modification, __pyx_t_8) < 0) __PYX_ERR(0, 354, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":358
 * 
 *     # Exit a parse tree produced by ModelicaParser#element_modification.
 *     def exitElement_modification(self, ctx:ModelicaParser.Element_modificationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Element_modificationContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_155exitElement_modification, 0, __pyx_n_s_ModelicaListener_exitElement_mod_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__158)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitElement_modification, __pyx_t_7) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":363
 * 
 *     # Enter a parse tree produced by ModelicaParser#element_redeclaration.
 *     def enterElement_redeclaration(self, ctx:ModelicaParser.Element_redeclarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Element_redeclarationContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_157enterElement_redeclaration, 0, __pyx_n_s_ModelicaListener_enterElement_re, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__160)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterElement_redeclaration, __pyx_t_9) < 0) __PYX_ERR(0, 363, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":367
 * 
 *     # Exit a parse tree produced by ModelicaParser#element_redeclaration.
 *     def exitElement_redeclaration(self, ctx:ModelicaParser.Element_redeclarationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Element_redeclarationContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_159exitElement_redeclaration, 0, __pyx_n_s_ModelicaListener_exitElement_red, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__162)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitElement_redeclaration, __pyx_t_8) < 0) __PYX_ERR(0, 367, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":372
 * 
 *     # Enter a parse tree produced by ModelicaParser#element_replaceable.
 *     def enterElement_replaceable(self, ctx:ModelicaParser.Element_replaceableContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Element_replaceableContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_161enterElement_replaceable, 0, __pyx_n_s_ModelicaListener_enterElement_re_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__164)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterElement_replaceable, __pyx_t_7) < 0) __PYX_ERR(0, 372, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":376
 * 
 *     # Exit a parse tree produced by ModelicaParser#element_replaceable.
 *     def exitElement_replaceable(self, ctx:ModelicaParser.Element_replaceableContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Element_replaceableContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_163exitElement_replaceable, 0, __pyx_n_s_ModelicaListener_exitElement_rep, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__166)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitElement_replaceable, __pyx_t_9) < 0) __PYX_ERR(0, 376, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":381
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_clause1.
 *     def enterComponent_clause1(self, ctx:ModelicaParser.Component_clause1Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Component_clause1Context); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_165enterComponent_clause1, 0, __pyx_n_s_ModelicaListener_enterComponent_4, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__168)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterComponent_clause1, __pyx_t_8) < 0) __PYX_ERR(0, 381, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":385
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_clause1.
 *     def exitComponent_clause1(self, ctx:ModelicaParser.Component_clause1Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Component_clause1Context); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_167exitComponent_clause1, 0, __pyx_n_s_ModelicaListener_exitComponent_c_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__170)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitComponent_clause1, __pyx_t_7) < 0) __PYX_ERR(0, 385, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":390
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_declaration1.
 *     def enterComponent_declaration1(self, ctx:ModelicaParser.Component_declaration1Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Component_declaration1Context); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_169enterComponent_declaration1, 0, __pyx_n_s_ModelicaListener_enterComponent_5, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__172)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterComponent_declaration1, __pyx_t_9) < 0) __PYX_ERR(0, 390, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":394
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_declaration1.
 *     def exitComponent_declaration1(self, ctx:ModelicaParser.Component_declaration1Context):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Component_declaration1Context); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_171exitComponent_declaration1, 0, __pyx_n_s_ModelicaListener_exitComponent_d_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__174)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitComponent_declaration1, __pyx_t_8) < 0) __PYX_ERR(0, 394, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":399
 * 
 *     # Enter a parse tree produced by ModelicaParser#short_class_definition.
 *     def enterShort_class_definition(self, ctx:ModelicaParser.Short_class_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Short_class_definitionContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_173enterShort_class_definition, 0, __pyx_n_s_ModelicaListener_enterShort_clas, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__176)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterShort_class_definition, __pyx_t_7) < 0) __PYX_ERR(0, 399, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":403
 * 
 *     # Exit a parse tree produced by ModelicaParser#short_class_definition.
 *     def exitShort_class_definition(self, ctx:ModelicaParser.Short_class_definitionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Short_class_definitionContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_175exitShort_class_definition, 0, __pyx_n_s_ModelicaListener_exitShort_class, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__178)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitShort_class_definition, __pyx_t_9) < 0) __PYX_ERR(0, 403, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":408
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_section.
 *     def enterEquation_section(self, ctx:ModelicaParser.Equation_sectionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Equation_sectionContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_177enterEquation_section, 0, __pyx_n_s_ModelicaListener_enterEquation_s, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__180)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterEquation_section, __pyx_t_8) < 0) __PYX_ERR(0, 408, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":412
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_section.
 *     def exitEquation_section(self, ctx:ModelicaParser.Equation_sectionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Equation_sectionContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_179exitEquation_section, 0, __pyx_n_s_ModelicaListener_exitEquation_se, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__182)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitEquation_section, __pyx_t_7) < 0) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":417
 * 
 *     # Enter a parse tree produced by ModelicaParser#algorithm_section.
 *     def enterAlgorithm_section(self, ctx:ModelicaParser.Algorithm_sectionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Algorithm_sectionContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_181enterAlgorithm_section, 0, __pyx_n_s_ModelicaListener_enterAlgorithm, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__184)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterAlgorithm_section, __pyx_t_9) < 0) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":421
 * 
 *     # Exit a parse tree produced by ModelicaParser#algorithm_section.
 *     def exitAlgorithm_section(self, ctx:ModelicaParser.Algorithm_sectionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Algorithm_sectionContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_183exitAlgorithm_section, 0, __pyx_n_s_ModelicaListener_exitAlgorithm_s, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__186)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitAlgorithm_section, __pyx_t_8) < 0) __PYX_ERR(0, 421, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":426
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_simple.
 *     def enterEquation_simple(self, ctx:ModelicaParser.Equation_simpleContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Equation_simpleContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_185enterEquation_simple, 0, __pyx_n_s_ModelicaListener_enterEquation_s_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__188)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterEquation_simple, __pyx_t_7) < 0) __PYX_ERR(0, 426, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":430
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_simple.
 *     def exitEquation_simple(self, ctx:ModelicaParser.Equation_simpleContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Equation_simpleContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_187exitEquation_simple, 0, __pyx_n_s_ModelicaListener_exitEquation_si, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__190)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitEquation_simple, __pyx_t_9) < 0) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":435
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_if.
 *     def enterEquation_if(self, ctx:ModelicaParser.Equation_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Equation_ifContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_189enterEquation_if, 0, __pyx_n_s_ModelicaListener_enterEquation_i, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__192)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterEquation_if, __pyx_t_8) < 0) __PYX_ERR(0, 435, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":439
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_if.
 *     def exitEquation_if(self, ctx:ModelicaParser.Equation_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Equation_ifContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_191exitEquation_if, 0, __pyx_n_s_ModelicaListener_exitEquation_if, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__194)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitEquation_if, __pyx_t_7) < 0) __PYX_ERR(0, 439, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":444
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_for.
 *     def enterEquation_for(self, ctx:ModelicaParser.Equation_forContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Equation_forContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_193enterEquation_for, 0, __pyx_n_s_ModelicaListener_enterEquation_f, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__196)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterEquation_for, __pyx_t_9) < 0) __PYX_ERR(0, 444, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":448
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_for.
 *     def exitEquation_for(self, ctx:ModelicaParser.Equation_forContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Equation_forContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_195exitEquation_for, 0, __pyx_n_s_ModelicaListener_exitEquation_fo, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__198)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitEquation_for, __pyx_t_8) < 0) __PYX_ERR(0, 448, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":453
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_connect_clause.
 *     def enterEquation_connect_clause(self, ctx:ModelicaParser.Equation_connect_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Equation_connect_clauseContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_197enterEquation_connect_clause, 0, __pyx_n_s_ModelicaListener_enterEquation_c, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__200)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterEquation_connect_clause, __pyx_t_7) < 0) __PYX_ERR(0, 453, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":457
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_connect_clause.
 *     def exitEquation_connect_clause(self, ctx:ModelicaParser.Equation_connect_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Equation_connect_clauseContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_199exitEquation_connect_clause, 0, __pyx_n_s_ModelicaListener_exitEquation_co, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__202)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitEquation_connect_clause, __pyx_t_9) < 0) __PYX_ERR(0, 457, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":462
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_when.
 *     def enterEquation_when(self, ctx:ModelicaParser.Equation_whenContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Equation_whenContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_201enterEquation_when, 0, __pyx_n_s_ModelicaListener_enterEquation_w, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__204)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterEquation_when, __pyx_t_8) < 0) __PYX_ERR(0, 462, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":466
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_when.
 *     def exitEquation_when(self, ctx:ModelicaParser.Equation_whenContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Equation_whenContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_203exitEquation_when, 0, __pyx_n_s_ModelicaListener_exitEquation_wh, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__206)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitEquation_when, __pyx_t_7) < 0) __PYX_ERR(0, 466, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":471
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation_function.
 *     def enterEquation_function(self, ctx:ModelicaParser.Equation_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Equation_functionContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_205enterEquation_function, 0, __pyx_n_s_ModelicaListener_enterEquation_f_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__208)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterEquation_function, __pyx_t_9) < 0) __PYX_ERR(0, 471, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":475
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation_function.
 *     def exitEquation_function(self, ctx:ModelicaParser.Equation_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Equation_functionContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_207exitEquation_function, 0, __pyx_n_s_ModelicaListener_exitEquation_fu, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__210)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitEquation_function, __pyx_t_8) < 0) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":480
 * 
 *     # Enter a parse tree produced by ModelicaParser#equation.
 *     def enterEquation(self, ctx:ModelicaParser.EquationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_EquationContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_209enterEquation, 0, __pyx_n_s_ModelicaListener_enterEquation, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__212)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterEquation, __pyx_t_7) < 0) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":484
 * 
 *     # Exit a parse tree produced by ModelicaParser#equation.
 *     def exitEquation(self, ctx:ModelicaParser.EquationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_EquationContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_211exitEquation, 0, __pyx_n_s_ModelicaListener_exitEquation, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__214)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitEquation, __pyx_t_9) < 0) __PYX_ERR(0, 484, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":489
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_component_reference.
 *     def enterStatement_component_reference(self, ctx:ModelicaParser.Statement_component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Statement_component_referenceCon); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_213enterStatement_component_reference, 0, __pyx_n_s_ModelicaListener_enterStatement, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__216)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterStatement_component_referen, __pyx_t_8) < 0) __PYX_ERR(0, 489, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":493
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_component_reference.
 *     def exitStatement_component_reference(self, ctx:ModelicaParser.Statement_component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Statement_component_referenceCon); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_215exitStatement_component_reference, 0, __pyx_n_s_ModelicaListener_exitStatement_c, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__218)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitStatement_component_referenc, __pyx_t_7) < 0) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":498
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_component_function.
 *     def enterStatement_component_function(self, ctx:ModelicaParser.Statement_component_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Statement_component_functionCont); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_217enterStatement_component_function, 0, __pyx_n_s_ModelicaListener_enterStatement_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__220)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterStatement_component_functio, __pyx_t_9) < 0) __PYX_ERR(0, 498, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":502
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_component_function.
 *     def exitStatement_component_function(self, ctx:ModelicaParser.Statement_component_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Statement_component_functionCont); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_219exitStatement_component_function, 0, __pyx_n_s_ModelicaListener_exitStatement_c_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__222)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitStatement_component_function, __pyx_t_8) < 0) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":507
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_break.
 *     def enterStatement_break(self, ctx:ModelicaParser.Statement_breakContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Statement_breakContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_221enterStatement_break, 0, __pyx_n_s_ModelicaListener_enterStatement_3, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__224)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterStatement_break, __pyx_t_7) < 0) __PYX_ERR(0, 507, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":511
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_break.
 *     def exitStatement_break(self, ctx:ModelicaParser.Statement_breakContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Statement_breakContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_223exitStatement_break, 0, __pyx_n_s_ModelicaListener_exitStatement_b, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__226)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitStatement_break, __pyx_t_9) < 0) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":516
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_return.
 *     def enterStatement_return(self, ctx:ModelicaParser.Statement_returnContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Statement_returnContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_225enterStatement_return, 0, __pyx_n_s_ModelicaListener_enterStatement_4, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__228)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterStatement_return, __pyx_t_8) < 0) __PYX_ERR(0, 516, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":520
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_return.
 *     def exitStatement_return(self, ctx:ModelicaParser.Statement_returnContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Statement_returnContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_227exitStatement_return, 0, __pyx_n_s_ModelicaListener_exitStatement_r, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__230)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitStatement_return, __pyx_t_7) < 0) __PYX_ERR(0, 520, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":525
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_if.
 *     def enterStatement_if(self, ctx:ModelicaParser.Statement_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Statement_ifContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_229enterStatement_if, 0, __pyx_n_s_ModelicaListener_enterStatement_5, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__232)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterStatement_if, __pyx_t_9) < 0) __PYX_ERR(0, 525, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":529
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_if.
 *     def exitStatement_if(self, ctx:ModelicaParser.Statement_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Statement_ifContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_231exitStatement_if, 0, __pyx_n_s_ModelicaListener_exitStatement_i, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__234)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitStatement_if, __pyx_t_8) < 0) __PYX_ERR(0, 529, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":534
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_for.
 *     def enterStatement_for(self, ctx:ModelicaParser.Statement_forContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Statement_forContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_233enterStatement_for, 0, __pyx_n_s_ModelicaListener_enterStatement_6, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__236)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterStatement_for, __pyx_t_7) < 0) __PYX_ERR(0, 534, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":538
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_for.
 *     def exitStatement_for(self, ctx:ModelicaParser.Statement_forContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Statement_forContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_235exitStatement_for, 0, __pyx_n_s_ModelicaListener_exitStatement_f, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__238)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitStatement_for, __pyx_t_9) < 0) __PYX_ERR(0, 538, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":543
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_while.
 *     def enterStatement_while(self, ctx:ModelicaParser.Statement_whileContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Statement_whileContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_237enterStatement_while, 0, __pyx_n_s_ModelicaListener_enterStatement_7, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__240)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterStatement_while, __pyx_t_8) < 0) __PYX_ERR(0, 543, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":547
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_while.
 *     def exitStatement_while(self, ctx:ModelicaParser.Statement_whileContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Statement_whileContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_239exitStatement_while, 0, __pyx_n_s_ModelicaListener_exitStatement_w, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__242)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitStatement_while, __pyx_t_7) < 0) __PYX_ERR(0, 547, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":552
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement_when.
 *     def enterStatement_when(self, ctx:ModelicaParser.Statement_whenContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Statement_whenContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_241enterStatement_when, 0, __pyx_n_s_ModelicaListener_enterStatement_8, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__244)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterStatement_when, __pyx_t_9) < 0) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":556
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement_when.
 *     def exitStatement_when(self, ctx:ModelicaParser.Statement_whenContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Statement_whenContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_243exitStatement_when, 0, __pyx_n_s_ModelicaListener_exitStatement_w_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__246)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitStatement_when, __pyx_t_8) < 0) __PYX_ERR(0, 556, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":561
 * 
 *     # Enter a parse tree produced by ModelicaParser#statement.
 *     def enterStatement(self, ctx:ModelicaParser.StatementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_StatementContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_245enterStatement, 0, __pyx_n_s_ModelicaListener_enterStatement_9, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__248)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterStatement, __pyx_t_7) < 0) __PYX_ERR(0, 561, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":565
 * 
 *     # Exit a parse tree produced by ModelicaParser#statement.
 *     def exitStatement(self, ctx:ModelicaParser.StatementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_StatementContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_247exitStatement, 0, __pyx_n_s_ModelicaListener_exitStatement, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__250)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitStatement, __pyx_t_9) < 0) __PYX_ERR(0, 565, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":570
 * 
 *     # Enter a parse tree produced by ModelicaParser#if_equation.
 *     def enterIf_equation(self, ctx:ModelicaParser.If_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_If_equationContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_249enterIf_equation, 0, __pyx_n_s_ModelicaListener_enterIf_equatio, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__252)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterIf_equation, __pyx_t_8) < 0) __PYX_ERR(0, 570, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":574
 * 
 *     # Exit a parse tree produced by ModelicaParser#if_equation.
 *     def exitIf_equation(self, ctx:ModelicaParser.If_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_If_equationContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_251exitIf_equation, 0, __pyx_n_s_ModelicaListener_exitIf_equation, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__254)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitIf_equation, __pyx_t_7) < 0) __PYX_ERR(0, 574, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":579
 * 
 *     # Enter a parse tree produced by ModelicaParser#if_statement.
 *     def enterIf_statement(self, ctx:ModelicaParser.If_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_If_statementContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_253enterIf_statement, 0, __pyx_n_s_ModelicaListener_enterIf_stateme, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__256)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterIf_statement, __pyx_t_9) < 0) __PYX_ERR(0, 579, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":583
 * 
 *     # Exit a parse tree produced by ModelicaParser#if_statement.
 *     def exitIf_statement(self, ctx:ModelicaParser.If_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_If_statementContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_255exitIf_statement, 0, __pyx_n_s_ModelicaListener_exitIf_statemen, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__258)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitIf_statement, __pyx_t_8) < 0) __PYX_ERR(0, 583, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":588
 * 
 *     # Enter a parse tree produced by ModelicaParser#for_equation.
 *     def enterFor_equation(self, ctx:ModelicaParser.For_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_For_equationContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_257enterFor_equation, 0, __pyx_n_s_ModelicaListener_enterFor_equati, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__260)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterFor_equation, __pyx_t_7) < 0) __PYX_ERR(0, 588, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":592
 * 
 *     # Exit a parse tree produced by ModelicaParser#for_equation.
 *     def exitFor_equation(self, ctx:ModelicaParser.For_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_For_equationContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_259exitFor_equation, 0, __pyx_n_s_ModelicaListener_exitFor_equatio, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__262)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitFor_equation, __pyx_t_9) < 0) __PYX_ERR(0, 592, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":597
 * 
 *     # Enter a parse tree produced by ModelicaParser#for_statement.
 *     def enterFor_statement(self, ctx:ModelicaParser.For_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_For_statementContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_261enterFor_statement, 0, __pyx_n_s_ModelicaListener_enterFor_statem, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__264)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterFor_statement, __pyx_t_8) < 0) __PYX_ERR(0, 597, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":601
 * 
 *     # Exit a parse tree produced by ModelicaParser#for_statement.
 *     def exitFor_statement(self, ctx:ModelicaParser.For_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_For_statementContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_263exitFor_statement, 0, __pyx_n_s_ModelicaListener_exitFor_stateme, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__266)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitFor_statement, __pyx_t_7) < 0) __PYX_ERR(0, 601, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":606
 * 
 *     # Enter a parse tree produced by ModelicaParser#for_indices.
 *     def enterFor_indices(self, ctx:ModelicaParser.For_indicesContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_For_indicesContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_265enterFor_indices, 0, __pyx_n_s_ModelicaListener_enterFor_indice, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__268)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterFor_indices, __pyx_t_9) < 0) __PYX_ERR(0, 606, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":610
 * 
 *     # Exit a parse tree produced by ModelicaParser#for_indices.
 *     def exitFor_indices(self, ctx:ModelicaParser.For_indicesContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_For_indicesContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_267exitFor_indices, 0, __pyx_n_s_ModelicaListener_exitFor_indices, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__270)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitFor_indices, __pyx_t_8) < 0) __PYX_ERR(0, 610, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":615
 * 
 *     # Enter a parse tree produced by ModelicaParser#for_index.
 *     def enterFor_index(self, ctx:ModelicaParser.For_indexContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_For_indexContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_269enterFor_index, 0, __pyx_n_s_ModelicaListener_enterFor_index, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__272)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterFor_index, __pyx_t_7) < 0) __PYX_ERR(0, 615, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":619
 * 
 *     # Exit a parse tree produced by ModelicaParser#for_index.
 *     def exitFor_index(self, ctx:ModelicaParser.For_indexContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_For_indexContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_271exitFor_index, 0, __pyx_n_s_ModelicaListener_exitFor_index, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__274)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitFor_index, __pyx_t_9) < 0) __PYX_ERR(0, 619, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":624
 * 
 *     # Enter a parse tree produced by ModelicaParser#while_statement.
 *     def enterWhile_statement(self, ctx:ModelicaParser.While_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_While_statementContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_273enterWhile_statement, 0, __pyx_n_s_ModelicaListener_enterWhile_stat, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__276)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterWhile_statement, __pyx_t_8) < 0) __PYX_ERR(0, 624, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":628
 * 
 *     # Exit a parse tree produced by ModelicaParser#while_statement.
 *     def exitWhile_statement(self, ctx:ModelicaParser.While_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_While_statementContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_275exitWhile_statement, 0, __pyx_n_s_ModelicaListener_exitWhile_state, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__278)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitWhile_statement, __pyx_t_7) < 0) __PYX_ERR(0, 628, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":633
 * 
 *     # Enter a parse tree produced by ModelicaParser#when_equation.
 *     def enterWhen_equation(self, ctx:ModelicaParser.When_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_When_equationContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_277enterWhen_equation, 0, __pyx_n_s_ModelicaListener_enterWhen_equat, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__280)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterWhen_equation, __pyx_t_9) < 0) __PYX_ERR(0, 633, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":637
 * 
 *     # Exit a parse tree produced by ModelicaParser#when_equation.
 *     def exitWhen_equation(self, ctx:ModelicaParser.When_equationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_When_equationContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_279exitWhen_equation, 0, __pyx_n_s_ModelicaListener_exitWhen_equati, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__282)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitWhen_equation, __pyx_t_8) < 0) __PYX_ERR(0, 637, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":642
 * 
 *     # Enter a parse tree produced by ModelicaParser#when_statement.
 *     def enterWhen_statement(self, ctx:ModelicaParser.When_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_When_statementContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_281enterWhen_statement, 0, __pyx_n_s_ModelicaListener_enterWhen_state, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__284)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterWhen_statement, __pyx_t_7) < 0) __PYX_ERR(0, 642, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":646
 * 
 *     # Exit a parse tree produced by ModelicaParser#when_statement.
 *     def exitWhen_statement(self, ctx:ModelicaParser.When_statementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_When_statementContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_283exitWhen_statement, 0, __pyx_n_s_ModelicaListener_exitWhen_statem, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__286)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitWhen_statement, __pyx_t_9) < 0) __PYX_ERR(0, 646, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":651
 * 
 *     # Enter a parse tree produced by ModelicaParser#connect_clause.
 *     def enterConnect_clause(self, ctx:ModelicaParser.Connect_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Connect_clauseContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_285enterConnect_clause, 0, __pyx_n_s_ModelicaListener_enterConnect_cl, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__288)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterConnect_clause, __pyx_t_8) < 0) __PYX_ERR(0, 651, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":655
 * 
 *     # Exit a parse tree produced by ModelicaParser#connect_clause.
 *     def exitConnect_clause(self, ctx:ModelicaParser.Connect_clauseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Connect_clauseContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_287exitConnect_clause, 0, __pyx_n_s_ModelicaListener_exitConnect_cla, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__290)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitConnect_clause, __pyx_t_7) < 0) __PYX_ERR(0, 655, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":660
 * 
 *     # Enter a parse tree produced by ModelicaParser#expression_simple.
 *     def enterExpression_simple(self, ctx:ModelicaParser.Expression_simpleContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Expression_simpleContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_289enterExpression_simple, 0, __pyx_n_s_ModelicaListener_enterExpression, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__292)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterExpression_simple, __pyx_t_9) < 0) __PYX_ERR(0, 660, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":664
 * 
 *     # Exit a parse tree produced by ModelicaParser#expression_simple.
 *     def exitExpression_simple(self, ctx:ModelicaParser.Expression_simpleContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Expression_simpleContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_291exitExpression_simple, 0, __pyx_n_s_ModelicaListener_exitExpression, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__294)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitExpression_simple, __pyx_t_8) < 0) __PYX_ERR(0, 664, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":669
 * 
 *     # Enter a parse tree produced by ModelicaParser#expression_if.
 *     def enterExpression_if(self, ctx:ModelicaParser.Expression_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Expression_ifContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_293enterExpression_if, 0, __pyx_n_s_ModelicaListener_enterExpression_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__296)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterExpression_if, __pyx_t_7) < 0) __PYX_ERR(0, 669, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":673
 * 
 *     # Exit a parse tree produced by ModelicaParser#expression_if.
 *     def exitExpression_if(self, ctx:ModelicaParser.Expression_ifContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Expression_ifContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_295exitExpression_if, 0, __pyx_n_s_ModelicaListener_exitExpression_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__298)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitExpression_if, __pyx_t_9) < 0) __PYX_ERR(0, 673, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":678
 * 
 *     # Enter a parse tree produced by ModelicaParser#simple_expression.
 *     def enterSimple_expression(self, ctx:ModelicaParser.Simple_expressionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Simple_expressionContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_297enterSimple_expression, 0, __pyx_n_s_ModelicaListener_enterSimple_exp, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__300)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterSimple_expression, __pyx_t_8) < 0) __PYX_ERR(0, 678, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":682
 * 
 *     # Exit a parse tree produced by ModelicaParser#simple_expression.
 *     def exitSimple_expression(self, ctx:ModelicaParser.Simple_expressionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Simple_expressionContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_299exitSimple_expression, 0, __pyx_n_s_ModelicaListener_exitSimple_expr, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__302)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitSimple_expression, __pyx_t_7) < 0) __PYX_ERR(0, 682, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":687
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_neg.
 *     def enterExpr_neg(self, ctx:ModelicaParser.Expr_negContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Expr_negContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_301enterExpr_neg, 0, __pyx_n_s_ModelicaListener_enterExpr_neg, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__304)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterExpr_neg, __pyx_t_9) < 0) __PYX_ERR(0, 687, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":691
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_neg.
 *     def exitExpr_neg(self, ctx:ModelicaParser.Expr_negContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Expr_negContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_303exitExpr_neg, 0, __pyx_n_s_ModelicaListener_exitExpr_neg, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__306)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitExpr_neg, __pyx_t_8) < 0) __PYX_ERR(0, 691, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":696
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_add.
 *     def enterExpr_add(self, ctx:ModelicaParser.Expr_addContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Expr_addContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_305enterExpr_add, 0, __pyx_n_s_ModelicaListener_enterExpr_add, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__308)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterExpr_add, __pyx_t_7) < 0) __PYX_ERR(0, 696, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":700
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_add.
 *     def exitExpr_add(self, ctx:ModelicaParser.Expr_addContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Expr_addContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_307exitExpr_add, 0, __pyx_n_s_ModelicaListener_exitExpr_add, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__310)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitExpr_add, __pyx_t_9) < 0) __PYX_ERR(0, 700, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":705
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_exp.
 *     def enterExpr_exp(self, ctx:ModelicaParser.Expr_expContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Expr_expContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_309enterExpr_exp, 0, __pyx_n_s_ModelicaListener_enterExpr_exp, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__312)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterExpr_exp, __pyx_t_8) < 0) __PYX_ERR(0, 705, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":709
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_exp.
 *     def exitExpr_exp(self, ctx:ModelicaParser.Expr_expContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Expr_expContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_311exitExpr_exp, 0, __pyx_n_s_ModelicaListener_exitExpr_exp, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__314)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitExpr_exp, __pyx_t_7) < 0) __PYX_ERR(0, 709, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":714
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_or.
 *     def enterExpr_or(self, ctx:ModelicaParser.Expr_orContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Expr_orContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_313enterExpr_or, 0, __pyx_n_s_ModelicaListener_enterExpr_or, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__316)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterExpr_or, __pyx_t_9) < 0) __PYX_ERR(0, 714, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":718
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_or.
 *     def exitExpr_or(self, ctx:ModelicaParser.Expr_orContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Expr_orContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_315exitExpr_or, 0, __pyx_n_s_ModelicaListener_exitExpr_or, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__318)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitExpr_or, __pyx_t_8) < 0) __PYX_ERR(0, 718, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":723
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_primary.
 *     def enterExpr_primary(self, ctx:ModelicaParser.Expr_primaryContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Expr_primaryContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_317enterExpr_primary, 0, __pyx_n_s_ModelicaListener_enterExpr_prima, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__320)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterExpr_primary, __pyx_t_7) < 0) __PYX_ERR(0, 723, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":727
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_primary.
 *     def exitExpr_primary(self, ctx:ModelicaParser.Expr_primaryContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Expr_primaryContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_319exitExpr_primary, 0, __pyx_n_s_ModelicaListener_exitExpr_primar, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__322)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitExpr_primary, __pyx_t_9) < 0) __PYX_ERR(0, 727, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":732
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_and.
 *     def enterExpr_and(self, ctx:ModelicaParser.Expr_andContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Expr_andContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_321enterExpr_and, 0, __pyx_n_s_ModelicaListener_enterExpr_and, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__324)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterExpr_and, __pyx_t_8) < 0) __PYX_ERR(0, 732, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":736
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_and.
 *     def exitExpr_and(self, ctx:ModelicaParser.Expr_andContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Expr_andContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_323exitExpr_and, 0, __pyx_n_s_ModelicaListener_exitExpr_and, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__326)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitExpr_and, __pyx_t_7) < 0) __PYX_ERR(0, 736, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":741
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_rel.
 *     def enterExpr_rel(self, ctx:ModelicaParser.Expr_relContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Expr_relContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_325enterExpr_rel, 0, __pyx_n_s_ModelicaListener_enterExpr_rel, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__328)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterExpr_rel, __pyx_t_9) < 0) __PYX_ERR(0, 741, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":745
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_rel.
 *     def exitExpr_rel(self, ctx:ModelicaParser.Expr_relContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Expr_relContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_327exitExpr_rel, 0, __pyx_n_s_ModelicaListener_exitExpr_rel, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__330)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitExpr_rel, __pyx_t_8) < 0) __PYX_ERR(0, 745, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":750
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_not.
 *     def enterExpr_not(self, ctx:ModelicaParser.Expr_notContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Expr_notContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_329enterExpr_not, 0, __pyx_n_s_ModelicaListener_enterExpr_not, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__332)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterExpr_not, __pyx_t_7) < 0) __PYX_ERR(0, 750, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":754
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_not.
 *     def exitExpr_not(self, ctx:ModelicaParser.Expr_notContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Expr_notContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_331exitExpr_not, 0, __pyx_n_s_ModelicaListener_exitExpr_not, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__334)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitExpr_not, __pyx_t_9) < 0) __PYX_ERR(0, 754, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":759
 * 
 *     # Enter a parse tree produced by ModelicaParser#expr_mul.
 *     def enterExpr_mul(self, ctx:ModelicaParser.Expr_mulContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Expr_mulContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_333enterExpr_mul, 0, __pyx_n_s_ModelicaListener_enterExpr_mul, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__336)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterExpr_mul, __pyx_t_8) < 0) __PYX_ERR(0, 759, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":763
 * 
 *     # Exit a parse tree produced by ModelicaParser#expr_mul.
 *     def exitExpr_mul(self, ctx:ModelicaParser.Expr_mulContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Expr_mulContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_335exitExpr_mul, 0, __pyx_n_s_ModelicaListener_exitExpr_mul, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__338)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitExpr_mul, __pyx_t_7) < 0) __PYX_ERR(0, 763, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":768
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_unsigned_number.
 *     def enterPrimary_unsigned_number(self, ctx:ModelicaParser.Primary_unsigned_numberContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Primary_unsigned_numberContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_337enterPrimary_unsigned_number, 0, __pyx_n_s_ModelicaListener_enterPrimary_un, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__340)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterPrimary_unsigned_number, __pyx_t_9) < 0) __PYX_ERR(0, 768, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":772
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_unsigned_number.
 *     def exitPrimary_unsigned_number(self, ctx:ModelicaParser.Primary_unsigned_numberContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Primary_unsigned_numberContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_339exitPrimary_unsigned_number, 0, __pyx_n_s_ModelicaListener_exitPrimary_uns, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__342)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitPrimary_unsigned_number, __pyx_t_8) < 0) __PYX_ERR(0, 772, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":777
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_string.
 *     def enterPrimary_string(self, ctx:ModelicaParser.Primary_stringContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Primary_stringContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_341enterPrimary_string, 0, __pyx_n_s_ModelicaListener_enterPrimary_st, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__344)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterPrimary_string, __pyx_t_7) < 0) __PYX_ERR(0, 777, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":781
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_string.
 *     def exitPrimary_string(self, ctx:ModelicaParser.Primary_stringContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Primary_stringContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 781, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_343exitPrimary_string, 0, __pyx_n_s_ModelicaListener_exitPrimary_str, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__346)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 781, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitPrimary_string, __pyx_t_9) < 0) __PYX_ERR(0, 781, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":786
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_false.
 *     def enterPrimary_false(self, ctx:ModelicaParser.Primary_falseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Primary_falseContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_345enterPrimary_false, 0, __pyx_n_s_ModelicaListener_enterPrimary_fa, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__348)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterPrimary_false, __pyx_t_8) < 0) __PYX_ERR(0, 786, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":790
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_false.
 *     def exitPrimary_false(self, ctx:ModelicaParser.Primary_falseContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Primary_falseContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_347exitPrimary_false, 0, __pyx_n_s_ModelicaListener_exitPrimary_fal, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__350)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitPrimary_false, __pyx_t_7) < 0) __PYX_ERR(0, 790, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":795
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_true.
 *     def enterPrimary_true(self, ctx:ModelicaParser.Primary_trueContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Primary_trueContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_349enterPrimary_true, 0, __pyx_n_s_ModelicaListener_enterPrimary_tr, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__352)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterPrimary_true, __pyx_t_9) < 0) __PYX_ERR(0, 795, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":799
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_true.
 *     def exitPrimary_true(self, ctx:ModelicaParser.Primary_trueContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Primary_trueContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_351exitPrimary_true, 0, __pyx_n_s_ModelicaListener_exitPrimary_tru, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__354)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitPrimary_true, __pyx_t_8) < 0) __PYX_ERR(0, 799, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":804
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_function.
 *     def enterPrimary_function(self, ctx:ModelicaParser.Primary_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Primary_functionContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_353enterPrimary_function, 0, __pyx_n_s_ModelicaListener_enterPrimary_fu, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__356)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterPrimary_function, __pyx_t_7) < 0) __PYX_ERR(0, 804, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":808
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_function.
 *     def exitPrimary_function(self, ctx:ModelicaParser.Primary_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Primary_functionContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_355exitPrimary_function, 0, __pyx_n_s_ModelicaListener_exitPrimary_fun, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__358)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitPrimary_function, __pyx_t_9) < 0) __PYX_ERR(0, 808, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":813
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_derivative.
 *     def enterPrimary_derivative(self, ctx:ModelicaParser.Primary_derivativeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Primary_derivativeContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_357enterPrimary_derivative, 0, __pyx_n_s_ModelicaListener_enterPrimary_de, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__360)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterPrimary_derivative, __pyx_t_8) < 0) __PYX_ERR(0, 813, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":817
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_derivative.
 *     def exitPrimary_derivative(self, ctx:ModelicaParser.Primary_derivativeContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Primary_derivativeContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_359exitPrimary_derivative, 0, __pyx_n_s_ModelicaListener_exitPrimary_der, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__362)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitPrimary_derivative, __pyx_t_7) < 0) __PYX_ERR(0, 817, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":822
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_initial.
 *     def enterPrimary_initial(self, ctx:ModelicaParser.Primary_initialContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Primary_initialContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_361enterPrimary_initial, 0, __pyx_n_s_ModelicaListener_enterPrimary_in, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__364)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterPrimary_initial, __pyx_t_9) < 0) __PYX_ERR(0, 822, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":826
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_initial.
 *     def exitPrimary_initial(self, ctx:ModelicaParser.Primary_initialContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Primary_initialContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_363exitPrimary_initial, 0, __pyx_n_s_ModelicaListener_exitPrimary_ini, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__366)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitPrimary_initial, __pyx_t_8) < 0) __PYX_ERR(0, 826, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":831
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_component_reference.
 *     def enterPrimary_component_reference(self, ctx:ModelicaParser.Primary_component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Primary_component_referenceConte); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_365enterPrimary_component_reference, 0, __pyx_n_s_ModelicaListener_enterPrimary_co, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__368)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterPrimary_component_reference, __pyx_t_7) < 0) __PYX_ERR(0, 831, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":835
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_component_reference.
 *     def exitPrimary_component_reference(self, ctx:ModelicaParser.Primary_component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Primary_component_referenceConte); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_367exitPrimary_component_reference, 0, __pyx_n_s_ModelicaListener_exitPrimary_com, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__370)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitPrimary_component_reference, __pyx_t_9) < 0) __PYX_ERR(0, 835, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":840
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_output_expression_list.
 *     def enterPrimary_output_expression_list(self, ctx:ModelicaParser.Primary_output_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Primary_output_expression_listCo); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_369enterPrimary_output_expression_list, 0, __pyx_n_s_ModelicaListener_enterPrimary_ou, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__372)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterPrimary_output_expression_l, __pyx_t_8) < 0) __PYX_ERR(0, 840, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":844
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_output_expression_list.
 *     def exitPrimary_output_expression_list(self, ctx:ModelicaParser.Primary_output_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Primary_output_expression_listCo); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_371exitPrimary_output_expression_list, 0, __pyx_n_s_ModelicaListener_exitPrimary_out, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__374)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitPrimary_output_expression_li, __pyx_t_7) < 0) __PYX_ERR(0, 844, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":849
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_expression_list.
 *     def enterPrimary_expression_list(self, ctx:ModelicaParser.Primary_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Primary_expression_listContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_373enterPrimary_expression_list, 0, __pyx_n_s_ModelicaListener_enterPrimary_ex, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__376)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterPrimary_expression_list, __pyx_t_9) < 0) __PYX_ERR(0, 849, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":853
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_expression_list.
 *     def exitPrimary_expression_list(self, ctx:ModelicaParser.Primary_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Primary_expression_listContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_375exitPrimary_expression_list, 0, __pyx_n_s_ModelicaListener_exitPrimary_exp, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__378)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitPrimary_expression_list, __pyx_t_8) < 0) __PYX_ERR(0, 853, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":858
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_function_arguments.
 *     def enterPrimary_function_arguments(self, ctx:ModelicaParser.Primary_function_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Primary_function_argumentsContex); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_377enterPrimary_function_arguments, 0, __pyx_n_s_ModelicaListener_enterPrimary_fu_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__380)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterPrimary_function_arguments, __pyx_t_7) < 0) __PYX_ERR(0, 858, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":862
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_function_arguments.
 *     def exitPrimary_function_arguments(self, ctx:ModelicaParser.Primary_function_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Primary_function_argumentsContex); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_379exitPrimary_function_arguments, 0, __pyx_n_s_ModelicaListener_exitPrimary_fun_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__382)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitPrimary_function_arguments, __pyx_t_9) < 0) __PYX_ERR(0, 862, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":867
 * 
 *     # Enter a parse tree produced by ModelicaParser#primary_end.
 *     def enterPrimary_end(self, ctx:ModelicaParser.Primary_endContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Primary_endContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_381enterPrimary_end, 0, __pyx_n_s_ModelicaListener_enterPrimary_en, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__384)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterPrimary_end, __pyx_t_8) < 0) __PYX_ERR(0, 867, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":871
 * 
 *     # Exit a parse tree produced by ModelicaParser#primary_end.
 *     def exitPrimary_end(self, ctx:ModelicaParser.Primary_endContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Primary_endContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_383exitPrimary_end, 0, __pyx_n_s_ModelicaListener_exitPrimary_end, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__386)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitPrimary_end, __pyx_t_7) < 0) __PYX_ERR(0, 871, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":876
 * 
 *     # Enter a parse tree produced by ModelicaParser#name.
 *     def enterName(self, ctx:ModelicaParser.NameContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_NameContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_385enterName, 0, __pyx_n_s_ModelicaListener_enterName, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__388)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterName, __pyx_t_9) < 0) __PYX_ERR(0, 876, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":880
 * 
 *     # Exit a parse tree produced by ModelicaParser#name.
 *     def exitName(self, ctx:ModelicaParser.NameContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_NameContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_387exitName, 0, __pyx_n_s_ModelicaListener_exitName, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__390)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitName, __pyx_t_8) < 0) __PYX_ERR(0, 880, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":885
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_reference_element.
 *     def enterComponent_reference_element(self, ctx:ModelicaParser.Component_reference_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Component_reference_elementConte); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_389enterComponent_reference_element, 0, __pyx_n_s_ModelicaListener_enterComponent_6, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__392)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterComponent_reference_element, __pyx_t_7) < 0) __PYX_ERR(0, 885, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":889
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_reference_element.
 *     def exitComponent_reference_element(self, ctx:ModelicaParser.Component_reference_elementContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Component_reference_elementConte); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_391exitComponent_reference_element, 0, __pyx_n_s_ModelicaListener_exitComponent_r, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__394)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitComponent_reference_element, __pyx_t_9) < 0) __PYX_ERR(0, 889, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":894
 * 
 *     # Enter a parse tree produced by ModelicaParser#component_reference.
 *     def enterComponent_reference(self, ctx:ModelicaParser.Component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Component_referenceContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_393enterComponent_reference, 0, __pyx_n_s_ModelicaListener_enterComponent_7, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__396)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterComponent_reference, __pyx_t_8) < 0) __PYX_ERR(0, 894, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":898
 * 
 *     # Exit a parse tree produced by ModelicaParser#component_reference.
 *     def exitComponent_reference(self, ctx:ModelicaParser.Component_referenceContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Component_referenceContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_395exitComponent_reference, 0, __pyx_n_s_ModelicaListener_exitComponent_r_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__398)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitComponent_reference, __pyx_t_7) < 0) __PYX_ERR(0, 898, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":903
 * 
 *     # Enter a parse tree produced by ModelicaParser#function_call_args.
 *     def enterFunction_call_args(self, ctx:ModelicaParser.Function_call_argsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Function_call_argsContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_397enterFunction_call_args, 0, __pyx_n_s_ModelicaListener_enterFunction_c, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__400)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterFunction_call_args, __pyx_t_9) < 0) __PYX_ERR(0, 903, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":907
 * 
 *     # Exit a parse tree produced by ModelicaParser#function_call_args.
 *     def exitFunction_call_args(self, ctx:ModelicaParser.Function_call_argsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Function_call_argsContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_399exitFunction_call_args, 0, __pyx_n_s_ModelicaListener_exitFunction_ca, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__402)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitFunction_call_args, __pyx_t_8) < 0) __PYX_ERR(0, 907, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":912
 * 
 *     # Enter a parse tree produced by ModelicaParser#function_arguments.
 *     def enterFunction_arguments(self, ctx:ModelicaParser.Function_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Function_argumentsContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_401enterFunction_arguments, 0, __pyx_n_s_ModelicaListener_enterFunction_a, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__404)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterFunction_arguments, __pyx_t_7) < 0) __PYX_ERR(0, 912, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":916
 * 
 *     # Exit a parse tree produced by ModelicaParser#function_arguments.
 *     def exitFunction_arguments(self, ctx:ModelicaParser.Function_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Function_argumentsContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_403exitFunction_arguments, 0, __pyx_n_s_ModelicaListener_exitFunction_ar, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__406)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitFunction_arguments, __pyx_t_9) < 0) __PYX_ERR(0, 916, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":921
 * 
 *     # Enter a parse tree produced by ModelicaParser#named_arguments.
 *     def enterNamed_arguments(self, ctx:ModelicaParser.Named_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Named_argumentsContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_405enterNamed_arguments, 0, __pyx_n_s_ModelicaListener_enterNamed_argu, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__408)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterNamed_arguments, __pyx_t_8) < 0) __PYX_ERR(0, 921, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":925
 * 
 *     # Exit a parse tree produced by ModelicaParser#named_arguments.
 *     def exitNamed_arguments(self, ctx:ModelicaParser.Named_argumentsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Named_argumentsContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_407exitNamed_arguments, 0, __pyx_n_s_ModelicaListener_exitNamed_argum, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__410)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitNamed_arguments, __pyx_t_7) < 0) __PYX_ERR(0, 925, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":930
 * 
 *     # Enter a parse tree produced by ModelicaParser#named_argument.
 *     def enterNamed_argument(self, ctx:ModelicaParser.Named_argumentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Named_argumentContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_409enterNamed_argument, 0, __pyx_n_s_ModelicaListener_enterNamed_argu_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__412)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterNamed_argument, __pyx_t_9) < 0) __PYX_ERR(0, 930, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":934
 * 
 *     # Exit a parse tree produced by ModelicaParser#named_argument.
 *     def exitNamed_argument(self, ctx:ModelicaParser.Named_argumentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Named_argumentContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_411exitNamed_argument, 0, __pyx_n_s_ModelicaListener_exitNamed_argum_2, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__414)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitNamed_argument, __pyx_t_8) < 0) __PYX_ERR(0, 934, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":939
 * 
 *     # Enter a parse tree produced by ModelicaParser#argument_function.
 *     def enterArgument_function(self, ctx:ModelicaParser.Argument_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Argument_functionContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_413enterArgument_function, 0, __pyx_n_s_ModelicaListener_enterArgument_f, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__416)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterArgument_function, __pyx_t_7) < 0) __PYX_ERR(0, 939, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":943
 * 
 *     # Exit a parse tree produced by ModelicaParser#argument_function.
 *     def exitArgument_function(self, ctx:ModelicaParser.Argument_functionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Argument_functionContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_415exitArgument_function, 0, __pyx_n_s_ModelicaListener_exitArgument_fu, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__418)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitArgument_function, __pyx_t_9) < 0) __PYX_ERR(0, 943, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":948
 * 
 *     # Enter a parse tree produced by ModelicaParser#argument_expression.
 *     def enterArgument_expression(self, ctx:ModelicaParser.Argument_expressionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Argument_expressionContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_417enterArgument_expression, 0, __pyx_n_s_ModelicaListener_enterArgument_e, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__420)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterArgument_expression, __pyx_t_8) < 0) __PYX_ERR(0, 948, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":952
 * 
 *     # Exit a parse tree produced by ModelicaParser#argument_expression.
 *     def exitArgument_expression(self, ctx:ModelicaParser.Argument_expressionContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Argument_expressionContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 952, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_419exitArgument_expression, 0, __pyx_n_s_ModelicaListener_exitArgument_ex, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__422)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 952, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitArgument_expression, __pyx_t_7) < 0) __PYX_ERR(0, 952, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":957
 * 
 *     # Enter a parse tree produced by ModelicaParser#output_expression_list.
 *     def enterOutput_expression_list(self, ctx:ModelicaParser.Output_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Output_expression_listContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_421enterOutput_expression_list, 0, __pyx_n_s_ModelicaListener_enterOutput_exp, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__424)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterOutput_expression_list, __pyx_t_9) < 0) __PYX_ERR(0, 957, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":961
 * 
 *     # Exit a parse tree produced by ModelicaParser#output_expression_list.
 *     def exitOutput_expression_list(self, ctx:ModelicaParser.Output_expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Output_expression_listContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_423exitOutput_expression_list, 0, __pyx_n_s_ModelicaListener_exitOutput_expr, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__426)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitOutput_expression_list, __pyx_t_8) < 0) __PYX_ERR(0, 961, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":966
 * 
 *     # Enter a parse tree produced by ModelicaParser#expression_list.
 *     def enterExpression_list(self, ctx:ModelicaParser.Expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Expression_listContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_425enterExpression_list, 0, __pyx_n_s_ModelicaListener_enterExpression_3, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__428)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterExpression_list, __pyx_t_7) < 0) __PYX_ERR(0, 966, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":970
 * 
 *     # Exit a parse tree produced by ModelicaParser#expression_list.
 *     def exitExpression_list(self, ctx:ModelicaParser.Expression_listContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_Expression_listContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_427exitExpression_list, 0, __pyx_n_s_ModelicaListener_exitExpression_3, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__430)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitExpression_list, __pyx_t_9) < 0) __PYX_ERR(0, 970, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":975
 * 
 *     # Enter a parse tree produced by ModelicaParser#array_subscripts.
 *     def enterArray_subscripts(self, ctx:ModelicaParser.Array_subscriptsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_Array_subscriptsContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_429enterArray_subscripts, 0, __pyx_n_s_ModelicaListener_enterArray_subs, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__432)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterArray_subscripts, __pyx_t_8) < 0) __PYX_ERR(0, 975, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":979
 * 
 *     # Exit a parse tree produced by ModelicaParser#array_subscripts.
 *     def exitArray_subscripts(self, ctx:ModelicaParser.Array_subscriptsContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_Array_subscriptsContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 979, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_431exitArray_subscripts, 0, __pyx_n_s_ModelicaListener_exitArray_subsc, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__434)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 979, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitArray_subscripts, __pyx_t_7) < 0) __PYX_ERR(0, 979, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":984
 * 
 *     # Enter a parse tree produced by ModelicaParser#subscript.
 *     def enterSubscript(self, ctx:ModelicaParser.SubscriptContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_SubscriptContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_433enterSubscript, 0, __pyx_n_s_ModelicaListener_enterSubscript, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__436)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterSubscript, __pyx_t_9) < 0) __PYX_ERR(0, 984, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":988
 * 
 *     # Exit a parse tree produced by ModelicaParser#subscript.
 *     def exitSubscript(self, ctx:ModelicaParser.SubscriptContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_SubscriptContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_435exitSubscript, 0, __pyx_n_s_ModelicaListener_exitSubscript, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__438)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitSubscript, __pyx_t_8) < 0) __PYX_ERR(0, 988, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":993
 * 
 *     # Enter a parse tree produced by ModelicaParser#comment.
 *     def enterComment(self, ctx:ModelicaParser.CommentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_CommentContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_437enterComment, 0, __pyx_n_s_ModelicaListener_enterComment, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__440)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterComment, __pyx_t_7) < 0) __PYX_ERR(0, 993, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":997
 * 
 *     # Exit a parse tree produced by ModelicaParser#comment.
 *     def exitComment(self, ctx:ModelicaParser.CommentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_CommentContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_439exitComment, 0, __pyx_n_s_ModelicaListener_exitComment, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__442)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitComment, __pyx_t_9) < 0) __PYX_ERR(0, 997, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":1002
 * 
 *     # Enter a parse tree produced by ModelicaParser#string_comment.
 *     def enterString_comment(self, ctx:ModelicaParser.String_commentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_String_commentContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_441enterString_comment, 0, __pyx_n_s_ModelicaListener_enterString_com, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__444)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterString_comment, __pyx_t_8) < 0) __PYX_ERR(0, 1002, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":1006
 * 
 *     # Exit a parse tree produced by ModelicaParser#string_comment.
 *     def exitString_comment(self, ctx:ModelicaParser.String_commentContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_8 = PyDict_New(); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_PyObject_GetAttrStr(__pyx_t_9, __pyx_n_s_String_commentContext); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyDict_SetItem(__pyx_t_8, __pyx_n_s_ctx, __pyx_t_7) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  __pyx_t_7 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_443exitString_comment, 0, __pyx_n_s_ModelicaListener_exitString_comm, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__446)); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_7, __pyx_t_8);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitString_comment, __pyx_t_7) < 0) __PYX_ERR(0, 1006, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;

  /* "pymola/generated/ModelicaListener.pyx":1011
 * 
 *     # Enter a parse tree produced by ModelicaParser#annotation.
 *     def enterAnnotation(self, ctx:ModelicaParser.AnnotationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_7 = PyDict_New(); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __pyx_t_9 = __Pyx_PyObject_GetAttrStr(__pyx_t_8, __pyx_n_s_AnnotationContext); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  if (PyDict_SetItem(__pyx_t_7, __pyx_n_s_ctx, __pyx_t_9) < 0) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  __pyx_t_9 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_445enterAnnotation, 0, __pyx_n_s_ModelicaListener_enterAnnotation, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__448)); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_9, __pyx_t_7);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_enterAnnotation, __pyx_t_9) < 0) __PYX_ERR(0, 1011, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;

  /* "pymola/generated/ModelicaListener.pyx":1015
 * 
 *     # Exit a parse tree produced by ModelicaParser#annotation.
 *     def exitAnnotation(self, ctx:ModelicaParser.AnnotationContext):             # <<<<<<<<<<<<<<
 *         pass
 * 
 */
  __pyx_t_9 = PyDict_New(); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_9);
  __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_ModelicaParser); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_7);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_7, __pyx_n_s_AnnotationContext); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
  if (PyDict_SetItem(__pyx_t_9, __pyx_n_s_ctx, __pyx_t_8) < 0) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_8 = __Pyx_CyFunction_NewEx(&__pyx_mdef_6pymola_9generated_16ModelicaListener_16ModelicaListener_447exitAnnotation, 0, __pyx_n_s_ModelicaListener_exitAnnotation, NULL, __pyx_n_s_pymola_generated_ModelicaListene, __pyx_d, ((PyObject *)__pyx_codeobj__450)); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_CyFunction_SetAnnotationsDict(__pyx_t_8, __pyx_t_9);
  __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
  if (PyObject_SetItem(__pyx_t_6, __pyx_n_s_exitAnnotation, __pyx_t_8) < 0) __PYX_ERR(0, 1015, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;

  /* "pymola/generated/ModelicaListener.pyx":9
 * 
 * # This class defines a complete listener for a parse tree produced by ModelicaParser.
 * class ModelicaListener(ParseTreeListener):             # <<<<<<<<<<<<<<
 * 
 *     # Enter a parse tree produced by ModelicaParser#stored_definition.
 */
  __pyx_t_8 = __Pyx_Py3ClassCreate(__pyx_t_2, __pyx_n_s_ModelicaListener, __pyx_t_1, __pyx_t_6, NULL, 0, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_ModelicaListener, __pyx_t_8) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "pymola/generated/ModelicaListener.pyx":1
 * # Generated from Modelica.g4 by ANTLR 4.7             # <<<<<<<<<<<<<<
 * from antlr4 import *
 * if __name__ is not None and "." in __name__:
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_1) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init pymola.generated.ModelicaListener", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init pymola.generated.ModelicaListener");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* RaiseArgTupleInvalid */
static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* Import */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* GetModuleGlobalName */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* ImportFrom */
  static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CalculateMetaclass */
  static PyObject *__Pyx_CalculateMetaclass(PyTypeObject *metaclass, PyObject *bases) {
    Py_ssize_t i, nbases = PyTuple_GET_SIZE(bases);
    for (i=0; i < nbases; i++) {
        PyTypeObject *tmptype;
        PyObject *tmp = PyTuple_GET_ITEM(bases, i);
        tmptype = Py_TYPE(tmp);
#if PY_MAJOR_VERSION < 3
        if (tmptype == &PyClass_Type)
            continue;
#endif
        if (!metaclass) {
            metaclass = tmptype;
            continue;
        }
        if (PyType_IsSubtype(metaclass, tmptype))
            continue;
        if (PyType_IsSubtype(tmptype, metaclass)) {
            metaclass = tmptype;
            continue;
        }
        PyErr_SetString(PyExc_TypeError,
                        "metaclass conflict: "
                        "the metaclass of a derived class "
                        "must be a (non-strict) subclass "
                        "of the metaclasses of all its bases");
        return NULL;
    }
    if (!metaclass) {
#if PY_MAJOR_VERSION < 3
        metaclass = &PyClass_Type;
#else
        metaclass = &PyType_Type;
#endif
    }
    Py_INCREF((PyObject*) metaclass);
    return (PyObject*) metaclass;
}

/* FetchCommonType */
  static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* CythonFunction */
  static PyObject *
__Pyx_CyFunction_get_doc(__pyx_CyFunctionObject *op, CYTHON_UNUSED void *closure)
{
    if (unlikely(op->func_doc == NULL)) {
        if (op->func.m_ml->ml_doc) {
#if PY_MAJOR_VERSION >= 3
            op->func_doc = PyUnicode_FromString(op->func.m_ml->ml_doc);
#else
            op->func_doc = PyString_FromString(op->func.m_ml->ml_doc);
#endif
            if (unlikely(op->func_doc == NULL))
                return NULL;
        } else {
            Py_INCREF(Py_None);
            return Py_None;
        }
    }
    Py_INCREF(op->func_doc);
    return op->func_doc;
}
static int
__Pyx_CyFunction_set_doc(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp = op->func_doc;
    if (value == NULL) {
        value = Py_None;
    }
    Py_INCREF(value);
    op->func_doc = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_name(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_name == NULL)) {
#if PY_MAJOR_VERSION >= 3
        op->func_name = PyUnicode_InternFromString(op->func.m_ml->ml_name);
#else
        op->func_name = PyString_InternFromString(op->func.m_ml->ml_name);
#endif
        if (unlikely(op->func_name == NULL))
            return NULL;
    }
    Py_INCREF(op->func_name);
    return op->func_name;
}
static int
__Pyx_CyFunction_set_name(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = op->func_name;
    Py_INCREF(value);
    op->func_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_qualname(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_qualname);
    return op->func_qualname;
}
static int
__Pyx_CyFunction_set_qualname(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = op->func_qualname;
    Py_INCREF(value);
    op->func_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_self(__pyx_CyFunctionObject *m, CYTHON_UNUSED void *closure)
{
    PyObject *self;
    self = m->func_closure;
    if (self == NULL)
        self = Py_None;
    Py_INCREF(self);
    return self;
}
static PyObject *
__Pyx_CyFunction_get_dict(__pyx_CyFunctionObject *op)
{
    if (unlikely(op->func_dict == NULL)) {
        op->func_dict = PyDict_New();
        if (unlikely(op->func_dict == NULL))
            return NULL;
    }
    Py_INCREF(op->func_dict);
    return op->func_dict;
}
static int
__Pyx_CyFunction_set_dict(__pyx_CyFunctionObject *op, PyObject *value)
{
    PyObject *tmp;
    if (unlikely(value == NULL)) {
        PyErr_SetString(PyExc_TypeError,
               "function's dictionary may not be deleted");
        return -1;
    }
    if (unlikely(!PyDict_Check(value))) {
        PyErr_SetString(PyExc_TypeError,
               "setting function's dictionary to a non-dict");
        return -1;
    }
    tmp = op->func_dict;
    Py_INCREF(value);
    op->func_dict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_globals(__pyx_CyFunctionObject *op)
{
    Py_INCREF(op->func_globals);
    return op->func_globals;
}
static PyObject *
__Pyx_CyFunction_get_closure(CYTHON_UNUSED __pyx_CyFunctionObject *op)
{
    Py_INCREF(Py_None);
    return Py_None;
}
static PyObject *
__Pyx_CyFunction_get_code(__pyx_CyFunctionObject *op)
{
    PyObject* result = (op->func_code) ? op->func_code : Py_None;
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_init_defaults(__pyx_CyFunctionObject *op) {
    int result = 0;
    PyObject *res = op->defaults_getter((PyObject *) op);
    if (unlikely(!res))
        return -1;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    op->defaults_tuple = PyTuple_GET_ITEM(res, 0);
    Py_INCREF(op->defaults_tuple);
    op->defaults_kwdict = PyTuple_GET_ITEM(res, 1);
    Py_INCREF(op->defaults_kwdict);
    #else
    op->defaults_tuple = PySequence_ITEM(res, 0);
    if (unlikely(!op->defaults_tuple)) result = -1;
    else {
        op->defaults_kwdict = PySequence_ITEM(res, 1);
        if (unlikely(!op->defaults_kwdict)) result = -1;
    }
    #endif
    Py_DECREF(res);
    return result;
}
static int
__Pyx_CyFunction_set_defaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyTuple_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__defaults__ must be set to a tuple object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_tuple;
    op->defaults_tuple = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_defaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_tuple;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_tuple;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_kwdefaults(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value) {
        value = Py_None;
    } else if (value != Py_None && !PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__kwdefaults__ must be set to a dict object");
        return -1;
    }
    Py_INCREF(value);
    tmp = op->defaults_kwdict;
    op->defaults_kwdict = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_kwdefaults(__pyx_CyFunctionObject *op) {
    PyObject* result = op->defaults_kwdict;
    if (unlikely(!result)) {
        if (op->defaults_getter) {
            if (__Pyx_CyFunction_init_defaults(op) < 0) return NULL;
            result = op->defaults_kwdict;
        } else {
            result = Py_None;
        }
    }
    Py_INCREF(result);
    return result;
}
static int
__Pyx_CyFunction_set_annotations(__pyx_CyFunctionObject *op, PyObject* value) {
    PyObject* tmp;
    if (!value || value == Py_None) {
        value = NULL;
    } else if (!PyDict_Check(value)) {
        PyErr_SetString(PyExc_TypeError,
                        "__annotations__ must be set to a dict object");
        return -1;
    }
    Py_XINCREF(value);
    tmp = op->func_annotations;
    op->func_annotations = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_CyFunction_get_annotations(__pyx_CyFunctionObject *op) {
    PyObject* result = op->func_annotations;
    if (unlikely(!result)) {
        result = PyDict_New();
        if (unlikely(!result)) return NULL;
        op->func_annotations = result;
    }
    Py_INCREF(result);
    return result;
}
static PyGetSetDef __pyx_CyFunction_getsets[] = {
    {(char *) "func_doc", (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "__doc__",  (getter)__Pyx_CyFunction_get_doc, (setter)__Pyx_CyFunction_set_doc, 0, 0},
    {(char *) "func_name", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__name__", (getter)__Pyx_CyFunction_get_name, (setter)__Pyx_CyFunction_set_name, 0, 0},
    {(char *) "__qualname__", (getter)__Pyx_CyFunction_get_qualname, (setter)__Pyx_CyFunction_set_qualname, 0, 0},
    {(char *) "__self__", (getter)__Pyx_CyFunction_get_self, 0, 0, 0},
    {(char *) "func_dict", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "__dict__", (getter)__Pyx_CyFunction_get_dict, (setter)__Pyx_CyFunction_set_dict, 0, 0},
    {(char *) "func_globals", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "__globals__", (getter)__Pyx_CyFunction_get_globals, 0, 0, 0},
    {(char *) "func_closure", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "__closure__", (getter)__Pyx_CyFunction_get_closure, 0, 0, 0},
    {(char *) "func_code", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "__code__", (getter)__Pyx_CyFunction_get_code, 0, 0, 0},
    {(char *) "func_defaults", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__defaults__", (getter)__Pyx_CyFunction_get_defaults, (setter)__Pyx_CyFunction_set_defaults, 0, 0},
    {(char *) "__kwdefaults__", (getter)__Pyx_CyFunction_get_kwdefaults, (setter)__Pyx_CyFunction_set_kwdefaults, 0, 0},
    {(char *) "__annotations__", (getter)__Pyx_CyFunction_get_annotations, (setter)__Pyx_CyFunction_set_annotations, 0, 0},
    {0, 0, 0, 0, 0}
};
static PyMemberDef __pyx_CyFunction_members[] = {
    {(char *) "__module__", T_OBJECT, offsetof(__pyx_CyFunctionObject, func.m_module), PY_WRITE_RESTRICTED, 0},
    {0, 0, 0,  0, 0}
};
static PyObject *
__Pyx_CyFunction_reduce(__pyx_CyFunctionObject *m, CYTHON_UNUSED PyObject *args)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromString(m->func.m_ml->ml_name);
#else
    return PyString_FromString(m->func.m_ml->ml_name);
#endif
}
static PyMethodDef __pyx_CyFunction_methods[] = {
    {"__reduce__", (PyCFunction)__Pyx_CyFunction_reduce, METH_VARARGS, 0},
    {0, 0, 0, 0}
};
#if PY_VERSION_HEX < 0x030500A0
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func_weakreflist)
#else
#define __Pyx_CyFunction_weakreflist(cyfunc) ((cyfunc)->func.m_weakreflist)
#endif
static PyObject *__Pyx_CyFunction_New(PyTypeObject *type, PyMethodDef *ml, int flags, PyObject* qualname,
                                      PyObject *closure, PyObject *module, PyObject* globals, PyObject* code) {
    __pyx_CyFunctionObject *op = PyObject_GC_New(__pyx_CyFunctionObject, type);
    if (op == NULL)
        return NULL;
    op->flags = flags;
    __Pyx_CyFunction_weakreflist(op) = NULL;
    op->func.m_ml = ml;
    op->func.m_self = (PyObject *) op;
    Py_XINCREF(closure);
    op->func_closure = closure;
    Py_XINCREF(module);
    op->func.m_module = module;
    op->func_dict = NULL;
    op->func_name = NULL;
    Py_INCREF(qualname);
    op->func_qualname = qualname;
    op->func_doc = NULL;
    op->func_classobj = NULL;
    op->func_globals = globals;
    Py_INCREF(op->func_globals);
    Py_XINCREF(code);
    op->func_code = code;
    op->defaults_pyobjects = 0;
    op->defaults = NULL;
    op->defaults_tuple = NULL;
    op->defaults_kwdict = NULL;
    op->defaults_getter = NULL;
    op->func_annotations = NULL;
    PyObject_GC_Track(op);
    return (PyObject *) op;
}
static int
__Pyx_CyFunction_clear(__pyx_CyFunctionObject *m)
{
    Py_CLEAR(m->func_closure);
    Py_CLEAR(m->func.m_module);
    Py_CLEAR(m->func_dict);
    Py_CLEAR(m->func_name);
    Py_CLEAR(m->func_qualname);
    Py_CLEAR(m->func_doc);
    Py_CLEAR(m->func_globals);
    Py_CLEAR(m->func_code);
    Py_CLEAR(m->func_classobj);
    Py_CLEAR(m->defaults_tuple);
    Py_CLEAR(m->defaults_kwdict);
    Py_CLEAR(m->func_annotations);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_XDECREF(pydefaults[i]);
        PyObject_Free(m->defaults);
        m->defaults = NULL;
    }
    return 0;
}
static void __Pyx_CyFunction_dealloc(__pyx_CyFunctionObject *m)
{
    PyObject_GC_UnTrack(m);
    if (__Pyx_CyFunction_weakreflist(m) != NULL)
        PyObject_ClearWeakRefs((PyObject *) m);
    __Pyx_CyFunction_clear(m);
    PyObject_GC_Del(m);
}
static int __Pyx_CyFunction_traverse(__pyx_CyFunctionObject *m, visitproc visit, void *arg)
{
    Py_VISIT(m->func_closure);
    Py_VISIT(m->func.m_module);
    Py_VISIT(m->func_dict);
    Py_VISIT(m->func_name);
    Py_VISIT(m->func_qualname);
    Py_VISIT(m->func_doc);
    Py_VISIT(m->func_globals);
    Py_VISIT(m->func_code);
    Py_VISIT(m->func_classobj);
    Py_VISIT(m->defaults_tuple);
    Py_VISIT(m->defaults_kwdict);
    if (m->defaults) {
        PyObject **pydefaults = __Pyx_CyFunction_Defaults(PyObject *, m);
        int i;
        for (i = 0; i < m->defaults_pyobjects; i++)
            Py_VISIT(pydefaults[i]);
    }
    return 0;
}
static PyObject *__Pyx_CyFunction_descr_get(PyObject *func, PyObject *obj, PyObject *type)
{
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    if (m->flags & __Pyx_CYFUNCTION_STATICMETHOD) {
        Py_INCREF(func);
        return func;
    }
    if (m->flags & __Pyx_CYFUNCTION_CLASSMETHOD) {
        if (type == NULL)
            type = (PyObject *)(Py_TYPE(obj));
        return __Pyx_PyMethod_New(func, type, (PyObject *)(Py_TYPE(type)));
    }
    if (obj == Py_None)
        obj = NULL;
    return __Pyx_PyMethod_New(func, obj, type);
}
static PyObject*
__Pyx_CyFunction_repr(__pyx_CyFunctionObject *op)
{
#if PY_MAJOR_VERSION >= 3
    return PyUnicode_FromFormat("<cyfunction %U at %p>",
                                op->func_qualname, (void *)op);
#else
    return PyString_FromFormat("<cyfunction %s at %p>",
                               PyString_AsString(op->func_qualname), (void *)op);
#endif
}
static PyObject * __Pyx_CyFunction_CallMethod(PyObject *func, PyObject *self, PyObject *arg, PyObject *kw) {
    PyCFunctionObject* f = (PyCFunctionObject*)func;
    PyCFunction meth = f->m_ml->ml_meth;
    Py_ssize_t size;
    switch (f->m_ml->ml_flags & (METH_VARARGS | METH_KEYWORDS | METH_NOARGS | METH_O)) {
    case METH_VARARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0))
            return (*meth)(self, arg);
        break;
    case METH_VARARGS | METH_KEYWORDS:
        return (*(PyCFunctionWithKeywords)meth)(self, arg, kw);
    case METH_NOARGS:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 0))
                return (*meth)(self, NULL);
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes no arguments (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    case METH_O:
        if (likely(kw == NULL || PyDict_Size(kw) == 0)) {
            size = PyTuple_GET_SIZE(arg);
            if (likely(size == 1)) {
                PyObject *result, *arg0 = PySequence_ITEM(arg, 0);
                if (unlikely(!arg0)) return NULL;
                result = (*meth)(self, arg0);
                Py_DECREF(arg0);
                return result;
            }
            PyErr_Format(PyExc_TypeError,
                "%.200s() takes exactly one argument (%" CYTHON_FORMAT_SSIZE_T "d given)",
                f->m_ml->ml_name, size);
            return NULL;
        }
        break;
    default:
        PyErr_SetString(PyExc_SystemError, "Bad call flags in "
                        "__Pyx_CyFunction_Call. METH_OLDARGS is no "
                        "longer supported!");
        return NULL;
    }
    PyErr_Format(PyExc_TypeError, "%.200s() takes no keyword arguments",
                 f->m_ml->ml_name);
    return NULL;
}
static CYTHON_INLINE PyObject *__Pyx_CyFunction_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    return __Pyx_CyFunction_CallMethod(func, ((PyCFunctionObject*)func)->m_self, arg, kw);
}
static PyObject *__Pyx_CyFunction_CallAsMethod(PyObject *func, PyObject *args, PyObject *kw) {
    PyObject *result;
    __pyx_CyFunctionObject *cyfunc = (__pyx_CyFunctionObject *) func;
    if ((cyfunc->flags & __Pyx_CYFUNCTION_CCLASS) && !(cyfunc->flags & __Pyx_CYFUNCTION_STATICMETHOD)) {
        Py_ssize_t argc;
        PyObject *new_args;
        PyObject *self;
        argc = PyTuple_GET_SIZE(args);
        new_args = PyTuple_GetSlice(args, 1, argc);
        if (unlikely(!new_args))
            return NULL;
        self = PyTuple_GetItem(args, 0);
        if (unlikely(!self)) {
            Py_DECREF(new_args);
            return NULL;
        }
        result = __Pyx_CyFunction_CallMethod(func, self, new_args, kw);
        Py_DECREF(new_args);
    } else {
        result = __Pyx_CyFunction_Call(func, args, kw);
    }
    return result;
}
static PyTypeObject __pyx_CyFunctionType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "cython_function_or_method",
    sizeof(__pyx_CyFunctionObject),
    0,
    (destructor) __Pyx_CyFunction_dealloc,
    0,
    0,
    0,
#if PY_MAJOR_VERSION < 3
    0,
#else
    0,
#endif
    (reprfunc) __Pyx_CyFunction_repr,
    0,
    0,
    0,
    0,
    __Pyx_CyFunction_CallAsMethod,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC,
    0,
    (traverseproc) __Pyx_CyFunction_traverse,
    (inquiry) __Pyx_CyFunction_clear,
    0,
#if PY_VERSION_HEX < 0x030500A0
    offsetof(__pyx_CyFunctionObject, func_weakreflist),
#else
    offsetof(PyCFunctionObject, m_weakreflist),
#endif
    0,
    0,
    __pyx_CyFunction_methods,
    __pyx_CyFunction_members,
    __pyx_CyFunction_getsets,
    0,
    0,
    __Pyx_CyFunction_descr_get,
    0,
    offsetof(__pyx_CyFunctionObject, func_dict),
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#endif
};
static int __pyx_CyFunction_init(void) {
    __pyx_CyFunctionType = __Pyx_FetchCommonType(&__pyx_CyFunctionType_type);
    if (__pyx_CyFunctionType == NULL) {
        return -1;
    }
    return 0;
}
static CYTHON_INLINE void *__Pyx_CyFunction_InitDefaults(PyObject *func, size_t size, int pyobjects) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults = PyObject_Malloc(size);
    if (!m->defaults)
        return PyErr_NoMemory();
    memset(m->defaults, 0, size);
    m->defaults_pyobjects = pyobjects;
    return m->defaults;
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsTuple(PyObject *func, PyObject *tuple) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_tuple = tuple;
    Py_INCREF(tuple);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetDefaultsKwDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->defaults_kwdict = dict;
    Py_INCREF(dict);
}
static CYTHON_INLINE void __Pyx_CyFunction_SetAnnotationsDict(PyObject *func, PyObject *dict) {
    __pyx_CyFunctionObject *m = (__pyx_CyFunctionObject *) func;
    m->func_annotations = dict;
    Py_INCREF(dict);
}

/* Py3ClassCreate */
      static PyObject *__Pyx_Py3MetaclassPrepare(PyObject *metaclass, PyObject *bases, PyObject *name,
                                           PyObject *qualname, PyObject *mkw, PyObject *modname, PyObject *doc) {
    PyObject *ns;
    if (metaclass) {
        PyObject *prep = __Pyx_PyObject_GetAttrStr(metaclass, __pyx_n_s_prepare);
        if (prep) {
            PyObject *pargs = PyTuple_Pack(2, name, bases);
            if (unlikely(!pargs)) {
                Py_DECREF(prep);
                return NULL;
            }
            ns = PyObject_Call(prep, pargs, mkw);
            Py_DECREF(prep);
            Py_DECREF(pargs);
        } else {
            if (unlikely(!PyErr_ExceptionMatches(PyExc_AttributeError)))
                return NULL;
            PyErr_Clear();
            ns = PyDict_New();
        }
    } else {
        ns = PyDict_New();
    }
    if (unlikely(!ns))
        return NULL;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_module, modname) < 0)) goto bad;
    if (unlikely(PyObject_SetItem(ns, __pyx_n_s_qualname, qualname) < 0)) goto bad;
    if (unlikely(doc && PyObject_SetItem(ns, __pyx_n_s_doc, doc) < 0)) goto bad;
    return ns;
bad:
    Py_DECREF(ns);
    return NULL;
}
static PyObject *__Pyx_Py3ClassCreate(PyObject *metaclass, PyObject *name, PyObject *bases,
                                      PyObject *dict, PyObject *mkw,
                                      int calculate_metaclass, int allow_py2_metaclass) {
    PyObject *result, *margs;
    PyObject *owned_metaclass = NULL;
    if (allow_py2_metaclass) {
        owned_metaclass = PyObject_GetItem(dict, __pyx_n_s_metaclass);
        if (owned_metaclass) {
            metaclass = owned_metaclass;
        } else if (likely(PyErr_ExceptionMatches(PyExc_KeyError))) {
            PyErr_Clear();
        } else {
            return NULL;
        }
    }
    if (calculate_metaclass && (!metaclass || PyType_Check(metaclass))) {
        metaclass = __Pyx_CalculateMetaclass((PyTypeObject*) metaclass, bases);
        Py_XDECREF(owned_metaclass);
        if (unlikely(!metaclass))
            return NULL;
        owned_metaclass = metaclass;
    }
    margs = PyTuple_Pack(3, name, bases, dict);
    if (unlikely(!margs)) {
        result = NULL;
    } else {
        result = PyObject_Call(metaclass, margs, mkw);
        Py_DECREF(margs);
    }
    Py_XDECREF(owned_metaclass);
    return result;
}

/* CodeObjectCache */
      static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
      #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntToPy */
      static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntFromPyVerify */
      #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntFromPy */
      static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* CIntFromPy */
      static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CStringEquals */
      static CYTHON_INLINE int __Pyx_StrEq(const char *s1, const char *s2) {
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    return *s1 == *s2;
}

/* CheckBinaryVersion */
      static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* InitStrings */
      static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
  #else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
  #endif
#else
  res = PyNumber_Int(x);
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
